#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>

#include "commmod.h"
#include "filesel.h"
#include "struct.h"
#include "editor.h"
#include "ident.h"
#include "netcl.h"
#include "node.h"

#include "xbefore.h"
#include <Xm/RowColumn.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include "xafter.h"

#ifndef NoExplicitInstanciation
#define NoExplicitInstanciation
#include "dialog.cc"
#include "filesel.cc"
#include "wrapper.cc"
#include "linked.cc"

DIAINSTANCIATE(CommModule, Editor*)
FSELINSTANCIATE(CommModule, Editor*)

CINSTANCIATE(CommDialog)
CINSTANCIATE(CommLogwin)
CINSTANCIATE(CommFilesel)
#endif

typedef char* charP;

//-----------------------------------------------------------------------------
// Die Klasse CommDialog enthlt zustzlich die Parameter-Struktur 
//
class CommDialog: public Dialog<CommModule, Editor* >
{
  Struct params;

  friend class CommModule;
  
public:
  CommDialog(CommModule* p, char* paramdef): Dialog<CommModule, Editor* >(p), 
                                             params(paramdef) {}
};


//-----------------------------------------------------------------------------
// Die Klasse CommLogwin benachrichtigt das CommModule von eingelesenen Zeichen
//
class CommLogwin: public LogWindow
{
  CommModule* parent;
  bool  command_mode;
  int  command_state;
  Text command;

  friend class CommModule;

public:
  CommLogwin(CommModule* p, int filedes): LogWindow(filedes), parent(p), 
                                          command_mode(true),command_state(1){}
	CommLogwin(CommModule* p, char* filen): LogWindow(filen), parent(p),
                                          command_mode(true),command_state(1){}
  
  void DialogCB(Dialog<LogWindow, bool> &, bool, Ident& button)
    { parent->current_logwin = this;
      parent->OnLogwinButton(button); }
  
  bool OnRecieveChar(char c)
    { parent->current_logwin = this;
      return parent->OwnOnLogwinChar(c); }
};


//-----------------------------------------------------------------------------
// Die Klasse CommFilesel enthlt zustzlich das letzte Directory
//
class CommFilesel: public FileSelector<CommModule, Editor* >
{
  Text last_dir;
  
  friend class CommModule;

public:
  CommFilesel(CommModule* p, char* dir): FileSelector<CommModule, Editor* >(p),
                                         last_dir(dir) {}
};

//-----------------------------------------------------------------------------
// Konstruktor
//
CommModule::CommModule(Net& net, char* name): Module(net, name), filesel(0),
  current_dialog(0), current_logwin(0), last_editor(0)
{
}

//-----------------------------------------------------------------------------
// Der Destruktor schliet alle offenen Fenster und gibt alle Resourcen frei.
//
CommModule::~CommModule()
{
  while (dialogs) { SetCurrentDialog(1); DestroyDialog(); }
  while (logwins) { SetCurrentLogwin(1); DestroyLogwin(); }
  if (filesel) DestroyFilesel();
}

//-----------------------------------------------------------------------------
// ModuleLock bestimmt, wann das Netz vom Modul gelockt wird.
//
bool CommModule::ModuleLock()
{
  Iterator<CommDialog*> dias = dialogs;
  for (CommDialog* d = dias; d; d=++dias)
    if (*d) return true;

  if (filesel && *filesel) return true;
  else return false;
}


//-----------------------------------------------------------------------------
// (Set)CurrentDialog gibt bzw. setzt den aktuellen Dialog
//
void CommModule::SetCurrentDialog(int i)
{
  CommDialog** cd = dialogs[i-1];
  if (cd) current_dialog = *cd;
}  

int CommModule::CurrentDialog()
{
  if (current_dialog) return dialogs.Index(current_dialog)+1;
  else return 0;
}      

//-----------------------------------------------------------------------------
// CreateDialog erzeugt einen Dialog mit den angegebenen Parametern.
//
void CommModule::CreateDialog(char* params)
{
  current_dialog = new CommDialog(this, params);
  dialogs.Append(current_dialog);
}

//-----------------------------------------------------------------------------
// OpenDialog ffnet einen Dialog. Vorher CreateDialog aufrufen!
//
void CommModule::OpenDialog(Editor& editor, char* title, char* buttons)
{
  if (!current_dialog || *current_dialog) return;

  current_dialog->Create("ObjectDialog", title, editor.Shell(), &editor,
                         OwnDialogCB, buttons);

  Widget form = XtVaCreateManagedWidget("InfoForm", xmFormWidgetClass,
                                        *current_dialog, 0);
  current_dialog->params.EnterDialog(form);
  current_dialog->Manage();
  last_editor = &editor;
}

//-----------------------------------------------------------------------------
// OpenMsgDialog ffnet eine Message-Dialog-Box.
//
void CommModule::OpenMsgDialog(Editor& editor, char* title, char* msg, 
                               bool errormsg, char* buttons)
{
  if (!current_dialog || *current_dialog) 
    CreateDialog();

  current_dialog->OpenMsg("Message", title, editor.Shell(), 
                          errormsg ? XmDIALOG_ERROR : XmDIALOG_INFORMATION, 
                          msg, &editor, OwnDialogCB, buttons);
  last_editor = &editor;
}

//-----------------------------------------------------------------------------
// CloseDialog schliet den aktuellen Dialog
//
void CommModule::CloseDialog()
{
  if (current_dialog)
    current_dialog->Close();
}

//-----------------------------------------------------------------------------
// DestroyDialog gibt den aktuellen Dialog frei.
//
void CommModule::DestroyDialog()
{
  if (current_dialog) {
    CloseDialog();
    dialogs.Remove(current_dialog);
    delete current_dialog;
    current_dialog = 0;
  }
}

//-----------------------------------------------------------------------------
// DialogVisible gibt an, ob der aktuelle Dialog zur Zeit geffnet ist.
//
bool CommModule::DialogVisible()
{
  if (current_dialog) return *current_dialog;
  else return 0;
}

//-----------------------------------------------------------------------------
// Param/GetParams gibt den (die) Parameter des aktuellen Dialogs.
//
Text CommModule::Param(char* name)
{
  if (current_dialog) {
    Value* val = current_dialog->params.GetMember(name);
    if (val) return (char*)(*val);
  }
  return "";
}

Struct* CommModule::GetParams()
{
  if (current_dialog)
    return &current_dialog->params;
  else return 0;
}

//-----------------------------------------------------------------------------
// Eigener Dialog-Callback.
//
void CommModule::OwnDialogCB(Dialog<CommModule, Editor*>& dialog,
                             Editor* editor, Ident& button)
{ 
  last_editor = editor;
  current_dialog = (CommDialog*) &dialog;
  current_dialog->params.EvalWidget();
  OnDialogButton(button); 
}

//-----------------------------------------------------------------------------
// Die Funktion OnDialogButton kann vom Client berschrieben werden.
// Der Default-Callback schliet den Dialog.
//
void CommModule::OnDialogButton(Ident&)
{
  CloseDialog();
}

//-----------------------------------------------------------------------------
// Exec startet einen neuen Prozess lenkt die Eingabe bzw. Ausgabe in die
// gewnschten Files um.
//
int CommModule::Exec(char* command, char* outfile, char* infile)
{
  int outf_fd;
  int inf_fd;
  int pid;

  pid = fork();
  if (pid == 0) {

    if (infile) {
      if ((inf_fd = open(infile, O_RDONLY, 420)) < 0) {
        perror ("Kann die Eingabedatei nicht oeffnen.");
        cout << infile << endl;
        exit (-1);
      }
      dup2 (inf_fd, 0);
    }

    if (outfile) {
      if ((outf_fd = open(outfile ,O_WRONLY|O_CREAT|O_TRUNC,420)) < 0) {
        perror ("Kann die Ausgabedatei nicht oeffnen.");
        cout << outfile << endl;
        exit (-1);
      }
      dup2 (outf_fd, 1);
    }

    char* arguments = new char[strlen(command)];
    char** argv = new charP[strlen(command)/2];
    char* args = arguments;
    
    int argc = 0; *args = 0; argv[0] = args;
    for (char* c=command; *c; c++) {
      if (!isspace(*c)) *args++ = *c;
      else if (args != argv[argc]) {
        *args++ = 0;
        argv[++argc] = args;
      }
    }
    *args++ = 0;
    argv[++argc] = 0;

    execv (argv[0], argv);
    perror ("Kann das externe Programm nicht starten.");
    cout << "Konnte das externe Programm nicht starten: " << argv[0] << endl;
    exit(0);
  }

  int dummy;
  waitpid(-1, &dummy, WNOHANG);
  return pid;
}

//-----------------------------------------------------------------------------
// CreateFilesel erzeugt einen Fileselector-Dialog
//
void CommModule::CreateFilesel(char* dir)
{
  if (!filesel) 
    filesel = new CommFilesel(this, dir);
}

//-----------------------------------------------------------------------------
// OpenFilesel ffnet den Fileselector
//
void CommModule::OpenFilesel(Editor& editor, char* title, char* mask)
{
  if (filesel) 
    filesel->Open(editor.Shell(), title, OwnFileselCB, &editor, 
                  filesel->last_dir, mask);
  last_editor = &editor;
}

//-----------------------------------------------------------------------------
// Eigener Fileselector-Callback.
//
void CommModule::OwnFileselCB(Editor* editor, char* file, bool ok)
{
  last_editor = editor;
  OnFilesel(file, ok); 
}

//-----------------------------------------------------------------------------
// Benutzerdefinierter Fileselector-Callback.
//
void CommModule::OnFilesel(char*, bool)
{
  CloseFilesel(); 
}

//-----------------------------------------------------------------------------
// CloseFilesel/DestroyFilesel schliet den Fileselector bzw. gibt ihn frei.
//
void CommModule::CloseFilesel()
{
  DestroyFilesel();
}

void CommModule::DestroyFilesel()
{
  if (filesel) {
    delete filesel;
    filesel = 0;
  }
}

//-----------------------------------------------------------------------------
// (Set)CurrentLogwin setzt bzw. gibt das aktuelle Ausgabefenster.
//
void CommModule::SetCurrentLogwin(int i)
{
  CommLogwin** cl = logwins[i-1];
  if (cl) current_logwin = *cl;
}  

int CommModule::CurrentLogwin()
{
  if (current_logwin) return logwins.Index(current_logwin)+1;
  else return 0;
}      

//-----------------------------------------------------------------------------
// CreateLogwin startet einen Logvorgang. Das File wird geffnet und alle 
// enhaltenen Kommandos werden ausgefhrt.
//
void CommModule::CreateLogwin(char* filename)
{
  current_logwin = new CommLogwin(this, filename);
  logwins.Append(current_logwin);
}

void CommModule::CreateLogwin2(int filedes)
{
  current_logwin = new CommLogwin(this, filedes);
  logwins.Append(current_logwin);
}

//-----------------------------------------------------------------------------
// Mit SetCommandLog kann man die Ausfhrung von Kommandos im Logfile ein- und 
// ausschalten.
//
void CommModule::SetCommandLog(bool newstate)
{
  if (current_logwin) 
    current_logwin->command_mode = newstate;
}

//-----------------------------------------------------------------------------
// OpenLogwin ffnet ein Logwindow. Vorher StartLog aufrufen!
//
void CommModule::OpenLogwin(Editor& editor, char* title, char* buttons)
{
  if (current_logwin)
    current_logwin->Open("LogWindow", title, editor.Shell(), buttons);
  last_editor = &editor;
}

//-----------------------------------------------------------------------------
// CloseLogwin schliet das Logwindow
//
void CommModule::CloseLogwin()
{
  if (current_logwin) 
    current_logwin->Close();
}

//-----------------------------------------------------------------------------
// DestroyLogwin zerstrt das Logwindow
//
void CommModule::DestroyLogwin()
{
  if (current_logwin) {
    logwins.Remove(current_logwin);
    delete current_logwin;
    current_logwin = 0;
  }
}

//-----------------------------------------------------------------------------
// Eigener Logwin-Callback.
// 
void CommModule::OwnLogwinCB(Dialog<CommModule, bool>& lw, bool, Ident& button)
{
  current_logwin = (CommLogwin*) &lw;
  OnLogwinButton(button);
}

//-----------------------------------------------------------------------------
// Kann vom Benutzer berschrieben werden.
// 
void CommModule::OnLogwinButton(Ident&)
{
  DestroyLogwin();
}

//-----------------------------------------------------------------------------
// OwnOnLogwinChar berwacht alle eingelesen Daten des Logwindows und filtert
// ggf. enthaltene Kommandos heraus.
//
bool CommModule::OwnOnLogwinChar(char c)
{
  if (OnLogwinChar(c)) return true;
  if (!current_logwin->command_mode) return false;

  int& state = current_logwin->command_state;
  Text& command = current_logwin->command;
  bool newline = c=='\n';
  
       if (state==0 && newline) state = 1;
  else if (state==1 && c=='!') { state = 2; return true; }
  else if (state==1 && !newline) state = 0;
  else if (state==2 && !newline) { command += c; return true; }
  else if (state==2 && newline) { 
    LogwinCommand(command); 
    state = 1; command = "";
    return true; 
  }
  
  return false;
}

//-----------------------------------------------------------------------------
// LogwinCommand fhrt im Log enthaltene Kommandos aus. Es sind folgende 
// Kommandos implementiert:
//
// !setattribute nodename attrname value
// !errormessage "message"
// !clear
// !deletelastline
//
void CommModule::LogwinCommand(Text& command_line)
{
  istrstream cmd(command_line);
  Ident command(false);

  cmd>>command;

  if (command=="setattribute") CommandSetAttribute(cmd);
  else if (command=="setnetattribute") CommandSetAttribute(cmd);
  else if (command=="showpage") CommandShowPage(cmd);
  else if (command=="showobject") CommandShowObject(cmd);
  else if (command=="errormessage") CommandErrorMessage(cmd);
  else if (command=="clear") current_logwin->SetContents("");
  else if (command=="deletelastline") CommandDeleteLastLine();
}

//-----------------------------------------------------------------------------
// CommandGetObject liest den Namen eines Netz-Objekts aus der Kommandozeile.
//
Object* CommModule::CommandGetObject(istrstream& cmd)
{
  Page* page = net.Prime();
  Node* node;
  Connector* conn;
  Ident name;
  char c;

  while (cmd) {
    cmd>>c; 
    if (isalpha(c)) cmd.putback(c);
 
    cmd>>name;
   	node = page->GetNode((char*) name);
	  conn = page->GetConnector((char*) name);

    if (!isalpha(cmd.peek()) && !isspace(cmd.peek())
        && node && node->SubPage()) 
      page = node->SubPage();
    else break;
  }
  if (node) return node;
  else return conn;
}

//-----------------------------------------------------------------------------
// Die Command-Methoden setzen die vom Logwindow erhaltenen Kommandos um.
//
void CommModule::CommandSetAttribute(istrstream& cmd)
{
  if (net.Locked()) return;

  Ident name;
  Object* obj = CommandGetObject(cmd);
  cmd>>name;
  if (obj && obj->Attr(name)) {
    cmd>> *obj->Attr(name);
    obj->GetPage()->Update();
  }    
}

void CommModule::CommandSetNetAttribute(istrstream& cmd)
{
  if (net.Locked()) return;
  
  Ident name;
  cmd>>name;
  if (net.Attr(name)) {
    cmd>> *net.Attr(name);
    net.Update();
  }    
}

void CommModule::CommandShowPage(istrstream& cmd)
{
  if (net.Locked()) return;
  Object* obj = CommandGetObject(cmd);
  Page* page = obj && obj->SubPage() ? obj->SubPage() : net.Prime();

  if (last_editor)
    net.ChangeEditor(*last_editor, page);
}

void CommModule::CommandShowObject(istrstream& cmd)
{
  if (net.Locked()) return;
  Object* obj = CommandGetObject(cmd);
  if (!obj) return;

  Page* page = obj->GetPage();
  page->Deselect();
  obj->Select(true);

  if (last_editor) {
    net.ChangeEditor(*last_editor, page);
    last_editor->ScrollVisible(obj->Mid(), true);
  }
}    

  
void CommModule::CommandErrorMessage(istrstream& cmd)
{
  Text message;
  cmd>>message;

  if (last_editor)
    OpenMsgDialog(*last_editor, "Error", message, true);
}

void CommModule::CommandDeleteLastLine()
{
  int i = 0;
  Text text = current_logwin->Contents();
  for (i = text.Length()-2; i>=0; i--)
    if (charP(text)[i]=='\n') break;

  text = text.Left(i+1);
  current_logwin->SetContents(text);
}


