#include "net.h"
#include "page.h"
#include "node.h"
#include "connect.h"
#include "object.h"
#include "editor.h"
#include "ident.h"
#include "netcl.h"
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>

#include "main.h"
#include "regex.h"
#include "metrics.h"

Page::Page(Net* n, NetClass* nc, Node* par):
	net(n),
	netcl(nc),
	parent(par),
	oiter(objects),
	niter(nodes),
	citer(connectors),
	curr_ob(0),
	action(0),
        curr_action(""),
        desel_allowed(true),
        melt_allowed(true)
{
}


Page::Page(Page& pg, Net* par, bool selection, Array<Editor*, Page*>* pages): 
	net(par),
	netcl(pg.netcl),
	parent(0),
	oiter(objects),
	niter(nodes),
	citer(connectors),
	curr_ob(0),
	action(0),
        curr_action(""),
        desel_allowed(true),
        melt_allowed(true)
{
	PageCopy(pg,selection,pages);
}


void Page::PageCopy(Page& src, bool selection, Array<Editor*, Page*>* pages)
{
	LinkedList<Page*> subsrc;
	Iterator<Page*> srciter(subsrc);
	LinkedList<Page*> subdest;
	Iterator<Page*> destiter(subdest);

	if (pages)                         // Wenn Editor-Seiten-Zuweisung uebergeben
		pages->ReplaceValues(&src,this); // dann Seitenkopien zuweisen

	// Erst die Nodes ..
	Iterator<Node*> niter(src.nodes);
	for (Node* n=niter; n; n=++niter)
		if (!selection || n->Selected())
		{
			Node* node = n->Copy(this);
			if (n->SubPage())
			{
				subdest.Append(node->SubPage());
				subsrc.Append(n->SubPage());
			}
		}

	// .. dann die Connectors ..
	Iterator<Connector*> citer(src.connectors);
	for (Connector* c=citer; c; c=++citer)
		if (!selection || c->Selected())
				c->Copy(this);

	// .. dann (wenn alle Objekte existieren) die fehlenden Unterseiten
	Page* nsrc,* ndest;
	for (nsrc=srciter,ndest=destiter; nsrc; nsrc=++srciter,ndest=++destiter)
		ndest->PageCopy(*nsrc,false,pages);
}


Page* Page::Copy(Net* par, bool selection, Array<Editor*, Page*>* pages)
{
	return new Page(*this,par,selection,pages);
}


Page::~Page()
{
	if (net) net->PageKilled(this);
	Object* ob;
	while ((ob = oiter)) delete ob;
}


void Page::Renumber()
{
	Object* ob;

	for (ob=oiter; ob; ob=++oiter)
		ob->Renumber();	

	for (ob=oiter; ob; ob=++oiter)
		if (ob->SubPage())
			ob->SubPage()->Renumber();
}


void Page::UpdateHierarchy()
{
	if (parent)
		parent->UpdateVirtualConnections();

	for (Node* sn=niter; sn; sn=++niter)
		if (sn->SubPage())
			sn->UpdateHierarchy();
}


Page* Page::SuperPage()
{
	return parent?parent->GetPage():0;
}


Page* Page::Prime()
{
	return net?net->Prime():this;
}


void Page::Update()
{
	if (net) net->Update(this);
}


void Page::AddObject(Object* ob)
{
	objects.Append(ob);

	if (ob->Class()->IsNodeClass())
		nodes.Append(NodeP(ob));
	else
		connectors.Append(ConnectorP(ob));

	Update();
}


void Page::RemoveObject(Object* ob)
{
	objects.Remove(ob);

	if (nodes.IsIn(NodeP(ob)))
		nodes.Remove(NodeP(ob));
	else
		connectors.Remove(ConnectorP(ob));

	if (curr_ob==ob)
		FinishAction();

	Update();
}


Object* Page::GetObject(unsigned id)
{
	for (Object* ob=oiter; ob; ob=++oiter)
		if (ob->ID()==id) return ob;
	cerr<<"Unknown object "<<id<<" requested!"<<endl;

	return 0;
}


Node* Page::GetNode(unsigned id)
{
	for (Node* ob=niter; ob; ob=++niter)
		if (ob->ID()==id) return ob;
	cerr<<"Unknown node "<<id<<" requested!"<<endl;

	return 0;
}


Node* Page::GetNode(const char* name)
{
	for (Node* ob=niter; ob; ob=++niter)
		if (ob->Name()==name) return ob;
	cerr<<"Unknown node "<<name<<" requested!"<<endl;

	return 0;
}


Connector* Page::GetConnector(unsigned id)
{
	for (Connector* ob=citer; ob; ob=++citer)
		if (ob->ID()==id) return ob;
	cerr<<"Unknown connector "<<id<<" requested!"<<endl;

	return 0;
}


Connector* Page::GetConnector(const char* name)
{
	for (Connector* ob=citer; ob; ob=++citer)
		if (ob->Name()==name) return ob;
	cerr<<"Unknown connector "<<name<<" requested!"<<endl;

	return 0;
}


Object* Page::NewObject(ObjectClass* obcl)
{
	if (obcl->IsNodeClass())
		return new Node(NodeClassP(obcl),this,""); // "" verhindert AutoName
	else
		return new Connector(ConnectorClassP(obcl),this,"");
}


void Page::Write(ostream& os)
{
	os<<*this;
	Page* sub;

	for (Object* ob=oiter; ob; ob=++oiter)
		if ((sub = ob->SubPage()))
		{
			os<<',';
			sub->Write(os);
		}
}


bool Page::Read(Scope &sn, istream& is)
{
	if (is>>*this)
	{
		Page* sub;
		Iterator<Node*> iter(nodes);
		for (Node* n=iter; n; n=++iter)
			if ((sub = n->SubPage()))
			{
				if (!sn.Get() || !sub->Read(sn,is))
				{
					cerr<<"SubPage of Node "<<n->Name()<<" expected!"<<endl;
					return false;
				}
			}

		return true;
	}
	return false;
}


istream& Page::Input(istream& is)
{
	Ident id(false);
	Scope sn(is);
	while (is && sn.GetId(id))
	{
		if (id=="objects")
		{
			Scope so(is);
			while (is && so.Get())
			{
				if (is>>id)
				{
					ObjectClass* obcl;
					if ((obcl = netcl->GetObjectClass(id)))
					{
						Object* ob = NewObject(obcl);
						if (!(is>>*ob))
							delete ob;
					}
					else
					{
						cerr<<"Unknown objectclass '"<<id<<"'!"<<endl;
						Scope sdum(is); sdum.Skip();
					}
				}
				else
				{
					cerr<<"Name of objectclass expected!"<<endl;
					so.Skip();
					break;
				}
			}
		}
		else if (id=="name")
		{
			Ident name;
			if (!(is>>name))
				cerr<<"Name of page expected!"<<endl;
		}
	}
	
	return is;
}


ostream& Page::Output(ostream& os)
{
	os<<"("<<endl; // <<"name = "<<Name()<<';'<<endl;

	os<<"objects = ("<<endl;
	bool first = true;

	for (Node* n=niter; n; n=++niter)
	{
		if (first) first = false; else os<<','<<endl;
		n->Output(os);
	}

	for (Connector *c=citer; c; c=++citer)
	{
		if (first) first = false; else os<<','<<endl;
		c->Output(os);
	}

	return os<<endl<<"))"<<endl;
}


Node* Page::NearestNode(Position pos, Position& nearest)
{
	double best = 1e300,d;
	Node* nob = 0; Position npos;

	for (Node* n=niter; n; n=++niter)
		if ((d = n->Nearest(pos,npos))<best) { best = d; nob = n; nearest = npos; }

	return nob;
}


Object* Page::UnderMouse(Display* dpy, Position& pos)
{
	Object* ob,* curr = 0;

	for (ob=oiter; ob; ob=++oiter)
		if (ob->IsUnder(dpy,pos))
			curr = ob;

	return curr;
}


Node* Page::NodeForMelt(Display* dpy, Position& pos)
{
	Node* ob,* curr = 0;

	for (ob=niter; ob; ob=++niter)
		if (!ob->Selected() && ob->IsUnder(dpy,pos))
			curr = ob;

	for (ob=niter; curr && ob; ob=++niter)
		if (ob->Selected() && !curr->Meltable(ob))
			curr = 0;

	return curr;
}


void Page::FinishAction()
{
	UpdateHierarchy();
	Update();
	action = 0;
        curr_action = "";
	actives.Clear();
	curr_ob = 0;
}


bool Page::Moving(Editor& editor, EventInfo& ev, bool undo_copy)
{
	switch (ev.event)
 	{
	case EventInit:
		if (action) return true;
		last_pos = ev.pos;
		action = Moving;
                curr_action = "Move";
                if (undo_copy)
		  net->UndoCopy();
		InitActives();
		editor.Message("Moving");
		Update();
		return true;

	case EventMotion:
	{
		EditorSettings set = editor.Settings();
		Position cpos = curr_ob->Mid(),
			npos = (ev.pos-last_pos+cpos).Grid(set.mvgrid)-cpos;

		last_pos -= cpos;
		for (Object* ob=oiter; ob; ob=++oiter)
			if (ob->Selected())
				ob->Move(npos);
		last_pos += curr_ob->Mid();

//		editor.Message(Text("Moving to Position ")+(char*)curr_ob->Mid());
		editor.ScrollVisible(ev.pos);
		Update();
		return true;
	}

	case EventQuickExit:
	case EventExit:
	{
          if (melt_allowed) {
                Node* mnode = NodeForMelt(editor.Dpy(),ev.pos);
		if (mnode)
			for (Node* nob=niter; nob; )
				if (nob->Selected())
				{
					mnode->Melt(nob);
					nob = niter;
				}
				else nob = ++niter;
          }
          else
            ResetMelting();
          FinishAction();
          editor.Message("");
          return false;
	}
	default:
		return true;
	}
}

bool Page::Rotating(Editor& editor, EventInfo& ev, bool)
{
	switch (ev.event)
	{
	case EventInit:
		if (action) return true;
		last_angle = (ev.pos-curr_ob->Mid()).Angle();
		action = Rotating;
                curr_action = "Rotate";
		net->UndoCopy();
		InitActives();
		editor.Message("Rotating");
		Update();
		return true;

	case EventMotion:
	{
		EditorSettings set = editor.Settings();
		double curr_angle = (ev.pos-curr_ob->Mid()).Angle();
		double xang = curr_ob->Angle()+curr_angle-last_angle,
			grid = set.rotgrid/180.0*M_PI,
			ang = grid?(grid*int(xang/grid+.5)):xang;

		for (Node* nob=niter; nob; nob=++niter)
			if (nob->Selected())
				nob->SetAngle(ang);

		last_angle = ang - curr_angle;
		Update();
		return true;
	}

	case EventQuickExit:
	case EventExit:
		editor.Message("");
		FinishAction();
		return false;
	
	default:
		return true;
	}
}

bool Page::Scaling(Editor& editor, EventInfo& ev, bool)
{
	switch (ev.event)
	{
	case EventInit:
		if (action) return true;
//		curr_scale = ev.pos.Distance;
		action = Scaling;
                curr_action = "Scale";
		net->UndoCopy();
		InitActives();
		editor.Message("Scaling");
		Update();
		return true;

	case EventMotion:
	{
		EditorSettings set = editor.Settings();
		Position scale = ((ev.pos-curr_ob->Mid())/100).Grid(set.scgrid/100.0);

		for (Node* nob=niter; nob; nob=++niter)
			if (nob->Selected())
				nob->SetScale(scale);
		
		Update();
		return true;
	}

	case EventQuickExit:
	case EventExit:
		editor.Message("");
		FinishAction();
		return false;
	
	default:
		return true;
	}
}


bool Page::DragboxInside(Position pos)
{
	Position upleft  = 
		dragbox_start.XYMin(dragbox_start+dragbox_dim-Position(1,1)),
	botrig = upleft + dragbox_dim.Abs();

	return pos.x>=upleft.x && pos.x<=botrig.x 
		&& pos.y>=upleft.y && pos.y<=botrig.y;
}	

void Page::Dragbox(Display *dpy, Drawable d, GC gc, double zoom)
{
	if (action==Dragging)
	{
		Position upleft  = 
			dragbox_start.XYMin(dragbox_start+dragbox_dim-Position(1,1))*zoom,
		dim = dragbox_dim.Abs()*zoom;

		XDrawRectangle(dpy,d,gc,int(upleft.x),int(upleft.y),int(dim.x),int(dim.y));
	}
}


bool Page::Dragging(Editor& editor, EventInfo& ev, bool)
{
	switch (ev.event)
	{
	case EventInit:
		if (action) return true;
		action = Dragging;
    curr_action = "Drag";
		dragbox_start = ev.pos;
		dragbox_dim = Position(1,1);
		InitActives();
		editor.Message("Dragging");
		Update();
		return true;

	case EventMotion:
		dragbox_dim = ev.pos-dragbox_start+Position(1,1);
		editor.ScrollVisible(ev.pos);

    if (desel_allowed) // fuer SPNLControl
    for (Object* ob=oiter; ob; ob=++oiter)
      ob->Select(DragboxInside(ob->Mid()));
    else
      ResetDeselection();

		InitActives();
		Update();
		return true;

	case EventQuickExit:
	case EventExit:
		editor.Message("");
		FinishAction();
		return false;
	
	default:
		return true;
	}
}

bool Page::ConnCreate(Editor& editor, EventInfo& ev, bool)
{
	EditorSettings set = editor.Settings();
	Connector*& conn = (Connector*&)curr_ob;
	EventInfo ev2 = ev;
	Display* dpy = editor.Dpy();
  ev2.pos = ev2.pos.Grid(set.mvgrid);
	ev2.button = ButtonSelect;
	
	switch (ev.event)
	{
	case EventInit:
	{
		if (action) return true;
		ConnectorClass* ccl = editor.CurrConnectorClass();
		if (ccl)
		{
			Position start;
			Node* obn = NearestNode(ev2.pos,start);
			if (!obn)
			{
				FinishAction();
				return false;
			}

			action = ConnCreate;
      curr_action = "ConnCreate";
			net->UndoCopy();
			conn = new Connector(ccl,this);
			conn->InsertPos(0,start);
			conn->InsertPos(1,ev2.pos);
			conn->SetConnections(obn,0);

      if (conn->Event(editor.Dpy(),ev2))
			{
				actives.Clear();
				actives.Append(curr_ob);
				editor.Message("Creating Connector");
				Update();
				return true;
			}
			else
				delete conn;
		}
		return false;
	}

	case EventMotion:
		conn->Event(dpy,ev2);
		conn->Rearrange();
		editor.ScrollVisible(ev2.pos);
		Update();
		return true;

	case EventQuickExit:
	case EventExit:		
	{
		curr_ob->Event(dpy,ev2);

		ConnectorClass* ccl = conn->Class();
		Node* in = conn->GetInput(),*out = conn->GetOutput();

		if (!ccl->AllowedConnection(in,out))
			if ((ccl = netcl->FittingConnectorClass(in,out)))
				conn->SetClass(ccl);
			else			
			{
				delete conn;
				conn = 0;
			}

		if (conn) 
			conn->Rearrange();
		editor.Message("");
		FinishAction();
		return false;
	}
	default:
		return true;
	}
}

bool Page::SubMoving(Editor& editor, EventInfo& ev, bool)
{
	EditorSettings set = editor.Settings();
	EventInfo ev2 = ev;
	Display* dpy = editor.Dpy();
	
	if (!curr_ob->IsNode())
		ev2.pos = ev2.pos.Grid(set.mvgrid);

	switch (ev.event)
	{
	case EventInit:
		if (action) return true;
		if (curr_ob->Event(dpy,ev))
		{
			action = SubMoving;
                        curr_action = "SubMove";
			net->UndoCopy();
			InitActives(false);
			editor.Message("Moving Relative Position");
			return true;
		}
		else return false;

	case EventMotion:
		curr_ob->Event(dpy,ev2);
		editor.ScrollVisible(ev.pos);
		Update();
		return true;

	case EventQuickExit:
	case EventExit:
		curr_ob->Event(dpy,ev2);
		editor.Message("");
		FinishAction();
		return false;	
	
	default:
		return true;
	}
}		


bool Page::Event(Editor& editor, EventInfo& ev)
{
	Display* dpy = editor.Dpy();
	EditorSettings set = editor.Settings();
	Widget shell = editor.Shell();

	switch (ev.event)
	{
	case EventInit:
		if (action)
			return (this->*action)(editor,ev);
		
		switch (ev.button)
		{
		case ButtonUser:
			if (set.rclick=="Scrolls")
			{
				editor.Scroll();
				return false;
			}
			else if (set.rclick=="Undoes")
			{
				net->Undo();
				return false;
			}
      else if (set.rclick=="Pastes")
      {
        net->UndoCopy();
        Paste(editor,dpy,ev.pos);
        return Moving(editor,ev,false);
      }
 			// else fall thru
		case ButtonSelect:
			if ((curr_ob = UnderMouse(dpy,ev.pos)))
			{
				if (ev.shift)
				{
          curr_ob->Select(!curr_ob->Selected());
					return false;
				}

        if (!curr_ob->Selected() && desel_allowed)
				{
					Deselect();
					curr_ob->Select(true);
				}
        ResetDeselection();
				
				if (ev.button!=ButtonSelect || !SubMoving(editor,ev))
				{
					if (ev.button==ButtonSelect)
						return Moving(editor,ev);
					else
					{
						if (set.rclick=="Rotates")
							return Rotating(editor,ev);
						else if (set.rclick=="Scales")
							return Scaling(editor,ev);
						else if (set.rclick=="Removes")
						{
							net->UndoCopy();
							Remove();
							return false;
						}
					}
				}
				else return true;
			}
			else
			{
				if (!ev.shift) Deselect();
				return Dragging(editor,ev);
			}
			return false;

		case ButtonCreate:
			if ((curr_ob = UnderMouse(dpy,ev.pos)))
			{
				if (curr_ob->IsNode())
					return ConnCreate(editor,ev);
				else
				{
					Deselect();
					curr_ob->Select(true);
					return SubMoving(editor,ev);
				}
			}
			else // Knoten erzeugen (machen wir gleich hier..)
			{
				if (!ev.shift) Deselect();
	
				net->UndoCopy();
				curr_ob = new Node(editor.CurrNodeClass(),this,0,ev.pos.Grid(set.mvgrid));
				Deselect();
				curr_ob->Select(true);
				
        return Moving(editor,ev,false);
			}
		}
		return false;

	case EventDouble:
		if (action)
			return (this->*action)(editor,ev);
	
		if ((curr_ob = UnderMouse(dpy,ev.pos)))
			switch (ev.button)
			{
			case ButtonSelect: {
        Iterator<Module*> miter = net->Modules();
        bool doneit = false;
        for (Module* m=miter; m; m=++miter)
          doneit = doneit || m->DoubleClick(editor, *this, curr_ob);

        if (doneit) break;
				else if (curr_ob->SubPage()) 
					net->ChangeEditor(editor,curr_ob->SubPage());       
				else
				{
					Deselect();
					curr_ob->Select(true);
					net->ObjectDialog(shell,curr_ob,set.expert_mode, false);
				}}
				break;
			default:
				return curr_ob->Event(dpy,ev);
			}

		curr_ob = 0;
		return true;

	case EventMotion:
	case EventQuickExit:
	case EventExit:
    if (action)
			return (this->*action)(editor,ev);

		if (ev.button==ButtonUser)
			editor.Scroll();

		return false;
	
	default:
		return true;
	}
}	


Text Page::FullName()
{
	return parent ? parent->FullName() : net->FullName();
}


void Page::InitActives(bool linked_conns)
{
	Node* n;

	actives.Clear();
	for (n=niter; n; n=++niter)
		if (n->Selected()) actives.Append(n);

	for (Connector* c=citer; c; c=++citer)
		if (c->Selected() ||
		(linked_conns && ((n = c->GetInput()) && n->Selected()) 
		|| ((n = c->GetOutput()) && n->Selected())))
		actives.Append(c);
}


void Page::Draw(Editor& e, Display* dpy, Drawable d, bool draw)
{
	EditorSettings set = e.Settings();
	bool active = action!=0;
	Iterator<Object*> iter(active?actives:objects);
	Object* ob;

	if (draw)
	{
		if (!active) e.Clear();

		for (ob=iter; ob; ob=++iter)
			ob->Draw(dpy,d,set.zoom,active);

		Dragbox(dpy,d,e.DragboxGC(),set.zoom);
	}
	else
	{
		Dragbox(dpy,d,e.DragboxGC(),set.zoom);

		if (active)
			for (ob=iter; ob; ob=++iter)
				ob->Draw(dpy,d,set.zoom,active);
	}
}


bool Page::NameExists(Object* nob, bool check)
{
	if (check || netcl->CheckNames())
	{
		for (Object* ob=oiter; ob; ob=++oiter)
			if (nob!=ob && nob->Name()==ob->Name())
				return true;
	}

	return false;
}


void Page::Remove()
{
	Object* ob,* next;

	for (ob=oiter; ob; ob=next)
		if (ob->Selected())
		{
			delete ob; next = oiter;
		}
		else next = ++oiter;
	UpdateHierarchy();
}


Page* Page::CurrSubPage()
{
	for (Node* n=niter; n; n=++niter)
		if (n->Selected() && (n->SubPage()))
			return n->SubPage();
	return 0;
}


void Page::Select(char* expr)
{
	Object* ob;
	
	if (expr)
		for (ob=oiter; ob; ob=++oiter)
			ob->Select(Matches(ob->Name(),expr));
	else
	{
		for (ob=oiter; ob; ob=++oiter)
			if (!ob->Selected()) break;
		if (ob)
		{
			for (; ob; ob=++oiter)
				if (!ob->Selected())
					ob->Select(true);
		}
		else
			for (ob=oiter; ob; ob=++oiter)
				if (ob->Selected())
					ob->Select(false);
	}
}


void Page::Copy()
{
	Page& cutbuf = netcl->cutbuffer;
	cutbuf.Clear();
	cutbuf.PageCopy(*this,true);
}


void Page::Paste(Editor& ed,Display* dpy,Position pos)
{
        EditorSettings set = ed.Settings();
        Object* ob;
        Page& cutbuf = netcl->cutbuffer;
        for (ob=oiter; ob; ob=++oiter)
                ob->Select(false);
        cutbuf.Renumber();
        PageCopy(cutbuf);

        // Mitte der Objektmenge finden
        Metrics m;
        for (ob=oiter; ob; ob=++oiter)
          if (ob->Selected()) 
            m.Compare(ob->GetMetrics(dpy));

        //cout << "middle";
        //m.Print();
        Position middle((m.x_beg+m.x_end)*0.5,(m.y_beg+m.y_end)*0.5);

        for (ob=oiter; ob; ob=++oiter)
          if (ob->Selected()) {
            ob->Move((pos-middle).Grid(set.mvgrid));
            curr_ob = ob; // Wird danach fuer Moving() gebraucht
          }
        UpdateHierarchy();
}


void Page::Paste()
{
	Object* ob;
	Page& cutbuf = netcl->cutbuffer;
	for (ob=oiter; ob; ob=++oiter)
		ob->Select(false);
	cutbuf.Renumber();
	PageCopy(cutbuf);
	for (ob=oiter; ob; ob=++oiter)
		if (ob->Selected())
			ob->Move(Position(10,10));
	UpdateHierarchy();
}


void Page::Clear()
{	
	Object* ob;
	while ((ob = oiter)) delete ob;
	UpdateHierarchy();
}


void Page::Deselect()
{
	for (Object* ob=oiter; ob; ob=++oiter)
		if (ob->Selected()) ob->Select(false);
}


Object* Page::NextSelected(Object* ob)
{
	for (Object* next=oiter; next; next=++oiter)
		if (!ob)
		{
			if (next->Selected()) return next;
		}		
		else if(next==ob)
		{
			if ((next = ++oiter))
				while (!next->Selected() && (next = ++oiter));
			
			return next;
		}
	
	return 0;
}


void Page::SelectNextObject(bool fwd)
{
	if(!objects) return;
	Iterator<Object*> iter(objects);
	Object* curr,* prev = iter.Last();
	
	for (curr=iter; curr; prev=curr,curr=++iter)
		if (curr->Selected())
		{
			Deselect();
			(fwd ? (curr=++iter) ? curr : iter.First() : prev)->Select(true);
			return;
		}
	
	iter.First()->Select(true);
}


