/******************************************************************************
* Programm  : gensol                                                          *
* Komponente: do_expand.c Funktion expand()                                   *
* Parameter : globale Variablen aus main.c                                    *
* Rueckgabe : 0 - alles OK, oder Programmtermination mit exit()               *
* Aufgabe   : expandieren des reduzierten Erreichbarkeitsgraphen              *
*             Der Aufrufkontext (hier main.c) hat bereits die Eingabedateien  *
*             gelesen und die Tabelle der alten und neuen Markierungen inital.*
* Projektleiter: Reinhard German, TU-Berlin                                   *
* Programmierer: Thomas Kuhlmann, TU-Berlin                                   *
* Datum:         Mon Jun 21 18:18:40 MET DST 1993                             *
******************************************************************************/
#include "expand.h"

extern int PNewMarkCompare(); /* totale Ordnungsfunktionen auf Listen */
extern int PredCompare    ();

/*
 *	Diese beiden Variablen werden haeufig benutzt und deshalb nicht als Para-
 * meter durch alle Funktionen gereicht, sondern hier global fuer alle
 * Funktionen aus expand.c verfuegbar gemacht:
 */
static new_mark_t *n_mark;     /* aktuell bearbeitete neue Markierung */
static phase_t    *phase_list; /* aktuell bearbeitete Phasenzaehler   */

/*
 *	Speicher fuer eine neue Markierung allozieren und als neue aktuelle
 * Markierung bekanntgeben:
 */
void AllocNewMark()
{
	n_mark = (new_mark_t*)emalloc(1,sizeof(new_mark_t));
	n_mark->phase = phase_list = (phase_t*)
			emalloc(no_det_trans, sizeof(phase_t));
	n_mark->new_no   = no_new_marks;
	n_mark->no_preds = (pred_cnt_t)0;
	n_mark->pred     = (pred_t*)0;
}

/*---------------------------------------------------------------------------
 *	Neue Markierung erzeugen und aktuelle als Vorgaenger der Neuen eintragen:
 * in:  n_mark, new_mark, no_new_marks, curr_new->next
 * out: n_mark, new_mark, no_new_marks, curr_new
 */
void AddPredecessor(rate)
double rate;
{
	new_mark_t **p_new;
	pred_t      *ppred;
	pred_t       pred;

	/*
	 *	Verweis in geordnete Liste von neuen Markierungen einfuegen:
	 */
	if (insert( (char*)&n_mark, (char**)&new_mark, &no_new_marks,
		         sizeof(new_mark_t*), PNewMarkCompare, (char**) &p_new) == NEW) {
		/*
		 *	Markierung ist neu, also in die Liste der noch nicht bearbeiteten:
		 */
		n_mark->next   = curr_new->next;
		curr_new->next = n_mark;
		/*
		 *	neuen Arbeitsspeicher fuer die naechste neue Markierung beschaffen:
		 */
		AllocNewMark();
#ifdef DEBUG
		printf("Adding new marking:\n");
		printf("%02d %02d\n", (*p_new)->new_no, (*p_new)->old_no );
#endif
	}

	/*
	 *	Aktuelle Markierung bei Neuer als Predecessor eintragen:
	 */
	curr_new->ratesum += (pred.rate = rate);
	pred.from_mark     = curr_new->new_no;
	if ((*p_new)->new_no == curr_new->new_no) {
		fprintf(stderr,"WARNING: selfloop dedected, solving will fail maybe!\n");
		fprintf(stderr,"%02d %02d\n", (*p_new)->new_no, (*p_new)->old_no );
	}
	if (insert( (char*)&pred, (char**)&((*p_new)->pred), &((*p_new)->no_preds),
		sizeof(pred_t), PredCompare, (char**)&ppred ) == FOUND) {
		/*
		 *	Kanten zusammenfassen (Raten addieren):
		 */
		ppred->rate += rate;
	} else no_new_edges++;
#ifdef DEBUG
	printf("new predecessor for:\n");
	printf("%02d %02d\n", (*p_new)->new_no, (*p_new)->old_no );
	printf(" from: %02d, with rate: %.15g\n", curr_new->new_no, ppred->rate );
#endif
}

/*---------------------------------------------------------------------------
 *	Phasenzaehler der deterministischen Transitionen abhaengig von der Schaltse-
 * mantik und den schaltfaehigen Transitionen neuinitialisieren. In UpdatePhase
 * wird der Zaehler der gerade schaltenden Transition noch nicht gezaehlt!
 * in:  curr_new->phase, old_mark[], n_mark->old_no, phase_list
 * out: n_mark->phase == phaselist
 */
void UpdatePhase()
{
	det_tra_no_t fireable_det;
	det_cnt_t    det_tra_no;
	det_tran_t  *det_tra;
	old_mark_t  *o_mark;

	switch(policy) {
	case RESAMPLING: /* Alle Zaehler auf den Startwert zuruecksetzen: */
		INIT_PHASES(phase_list);
		break;
	case ENABLING:
		INIT_PHASES(phase_list); /* erstmal alle zuruecksetzen */
		/*
		 *	Die Phasenzaehler der determinisischen Transitionen, die in
		 * der neuen Markierung schaltfaehig sind werden uebernommen:
		 * War eine det. Transition bisher nicht schaltfaehig, so wird ihr
		 * Phasenzaehler auch uebernommen, da er in diesem Fall vorher bereits
		 * auf den Startwert zurueckgesetzt wurde.
		 */
		o_mark  = &(old_mark[n_mark->old_no]);
		det_tra = o_mark->det_tran;
		for( det_tra_no = o_mark->no_det_trans; det_tra_no--; ) {
			fireable_det = det_tra[det_tra_no].tra_no;
			phase_list[fireable_det] = curr_new->phase[fireable_det];
		}
		break;
	default /*AGE*/: /* alle Phasenzaehler unveraendert uebernehmen */
		PHASELIST_COPY(phase_list, curr_new->phase);
		break;
	}
}

/******************************************************************************
* Hauptfunktion, die das Expandieren des reduzierten Erreichbarkeitsgraphen   *
* durchfuehrt.                                                                *
* Es werden solange neue Markierungen erzeugt, bis alle Neuen bearbeitet sind *
******************************************************************************/
int expand()
{
	old_mark_t    *o_mark;
	exp_path_t    *exp_path;
	det_tran_t    *det_tran;
	det_path_t    *det_path;
	double         rate;
	exp_path_no_t  exp_path_no;
	det_cnt_t      det_tra_cnt;
	det_tra_no_t   fireable_det;
	det_path_no_t  det_path_no;
	phase_t        phase;
	bool_t         cox_nach_erster;

	AllocNewMark(); /* Platz fuer aktuelle Arbeitsmarkierung beschaffen */
	while (curr_new) {
		o_mark = &(old_mark[curr_new->old_no]);
		curr_new->ratesum  = (double)0.0;
		/*
		 *	Create all exponential Edges for curr_new:
		 */
		exp_path = o_mark->exp_path;
		for( exp_path_no = o_mark->no_exp_paths; exp_path_no--; ) {
			/*
			 *	Die neue Markierung wird der Alten zugeordnet, die durch Schalten
			 * der Transition im alten RRG erreicht wuerde.
			 */
			n_mark->old_no = exp_path[exp_path_no].to_mark;
			rate           = exp_path[exp_path_no].rate;
			UpdatePhase();
			AddPredecessor(rate);
		}
		/*
		 *	Create all deterministic Edges for curr_new:
		 */
		for ( det_tra_cnt = o_mark->no_det_trans; det_tra_cnt--; ) {
			det_tran     = &(o_mark->det_tran[det_tra_cnt]);
			fireable_det = det_tran->tra_no;
			phase        = curr_new->phase[fireable_det];
			cox_nach_erster = ((phase_type==COX) && (phase==no_phases));
			if (!(--phase) || cox_nach_erster) {
				/*
				 * Phasenzaehler abgelaufen, oder COX nach der ersten Phase
				 *	fuer jeden Pfad eine Kante AUFSPANNEN:
				 */
				for( det_path_no = det_tran->no_paths; det_path_no--; ) {
					det_path = &(det_tran->det_path[det_path_no]);
					n_mark->old_no = det_path->to_mark;
					UpdatePhase();
					/*
					 *	Phasenzaehler neu initialisieren:
					 */
					phase_list[fireable_det] = no_phases;
					AddPredecessor( (cox_nach_erster)?
					                det_path->rate * eins_minus_a:
					                det_path->rate );
				}
			}
			if (phase) {
				/*
				 * nicht in der letzten Phase, dann immer mind. EINE Kante:
				 */
				n_mark->old_no = curr_new->old_no; /* selbe zugeordnete alte */
				UpdatePhase();
				phase_list[fireable_det] = phase;
				AddPredecessor( (cox_nach_erster) ?
				                det_tran->rate * a:
				                det_tran->rate );
			}
		} /* end for(det_tra_cnt.. */
		/*
		 *	bearbeitete neue Markierung ausketten:
		 */
		curr_new = curr_new->next;
	}
	return (0);
}
