/******************************************************************************
* Programm  : gensol                                                          *
* Komponente: read.c (read_rrg())                                             *
* Parameter : FILE *rrg_file                                                  *
* Rueckgabe : 0 - alles OK, Datenstrukturen initialisiert, 1 - Fehler in .rrg *
* Aufgabe   : Einlesen des reduzierten Erreichbarkeitsgraphen und fuellen der *
*             globalen Datenstrukturen (Liste der alten Markierungen)         *
* Projektleiter: Reinhard German, TU-Berlin                                   *
* Programmierer: Thomas Kuhlmann, TU-Berlin                                   *
* Datum:         Mon Jun 21 18:18:40 MET DST 1993                             *
* nderung:      Tue Mar  8 18:25:58 MET 1994                                 *
******************************************************************************/
#include "expand.h"

extern int ExpPathCompare ();     /* Ordnungsrelationen auf diversen Listen */
extern int DetTraCompare  ();
extern int DetPathCompare ();
extern int TimedTraCompare();
extern det_tra_no_t det_tra_no();

/* ---------------------------------------------------------------------------
 *	rrg-File einlesen:
 * dabei werden folgende globale Variablen definiert:
 *		no_det_trans  - Anzahl der global vorhandenen deterministischen Trans.
 *		no_old_marks  - Anzahl der Marken im alten RRG
 *		old_mark      - alter RRG
 * Probleme gabs mit fscanf() bei %f mit double! -> %lf
 */
read_rrg(frrg)
FILE *frrg;
{
	old_mark_no_t from, to;
	unsigned      tra_id;
	char          type;
	double        delay, prob;
	old_mark_t   *mark;
	exp_path_t    exp_path;
	exp_path_t   *p_new_exp;
	det_tran_t    det_tran;
	det_tran_t   *p_new_det;
	det_path_t    det_path;
	det_path_t   *p_new_path;
	double        faktor; /* ERLANG: no_phases, COX: 1 + (r-1)*a */
	timed_tra_no_t timed_tran_no;
	timed_tra_t  timed_transition, *p_timed;
	int           edges;
	char dummy[255], serverType[255];

	if (skip_comment_lines(frrg)) return 1;
	if (fscanf(frrg,"%d\n",&no_old_marks)<1) return 1;
	old_mark = (old_mark_t *) ecalloc(no_old_marks, sizeof(old_mark_t));
	no_det_trans = 0;
	if (skip_comment_lines(frrg)) return 1;
	if (fscanf(frrg,"%d\n",&no_old_edges)<1) return 1;
	if (skip_comment_lines(frrg)) return 1;
	if (fscanf(frrg,"%u\n",&timed_tran_no)<1) return 1;
	/*
	 *	read transition description into array timed_tran[] sorted by timed_no
	 */
	while (timed_tran_no--) {
		if (skip_comment_lines(frrg)) return 1;
		if (fscanf(frrg,"%s %u %c %s %s\n", dummy, &(timed_transition.timed_no),
			&type, serverType, dummy )!=5) return 1;
		switch(type){
		case 'e': timed_transition.type = EXPONENTIAL;   break;
		case 'd': timed_transition.type = DETERMINISTIC; break;
		default:
			fprintf(stderr,"%s: transiton type '%c' in <model>.rrg not supported\n"
				,progname,type);
			exit(1);
		}
		if (strcmp(serverType,"SS")) {
			fprintf(stderr,"%s: serverType '%s' in <model>.rrg not supported\n",
				progname, serverType );
			return 1;
		}
		if (insert(&timed_transition,&timed_tran,&no_timed_trans,
				sizeof(timed_tra_t),TimedTraCompare,&p_timed)==FOUND) {
			fprintf(stderr,"%s transition number %u twice in <model>.rrg\n",
					progname, timed_transition.timed_no );
			exit(1);
		}
	}
	if (phase_type == ERLANG)
		faktor = (double)no_phases;
	else
		faktor = ((double)1.0) + ((double)(no_phases-1))*a;
	/*
	 *	read edges, format: from to tranNo delay prob serverMult
	 */
	edges = no_old_edges;
	while (edges--) {
		if (skip_comment_lines(frrg)) return 1;
		if (fscanf(frrg,"%u %u %u %lf %lf\n",
				&from, &to, &tra_id, &delay, &prob)!=5) return 1;
		if (from>=no_old_marks || to>=no_old_marks) {
			fprintf(stderr,"number of marking out of range:%d!\n",MAXIMUM(from,to));
			return (1);
		}
		/*
		 *	delay ueberpruefen, damit beim Teilen dadurch kein Problem auftritt
		 */
		if (delay < EPSILON) {
			fprintf(stderr,"delay %.15g evaluates to zero!\n",delay);
			return 1;
		}
		/*
		 *	transition ermitteln (f"ur type)
		 */
		timed_transition.timed_no = tra_id;
		if (!(p_timed=(timed_tra_t*)bsearch2(&timed_transition,timed_tran,
					no_timed_trans,sizeof(timed_tra_t),TimedTraCompare,&p_timed))) {
			fprintf(stderr,"%s: can't find transiton number %d in <mode>.rrg\n",
					progname, tra_id );
			return 1;
		}
		/*
		 *	 we don't care about INFINITE servers yet!
		 */
		mark = &(old_mark[from]);
		if (p_timed->type == DETERMINISTIC) {
			/*
			 *	add fireable deterministic transition and one path
			 */
			det_tran.rate     = faktor/delay;
			det_tran.tra_no   = det_tra_no(tra_id);
			det_tran.prob_sum = prob;
			det_tran.det_path = (det_path_t *)0;
			det_tran.no_paths = (det_path_no_t)0;
			/*
			 *	Transition der Liste der schaltfaehigen deterministischen zufuegen
			 */
			if (insert( (char*)&det_tran, (char**)&(mark->det_tran),
					&(mark->no_det_trans), sizeof(det_tran_t), DetTraCompare,
					(char**)&p_new_det ) == FOUND) {
				if (p_new_det->rate != det_tran.rate) {
					fprintf(stderr,"WARNING:%s(%d):\n",__FILE__,__LINE__);
					fprintf(stderr,"the same det transition with two different rates is fireable in marking %d\n", from);
					return 1;
				}
				p_new_det->prob_sum += prob; /* Summe muss am Schluss 1 sein */
			}
			/*
			 *	Schaltpfad hinzufuegen:
			 */
			det_path.rate    = p_new_det->rate * prob; /* rate gleich anpassen */
			det_path.to_mark = to;
			if (insert( (char*)&det_path, (char**)&(p_new_det->det_path),
					&(p_new_det->no_paths), sizeof(det_path_t), DetPathCompare,
					(char**) &p_new_path ) == FOUND) {
				fprintf(stderr,"the same det transition fires twice from marking %d to marking %d!\n",from,to);
				return 1;
			}
		} else {
/*
			if (type != 'e') {
				fprintf(stderr,"unknown transition type:%c\n",type);
				return 1;
			}
*/
			exp_path.rate    = prob/delay;
			exp_path.to_mark = to;
			if (insert( (char*)&exp_path, (char**)&(mark->exp_path),
					&(mark->no_exp_paths), sizeof(exp_path_t), ExpPathCompare,
					(char**)&p_new_exp ) == FOUND) {
				/*
				 *	Element gefunden: (Kante gab es schon) rate anpassen:
				 */
				printf("HINT:there is more then one exponential transition that fires from marking %d to %d (eleminating one by adding rates..)\n",from,to);
                                fflush(stdout);
				p_new_exp->rate += exp_path.rate;
			}
		}
	}
	/*
	 *	Fuer jede Markierung und jede deterministische Transition checken, ob
	 * die Summe der Wahrscheinlichkeiten der einzelnen Schaltpfade eins ist:
	 *		passiert in initNewMarks()
	 */
	return (0);
}
