/*************************************************************************
*                                                                        *
* KOMPONENTE  :   solve_LGS.c                                            *
* PROGRAMM    :   gensol                                                 *
*                                                                        *
* ZWECK       :   1. Auswahl des Berechnungsverfahrens zum Loesen des    *
*                    Gleichungssystems                                   *
*		  2. Berechnung aufrufen                                 *
*		  3. Bereitstellen des Loesungsvektors                   *
*                                                                        *
* PARAMETER   :   Loesungsmethode                                        *
*                 Genauigkeit                                            *
*                 max. Anzahl auszufuehrender Iterationen                *
* ====================================================================== *
* erstellt am    :   Juni 1994                                           *
* Programmierer  :   Anja Seibt                                          *
* Betreuer       :   Reinhard German                                     *
*************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include "expand.h"
#include "struct_types.h"


extern  int      PNewMarkNewNoComp();
extern  double*  p_gauss_skal_Pivot();
extern  double*  p_gauss_voll_Pivot();
extern  double*  p_iteration();


/* -------------------------------------------- Globale Variablen ----- */
Matrix_T  *Matrix;    /* die Matrix */
unsigned   Dm;        /* Dimension der Matrix */
double     Droptol;
double     Prec;
int        Nachiter;
int        OK;


/* -----------------------------------------------------------------------
   Abschaetzen des mindest notwendigen Speichers fuer die Matrix
----------------------------------------------------------------------- */
void  p_speicher_schaetzen( nonzeros)
unsigned  nonzeros;
{
  int  gr;


  gr = ((int) ((Dm*Dm)/2)) * sizeof(double) 
       + Dm * sizeof(double)
       + Dm * sizeof(Matrix_T)
       + 3 * Dm * sizeof(int);

  printf("required space for P-Matrix: %i byte\n",gr);

}  /* p_speicher_schaetzen */



/* ------------------------------------------------------------ main -----
   NAME  :  p_solve_LGS

   ZWECK :  - Gewuenschtes Losungsverfahren anwenden
            - Konvertierung zu DSPN Zustandswahrscheinlichkeiten

   EINGABEPARAMETER :
      sol_typ  - Berechnungsmethode (autom., iterativ, direct)
      prec     - Berechnungsgenauigkeit
      max_iter - Anzahl der max. auszufuehrenden Iterationen
----------------------------------------------------------------------- */

void  p_solve_LGS( sol_type, drop_tol, precision, iterations, nach)
solution_t  sol_type;
double      drop_tol;
double      precision;
int         iterations;
int         nach;
{
  double     *X, *Xstart;
  unsigned    zeros_caused_by_normation;
  unsigned    nonzeros;
  double      maxdiagonal, one_diagonal;
  unsigned    i, j, no;
  int         fd;
  int         exact;
  new_mark_t *pnmark;
  pred_t     *ppred;



  puts ("\nSOLVING THE LINEAR SYSTEM OF GLOBAL BALANCE EQUATIONS (new)... \n");
  fflush(stdout);  

  if (0 == no_new_marks)
    {
      fprintf(stderr,"Dimension des Erreichbarkeitsgraphen = 0!\n");
      exit(1);
    }
  /* ....... Graphen nach aufsteigender neuer Markierungsnummer sortieren,
   * .. damit die Zeilen der P-Matrix in richtiger Reihenfolge sind ... */
  (void) qsort((char*)new_mark,
	       (int)no_new_marks,
	       sizeof(new_mark_t*),
	       PNewMarkNewNoComp );
  
  /* .. betragsmaessig groesstes Diagonalelement ermitteln und zaehlen, 
     .. wie haeufig es vorkommt. Denn soviele Elemente werden zusaetzlich 
     .. in der P-Matrix Null sein.
   */
  maxdiagonal               = (double)0.0;
  zeros_caused_by_normation = (unsigned)0;
  for(i = no_new_marks; i--; )
    {
      pnmark = new_mark[i];
      if( (one_diagonal=fabs(pnmark->ratesum)) > maxdiagonal)
        {
          maxdiagonal               = one_diagonal;
          zeros_caused_by_normation = (unsigned)1;
        }
      else if (pnmark->ratesum == maxdiagonal)
	zeros_caused_by_normation++;
    }

  /* Anzahl der Nicht-Null-Elemente = Anz( Kanten )
   *                                  + Anz( Diagonalelemente )
   *                                  - Anz( Diagonal-Null-Elemente )
   */
  nonzeros=(unsigned)(no_new_edges 
		      + no_new_marks 
		      - zeros_caused_by_normation);

  Dm = no_new_marks;                               /* globale Variablen */
  printf ("\n%-31s %d\n","Number of nonzeros :", nonzeros);
  printf ("%-31s %.2lf\n","Number of nonzeros per row :", (double)nonzeros/(double)Dm);
  fflush(stdout);
  
  Droptol = drop_tol;
  if (0.0 == Droptol)
    exact = 1;
  else
    exact = 0;
  Prec = precision;
  /*  printf("precision : %e\n", Prec); */
  Nachiter = nach;


  switch (sol_type) 
    {
    case AUTOMATIC:
      OK = FALSE;
      Nachiter = TRUE;

      if (NULL == (Xstart = (double*) malloc ((int)Dm * sizeof(double))))
	{
	  perror("p_solve_LGS  --  malloc: vector Xstart\n");
	  exit(1);
	}
      for (i = 0; i < Dm; i++)
	Xstart[i] = 1.0/(double)Dm;
	  
      X = p_iteration( maxdiagonal, iterations, Xstart, exact, ITERATION);
      
      if (!OK)
	{
	  fprintf(stderr,"iterativ solution method FAILED!\n");
	  X = p_gauss_skal_Pivot( maxdiagonal, exact);
	}
      if (!OK)
	{
	  fprintf(stderr,"gauss-elimination - method FAILED!\n");
	  X = p_gauss_voll_Pivot( maxdiagonal);
	}
      if (OK && Nachiter)
	{
	  X = p_iteration( maxdiagonal, iterations, X, exact, NACHITERATION);
	}
      if (!OK)
	{ 
	  fprintf(stderr,"all methods FAILED!\n");
	  exit(1);
	}
      break;

    case DIRECT:
      OK = FALSE;
      /* p_speicher_schaetzen(nonzeros); */
      X = p_gauss_skal_Pivot( maxdiagonal, exact);
      if (OK && Nachiter)
        {
	X = p_iteration( maxdiagonal, iterations, X, exact, NACHITERATION);
        }
      break;

    case STABIL:
      OK = FALSE;
      /* p_speicher_schaetzen(nonzeros); */
      X = p_gauss_voll_Pivot( maxdiagonal);
      if (OK && Nachiter)
	X = p_iteration( maxdiagonal, iterations, X , exact, NACHITERATION);
      break;

    case ITERATIV: 
      OK = FALSE;
      if (NULL == (Xstart = (double*) malloc ((int)Dm * sizeof(double))))
	{
	  perror("p_solve_LGS  --  malloc: vector Xstart\n");
	  exit(1);
	}
      for (i = 0; i < Dm; i++)
	Xstart[i] = 1.0/(double)Dm;

      X = p_iteration( maxdiagonal, iterations, Xstart, exact, ITERATION);
      break;
    default:
      fprintf(stderr,"UNKOWN SOLUTIONTYPE: %d\n",sol_type);
      exit(1);
    }


  /* ------------------ Zustandswahrscheinlichkeiten aufsummieren ----- */
  for(i = no_new_marks; i--; )
    {
      pnmark = new_mark[i];
      old_mark[pnmark->old_no].time_spent += X[pnmark->new_no];
    }

}  /* p_solve_LGS */










