/*****************************************************************************/
/* Programm   :            load_dtn.c                                        */
/* File       :            create_dtp.c                                      */
/* Purpose    :	Routinen fuer allgemein verteilte Transitionen               */
/* Date       : April 1997                                                   */
/* Programmer : Heiko Bergmann                                               */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "parsdfunc.h"
#include "netio.h"

#define _DTPDEBUG
#define _PDFDEBUG
#define _INTDEBUG
#define _NORDEBUG
#define MEANDEBUG
/*****************************************************************************/
/*avoid 0^0 -Problem                                                         */
/*****************************************************************************/
double hb_pow(a,b)
     double a,b;
{
  if ((a==0.0) && (b==0.0))
    return (1.0);
  return pow(a,b);
}

/*****************************************************************************/
/*Romberg Integration of Normal Distribution; Numerical Recipes in C p.140   */
/*****************************************************************************/
#define EPS 1.0e-07
#define JMAX 20
#define JMAXP (JMAX+1)
#define K 5
#define NR_END 1
#define FREE_ARG char*
#define PI 3.141592654
#define NEG_INFINITY -999.0

double *vector(long nl, long nh)
{
  double *v;
  
  v=(double *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(double)));
  if (!v) fprintf(stdout,"allocation failure in vector()");
  return v-nl+NR_END;
}

void free_vector(double *v, long nl, long nh)
{
  free((FREE_ARG) (v+nl-NR_END));
}

void polint(double xa[], double ya[], int n, double x, double *y, double *dy)
{
  int i,m,ns=1;
  double den,dif,dift,ho,hp,w;
  double *c,*d;
  dif=fabs(x-xa[1]);
  c=vector(1,n);
  d=vector(1,n);
  for (i=1; i<=n; i++) {
    if ( (dift=fabs(x-xa[i])) < dif) {
      ns=i;
      dif=dift;
    }
    c[i]=ya[i];
    d[i]=ya[i];
  }
  *y=ya[ns--];
  for (m=1;m<n;m++) {
    for (i=1; i<=n-m;i++) {
      ho=xa[i]-x;
      hp=xa[i+m]-x;
      w=c[i+1]-d[i];
      if ( (den=ho-hp) == 0.0) fprintf(stdout,"Error in routine polint");
      den=w/den;
      d[i]=hp*den;
      c[i]=ho*den;
    }
    *y += (*dy=(2*ns < (n-m) ? c[ns+1] : d[ns-1]));
  }
  free_vector(d,1,n);
  free_vector(c,1,n);
}

double normfunc(double x, double sigma, double mu)
{
#ifdef NORDEBUG
      fprintf(stdout,"\nNormalwert(%lg)=%lg",x,1.0/(sigma*sqrt(2.0*PI))*exp(-hb_pow(x-mu,2.0)/(2.0*hb_pow(sigma,2.0))) );
#endif
  return  1.0/(sigma*sqrt(2.0*PI))*exp(-hb_pow(x-mu,2.0)/(2.0*hb_pow(sigma,2.0))); 
}

double trapzd(double a, double b, int n, double sigma, double mu)
{
  double x,tnm,sum,del;
  static double s;
  int it,j;
  
  if (n == 1) {
    return (s=0.5*(b-a)*(normfunc(a,sigma,mu)+normfunc(b,sigma,mu)));
  } else {
    for (it=1, j=1; j<n-1; j++) it <<= 1;
    tnm=it;
    del=(b-a)/tnm;
    x=a+0.5*del;
    for (sum=0.0,j=1;j<=it;j++,x+=del) sum += normfunc(x,sigma,mu);
    s=0.5*(s+(b-a)*sum/tnm);
    return s;
  }
}

double qromb( double a, double b , double sig, double m)
{
  double ss,dss;
  double s[JMAXP+1],h[JMAXP+1];
  int j;
  
  h[1]=1.0;
  for (j=1;j<=JMAX;j++) {
    s[j]=trapzd(a,b,j,sig,m);
#ifdef INTDEBUG
      fprintf(stdout,"\ns(%d)=%lg",j,s[j] );
#endif
    if (j>= K) {
      polint(&h[j-K],&s[j-K],K,0.0,&ss,&dss);
     if (fabs(dss) <=EPS*fabs(ss)) return ss;
    }
    s[j+1]=s[j];
    h[j+1]=0.25*h[j];
  }
  fprintf(stdout,"To many steps in Integration routine");
  exit(1);
  return 0.0;
}

/*****************************************************************************/
/*Funktionswert der pmf(t)                                                   */
/*Uebergeben wird Exponential Struktur und Zeit, Omega(Phase/timestep)       */
/*****************************************************************************/
double pmf(gexp,pt,omega,pdf_old)
time_l *gexp;
double pt, omega;
double *pdf_old;
{ 
  time_l *expoly;
  power_m *power;
  expo_n *expo;
  double a,b,t,pdf_new,prob;
  unsigned p,e;

  expoly = gexp;
  /* tp*omega reelle Zeit in der Verteilung */
  t = pt*omega;
  if (expoly->type == UNIFORM) {
    while( ((t > expoly->x2) || 
	    ((t == expoly->x2)&&(expoly->power->expo->c == 0.0))  )&&
	   (expoly->next) )
      expoly = expoly->next;

    if (expoly->power->expo->c != 0.0) {
          a = expoly->x1/omega;
	  b = expoly->x2/omega;
	  prob = 1/(b-a+1);      
	  pdf_new = *pdf_old - prob; 
    }
    else 
      {
	prob = 0.0;
	pdf_new = *pdf_old;
      }
  }
  else  
    if (expoly->type == WEIBULL) {
      while( ((t > expoly->x2) || 
	      ((t == expoly->x2)&&(expoly->mu == 0.0))  )&&
	     (expoly->next) )
	expoly = expoly->next;
      
      if (expoly->mu != 0.0) { /*definition intervall*/
	/*F(t)= 1-exp[-(t/sigma)^theta], but here 1-F(x)*/
	pdf_new = exp(-hb_pow(t/expoly->sigma,expoly->mu));
	/*diskrete pmf-value*/
	prob = *pdf_old - pdf_new; 
#ifdef PDFDEBUG
      fprintf(stdout,"\nWeibull def:\n1-cdf(t=%lg)=%lg",t,pdf_new);
#endif

      }
      else
      {
	prob = 0.0;
	pdf_new = *pdf_old;
      }
    }
  else
    if (expoly->type == NORMAL) {
      while( ((t > expoly->x2) || 
	      ((t == expoly->x2)&&(expoly->sigma == 0.0))  )&&
	     (expoly->next) )
	expoly = expoly->next;
      
      if (expoly->sigma != 0.0) { /*definition intervall*/
	/*1-cdf(t) ... INT(0..t) f(x)dx */
 	pdf_new = 1.0-qromb(NEG_INFINITY, t, expoly->sigma ,expoly->mu);
 	/*diskrete pmf-value*/
 	prob = *pdf_old - pdf_new; 
#ifdef PDFDEBUG
      fprintf(stdout,"\nNormal def:\n1-cdf(t=%lg)=%lg
                      \npmf(%lg)=%lg",t,pdf_new,t,prob);
#endif

      }
      else 
      {
	prob = 0.0;
	pdf_new = *pdf_old;
      } 
    }
  else 
    if(expoly->type == DETERMINISTIC) {
      if((t == expoly->x1)||(t == expoly->x2)){
	prob = 1.0;
	pdf_new = *pdf_old - prob;
      }
      else 
      {
	prob = 0.0;
	pdf_new = *pdf_old;
      }
    }
    else /* expolynomial distribution */
    {
    while( (t > expoly->x2) && (expoly->next) )
      expoly = expoly->next;
#ifdef PDFDEBUG
      fprintf(stdout,"\nexpolynomial def:\n1-cdf(t=%lg)=0.0",t);
#endif
      pdf_new=0.0;
      for (p = 0, power = expoly->power; power; power = power->next, p++)
	for (e = 0, expo = power->expo; expo; expo = expo->next, e++)
	{
	  pdf_new += (expo->b * hb_pow(t,(double)power->m) * exp(-expo->lambda*t));
#ifdef PDFDEBUG
	  fprintf(stdout,"\n(p:%d e:%d)+ %lg * %lg^%d * e^(-%lg * %lg) ...(1-cdf=%lg so far)",
                p,e,expo->b,t,power->m,expo->lambda,t,pdf_new);
#endif	  
	}
      
      prob = *pdf_old - pdf_new;
#ifdef PDFDEBUG
      fprintf(stdout,"\nprob(t=%lg)=%lg\n",t,prob);
#endif

      if(t == expoly->x2) /*could be an additional pointmass at t */
      {
	prob += expoly->next->p;
	pdf_new = *pdf_old - prob; 
#ifdef PDFDEBUG
	fprintf(stdout,"\npointmass at %lg: %lg\n prob(t=%lg)=%lg\n",t,expoly->next->p,t,prob);
#endif
      }
#ifdef PDFDEBUG
      fprintf(stdout,"\n1-cdf(t=%lg)=%lg\n",t,pdf_new);
#endif
    }
  if((prob > 1.0)||(pdf_new > *pdf_old)) /*something ist wrong*/
  {
#ifdef PDFDEBUG
      fprintf(stdout,"\nTROUBLE prob=%lg \t new_cdf=%lg \t old_cdf=%lg",prob,pdf_new,*pdf_old);
#endif
    return -1.0;
  }
  else{
    *pdf_old = pdf_new;
    return prob;
  }
}

/*****************************************************************************/
/*Generation der dtp und Bestimmung der phasenanzahl,                        */
/*Uebergeben wird Exponential Struktur und die Transition,                   */
/*neue phasenanzahl und dtp stehen in der Transitionstruktur                 */
/*****************************************************************************/
int create_dtp(exp,t,uts)
time_l *exp;
Transition *t;
int uts;
{
  double *dtp_matrix, *new_dtp, *old_dtp;
  int k,new_phases,o,n;
  int c = t->phases;
  time_l *expoly;
  double sum_pmf,i,rest,pmf_val,b,pdf;
  unsigned l;

#ifdef MEANDEBUG
  double mean_val = 0.0;
#endif

  expoly = exp;

  if (expoly->type == UNIFORM)
  { 
    while(expoly->power->expo->c == 0){
      expoly = expoly->next;
      if (!expoly->next){
	fprintf(stdout,"\n No Definition found  ...\n");
	return 0;
      }
    }
    if (expoly->x1*uts < 1.0) { 
      fprintf(stdout,"\nBegin time*underlying timestep of the Definiton interval must be >= 1.0 \n");
      return 0;
    }  
  }
  else c = 1;            /*No phase implementation for other distributions ...*/
  expoly = exp;
  pdf = 1.0;
  /* Check der Time Intervals */
  for (l = 0; expoly; expoly = expoly->next, l++) 
  {
    if ((expoly->x1*uts < 1.0)&&(l>0)) { 
      fprintf(stdout,"\nBegin time*underlying timestep of the %d.interval must be >= 1.0 \n",l);
      return 0;
    }
    if(expoly->next) {
      if ((modf(expoly->x1*uts,&i) != 0.0 ) || 
	  (modf(expoly->x2*uts,&i) != 0.0)) {
	fprintf(stdout,"\nStart/End times*underlying timestep of the %d.interval must be Cardinals \n",l);
	return 0;
      }
      /* Check ob Phasen/Omega moeglich */
      if ((modf((expoly->x1*uts)/c,&i)) || (modf((expoly->x2*uts)/c,&i))) {
	fprintf(stdout,"\n Phasen(%d) of Transition %s does not fit 
                \na*uts/Phasen and b*uts/ Phasen must be Cardinals\n",c, t->name);
	return 0;
      }
#ifdef DTPDEBUG
      printf("\nCheck of Interval %d:%lg...%lg",l,expoly->x1,expoly->x2);
#endif
    }
    else /*last Interval*/ {      
      if (modf(expoly->x1*uts,&i) != 0.0 ) {
	fprintf(stdout,"\nTimes*underlying timestep of the %d.interval must be Cardinals \n",l);
	return 0;
      }
      /* Check ob Phasen/Omega moeglich */
      if (modf((expoly->x1*uts)/c,&i)) {
	fprintf(stdout,"\n Phasen(%d) of Transition %s does not fit 
                \na*uts/Phasen must be Cardinal\n",c, t->name);
	return 0;
      }
#ifdef DTPDEBUG
      printf("\nCheck of last interval %d:%lg...infinity",l,expoly->x1);
      printf("\n Number of Phases:expoly->x1*uts = %lg\n",expoly->x1*uts);
#endif
      break;
    } 
  }
  /*Anzahl der Phasen gleich Endwert des Intervalls * timestep + 1 */
  /*Groesste Phase Endwert des Intervalls * timestep*/
  t->phases = (int)(expoly->x1*uts);
  dtp_matrix = (double  *)calloc(t->phases+1,sizeof(double));
  sum_pmf = 0.0;
  b = expoly->x1*uts/c;
  for( k=t->phases; k>0; k--) {
    rest = modf(b - (double)(k-1)/c, &i);
#ifdef DTPDEBUG
    printf("\nk:%d, rest:%lg, i:%lg\n",k,rest,i);
#endif
    if ((i >= 0.0) && (rest == 0.0)) {
      pmf_val = pmf(exp,i,(double)(c)/uts,&pdf);
      if (pmf_val == -1.0) {
	fprintf(stdout,"\n Trouble with pmf definition... \n");
	return 0;
      }
      /*last phase, special case for truncated distribution */
      if((exp->type == WEIBULL)||(exp->type == NORMAL))
	if(k == 1) pmf_val = 1 - sum_pmf;
      if((1-sum_pmf)>0.0)
	dtp_matrix[k] = (double)(pmf_val/(1-sum_pmf));
      else dtp_matrix[k]=1.0;
#ifdef DTPDEBUG
      printf("\npmf_val:%lg, dtp_matrix[%d]:%lg,omega:%lg,sum_pmf:%lg,1-sum_pmf:%lg \n",pmf_val,k,dtp_matrix[k],(double)(c)/uts,sum_pmf,1-sum_pmf);
#endif
#ifdef MEANDEBUG      
if(dtp_matrix[k] != 0.0)
	mean_val += i*((double)(c)/uts)*pmf_val;
if((k == 1)||(dtp_matrix[k] >= 1.0))
	printf("mean_val:%lg\n",mean_val);
#endif
      /* prob 1.0 should be enough to fire */ 
      if (dtp_matrix[k] > 1.0) dtp_matrix[k] = 1.0;

	
      sum_pmf += pmf_val;
    }
    /*dtp is already created, no prob left, but dtp to big*/
    if((k!=1)&&(dtp_matrix[k]==1.0)) 
    {
      new_phases=(t->phases-k)+1;
      new_dtp =(double  *)calloc(new_phases+1,sizeof(double));
      for(o=t->phases,n=new_phases;n>0;n--,o--)
	new_dtp[n]=dtp_matrix[o];
      t->phases=new_phases;
      old_dtp=dtp_matrix;
      dtp_matrix=new_dtp;
      free(old_dtp);
      break;
    }
  }
  t->dtp_array = dtp_matrix;
  return 1;
}

void print_dtp(t)
Transition *t;
{
  int k;  
  printf("Number of Phases:%d \n",t->phases);
  for( k=t->phases; k >0; k--) 
    if (t->dtp_array[k] != 0.0) printf("Phase:%d FireProb:%.16g\n",k,t->dtp_array[k]);  
/*     printf("Phase:%d FireProb:%.16g\n",k,t->dtp_array[k]); */
}
