/***************************************************************************

 File:		expoly.c

 Purpose:	routines for manipulating expolynomial distributions
		and for computing the factors needed by the
		randomization procedure

 Programmer:	Reinhard German

 History:	Release 1: June 2, 1993	
                Heiko Bergmann, August 1997
                Change for time discrete pmf of uniform transitions, 
		and definition of Weibull and Normal distribution
***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "expoly.h"

int	mem_bit;

/*	emalloc tries to allocate required memory space and
	tests if it was successfully allocated; if not => exit */
char *emalloc(bytes, account)
	int bytes, account;
{
	int *point;

	if (!bytes)
		return NULL;
#ifdef MEMDEBUG
	point = (int *)malloc((unsigned)(bytes + 2*sizeof(int)));
#else
	point = (int *)malloc((unsigned)(bytes));
#endif
	if (point == NULL)
	{	printf("The reduced reachability graph of this DSPN\n");
		printf("cannot be generated because of lacking main memory.\n");
		if (mem_bit)
			printf("Retry on a workstation with more main memory.\n");
		exit(2);
	}

#ifdef MEMDEBUG
	*(point++) = account;
	*(point++) = bytes;
	now[account] += bytes;
	sum = 0;
	for(i = 0; i < NO_ACCOUNTS; i++)
		sum += now[i];
	if (sum > max_sum)
	{	max_sum = sum;
		for (i = 0; i < NO_ACCOUNTS; i++)
			max[i] = now[i];
	}
#endif
	return (char *)point;	
}

/* set deterministic distribution */
time_l *set_det_expoly(tau)
double tau;
{
  time_l *expoly, *return_expoly;
  power_m *power;
  expo_n *expo;

  return_expoly=(time_l *) emalloc(sizeof(time_l));
  expoly = return_expoly;

  if(tau>0)
  {
      expoly->type = DETERMINISTIC;
      expoly->x1 = 0.0;
      expoly->x2 = tau;
      expoly->p = 0.0;
      expoly->power = (power_m *) emalloc(sizeof(power_m));
        power = expoly->power;
        power->m = 0;
        power->expo = (expo_n *) emalloc(sizeof(expo_n));
          expo = power->expo;
          expo->lambda = 0.0;
          expo->b = 0.0; /* adapt: vorher 1.0 */
          expo->c = 0.0;
          expo->next = NULL;
        power->next = NULL;
      expoly->next = (time_l *) emalloc(sizeof(time_l));
      expoly = expoly->next;
  }
  expoly->type = DETERMINISTIC;
  expoly->x1 = tau;
  expoly->x2 = UNENDLICH;
  expoly->p = 1.0;
  expoly->power = (power_m *) emalloc(sizeof(power_m));
    power = expoly->power;
    power->m = 0;
    power->expo = (expo_n *) emalloc(sizeof(expo_n));
      expo = power->expo;
      expo->lambda = 0.0;
      expo->b = 1.0; /* adapt: alt 0.0 */
      expo->c = 0.0;
      expo->next = NULL;
    power->next = NULL;
  expoly->next = NULL;

  return return_expoly;
} /* set_det_expoly */


/* set uniform distribution */
time_l *set_uniform_expoly(t1, t2)
double t1, t2;
{
  time_l *expoly, *return_expoly;
  power_m *power;
  expo_n *expo;

  return_expoly=(time_l *) emalloc(sizeof(time_l));
  expoly = return_expoly;

  if(t1>0)
  {
    expoly->type = UNIFORM;
    expoly->x1 = 0.0; 
    expoly->x2 = t1; 
    expoly->p = 0.0;
    expoly->power = (power_m *) emalloc(sizeof(power_m));
    power = expoly->power;
    power->m = 0;
    power->expo = (expo_n *) emalloc(sizeof(expo_n));
    expo = power->expo; 
      expo->lambda = 0.0;
      expo->b = 0.0; /* adapt: vorher 1.0 */
      expo->c = 0.0;
      expo->next = NULL;
    power->next = NULL;
    expoly->next = (time_l *) emalloc(sizeof(time_l));
    expoly = expoly->next;
  }
  expoly->type = UNIFORM;
  expoly->x1 = t1;
  expoly->x2 = t2;
  expoly->p = 0.0;
  expoly->power = (power_m *) emalloc(sizeof(power_m));
    power = expoly->power;
    power->m = 0;
    power->expo = (expo_n *) emalloc(sizeof(expo_n));
      expo = power->expo;
      expo->lambda = 0.0;
      expo->b = -t1/(t2-t1);  /* adapt: alt: 1+t1/(): */
      expo->c = 1/(t2-t1+1);  /*new: discrete time pmf */ 
      expo->next = NULL;
    power->next = (power_m *) emalloc(sizeof(power_m));
    power = power->next;
    power->m = 1;
    power->expo = (expo_n *) emalloc(sizeof(expo_n));
      expo = power->expo;
      expo->lambda = 0.0;
      expo->b = 1/(t2-t1);    /* adapt: unklar wie b errechnet wird ||| */
      expo->c = 0.0;          /* alt: -1/...*/
      expo->next = NULL;
    power->next = NULL;
  expoly->next = (time_l *) emalloc(sizeof(time_l));
  expoly = expoly->next;

  expoly->type = UNIFORM;
  expoly->x1 = t2;
  expoly->x2 = UNENDLICH;
  expoly->p = 0.0;
  expoly->power = (power_m *) emalloc(sizeof(power_m));
    power = expoly->power;
    power->m = 0;
    power->expo = (expo_n *) emalloc(sizeof(expo_n));
      expo = power->expo;
      expo->lambda = 0.0;
      expo->b = 1.0;/* adapt: alt 0.0 */
      expo->c = 0.0;
      expo->next = NULL;
    power->next = NULL;
  expoly->next = NULL;

  return return_expoly;
} /* set_uniform_expoly */

/* set weibull distribution */
time_l *set_weibull(t1, t2, sigma, theta)
double t1, t2, sigma, theta;
{
  time_l *expoly, *return_expoly;
  power_m *power;
  expo_n *expo;

  return_expoly=(time_l *) emalloc(sizeof(time_l));
  expoly = return_expoly;

  if(t1>0)
  {
    expoly->type = WEIBULL;
    expoly->x1 = 0.0; 
    expoly->x2 = t1; 
    expoly->p = 0.0;
    expoly->sigma = 0.0;
    expoly->mu = 0.0;
    expoly->power = NULL;
    expoly->next = (time_l *) emalloc(sizeof(time_l));
    expoly = expoly->next;
  }
  expoly->type = WEIBULL;
  expoly->x1 = t1;
  expoly->x2 = t2;
  expoly->p = 0.0;
  expoly->sigma = sigma;
  expoly->mu = theta;
  expoly->power =  NULL;
  expoly->next = (time_l *) emalloc(sizeof(time_l));
  expoly = expoly->next;

  expoly->type = WEIBULL;
  expoly->x1 = t2;
  expoly->x2 = UNENDLICH;
  expoly->p = 0.0;    
  expoly->sigma = 0.0;
  expoly->mu = 0.0;
  expoly->power = NULL;
  expoly->next = NULL;

  return return_expoly;
} /* set_weibull */

/* set normal distribution */
time_l *set_normal(t1, t2, sigma, mu)
double t1, t2, sigma, mu;
{
  time_l *expoly, *return_expoly;
  power_m *power;
  expo_n *expo;

  return_expoly=(time_l *) emalloc(sizeof(time_l));
  expoly = return_expoly;

  if(t1>0)
  {
    expoly->type = NORMAL;
    expoly->x1 = 0.0; 
    expoly->x2 = t1; 
    expoly->p = 0.0;
    expoly->sigma = 0.0;
    expoly->mu = 0.0;
    expoly->power = NULL;
    expoly->next = (time_l *) emalloc(sizeof(time_l));
    expoly = expoly->next;
  }
  expoly->type = NORMAL;
  expoly->x1 = t1;
  expoly->x2 = t2;
  expoly->p = 0.0;
  expoly->sigma = sigma;
  expoly->mu = mu;
  expoly->power =  NULL;
  expoly->next = (time_l *) emalloc(sizeof(time_l));
  expoly = expoly->next;

  expoly->type = NORMAL;
  expoly->x1 = t2;
  expoly->x2 = UNENDLICH;
  expoly->p = 0.0;    
  expoly->sigma = 0.0;
  expoly->mu = 0.0;
  expoly->power = NULL;
  expoly->next = NULL;

  return return_expoly;
} /* set_normal */


/* print an expolynomial distribution: */
void print_expoly(expoly)
  time_l *expoly;
{
  power_m *power;
  expo_n *expo;
  unsigned l;

  for (l = 0; expoly; expoly = expoly->next, l++) 
  {
    printf("---------------------------------------------------------------\n"); 
    printf("interval %d: ",l);
    switch( expoly->type)
    {
      case DETERMINISTIC :
	printf("\t type: DETERMINISTIC ");
	break;
      case UNIFORM :
	printf("\t type: UNIFORM ");
	break;
      case WEIBULL :
	printf("\t type: WEIBULL ");
	break;
      case NORMAL :
	printf("\t type: NORMAL ");
	break;      
      case EXPOLY :
	printf("\t type: EXPOLY ");
	break;
      default:
	printf("\t type: EXPOLY ");
	break;
    }
    printf("\tx1=%.16g, ", expoly->x1);
    if (expoly->next)
      printf("x2=%.16g, ", expoly->x2);
    else 
      printf("x2=infinity, ");
    if (expoly->type == WEIBULL)
    {
    printf("sigma=%.16g\n", expoly->sigma);
    printf("theta=%.16g\n", expoly->mu);
    printf("p=%.16g\n", expoly->p);
    }  
    else if (expoly->type == NORMAL)
    {
    printf("sigma=%.16g\n", expoly->sigma);
    printf("mu=%.16g\n", expoly->mu);
    printf("p=%.16g\n", expoly->p);
    }  
    else
    printf("p=%.16g\n", expoly->p);
    for (power = expoly->power; power; power = power->next)
    {
      printf("power m=%d:", power->m); 
      for (expo = power->expo; expo; expo = expo->next)
      {
	  if(expo == power->expo)
	      printf("\tlambda=%.16g, b=%.16g, c=%.16g\n",
		     expo->lambda, expo->b, expo->c);
	  else
	      printf("          \tlambda=%.16g, b=%.16g, c=%.16g\n",
		     expo->lambda, expo->b, expo->c);
    
      }
    }
  }
  printf("---------------------------------------------------------------\n"); 
} /* print_expoly */
