/*******************************************************************************
* Program:	genrrg
*
* File:		genrrg_graph.c
*
* Purpose:	Graph generation components.
*
* Programmer:	Robert Zijal
*
* History:	April 1995
* 1.8.96 
* - improved detection of initial vanishing states
* - changed weight handling to meet zero (ie. no) weight-group
* - improved confusion message
* - introduced genrrg_hlp(s,s_ref) to help reduce stack usage (was sort of bug)
* 15.8.96 
* - added (post selection) priorities & m_dep input arc for flushing tokens
* - corrected absorbing state stop at adv_phases
*******************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <alloca.h>
#include <memory.h>
#include "genrrg.h"
#include "genrrg_tree.h"


/*---------------------*/
/* COMPILE-SWITCH-KEYS */
/*---------------------*/

#define  _VPDEBUG 
#define  _ECDEBUG 
#define  _FCDEBUG 
#define  _ADVDEBUG 
#define  _CHDEBUG 
#define  _FINEDEBUG 
#define  _RRGDEBUG 

/*
#define _VPDEBUG 
#define ECDEBUG 
#define FCDEBUG 
#define ADVDEBUG 
#define CHDEBUG 
#define _FINEDEBUG 
#define _RRGDEBUG 
*/

/*-----------*/
/* CONSTANTS */
/*-----------*/

#define	MAX_T_COUNT	0xFF	/* max no of van. trans-firings */
#define NO_TRANS	-1 	/* no transition fired */
#define DUMMY_TRANS	0xFF	/*(255); max. legal indices are: 0..FE(254) */
				/* => max. num. of ind. =  FF(255)=MAX_TRANS */

/*-------*/
/* TYPES */
/*-------*/

/* ----------------------- Confusion check ---------------------------------*/
typedef struct confusion{	/* used for confusion message */
  Bool	trans,			/* diff. transition firing counts */
	prob,			/* diff. sum of path prob. to one tang. state */
	state;			/* diff. reachable states */
} Confusion;


/* ---------------------- misc ---------------------------------------------*/
typedef struct ectrans{ 	/* enabled & candidate transitions */
     byte  num_en,		
           *enabled,	      
           num_cand,		
           *cand;	      
} ECtrans;		


/*-----------*/
/* VARIABLES */
/*-----------*/

ECtrans *tmp_ec, *hlp_cnd;
int DEBUG_level = 0;
/* global for act_marking() */
word *act_m_vector;


/*----------- make dumb proof: md_arc_mult(),  md_rate() --------------------*/
/* from <model>_MDF.c */
extern int md_arc_mult(/*int id*/);
extern double  md_rate(/*int id*/);

int Smd_arc_mult(id)
int id;
{
 int arc_mult;

 arc_mult = md_arc_mult(id);

 if (arc_mult < 0) {
  fprintf(stdout, 
               "\n\nERROR: md-arc (id %d) multiplicity = %d.\n", id, arc_mult);
  clean_up();
  dtnet->error_func("Computed negative marking dependent arc multiplicity.\n",TRUE);
 }

 return (arc_mult);
}


double Smd_rate(id)
int id;
{
 double rate;

 rate = md_rate(id+1); /* md-delay-id starts at 1 ! */

 if (rate <= 0.0) {
  fprintf(stdout, 
               "\n\nERROR: md-delay (id %d) = %g.\n", id, rate);
  clean_up();
  dtnet->error_func("Computed illegal marking dependent firing delay.\n",TRUE);
 }

 return (rate);
}

/*---------------------- extern for <model>_MDF.c ---------------------------*/
int act_marking(pla)
int pla;
{
 if ((pla<0) || (pla >= dtnet->num_places)) {
  fprintf(stdout, 
               "\n\nERROR in act_marking(): place id >%d< is not in range.\n", pla);
  clean_up();
  dtnet->error_func("Tried to access illegal place index for MD-function.\n",TRUE);
 }

 return ((int)(act_m_vector[dtnet->place[pla].r_index]));
}


/*---------------------- misc functions -------------------------------------*/
void dbspc() 
{
 int i;
 for (i=0; i < DEBUG_level; i++) printf("  ");
}

void init_graph()
{
 /* init temporary var array for enabled&cand. trans. */
 tmp_ec = (ECtrans *)nMalloc(sizeof(ECtrans));
 tmp_ec->enabled  = 
  (byte *)nMalloc( 2 * (dtnet->num_trans) * sizeof(byte));
 tmp_ec->cand  = tmp_ec->enabled + dtnet->num_trans;
 tmp_ec->num_en = tmp_ec->num_cand = 0; 

 /* init help var for cand. trans. determination */
 hlp_cnd = (ECtrans *)nMalloc(sizeof(ECtrans));
 hlp_cnd->enabled  = NULL; /* not needed here */
 hlp_cnd->cand  = 
  (byte *)nMalloc( (dtnet->num_trans) * sizeof(byte));
 hlp_cnd->num_en = hlp_cnd->num_cand = 0; 
}


void set_tmp_EC(s)
State *s;
{
 byte t, i, j;
 Transition *trans, *trans_j;
 int a;
 Bool enabled, candidate;
 
 tmp_ec->num_en = tmp_ec->num_cand = hlp_cnd->num_cand = 0; 
 
 Copy_m(s, act_m_vector);     /* global for act_marking() */

 /* find enabled transitions */
 for(t=0; t < dtnet->num_trans; t++) {
   trans = dtnet->t_ranks[t];
   /* check input cardinality */
   for(enabled=TRUE, a=0; (enabled==TRUE) && (a < trans->num_inarcs); a++) {
     if (!trans->in_arc[a].is_value) { /* marking dependent */
       if (s->m[(trans->in_arc[a].r_index)] < Smd_arc_mult(trans->in_arc[a].md_id))
	 enabled = FALSE;
     }
     else if (s->m[(trans->in_arc[a].r_index)] <  (trans->in_arc[a].multipl.value))
       enabled = FALSE;
   };
   /* check  inhibition cardinality */
   for(a=0; (enabled==TRUE) && (a < trans->num_inharcs); a++) {
     if (!trans->inh_arc[a].is_value) { /* marking dependent */
       if (
	   ( s->m[(trans->inh_arc[a].r_index)] >= Smd_arc_mult(trans->inh_arc[a].md_id) )
	   && ( Smd_arc_mult(trans->inh_arc[a].md_id) != 0 )
       )
	 enabled = FALSE;
     }
     else if (s->m[(trans->inh_arc[a].r_index)] >=  (trans->inh_arc[a].multipl.value))
       enabled = FALSE;
   };

   if(enabled) {
     tmp_ec->enabled[ (tmp_ec->num_en)++ ] = t; 

     /* check first candidate condition (RFT=0) */
     if ((t>=dtnet->num_timed_trans) || (s->ph[t]==0)) 
       hlp_cnd->cand[ (hlp_cnd->num_cand)++ ] = t;
   }
   else 
     /* apply RA/RE race policies to timed trans. */
     if (t<dtnet->num_timed_trans) {
       if ((trans->f_policy==RE) || (s->ph[t]==0))
       /* reset phases of not enabled RE-trans. */
       /* 0-phase RA-trans. are also reset, because they are not tangible */
       s->ph[t]=B_ph;
     }
 }


 /* find final candidate transitions */
 for(i=0; i < hlp_cnd->num_cand; i++) {

   trans = dtnet->t_ranks[ (hlp_cnd->cand[i]) ];
   candidate = TRUE;

   /* any other candidates with higher (post-selection) priority ? */
   /* priority <= 1 means no priority defined */
   if (trans->priority > 1) {
     for(j=0; ((candidate==TRUE) && (j < tmp_ec->num_cand)); j++) {
         trans_j = dtnet->t_ranks[ (tmp_ec->cand[j]) ];
	 if (trans->priority < trans_j->priority )
	   candidate = FALSE;
     }

     for(j=i+1; ((candidate==TRUE) && (j < hlp_cnd->num_cand)); j++) {
         trans_j = dtnet->t_ranks[ (hlp_cnd->cand[j]) ];
	 if (trans->priority < trans_j->priority )
	   candidate = FALSE;
     }
   }

   if(candidate) 
     tmp_ec->cand[ (tmp_ec->num_cand)++ ] = (hlp_cnd->cand[i]);
   }

}/*end set_tmp_EC(s)*/


/* Get enabled&candidates on callers stack frame */
#define Get_ST_EC(ec, s) \
{ \
 set_tmp_EC(s); \
 \
 tmp_size = sizeof(ECtrans); \
 if( (ec = (ECtrans *)alloca(tmp_size)) == NULL ){ \
 fprintf(stdout,"Get_ST_EC_a: Unable to alloca %u bytes on Stack\n", tmp_size);\
   exit( -1 ); \
 }; \
 ec->num_en = tmp_ec->num_en; ec->num_cand = tmp_ec->num_cand; \
 \
 tmp_size = (ec->num_en + ec->num_cand) * sizeof(byte); \
 if (tmp_size > 0) { \
  if( (ec->enabled = (byte *)alloca(tmp_size)) == NULL ){ \
  fprintf(stdout,"Get_ST_EC_b: Unable to alloca %u bytes on Stack\n", tmp_size);\
   exit( -1 ); \
  }; \
  memcpy(ec->enabled, tmp_ec->enabled, ec->num_en * sizeof(byte)); \
  if (ec->num_cand > 0) {\
    ec->cand = ec->enabled + ec->num_en; \
    memcpy(ec->cand, tmp_ec->cand, ec->num_cand * sizeof(byte)); \
 }} \
}



#ifdef ECDEBUG 
void dump_ST_EC(ec, s)   
State *s;
ECtrans *ec;
{
 int i;
  printf(" Dumping ST_EC of "); print_s(stdout, s); 
  printf("  %d Enbl:", ec->num_en);
  for (i = 0; i < ec->num_en; i++) 
    printf(" %s", dtnet->t_ranks[ec->enabled[i]]->name); 
  printf("\n  %d Cand:", ec->num_cand);
  for (i = 0; i < ec->num_cand; i++) 
    printf(" %s", dtnet->t_ranks[ec->cand[i]]->name); nl(stdout);
}
#endif 


/*------------------ Confusion check functions ------------------------------*/
Van_paths *init_van_paths(s)
State_ref *s;
{
 Van_paths *vp;
 vp = (Van_paths *)nMalloc(sizeof(Van_paths));

 vp->to_s = s;  
 vp->next = NULL;

 vp->mult_arcs = (Van_arcs *)nMalloc(sizeof(Van_arcs));  

 vp->mult_arcs->prob  = 1.0; 
 vp->mult_arcs->next_arc  = NULL;

 vp->mult_arcs->count = (byte *)nMalloc(dtnet->num_trans * sizeof(byte));  
  /* nMalloc sets all bytes to 0 */

 return vp;
}		    


void Free_vp(vp)
Van_paths *vp;
{
 Van_paths *tmp;
 Van_arcs  *va, *nxt;

 while(vp!=NULL){
   for(va=vp->mult_arcs; va!=NULL; va = nxt){
     nxt = va->next_arc;
     Free(va->count);
     Free(va);
   }
   tmp = vp;
   vp = vp->next;
   Free(tmp);
 }
}


void dump_vp(vp)
Van_paths *vp;
{
 Van_arcs  *va;
 int t;

 printf("Dumping Vanishing Paths\n");
 while(vp!=NULL){
   printf(" To s(%ld): [",vp->to_s->index);
   for(t=0; t<dtnet->num_trans; t++) printf(" %s,", dtnet->t_ranks[t]->name);
   printf(" ]\n");
   for(va=vp->mult_arcs; va!=NULL; va = va->next_arc) {
     printf("   arc: ");
     for(t=0; t<dtnet->num_trans; t++) printf("%d ", (int)va->count[t]);
     printf("(prob=%g)\n",va->prob);
   }
   vp = vp->next;
 }
}



void add_van_arcs(subva, va)
Van_arcs *subva, **va;
{
 Van_arcs  *tmp;
 int cnt_diff;


 while (subva != NULL) {
   /* find equal van. arcs (count-vectors) */
   do {
     if (*va == NULL) cnt_diff = -1;
     else {
       cnt_diff = memcmp(subva->count, (*va)->count, dtnet->num_trans);
       if (cnt_diff>0) va = &( (*va)->next_arc); 
     }
   } while (cnt_diff>0);
   if (cnt_diff==0) {	/* add count-vec firing probability */
     (*va)->prob += subva->prob; 
     va = &( (*va)->next_arc);
     tmp = subva;
     subva = subva->next_arc;
     Free(tmp->count);
     Free(tmp);	/* not needed anymore */
   }
   else {		/* prepend new van. arc */
     tmp = *va;
     *va = subva;
     if   (tmp==NULL) subva = NULL; /* append the rest */
     else {
       subva = subva->next_arc;
       (*va)->next_arc = tmp;
       va = &( (*va)->next_arc);
     }
   }
 } /*end while(subva...)*/		    
} /*end add_van_arcs*/             



void add_van_paths(t, prob, subvp, vp)
int t;
double *prob; /* pointer for effiency */
Van_paths *subvp, **vp;
{
 Van_arcs  *arc;
 Van_paths  *tmp;
 long int inx_diff;

#ifdef VPDEBUG 
 Van_paths  **first = vp;
 dbspc(); printf(" SUB_VPpre:"); dump_vp(subvp);
#endif

 for (tmp = subvp; tmp != NULL; tmp = tmp->next) {
   /* inc. trans. firing count add firing probability to sub_van. path */
   for(arc = tmp->mult_arcs; arc != NULL; arc = arc->next_arc) {
     if (t != NO_TRANS) {
       if (arc->count[t]== MAX_T_COUNT) {
         fprintf(stdout,"ERROR: in trans %d. %s.\n",t,dtnet->t_ranks[t]->name);
	 clean_up();
         dtnet->error_func(
         "Max num. of firing counts (255) in vanishing graph exceeded.\n",TRUE);
       }
       else (arc->count[t])++;  
     }
     arc->prob *= *prob;
   } 
 }

#ifdef VPDEBUG 
 dbspc(); printf(" SUB_VPpost:"); dump_vp(subvp);
 dbspc(); printf(" VP_pre:"); dump_vp(*vp);
#endif

 while (subvp != NULL) {
   /* find van. paths to equal tang. states */
   do {
     if (*vp == NULL) inx_diff = -1;
     else {
       inx_diff = subvp->to_s->index - (*vp)->to_s->index;
       if (inx_diff>0) vp = &( (*vp)->next); 
     }
   } while (inx_diff>0);
   if (inx_diff==0) {	/* merge equal van. paths, append new ones */
     add_van_arcs(subvp->mult_arcs, &((*vp)->mult_arcs));
     vp = &( (*vp)->next);
     tmp = subvp;
     subvp = subvp->next;
     Free(tmp);	/* not needed anymore */
     /* !!! Don't free arcs. They are reused or freed in add_van_arcs */
   }
   else {		/* prepend new van. paths to new tang. state */
     tmp = *vp;
     *vp = subvp;
     if   (tmp==NULL) subvp = NULL; /* append the rest */
     else {
       subvp = subvp->next;
       (*vp)->next = tmp;
       vp = &( (*vp)->next);
     }
   }
 } /*end while(subvp...)*/		    

#ifdef VPDEBUG 
 dbspc(); printf(" VP_post:"); dump_vp(*first); printf("\n");
#endif
} /*end add_van_paths*/             

 
Confusion comp_vp(vp_a, vp_b)
Van_paths *vp_a, *vp_b;
{
 Confusion conf;
 Van_arcs *va_a, *va_b;
 double sum_a, sum_b;

 conf.trans = conf.prob = conf.state = FALSE;

 while (!conf.state){ 
   if ( (vp_a==NULL) || (vp_b==NULL) ) {
     if (vp_a != vp_b)  conf.state = TRUE; 
     break;
   }
   if ( (vp_a->to_s != vp_b->to_s) ) 
     conf.state = TRUE;
   else {	/* paths go to same tang. state */
     va_a = vp_a->mult_arcs; va_b = vp_b->mult_arcs;
     sum_a = sum_b = 0;

     while (!conf.prob || !conf.trans) {
       if ( (va_a==NULL) || (va_b==NULL) ) {
	 if (va_a != va_b)  conf.trans = TRUE; 
	 else {
	   if (sum_a != sum_b) conf.prob = TRUE;
	   break;
	 }
       }
      if (!conf.trans && 
	  (memcmp(va_a->count, va_b->count, dtnet->num_trans) != 0)
	 )
	conf.trans = TRUE;
      if (!conf.prob) {
	if (va_a != NULL) sum_a += va_a->prob; 
	if (va_b != NULL) sum_b += va_b->prob;
      }
      if (va_a != NULL) va_a = va_a->next_arc;
      if (va_b != NULL) va_b = va_b->next_arc;
     }/*while*/

     vp_a = vp_a->next; vp_b = vp_b->next;  
   }/*else*/

 }/*end while*/ 

 return conf;
}



/*----------------------------------------------------------------------------*/

void check_state();

void adv_X_phases(z, z_x, vp, ec, i_en, x_prob, step_size)
State *z, *z_x;
Van_paths **vp;
ECtrans *ec;
int i_en;
double x_prob;
word step_size; 
{
 double total_prob, fire_prob;
 int t,act_phase;
 Bool FINISHED = FALSE, FIRST_LOOP = TRUE;

 t = (int) ec->enabled[i_en];

#ifdef ADVDEBUG 
 dbspc(); printf(" Advancing transition %s (phase: %d Stepsize: %d Prob: %lg) \n", 
	dtnet->t_ranks[t]->name, z_x->ph[t], step_size, x_prob); 
#endif 

 if (t >= dtnet->num_timed_trans) {
  fprintf(stdout,"ERROR: Inconsistency in adv_X_phases."); clean_up();
  dtnet->error_func("Tried to advance immediate transition.\n",TRUE);
 }

 z_x->ph[t] = z->ph[t];

 if (z_x->ph[t] == 0) {
  fprintf(stdout,"ERROR: Inconsistency in adv_X_phases."); clean_up();
  dtnet->error_func("Tried to advance 0-phase.\n",TRUE);
 }

 do {
   switch( dtnet->t_ranks[t]->type ) {
     case DET:    
       if (z_x->ph[t] == B_ph)
	 z_x->ph[t] = dtnet->t_ranks[t]->phases;
       else 
	 z_x->ph[t] = z_x->ph[t] - step_size;
       total_prob = x_prob; /* prob = 1 */
       FINISHED = TRUE;
       break;      
     case EXP:    
       if (z_x->ph[t] == B_ph) {
	 z_x->ph[t] = dtnet->t_ranks[t]->phases;
	 total_prob = x_prob;  /* prob = 1 */
	 FINISHED = TRUE; }
       else if (z_x->ph[t] > step_size) {
	 z_x->ph[t] = z_x->ph[t] - step_size;
	 total_prob = x_prob; /* prob = 1 */
	 FINISHED = TRUE; }
       else if (z_x->ph[t] == step_size) { /* fire */
	 z_x->ph[t] = z_x->ph[t] - step_size;
	 if ( dtnet->t_ranks[t]->prob == 0.0 ) { /* marking dependent */
	   /* global for act_marking() */
	   Copy_m(z, act_m_vector);

	   fire_prob = ((double) dtnet->t_ranks[t]->phases)/ 
	               ( Smd_rate(dtnet->t_ranks[t]->index) * ((double)but_step) );

	   if (fire_prob >= 1) 
	     geom_usage(dtnet, dtnet->t_ranks[t], Smd_rate(dtnet->t_ranks[t]->index));

	   total_prob = x_prob * fire_prob;
	   if (fire_prob == 1.0) /* exceptional case, equiv. to DET */
	     FINISHED = TRUE;
	 }
	 else {
	   total_prob = x_prob * dtnet->t_ranks[t]->prob;
	   if (dtnet->t_ranks[t]->prob == 1.0) /* exceptional case, equiv. to DET */
	     FINISHED = TRUE;
	 }
       }
       else if (z_x->ph[t] == 0) { /* not fire */
	 /* second loop from above */
	 z_x->ph[t] = dtnet->t_ranks[t]->phases;
	 if ( dtnet->t_ranks[t]->prob == 0.0 )   /* marking dependent */
	   total_prob = x_prob * (1.0 - fire_prob);
	 else
	   total_prob = x_prob * (1.0 - dtnet->t_ranks[t]->prob); 
#ifdef FCDEBUG 
	 dbspc(); printf(" NO FIRE transition %s\n", dtnet->t_ranks[t]->name); 
#endif 
	 FINISHED = TRUE;
       }
       else 
	 dtnet->error_func("Illegal phase to be advanced.\n",TRUE);
       break;
     case GEN:
       if (z_x->ph[t] == B_ph) {
	 z_x->ph[t] = dtnet->t_ranks[t]->phases;
	 total_prob = x_prob;  /* prob = 1 */
	 FINISHED = TRUE; }
       else {	 
	 if ((FIRST_LOOP) && (dtnet->t_ranks[t]->dtp_array[z_x->ph[t] - step_size + 1] == 0.0)) { 	 
	   z_x->ph[t] = z_x->ph[t] - step_size;
	   total_prob = x_prob;  /* prob = 1 */
	   FINISHED = TRUE;
	   }
	 else 
	   if (z_x->ph[t] != 0) {/* fire */
	     fire_prob = dtnet->t_ranks[t]->dtp_array[z_x->ph[t] - step_size + 1];
	     act_phase = z_x->ph[t];
	     FIRST_LOOP = FALSE;
	     z_x->ph[t] = 0;
	     total_prob = x_prob*fire_prob;
	     if (fire_prob == 1.0 ) FINISHED = TRUE;
	   }
	   else {/* not fire */
	     /* second loop from above */	     
	     total_prob = x_prob*(1.0 - fire_prob);
	     z_x->ph[t] = act_phase - step_size;
#ifdef FCDEBUG 
	 dbspc(); printf(" NO FIRE transition %s\n", dtnet->t_ranks[t]->name); 
#endif 
	     FINISHED = TRUE;
	   }
       }
       break;
   default:
     dtnet->error_func("Unknown transition-type to be advanced.\n",TRUE);
     break;      
   }

   if ((i_en+1) < ec->num_en) {
     adv_X_phases(z, z_x, vp, ec, (i_en+1), total_prob, step_size);
   }
   else {
#ifdef ADVDEBUG 
     dbspc(); printf(" All phases advanced \n"); 
#endif 
     check_state(&total_prob, z_x, NO_TRANS, vp);
   }

 } while (!FINISHED);

}/*end adv_X_phases*/ 


void adv_phases(z, vp, ec, ALL_phases, stepsize)
State *z;
Van_paths **vp;
ECtrans *ec;
Bool ALL_phases;
word stepsize;
{
 State *z_x;
 int en_cnt, adv_cnt;
 DEBUG_level++;

 New_St_s(z_x);
 Copy_s(z, z_x);

 if (!ALL_phases) {
   /* only B-phases */
   /* reuse ec->enabled for reduction of # of recursive calls */
   for (en_cnt=adv_cnt=0; en_cnt < ec->num_en; en_cnt++)
     if ((ec->enabled[en_cnt] < dtnet->num_timed_trans) && 
         (z->ph[ec->enabled[en_cnt]] == B_ph))
   		ec->enabled[adv_cnt++] = ec->enabled[en_cnt];
   ec->num_en = adv_cnt;
 }

 if (ec->num_en == 0) {
  fprintf(stdout,"ERROR: in ");  print_s(stdout, z); clean_up();
  dtnet->error_func("Absorbing state.\n",TRUE); 
 }

 *vp = NULL; 

 adv_X_phases(z, z_x, vp, ec, 0, 1.0, stepsize);

 DEBUG_level--; 
}/*end adv_phases*/ 


void fire_C(z, vp, ec)
State *z;
Van_paths **vp;
ECtrans *ec;
{
 State *z_f;
 Van_paths *vp_G;
 int  G_cand, f_cand, G_t, G_t_old, group, t_f, a, p;
 int i, tmp;
 Transition *trans;
 double w_G_sum, w_prob;
 Confusion conf;
 DEBUG_level++;
 
 *vp   = NULL; 

 New_St_s(z_f);

 /* loop over all weight-groups */
 for (G_cand=0; G_cand < ec->num_cand; G_cand++) {
   G_t = (int) ec->cand[G_cand];
   if (G_t == DUMMY_TRANS)
     continue;	 /* ignore already fired groups of transitions */

   /* first group-transition represents whole group */
   group = dtnet->t_ranks[G_t]->group;
   vp_G = NULL;


   /* loop over all candidate transitions of group G, calc. weight-sum */ 
   w_G_sum = 0.0;
   f_cand=G_cand;
   do {
     t_f = (int) ec->cand[f_cand];
     if ((t_f != DUMMY_TRANS) && (dtnet->t_ranks[t_f]->group == group))
       w_G_sum += dtnet->t_ranks[t_f]->group_weight; 
   } while ( (group!=DEFAULT_GRP) && (++f_cand < ec->num_cand) );


   /* loop over all candidate transitions of group G, fire group-candidates */ 
   f_cand=G_cand;
   do {
     t_f = (int) ec->cand[f_cand]; 
     if ((t_f == DUMMY_TRANS) || (dtnet->t_ranks[t_f]->group != group))
       continue; /* ignore already fired and different groups of transitions */
     else 
       ec->cand[f_cand] = DUMMY_TRANS; /* mark transition to fire */

     trans = dtnet->t_ranks[t_f];

     /* fire single candidate */
     Copy_s(z, z_f);
     /* global for act_marking() */
     Copy_m(z, act_m_vector);

        /* remove tokens */
     for(a=0; a < trans->num_inarcs; a++) {
       p = trans->in_arc[a].r_index;
       if (!trans->in_arc[a].is_value) /* marking dependent */
	 z_f->m[p] = z_f->m[p] - Smd_arc_mult(trans->in_arc[a].md_id);
       else 
	 z_f->m[p] = z_f->m[p] - trans->in_arc[a].multipl.value;
     };

     /* add tokens */
     for(a=0; a < trans->num_outarcs; a++) {
       p = trans->out_arc[a].r_index;
       if (!trans->out_arc[a].is_value)  /* marking dependent */
	 tmp = ((int)(z_f->m[p])) + Smd_arc_mult(trans->out_arc[a].md_id);
       else 
         tmp = ((int)(z_f->m[p])) + trans->out_arc[a].multipl.value;

       if (tmp > MAX_TOKEN) {
          fprintf(stdout,"ERROR: firing transition %s: %d > %d\n", 
	      trans->name, tmp , MAX_TOKEN);  clean_up();
          dtnet->error_func("Max. no. of tokens exceeded.\n",TRUE);
       }

       z_f->m[p] = (word)(tmp);
     };

     if (t_f < dtnet->num_timed_trans)
       z_f->ph[t_f] = B_ph; /* reset firing time distribution */
     
     /* HERE is the place for different race policies */

     /* DEFAULT race policy: enabling memory  */ 
     /* not enabled phases are reset to B_ph at set_tmp_EC(s) in check_state()*/
 
     w_prob = trans->group_weight / w_G_sum;

#ifdef FCDEBUG 
     dbspc(); printf(" FIRE transition %s with prob %g\n", trans->name,w_prob); 
#endif 

     check_state(&w_prob, z_f, t_f, &vp_G);

   } while ( (group!=DEFAULT_GRP) && (++f_cand < ec->num_cand) );
   /*end for all trans element of group G*/




   /* CONFUSION - check */
   if (*vp == NULL) {/* first group-transition loop */
     *vp   = vp_G; G_t_old = G_t;
   }
   else {
     conf = comp_vp(*vp, vp_G);
     if (conf.trans || conf.prob || conf.state) {
       fprintf(stdout, "\nDETECTED: Conflict or Confusion leading to ");

       if (conf.trans) fprintf(stdout, 
		"impuls reward (throughput) - confusion,\n");
       if (conf.prob)  fprintf(stdout, 
		"state transition probability - confusion,\n");
       if (conf.state) fprintf(stdout, 
		"tangible state reachability - confusion,\n");

       fprintf(stdout, "between ");

       if (dtnet->t_ranks[G_t_old]->group != DEFAULT_GRP) { 
	 fprintf(stdout, "group %d: {", dtnet->t_ranks[G_t_old]->group);
         for(i=0; i<dtnet->num_trans; i++)
	   if (dtnet->trans[i].r_index == G_t_old)
	     fprintf(stdout, " (%s)", dtnet->trans[i].name);	 
     	   else if (dtnet->t_ranks[G_t_old]->group == dtnet->trans[i].group)
	     fprintf(stdout, " %s", dtnet->trans[i].name);	 
       }
       else
	 fprintf(stdout, "{ (%s)", dtnet->t_ranks[G_t_old]->name);       

       fprintf(stdout, " } and ");

       if (dtnet->t_ranks[G_t]->group != DEFAULT_GRP) { 
	 fprintf(stdout, "group %d: {", dtnet->t_ranks[G_t]->group);
         for(i=0; i<dtnet->num_trans; i++)
	   if (dtnet->trans[i].r_index == G_t)
	     fprintf(stdout, " (%s)", dtnet->trans[i].name);	 
     	   else if (dtnet->t_ranks[G_t]->group == dtnet->trans[i].group)
	     fprintf(stdout, " %s", dtnet->trans[i].name);
       }
       else
	 fprintf(stdout, "{ (%s)", dtnet->t_ranks[G_t]->name);       

       fprintf(stdout, " }\n in vanishing state with \n");
       
       print_s(stdout, z);

       fflush(stdout);fflush(stderr); 

       fprintf(stdout, "\n\nDTSPN not well defined!\n");
       fprintf(stdout, "Please define priorities or (re)-group transitions.\n\n");
       clean_up();
       dtnet->error_func("\n",TRUE);
     }
     else /* no confusion */
       Free_vp(vp_G);
   }

 } /*end for all groups*/

 DEBUG_level--; 
}/*end fire_C*/



void check_state(prob, z_nxt, t, vp_z)
State *z_nxt;
double *prob;
int t;
Van_paths **vp_z;
{
 ECtrans *ec;
 Bool NO_en_b_ph;
 int i;
 Van_paths *vp_z_nxt;


#ifdef CHDEBUG 
 dbspc(); 
 printf(" Checking  reached with prob=%g ", *prob); print_s(stdout, z_nxt);
#endif 

 Get_ST_EC(ec, z_nxt);

#ifdef ECDEBUG 
 dump_ST_EC(ec, z_nxt);
#endif 


 if (ec->num_en == 0) {
  fprintf(stdout,"ERROR: in ");  print_s(stdout, z_nxt); clean_up();
  dtnet->error_func("Absorbing state.\n",TRUE); 
 }
 
 /* check for vanishing B-phases */
 for(i=0, NO_en_b_ph=TRUE; NO_en_b_ph && (i < ec->num_en); i++) {
   if ((ec->enabled[i] < dtnet->num_timed_trans) && 
       (z_nxt->ph[ec->enabled[i]] == B_ph))
     NO_en_b_ph=FALSE;
 }

 if ((ec->num_cand == 0) && NO_en_b_ph) {
   /* reached tangible state z_nxt */
   vp_z_nxt = init_van_paths( ch_add_S(z_nxt) );
#ifdef VPDEBUG 
 dbspc(); printf(" Checking  was reached with p=%g, t=%d (tang)\n", *prob, t);
#endif
   add_van_paths(t, prob, vp_z_nxt, vp_z);
 }
 else {
   /* traverse vanishing state z_nxt */
   if (NO_en_b_ph)
     fire_C(z_nxt, &vp_z_nxt, ec);
   else 
     adv_phases(z_nxt, &vp_z_nxt, ec, FALSE, 1);
#ifdef VPDEBUG 
   dbspc(); printf(" Checking  was reached with p=%g, t=%d\n", *prob, t);
#endif
   add_van_paths(t, prob, vp_z_nxt, vp_z);
 }   
}/*end check_state*/



/* help procedure for genrrg() due to EC on stack -> autom. removal of ec */
void genrrg_hlp(s,s_ref)
 State *s;
 State_ref *s_ref;
{
 int i,t,z;
 ECtrans *ec;
 Van_paths *vp;
 double C_entry;
 word stepsize;

   Get_ST_EC(ec, s);

#ifdef ECDEBUG 
   dump_ST_EC(ec, s);
#endif 

   if (Embedding) {
     /* only non-B, enabled, phases in a tangible state are advanced */
     /* chance for embedding :-) */
     /* determine min. step-size of all enabled phases */
     stepsize = MAX_PHASE;
     for (i=0; i < ec->num_en; i++) {
       t = (int) ec->enabled[i];
       if (t >= dtnet->num_timed_trans) {
         fprintf(stdout,"ERROR: Inconsistency in next tangible state."); clean_up();
         dtnet->error_func("Enabled immediate transition.\n",TRUE);
       }

       if (stepsize > s->ph[t]) 
	 stepsize = s->ph[t];
       if (dtnet->t_ranks[t]->type == GEN) 
	 for (z=0; z<(int)stepsize; z++) 
	   if(dtnet->t_ranks[t]->dtp_array[s->ph[t]-z] > 0.0){
	     stepsize = (word)(z + 1);
	     break;
	   }
     }
     /*if (stepsize > 1) stepsize--;  / was for security, one more state */
     C_entry = (double) stepsize; 
   } 
   else { 
     stepsize = 1;
     C_entry = 1.0;
   }

   adv_phases(s, &vp, ec, TRUE, stepsize);

#ifdef RRGDEBUG
   dump_vp(vp); 
#endif 
   /* extract state transition probs&measures from vp and store 'em*/
   extr_res_store(vp, s, s_ref, C_entry);

   /* all information has been extracted discharge van. path */
   Free_vp(vp);
   vp = NULL;

}/*end genrrg_hlp*/



void genrrg()
{
 int i;
 State *s;
 State_ref *s_ref;
 ECtrans *ec;
 Van_paths *vp;


 init_S();
 init_file_handles();
 init_graph();

 /* global for act_marking() */
 act_m_vector = New_m();

 s = New_s();

 /* init first vanishing state */
 for(i=0; i<dtnet->num_places; i++)      s->m[i]  = dtnet->p_ranks[i]->m_0;
 for(i=0; i<dtnet->num_timed_trans; i++) s->ph[i] = B_ph;

 /* create initial tangible states reachable from s */
 /* is ideally equal to one with p=1 */
 /* important for transient solution for first sojourn probability */
 Get_ST_EC(ec, s);

#ifdef ECDEBUG 
 dump_ST_EC(ec, s);
#endif 

 printf("Initial resampling of firing time-phases.\n\n");
#ifdef RRGDEBUG
 print_s(stdout, s); 
#endif
 if (ec->num_en > 0 && ec->enabled[ec->num_en-1] >= dtnet->num_timed_trans)
   printf("WARNING: initial state is vanishing\n");
 
 if (ec->num_en > 0 && ec->enabled[0] >= dtnet->num_timed_trans)
   /* NO timed trans. enabled and ONLY immediate trans. enabled 
      => we MUST have at least one candidate! */
   fire_C(s, &vp, ec);
 else
   /* - there are initially enabled timed transitions to advance 
        (or no enabled transitions)
      - immediates might be enabled and will fire after initialization
        of enabled timed trans. */
   adv_phases(s, &vp, ec, FALSE,1);

#ifdef FINEDEBUG 
 dump_Stree();
 dump_nxtS();
#endif

 /* extract initial sojourn probs from vp for transient solution */
 extr_init_prob(vp);
 Free_vp(vp);
 vp = NULL;

 /* ----------------  TRAVERSE ALL TANGIBLE STATES s -------------------*/
 while (get_nxtS(s, &s_ref)) {
#ifdef RRGDEBUG
   printf("\n----------------\nNEXT tangible is          "); 
   printf("[%ld] ", s_ref->index);
   print_s(stdout, s); 
#endif
#ifdef FINEDEBUG 
   dump_Stree();
   dump_nxtS();
#endif

   genrrg_hlp(s,s_ref);

 }/*end while*/


 finish_storage();
 

#ifdef FINEDEBUG 
 nl(stdout); dump_Stree();
 nl(stdout); dump_S();
 nl(stdout); dump_nxtS();
#endif 
 
}/*end genrrg*/



/*end genrrg_graph.c -------------------------------------------------------- */

