/*******************************************************************************
* Program:	genrrg
*
* File:		genrrg_tree.c
*
* Purpose:	Access and efficient storage (multiple-tree) of the rrg.
*
* Programmer:	Robert Zijal
*
* History:	March 1995
* 1.8.96 
* - beautified printing of states
* - storing probabilities of initial (vanishing) state for transient solution
* 18.9.96
* - extended phase storage to additional unbound mode (makes sense when embedding)
*******************************************************************************/

#include <stdio.h>
#include "genrrg.h"
#include "genrrg_tree.h"
#include "matrices_P_C.h"


/*---------------------*/
/* COMPILE-SWITCH-KEYS */
/*---------------------*/

#define SAVE_MEM	/* has to be canceled when debugging !!! */

/*
#define MTREEDEBUG 
#define PHTREEDEBUG 
#define STREEDEBUG 
#define MXDEBUG
*/
#define _PHTREEDEBUG
/*-----------*/
/* CONSTANTS */
/*-----------*/

#define MAX_MT_ENTRY	MAX_TOKEN   /* max no of entries in Mt_unbound_node */
#define MIN_MT_ENTRY	8 	    /* min no of entries in Mt_unbound_node */
#define MAX_PH_ENTRY	MAX_PHASE   /* max no of entries in PHt_unbound_node */

/*-----------*/
/* VARIABLES */
/*-----------*/

/* Trigger if all token-numbers fit into one byte variables */
static Bool m_is_byte = FALSE;

/* Trigger if all phase-numbers fit into one byte variables */
static Bool ph_is_byte = FALSE;

/* Tree of states S. Global access in file (more efficient) */
static void *S_tree = NULL;

static unsigned long M_cnt = 0;		/* Marking counter */

static unsigned long S_cnt = 0;		/* counter of all states */
static unsigned long check_S_cnt = 0;	/* check counter of all states */
static unsigned long S_chain_cnt = 0;	/* counter of all visited states */
static unsigned long nxtS_chain_cnt = 0;/* counter of all unvisited states */

/* pointer to first/last of chain of all states S */
static Ph_rfts *f_S_chain = NULL;
static Ph_rfts *l_S_chain = NULL;

/* pointer to first of chain of next states to visit */
/* starts with *f_S_chain and advances till *l_S_chain */
static Ph_rfts *f_nxtS_chain = NULL;

/* file handles for state&measure storage */
static FILE  *emc_fp, *conv_fp, *rg_fp, *rates_fp;
static char  *emc_filename, *conv_filename, *rg_filename, *rates_filename;
static int   *tmp_marking;                
static double *tmp_rates;

/* states and results are stored immediately during genrrg() */
void init_file_handles()
{
  int   m_len = strlen(dtnet->netname) + 10;

  /* prepare file handles */
  conv_filename = (char *)nMalloc(m_len * sizeof(char));
  emc_filename  = (char *)nMalloc(m_len * sizeof(char));
  rg_filename   = (char *)nMalloc(m_len * sizeof(char));
  rates_filename= (char *)nMalloc(m_len * sizeof(char));
  tmp_marking   = (int  *)nMalloc(dtnet->num_places  * sizeof(int));
  tmp_rates     = (double *)nMalloc(dtnet->num_trans * sizeof(double));

  sprintf (conv_filename, "%s.CONV00", dtnet->netname);
  sprintf (emc_filename,  "%s.EMC00",  dtnet->netname);
  sprintf (rg_filename,   "%s.RG",     dtnet->netname);
  sprintf (rates_filename,"%s.RATES",  dtnet->netname);

  if ((emc_fp  = fopen(emc_filename, "w")) == NULL){
    fprintf(stdout,"ERROR: in file %s \n", emc_filename);
    dtnet->error_func("Could not open file for writing.\n",TRUE);
  }
  if ((conv_fp = fopen(conv_filename, "w")) == NULL) {
    fprintf(stdout,"ERROR: in file %s \n", conv_filename);
    dtnet->error_func("Could not open file for writing.\n",TRUE);
  }
  if ((rg_fp   = fopen(rg_filename, "w")) == NULL) {
    fprintf(stdout,"ERROR: in file %s \n", rg_filename);
    dtnet->error_func("Could not open file for writing.\n",TRUE);
  }
  if ((rates_fp   = fopen(rates_filename, "w")) == NULL) {
    fprintf(stdout,"ERROR: in file %s \n", rates_filename);
    dtnet->error_func("Could not open file for writing.\n",TRUE);
  }
}/*init_file_handles*/


/* close open file handles */
void clean_up()
{
  fclose(conv_fp);
  fclose(emc_fp);
  fclose (rg_fp);
  fclose (rates_fp);
}

/*---------------------- Aux. functions --------------------------------------*/

void Copy_s(from, to)
State *from, *to;
{
  tmp_size = (dtnet->num_places + dtnet->num_timed_trans) * sizeof(word); 
  memcpy(to->m, from->m, tmp_size); 
  to->ph = to->m + dtnet->num_places; 
}

void Copy_m(from, to)
State *from;
word *to;
{
  tmp_size = (dtnet->num_places) * sizeof(word); 
  memcpy(to, from->m, tmp_size); 
}

State *New_s()	
{
 State *s;
 s = (State *)nMalloc(sizeof(State));
 s->m  = 
  (word *)nMalloc( (dtnet->num_places + dtnet->num_timed_trans) * sizeof(word));
 s->ph = s->m + dtnet->num_places; 
 return s;
}

void Free_s(s)	
State *s;
{ Free(s->m); Free(s); }

word *New_m()	
{
 word *m;
 m  = (word *)nMalloc(dtnet->num_places * sizeof(word));
 return m;
}

word *New_ph()	
{
 word *ph;
 ph  = (word *)nMalloc(dtnet->num_timed_trans * sizeof(word));
 return ph;
}

void print_s(stream, s) 
FILE *stream;
State *s;
{
 int i;
 fprintf(stream, "Marking:");
 for(i=0; i<dtnet->num_places; i++) 
    fprintf(stream, "  %s=%u", 
	    dtnet->p_ranks[i]->name, s->m[i]);

 fprintf(stream, "\nRFTs    :");
 for(i=0; i<dtnet->num_timed_trans; i++) 
    if (s->ph[i] != B_ph)
      fprintf(stream, "  %s=%u", 
	    dtnet->t_ranks[i]->name, s->ph[i]);
 fprintf(stream, "\n");
}

void print_s_simple(stream, s) 
FILE *stream;
State *s;
{
 int i;
 fprintf(stream, "state: (");
 for(i=0; i<dtnet->num_places; i++)  fprintf(stream, " %u",s->m[i]);
 fprintf(stream, " | ");
 for(i=0; i<dtnet->num_timed_trans; i++)  fprintf(stream, " %u",s->ph[i]);
 fprintf(stream, " )\n");
}

void print_m(stream, m)	
FILE *stream;
word *m;
{
 int p;
 fprintf(stream, "marking: (");
 for(p=0; p<dtnet->num_places; p++)  fprintf(stream, " %u",m[p]);
 fprintf(stream, " )");
}

void print_ph(stream, ph)       
FILE *stream;
word *ph;
{
 int t;
 fprintf(stream, "phases : (");
 for(t=0; t<dtnet->num_timed_trans; t++)  fprintf(stream, " %u",ph[t]);
 fprintf(stream, " )\n");
}

void get_m(mt_leaf, m)	
Mt_marking *mt_leaf;
word *m;
{
 int num_places, p;
 num_places = dtnet->num_places;      

 if (mt_leaf->m.b == NULL) {
   fprintf(stdout,"ERROR: mt_leaf-marking empty.\n"); clean_up();
   dtnet->error_func("Inconsistent m-vector pointer.\n",TRUE);
 }
 if (m_is_byte) 
   for(p=0; p<num_places; p++)  m[p] = (word) mt_leaf->m.b[p];
 else  /* m_is_word*/
   for(p=0; p<num_places; p++)  m[p] = mt_leaf->m.w[p];
}

void get_ph(ph_leaf, ph)	
Ph_rfts *ph_leaf;
word *ph;
{
 int num_trans, t;
 num_trans = dtnet->num_timed_trans;      

 if (ph_leaf->ph.b == NULL) {
   fprintf(stdout,"ERROR: ph_leaf-phases empty.\n"); clean_up(); 
   dtnet->error_func("Inconsistent phi-vector pointer.\n",TRUE);
 }
 if (ph_is_byte) 
   for(t=0; t<num_trans; t++)  ph[t] = (word) ph_leaf->ph.b[t];
 else  /* ph_is_word*/
   for(t=0; t<num_trans; t++)  ph[t] = ph_leaf->ph.w[t];
}

#ifdef MTREEDEBUG
void print_mt_leaf(stream, mt_leaf)
FILE *stream;
Mt_marking *mt_leaf;
{
 State *s;
 New_St_s(s);  get_m(mt_leaf, s->m);  print_m(stream, s->m); nl(stream);
}
#endif


#ifdef PHTREEDEBUG
void print_ph_leaf(stream, ph_leaf)
FILE *stream;
Ph_rfts *ph_leaf;
{
 State *s;
 New_St_s(s);
 get_ph(ph_leaf, s->ph);  print_ph(stream, s->ph);
}
#endif


/*---------------------- Marking tree functions ------------------------------*/

Mt_marking *init_mt_leaf(m)
word *m;
{
 int num_places, p;
 Mt_marking *mt_leaf;

 num_places = dtnet->num_places;      

 mt_leaf    = (Mt_marking *)nMalloc(sizeof(Mt_marking));

 if (m_is_byte) {
   mt_leaf->m.b = (byte *)nMalloc(num_places * sizeof(byte));
   for(p=0; p<num_places; p++)  mt_leaf->m.b[p] = (byte) m[p]; 
 }
 else { /* m_is_word*/
   mt_leaf->m.w = (word *)nMalloc(num_places * sizeof(word));
   for(p=0; p<num_places; p++)  mt_leaf->m.w[p] = m[p];
 }

 mt_leaf->Ph_tree = NULL;	/* create new phase-tree */     
 M_cnt ++;
 
#ifdef MTREEDEBUG
 printf("\talloc leaf (%ld bytes) with ", 
   num_places*(m_is_byte ? sizeof(byte) : sizeof(word)) + sizeof(Mt_marking)); 
 print_mt_leaf(stdout, mt_leaf);
#endif
#ifdef SHOWMEMUSAGE
 if (m_is_byte)
   m_state_size += ( sizeof(Mt_marking) + (num_places * sizeof(byte)) );
 else
   m_state_size += ( sizeof(Mt_marking) + (num_places * sizeof(word)) );
#endif

 return mt_leaf;
} /*end init_mt_leaf*/


Mt_bnd_node *init_bnd_Mnode(p)
int p;
{
 Mt_bnd_node *bnd_node;
 long b, num_tok;

 num_tok = dtnet->p_ranks[p]->bound + 1; /* + 0 tokens */
 bnd_node = (Mt_bnd_node *)nMalloc(num_tok*sizeof(Mt_bnd_node));
 for(b=0; b<num_tok; bnd_node[b++] = NULL);

#ifdef MTREEDEBUG
 printf("\talloc bnd marking node (%ld bytes) with size %ld\n",
	num_tok*sizeof(Mt_bnd_node), num_tok);
#endif
#ifdef SHOWMEMUSAGE
 m_bnd_size += ( num_tok*sizeof(Mt_bnd_node) );
#endif

 return bnd_node;
} /*end init_bnd_Mnode*/


Mt_unbnd_node *init_unbnd_Mnode()
{
 Mt_unbnd_node *unbnd_node;

 unbnd_node = (Mt_unbnd_node *)nMalloc(sizeof(Mt_unbnd_node));
 unbnd_node->size  = MIN_MT_ENTRY;
 unbnd_node->num_e = 0;
 unbnd_node->entry = (Mt_entry *)nMalloc(unbnd_node->size*sizeof(Mt_entry));

#ifdef MTREEDEBUG
 printf("\talloc UNbnd marking node (%ld bytes) with size %u\n",
   sizeof(Mt_unbnd_node)+unbnd_node->size*sizeof(Mt_entry),unbnd_node->size);
#endif
#ifdef SHOWMEMUSAGE
 m_unbnd_size += ( sizeof(Mt_unbnd_node) + (unbnd_node->size*sizeof(Mt_entry)) );
 m_unbnd_mxentries += unbnd_node->size;
#endif

 return unbnd_node;
} /*end init_unbnd_Mnode*/



/*
 Performs 2 actions at the same time for greater time-efficiency
  a. Check if entry is element of unbound node (get index (where it should be)).
  b. Store entry if not element of unbound node (reuse index).
*/
void **ch_add_M_unb_entry(node, m_p)
Mt_unbnd_node *node;
word m_p;  		/* current place-token count */
{
 Mt_entry *entry; 
 long  num_e, min, max, here = 0;
 Bool  is_el = FALSE;

 num_e = node->num_e;
 entry = node->entry;

 if (num_e != 0) {
   min = 0;  max = num_e-1;  	/* inv: (num >= 1) */
   if (num_e > 2)
     while( (!is_el) && ((min+1)!=max) ) {	/* inv: (num > 2) */
       here = (min+max)/2;
       if   (m_p > entry[here].no_tokens)    min = here;
       else if (m_p < entry[here].no_tokens) max = here;
       else is_el = TRUE;
     					/* inv: !is_el && (num == 2) */
					/*    || is_el && (num >= 2) */
     }

   if (!is_el) {			/* inv: (1 <= num <= 2) */
     if      (m_p == entry[min].no_tokens) 
      { here = min; is_el = TRUE; }
     else if (m_p == entry[max].no_tokens) 
      { here = max; is_el = TRUE; }
     else if (m_p <  entry[min].no_tokens) 
       here = min;
     else if (m_p <  entry[max].no_tokens) 
       here = max;
     else /* (entry[max].no_tokens < m_p) */ 
       here = max+1;
   }

   if (!is_el) {
     /* provide new memory if necessary */
     if (num_e == node->size) {
       if (node->size == MAX_MT_ENTRY) {
         fprintf(stdout,"ERROR: while storing tokencount %d.\n", m_p); 
	 clean_up();
         dtnet->error_func(
          "Unbound M-node has more entries than max. no of tokens.\n",TRUE);
       }
#ifdef MTREEDEBUG
       printf("\t(REALLOC memory)\n");
#endif
#ifdef SHOWMEMUSAGE
 no_mtree_reallocs++;
 m_unbnd_size-= (node->size*sizeof(Mt_entry));
 m_unbnd_mxentries-= node->size;
#endif

       if ( (node->size + MIN_MT_ENTRY) >= MAX_MT_ENTRY) 
	 node->size  = MAX_MT_ENTRY;
       else
	 node->size  = node->size + MIN_MT_ENTRY;
       node->entry = 
	 (Mt_entry *)nRealloc(entry, node->size*sizeof(Mt_entry));
       entry = node->entry;

#ifdef SHOWMEMUSAGE
 m_unbnd_size+= (node->size*sizeof(Mt_entry));
 m_unbnd_mxentries+= node->size;
#endif
     }

     /* make room for new entry, right-shift old (larger) entries */
     for (max=num_e-1; max>=here; max--){
#ifdef MTREEDEBUG
       printf("here:%ld, max:%ld\n",here, max) ;
#endif
       entry[max+1].no_tokens = entry[max].no_tokens;
       entry[max+1].next_node = entry[max].next_node; 
     }
   }
 }/*endif (num_e != 0)*/

 /* store new entry */
 if (!is_el) {
   entry[here].no_tokens = m_p;
   entry[here].next_node = NULL; 
   node->num_e++; 
#ifdef SHOWMEMUSAGE
 m_unbnd_entries++;
#endif
 }

 if (is_el && (entry[here].next_node == NULL)) {
   fprintf(stdout,"ERROR: in existing tokencount %d.\n", m_p); clean_up();
         dtnet->error_func(
          "Lost pointer while checking/adding marking (unbound) to S.\n",TRUE);
 }

 return (&entry[here].next_node);
} /*end ch_add_M_unb_entry*/



/*
 Performs 2 actions at the same time for greater time-efficiency
  a. Check if marking is element of S (tree-traversal).
  b. Store marking if not element of S (continue tree-traversal from a. ).
*/
Mt_marking *ch_add_M(m)
word *m;
{
 void		**next;

 int p, num_places;
 Bool new_state = FALSE;

#ifdef MTREEDEBUG
 printf("ch_add_M: "); print_m(stdout, m); nl(stdout);
#endif

 num_places = dtnet->num_places;	/* no of tree-levels */

 next = &S_tree;

 for(p =0; p <= num_places; p++) {

   if ((*next != NULL) && new_state) {
     fprintf(stdout,"ERROR: in level %d, ", p);  print_m(stdout, m); nl(stdout);
     clean_up();
     dtnet->error_func(
	  "Illegal pointer while checking/adding marking to S.\n",TRUE);
   }

   /* leaf node */
   if (p == num_places) {
     if (*next == NULL) {
       new_state = TRUE;
       *next = (void*) init_mt_leaf(m);
     }
   }
   /* bound node */
   else if (dtnet->p_ranks[p]->is_bound) {
#ifdef MTREEDEBUG 
     printf(" Bnd level: %d \n", p); 
#endif
     if (m[p] > dtnet->p_ranks[p]->bound) {
       fprintf(stdout,"ERROR: in place %s (l%d:%d>%d), ", 
	 dtnet->p_ranks[p]->name, p, m[p], dtnet->p_ranks[p]->bound);
       print_m(stdout, m); nl(stdout); clean_up();   
       dtnet->error_func(
	  "Token-Bound violation while checking/adding marking to S.\n",TRUE);
     }     
     if (*next == NULL) {
       new_state = TRUE;
       *next = (void*) init_bnd_Mnode(p);
     }
     next =   (void *)  ( ((Mt_bnd_node*)(*next)) + m[p] );
   }
   /* unbound node */
   else {
#ifdef MTREEDEBUG 
     printf(" Unbnd level: %d \n", p); 
#endif
     if (*next == NULL) {
       new_state = TRUE;
       *next = (void*) init_unbnd_Mnode();
     }
     next = ch_add_M_unb_entry(((Mt_unbnd_node*)(*next)), m[p]); 
   }


 } /*end for*/

 if ((Mt_marking *)(*next) == NULL) {
   fprintf(stdout,"ERROR: in "); print_m(stdout, m); nl(stdout); clean_up();   
   dtnet->error_func(
          "Tree-leaf not reached while checking/adding marking to S.\n",TRUE);
 }

#ifdef MTREEDEBUG
         printf("\t\tleaf reached with ");
 	 print_mt_leaf(stdout, (Mt_marking *)(*next));
#endif

 return ((Mt_marking *)(*next));

} /*end ch_add_M*/



/*------------------------ Phase tree functions ------------------------------*/

Ph_rfts *init_ph_leaf(ph)
word *ph;
{
 int num_trans, t;
 Ph_rfts *ph_leaf;

 num_trans = dtnet->num_timed_trans;      

 ph_leaf    = (Ph_rfts *)nMalloc(sizeof(Ph_rfts));

 if (ph_is_byte) {
   ph_leaf->ph.b = (byte *)nMalloc(num_trans * sizeof(byte));
   for(t=0; t<num_trans; t++)  ph_leaf->ph.b[t] = (byte) ph[t]; 
 }
 else { /* ph_is_word*/
   ph_leaf->ph.w = (word *)nMalloc(num_trans * sizeof(word));
   for(t=0; t<num_trans; t++)  ph_leaf->ph.w[t] = ph[t];
 }

 ph_leaf->mt_leaf = NULL;	/* new phase-tree leaf */     
 ph_leaf->chain = NULL;             

 ph_leaf->index = S_cnt++;

 /* show the user some action */
 if ((S_cnt%1000) == 0) {printf("*"); fflush(stdout);} 

#ifdef PHTREEDEBUG
 printf("\talloc %lu. leaf (%ld bytes) with ", ph_leaf->index, 
   num_trans*(ph_is_byte ? sizeof(byte) : sizeof(word)) + sizeof(Ph_rfts)); 
 print_ph_leaf(stdout, ph_leaf);
#endif
#ifdef SHOWMEMUSAGE
 if (ph_is_byte)
   ph_state_size += ( sizeof(Ph_rfts) + (num_trans * sizeof(byte)) );
 else
   ph_state_size += ( sizeof(Ph_rfts) + (num_trans * sizeof(word)) );
#endif

 return ph_leaf;
} /*end init_ph_leaf*/


Ph_bnd_node *init_bnd_PHnode(t)
int t;
{
 Ph_bnd_node *bnd_node;
 long b, num_phases;

 num_phases = dtnet->t_ranks[t]->bound + 1; /* + phase 0 = B_ph */
 bnd_node = (Ph_bnd_node *)nMalloc(num_phases*sizeof(Ph_bnd_node));
 for(b=0; b<num_phases; bnd_node[b++] = NULL);

#ifdef PHTREEDEBUG
 printf("\talloc bnd phase node (%ld bytes) with size %ld\n",
	num_phases*sizeof(Ph_bnd_node), num_phases);
#endif
#ifdef SHOWMEMUSAGE
 ph_bnd_size += ( num_phases*sizeof(Ph_bnd_node) );
#endif

 return bnd_node;
} /*end init_bnd_PHnode*/


PHt_unbnd_node *init_unbnd_PHnode(min_ph_entry)
word min_ph_entry;
{
 PHt_unbnd_node *unbnd_node;

 unbnd_node = (PHt_unbnd_node *)nMalloc(sizeof(PHt_unbnd_node));
 unbnd_node->size  = min_ph_entry;
 unbnd_node->num_e = 0;
 unbnd_node->entry = (PHt_entry *)nMalloc(unbnd_node->size*sizeof(PHt_entry));

#ifdef PHTREEDEBUG
 printf("\talloc UNbnd phase node (%ld bytes) with size %u\n",
   sizeof(PHt_unbnd_node)+unbnd_node->size*sizeof(PHt_entry),unbnd_node->size);
#endif
#ifdef SHOWMEMUSAGE
 ph_unbnd_size += ( sizeof(PHt_unbnd_node) + (unbnd_node->size*sizeof(PHt_entry)) );
 ph_unbnd_mxentries += unbnd_node->size;
#endif

 return unbnd_node;
} /*end init_unbnd_PHnode*/



/*
 Performs 2 actions at the same time for greater time-efficiency
  a. Check if entry is element of unbound node (get index (where it should be)).
  b. Store entry if not element of unbound node (reuse index).
*/
void **ch_add_PH_unb_entry(node, t_ph, min_ph_entry, max_t_ph_entry)
PHt_unbnd_node *node;
word t_ph;  		/* current place-token count */
word min_ph_entry;
word max_t_ph_entry;
{
 PHt_entry *entry; 
 long  num_e, min, max, here = 0;
 Bool  is_el = FALSE;

 num_e = node->num_e;
 entry = node->entry;

 if (num_e != 0) {
   min = 0;  max = num_e-1;  	/* inv: (num >= 1) */
   if (num_e > 2)
     while( (!is_el) && ((min+1)!=max) ) {	/* inv: (num > 2) */
       here = (min+max)/2;
       if   (t_ph > entry[here].no_phases)    min = here;
       else if (t_ph < entry[here].no_phases) max = here;
       else is_el = TRUE;
     					/* inv: !is_el && (num == 2) */
					/*    || is_el && (num >= 2) */
     }

   if (!is_el) {			/* inv: (1 <= num <= 2) */
     if      (t_ph == entry[min].no_phases) 
      { here = min; is_el = TRUE; }
     else if (t_ph == entry[max].no_phases) 
      { here = max; is_el = TRUE; }
     else if (t_ph <  entry[min].no_phases) 
       here = min;
     else if (t_ph <  entry[max].no_phases) 
       here = max;
     else /* (entry[max].no_phases < t_ph) */ 
       here = max+1;
   }

   if (!is_el) {
     /* provide new memory if necessary */
     if (num_e == node->size) {
       if (node->size == max_t_ph_entry) {
         fprintf(stdout,"ERROR: while storing phasecount %d.\n", t_ph); 
	 clean_up();
         dtnet->error_func(
          "Unbound PH-node has more entries than # of phases.\n",TRUE);
       }
#ifdef PHTREEDEBUG
       printf("\t(REALLOC memory)\n");
#endif
#ifdef SHOWMEMUSAGE
 no_phtree_reallocs++;
 ph_unbnd_size-= (node->size*sizeof(PHt_entry));
 ph_unbnd_mxentries-= node->size;
#endif
       if ( (node->size + min_ph_entry) >= max_t_ph_entry) 
	 node->size  = max_t_ph_entry;
       else {
	 node->size  = node->size + min_ph_entry;
	 if ( (node->size + min_ph_entry) > max_t_ph_entry) 
	   /* leave no small rests, take rather all, reducing # of reallocs */
	   node->size  = max_t_ph_entry; 
       }
       node->entry = 
	 (PHt_entry *)nRealloc(entry, node->size*sizeof(PHt_entry));
       entry = node->entry;

#ifdef SHOWMEMUSAGE
 ph_unbnd_size+= (node->size*sizeof(PHt_entry));
 ph_unbnd_mxentries+= node->size;
#endif
     }

     /* make room for new entry, right-shift old (larger) entries */
     for (max=num_e-1; max>=here; max--){
#ifdef PHTREEDEBUG
       printf("here:%ld, max:%ld\n",here, max) ;
#endif
       entry[max+1].no_phases = entry[max].no_phases;
       entry[max+1].next_node = entry[max].next_node; 
     }
   }
 }/*endif (num_e != 0)*/

 /* store new entry */
 if (!is_el) {
   entry[here].no_phases = t_ph;
   entry[here].next_node = NULL; 
   node->num_e++; 
#ifdef SHOWMEMUSAGE
 ph_unbnd_entries++;
#endif
 }

 if (is_el && (entry[here].next_node == NULL)) {
   fprintf(stdout,"ERROR: in existing phasecount %d.\n", t_ph); clean_up();
         dtnet->error_func(
          "Lost pointer while checking/adding phases (unbound) to S.\n",TRUE);
 }

 return (&entry[here].next_node);
} /*end ch_add_PH_unb_entry*/






/*
 Performs 2 actions at the same time for greater time-efficiency
  a. Check if phase-vector is element of S (tree-traversal).
  b. Store vector if not element of S (continue tree-traversal from a. ).
*/
Ph_rfts *ch_add_Ph(mt_leaf, ph)
Mt_marking *mt_leaf;
word *ph;
{
 void		**next;

 int t, num_trans;
 Bool new_state = FALSE;
 word phase;
 word min_ph_entry;

#ifdef PHTREEDEBUG
 printf("ch_add_Ph: "); print_ph(stdout, ph);
#endif

 num_trans = dtnet->num_timed_trans;	/* no of tree-levels */


 next = &mt_leaf->Ph_tree;

 for(t =0; t <= num_trans; t++) {


   if ((*next != NULL) && new_state) {
     fprintf(stdout,"ERROR: in level %d, ", t);  print_ph(stdout, ph);
     clean_up();   
     dtnet->error_func(
	  "Illegal pointer while checking/adding phases to S.\n",TRUE);
   }

   /* leaf node */
   if (t == num_trans) {
     if (*next == NULL) {
       new_state = TRUE;
       *next = (void*) init_ph_leaf(ph);
     }
   }
   /* leaf not reached */
   else {
     /* 0-phase not allowed in tangible state, reserved for accessing B_ph */
     if (ph[t] == 0) {
       fprintf(stdout,"ERROR: in level %d, ", t);  print_ph(stdout, ph);
       clean_up();   
       dtnet->error_func(
	 "Illegal 0 phase while checking/adding phases to S.\n",TRUE);
     }

     /* convert B_ph to 0 only while accessing levels */
     if (ph[t] == B_ph) phase = 0; 
     else 	 	phase = ph[t];

     if (phase > dtnet->t_ranks[t]->bound) {
       fprintf(stdout,"ERROR: in trans %s (l%d:%d>%d), ", 
	       dtnet->t_ranks[t]->name, t, phase, dtnet->t_ranks[t]->bound);    
       print_ph(stdout, ph); clean_up();   
       dtnet->error_func(
	 "Phase-Bound violation while checking/adding phases to S.\n",TRUE);
     }     

     /* bound node ? */
     if (dtnet->t_ranks[t]->is_bound) {
#ifdef PHTREEDEBUG 
       printf(" Bnd level: %d \n", t); 
#endif
       if (*next == NULL) {
         new_state = TRUE;
         *next = (void*) init_bnd_PHnode(t);
       }
       next =   (void *)  ( ((Ph_bnd_node*)(*next)) + phase );
     }

     /* unbound node */
     else {
#ifdef PHTREEDEBUG 
       printf(" Unbnd level: %d \n", t); 
#endif

       /* crete adaptive MIN_PH_ENTRY, since phases are always bound */
       if (dtnet->t_ranks[t]->type == DET || dtnet->t_ranks[t]->type == EXP) {

#ifdef DID_NOT_WORK_WELL 
	 if (dtnet->t_ranks[t]->bound <= 10) 
	   min_ph_entry = (word) ((dtnet->t_ranks[t]->bound) / 2);
	 else if (dtnet->t_ranks[t]->bound <= 100) { 
	   if (SaveSpace == TRUE)
	     min_ph_entry = 5;
	   else 
	     min_ph_entry = 10;
	 }
	 else {		/* dtnet->t_ranks[t]->bound > 100 */
	   if (SaveSpace == TRUE)
	     min_ph_entry = 10;
	   else 
	     min_ph_entry = (word) ((dtnet->t_ranks[t]->bound) / 10);
	 }
#endif
       /*  begin replacing DID_NOT_WORK_WELL */
	 if (SaveSpace == TRUE)
	     min_ph_entry = 1;
	 else if (dtnet->t_ranks[t]->bound <= 10) 
	   min_ph_entry = (word) ((dtnet->t_ranks[t]->bound) / 2);
	 else if (dtnet->t_ranks[t]->bound <= 100) 
	     min_ph_entry = 5;
	 else 		/* dtnet->t_ranks[t]->bound > 100 */
	     min_ph_entry = 10;
       /*  end replacing DID_NOT_WORK_WELL */

       }
       /* NOTE, use something like this, but more complex, for UNIFORM */
       else if (dtnet->t_ranks[t]->type == GEN) {
#ifdef DID_NOT_WORK_WELL 
	 min_ph_entry = (word) ((dtnet->t_ranks[t]->bound) /2);
#endif
       /*  begin replacing DID_NOT_WORK_WELL */
	 if (SaveSpace == TRUE)
	     min_ph_entry = 1;
	 else if (dtnet->t_ranks[t]->bound <= 10) 
	   min_ph_entry = (word) ((dtnet->t_ranks[t]->bound) / 2);
	 else if (dtnet->t_ranks[t]->bound <= 100) 
	     min_ph_entry = 5;
	 else 		/* dtnet->t_ranks[t]->bound > 100 */
	     min_ph_entry = 10;
       /*  end replacing DID_NOT_WORK_WELL */
       }
       else {
	 printf("Error: Illegal transition type in \"%s\"\n", 
		dtnet->t_ranks[t]->name);
         dtnet->error_func("While checking/adding phases to S.\n",TRUE);
       }
       
       if (*next == NULL) {
         new_state = TRUE;
         *next = (void*) init_unbnd_PHnode(min_ph_entry);
       }
       next = 
	 ch_add_PH_unb_entry(((PHt_unbnd_node*)(*next)), phase, 
			     min_ph_entry,
			     (dtnet->t_ranks[t]->bound + 1)); /* + B_ph=0 */
     }
   }

 } /*end for*/


 if ((Ph_rfts *)(*next) == NULL) {
   fprintf(stdout,"ERROR: in "); print_ph(stdout, ph); clean_up();   
   dtnet->error_func(
          "Tree-leaf not reached while checking/adding phases to S.\n",TRUE);
 }

#ifdef PHTREEDEBUG
         printf("\t\tleaf reached with ");
 	 print_ph_leaf(stdout, (Ph_rfts *)(*next));
#endif

 return ((Ph_rfts *)(*next));

} /*end ch_add_Ph*/





/*------------------------ State tree functions ------------------------------*/


/*
  Store state (m,ph) in S if not already element of S. 
*/
State_ref *ch_add_S(s)
State *s;
{
 Mt_marking     *mt_leaf = NULL; 
 Ph_rfts     *ph_leaf = NULL;
 Bool new_state = FALSE;


#ifdef STREEDEBUG
 printf("ch_add_S: "); print_m(stdout, s->m);
 printf("\t"); print_ph(stdout, s->ph); 
#endif

 mt_leaf =  ch_add_M(s->m); 
 if (mt_leaf->Ph_tree == NULL)
   new_state = TRUE;

 ph_leaf =  ch_add_Ph(mt_leaf, s->ph);
 if (ph_leaf->mt_leaf == NULL){
   new_state = TRUE;
   ph_leaf->mt_leaf = mt_leaf;
 }

#ifdef STREEDEBUG
 printf("          index=%ld", ph_leaf->index);
 if (new_state) printf(" (new)\n"); else printf(" (old)\n");  
#endif

 /* append to S-chain if new */
 if (new_state) {
   if (S_cnt == 1) {
     /* init S-chain */
     if ((f_S_chain != NULL) || (l_S_chain != NULL) ) {
       fprintf(stdout,"ERROR: Inconsistency in "); print_s(stdout, s); 
       clean_up();   
       dtnet->error_func(
       "Pointers != NULL  while appending new s to empty S-chain.\n",TRUE);
     }
     f_S_chain = l_S_chain = ph_leaf;
   }
   else {
     if (l_S_chain->chain != NULL) {
       fprintf(stdout,"ERROR: Inconsistency in "); print_s(stdout, s); 
       clean_up();   
       dtnet->error_func(
       "Last Pointer != NULL  while appending new s to S-chain.\n",TRUE);
     }
     l_S_chain->chain = ph_leaf;
     l_S_chain = ph_leaf;     
   }

   if (nxtS_chain_cnt == 0) {
     /* init nxtS-chain */
     if (f_nxtS_chain != NULL) {
       fprintf(stdout,"ERROR: Inconsistency in "); print_s(stdout, s); 
       clean_up();   
       dtnet->error_func(
       "nxtPointer != NULL  while appending new s to empty nxtS-chain.\n",TRUE);
     }
     f_nxtS_chain = l_S_chain;
   }

   nxtS_chain_cnt++;
 }

 return (ph_leaf);

} /*end ch_add_S(s)*/




void init_S()
{
 int num_places, num_trans, i;

 S_cnt = M_cnt = S_chain_cnt = nxtS_chain_cnt = check_S_cnt = 0;

 num_places = dtnet->num_places;      
 num_trans  = dtnet->num_timed_trans;      

 /* try to save memory */
 for(m_is_byte=TRUE, i=0  ;  m_is_byte && (i<num_places)  ; 
     m_is_byte=(dtnet->p_ranks[i++]->bound<=MAX_BYTE_LIM)); 

 for(ph_is_byte=TRUE, i=0  ;  ph_is_byte && (i<num_trans)  ; 
     ph_is_byte=(dtnet->t_ranks[i++]->bound<MAX_BYTE_LIM)); 

 if (ph_is_byte) B_ph = MAX_BYTE_LIM;
 else            B_ph = MAX_PHASE+1; /* MAX_WORD_LIM */

#ifdef MTREEDEBUG
   printf("INIT-TREE: \n"); 
   if (m_is_byte) 
     printf("All places are bound (<=%d), use bytes.\n", MAX_BYTE_LIM);
   if (ph_is_byte) 
     printf("All phases are <%d, use bytes.\n", MAX_BYTE_LIM);
   printf("Initial Phase is set to %d.\n", B_ph);
#endif

}/*end init_S*/


Bool get_nxtS(s, s_ref)
State *s;
State_ref **s_ref;
{
 if (nxtS_chain_cnt == 0){
   if ( (f_nxtS_chain != NULL) || (S_chain_cnt != S_cnt)) {
     fprintf(stdout,"ERROR: Inconsistency in "); print_s(stdout, s); 
     clean_up();   
     dtnet->error_func("S-Chain handling mismatch.\n",TRUE);
   }
   *s_ref=NULL;
   return FALSE;
 }
 else {
   if (f_nxtS_chain == NULL) {
     fprintf(stdout,"ERROR: Inconsistency in "); print_s(stdout, s); 
     clean_up();   
     dtnet->error_func("nxtS-Chain empty.\n",TRUE);
   }
    /* get state from next_S-chain */
   get_m( f_nxtS_chain->mt_leaf, s->m);
   get_ph(f_nxtS_chain,          s->ph);  
   *s_ref=f_nxtS_chain;

   /* move to next unvisited state */
   f_nxtS_chain = f_nxtS_chain->chain;
   nxtS_chain_cnt--;
   S_chain_cnt++;

   return TRUE;
 }
}/*get_nxtS(s)*/


void dump_chain(h, exp_num)
Ph_rfts *h;
unsigned long exp_num;
{
 State *s;
 unsigned long int i;

 New_St_s(s);

 printf("(");
 for (i=0; i < dtnet->num_places; i++) 
   printf(" %s", dtnet->p_ranks[i]->name);
 printf(" |");
 for (i=0; i < dtnet->num_timed_trans; i++) 
   printf(" %s", dtnet->t_ranks[i]->name);
 printf(" )\n");


 for (i = 0;   i < exp_num;    i++, h = h->chain) {
   if (h == NULL) {
     fprintf(stdout,"ERROR: Inconsistency while dumping chain."); 
     clean_up();   
     dtnet->error_func("nxtS-Chain empty.\n",TRUE);
   }
   get_m( h->mt_leaf, s->m);
   get_ph(h,          s->ph);  
   printf("%ld. ", h->index); print_s(stdout, s);
 }
 printf("Expected number: %ld.\n\n", exp_num);
}

void dump_nxtS()
{
 printf("Elements of next_S (%ld):\n", nxtS_chain_cnt);
 dump_chain(f_nxtS_chain, nxtS_chain_cnt);
}

void dump_S()
{
 printf("Elements of S (%ld):\n", S_chain_cnt);
 dump_chain(f_S_chain, S_chain_cnt);
}

void dump_ph_tree(node)
void *node;
{
 long b;
 static State *s, *sph;
 static int level = 0;

 if (level == 0){ 
   New_St_s(sph);   
   New_St_s(s);   
 }

 if (node == NULL){
   printf("\nUnexpected empty node in ph-tree level %d.\n", level);
   return;
 }

 /* leaf */
 if (level == dtnet->num_timed_trans){
   print_ph(stdout, sph->ph);
   get_m(  ((Ph_rfts *)node)->mt_leaf,  s->m);
   get_ph( ((Ph_rfts *)node)         ,  s->ph);  
   printf("\tReached "); print_s(stdout, s);
 }
  /* node */
 else {
    /* bound ?*/
   if ( dtnet->t_ranks[level]->is_bound ) {
     for(b=0; b <= dtnet->t_ranks[level]->bound; b++) 
       if ( ((Ph_bnd_node *)node)[b] != NULL ) {
	 sph->ph[level++]=(word)b;
         dump_ph_tree(((Ph_bnd_node *)node)[b] );
 	 sph->ph[--level]=0xFFFF; /*dummy */
     }
   }
   /* unbound */
   else  
     for(b=0; b<((PHt_unbnd_node *)node)->num_e ; b++) {
       sph->ph[level++]=((PHt_unbnd_node *)node)->entry[b].no_phases;
       dump_ph_tree( ((PHt_unbnd_node *)node)->entry[b].next_node );
       sph->ph[--level]=0xFFFF; /*dummy */
     }
 
 }

} /*end dump_ph_tree(node)*/


void dump_m_tree(node)
void *node;
{
 long b;
 static State *s;
 static int level = 0;

 if (level == 0){
   printf("Dumping S-tree (depth first search)\n");
   New_St_s(s);   
 }

 if (node == NULL){
   printf("Unexpected empty node in m-tree level %d.\n", level);
   return;
 }

 /* leaf */
 if (level == dtnet->num_places){
   printf("Path-"); print_m(stdout, s->m);
   dump_ph_tree( ((Mt_marking *)node)->Ph_tree ); 
 }
 /* node */
 else {
    /* bound */
   if ( dtnet->p_ranks[level]->is_bound ) {
     for(b=0; b <= dtnet->p_ranks[level]->bound; b++) 
       if ( ((Mt_bnd_node *)node)[b] != NULL ) {
	 s->m[level++]=(word)b; 
         dump_m_tree( ((Mt_bnd_node *)node)[b] );
 	 s->m[--level]=0xFFFF; /*dummy */
      }
   }  
   /* unbound */
   else  
     for(b=0; b<((Mt_unbnd_node *)node)->num_e ; b++) {
       s->m[level++]=((Mt_unbnd_node *)node)->entry[b].no_tokens;
       dump_m_tree( ((Mt_unbnd_node *)node)->entry[b].next_node );
       s->m[--level]=0xFFFF; /*dummy */
     }
 }

} /*end dump_m_tree(node)*/

void dump_Stree()
{
 dump_m_tree(S_tree);
} /*end dump_Stree()*/



/* 
  extract initial probs from vp for transient solution 
  and write the probability vector in sparse format to *.DTINI 
    code is taken and adapted from: extr_res_store() 
*/
void extr_init_prob(initial_vp)
Van_paths *initial_vp;
{
  /* file handle for initial vector storage */
  FILE  *dtini_fp;
  char  *dtini_filename;
  int   m_len = strlen(dtnet->netname) + 10;

  Van_paths *vp;
  Van_arcs  *va;
  unsigned long int num_s_trans;
  double check_sum; 


  /* prepare file handles and open file */
  dtini_filename = (char *)nMalloc(m_len * sizeof(char));

  sprintf (dtini_filename, "%s.DTINI", dtnet->netname);

  if ((dtini_fp  = fopen(dtini_filename, "w")) == NULL){
    fprintf(stdout,"ERROR: in file %s \n", dtini_filename);
    dtnet->error_func("Could not open file for writing.\n",TRUE);
  }


 /* get number of state transitions (no. of subsequently reachable states) */
 for (vp=initial_vp, num_s_trans=0; vp != NULL ; vp=vp->next)  num_s_trans++;

 /* we reuse the classical matrix sparse representation just for one row! */
 /*------------------------------*/
 /* write header to *.DTINI file */
 /*------------------------------*/
 P_row_header.row_index = 1;
 P_row_header.entries_in_row = num_s_trans;
#ifdef MXDEBUG
 printf("INITIAL vector Row %u has %u entries (Col/prob):\n", 
	  P_row_header.row_index, P_row_header.entries_in_row);
#endif
 if (fwrite (&P_row_header, SIZEOF_P_row_header, 1, dtini_fp) != 1){
   fprintf(stdout,"ERROR: in file %s \n", dtini_filename); clean_up();
   dtnet->error_func("Could not write header to file.\n",TRUE);
 }

 /*------------------------------------*/
 /* write state transitions to *.DTINI */
 /*------------------------------------*/
 /* traverse all state transitions */
 check_sum = 0;
 for (vp=initial_vp;   vp != NULL;   vp=vp->next) {
   P_row_entry.column_index = vp->to_s->index+1;
   /* sum up multiple arc probabilities to same state */
   for(va=vp->mult_arcs, P_row_entry.probability=0; va!=NULL; va=va->next_arc){
     P_row_entry.probability += va->prob;
   }
   check_sum += P_row_entry.probability;
#ifdef MXDEBUG
   printf(" %u/%g", 
          P_row_entry.column_index, P_row_entry.probability);
#endif
   if (fwrite (&P_row_entry, SIZEOF_P_row_entry, 1, dtini_fp) != 1){
     fprintf(stdout,"ERROR: in file %s \n", dtini_filename);  
     fclose(dtini_fp);
     clean_up();
     dtnet->error_func("Could not write entry to file.\n",TRUE);
   }
 }/*end for*/

#ifdef MXDEBUG
 printf("\t (sum=%g)\n", check_sum);
#endif
 if ((1.0 - check_sum) > DEB_accuracy){
   fprintf(stdout,"ERROR: accuracy violation in initial state, sum= %.20g, accuracy= %g\n", 
             check_sum, DEB_accuracy); 
   fclose(dtini_fp);
   clean_up();
   dtnet->error_func(
      "Inconsistency: Sum of state-trans-probabilities not equal to 1.\n",TRUE);
 }


  /* clean up */
 fclose(dtini_fp);

}



/* 
 -extract state transition probs from vp 
 -write the P-matrix to *.EMC00 and 
 -write the C-matrix to *.CONV00 
 -write tangible markings to *.RG 
*/
void extr_res_store(vp_head, s, s_ref, step_size)
Van_paths *vp_head;
State *s;
State_ref *s_ref;
double step_size;
{
 Van_paths *vp;
 Van_arcs  *va;
 unsigned long int num_s_trans;
 double check_sum; 
 int num_places, num_trans, p, t;
 byte trans_cnt;

 num_places = dtnet->num_places;      
 num_trans  = dtnet->num_trans;      

 /* get number of state transitions (no. of subsequently reachable states) */
 for (vp=vp_head, num_s_trans=0; vp != NULL ; vp=vp->next)  num_s_trans++;

 /*------------------------------*/
 /* write header to *.EMC00 file */
 /*------------------------------*/
 P_row_header.row_index = s_ref->index+1;
 P_row_header.entries_in_row = num_s_trans;
#ifdef MXDEBUG
 printf("Row %u has %u entries (Col/prob):\n", 
	  P_row_header.row_index, P_row_header.entries_in_row);
#endif
 if (fwrite (&P_row_header, SIZEOF_P_row_header, 1, emc_fp) != 1){
   fprintf(stdout,"ERROR: in file %s \n", emc_filename); clean_up();
   dtnet->error_func("Could not write header to file.\n",TRUE);
 }

 /*------------------------------------*/
 /* write state transitions to *.EMC00 */
 /*------------------------------------*/
 /* traverse all state transitions */
 check_sum = 0;
 for (vp=vp_head;   vp != NULL;   vp=vp->next) {
   P_row_entry.column_index = vp->to_s->index+1;
   /* sum up multiple arc probabilities to same state */
   for(va=vp->mult_arcs, P_row_entry.probability=0; va!=NULL; va=va->next_arc){
     P_row_entry.probability += va->prob;
   }
   check_sum += P_row_entry.probability;
#ifdef MXDEBUG
   printf(" %u/%g", 
          P_row_entry.column_index, P_row_entry.probability);
#endif
   if (fwrite (&P_row_entry, SIZEOF_P_row_entry, 1, emc_fp) != 1){
     fprintf(stdout,"ERROR: in file %s \n", emc_filename); clean_up();
     dtnet->error_func("Could not write entry to file.\n",TRUE);
   }
 }/*end for*/

#ifdef MXDEBUG
 printf("\t (sum=%g)\n", check_sum);
#endif
 if ((1.0 - check_sum) > DEB_accuracy){
   fprintf(stdout,"ERROR: in state %ld, sum= %.20g, accuracy= %g\n", 
             s_ref->index, check_sum, DEB_accuracy); clean_up();
   dtnet->error_func(
      "Inconsistency: Sum of state-trans-probabilities not equal to 1.\n",TRUE);
 }

 /*----------------------------------*/
 /* write *.CONV file diagonal entry */
 /*----------------------------------*/
 C0_entry.sojourn_time = step_size;
 C0_entry.state_number = s_ref->index+1;
#ifdef MXDEBUG
 printf("Conv_row %u/%g\n",
          C0_entry.state_number, C0_entry.sojourn_time);
#endif
 if (fwrite (&C0_entry, SIZEOF_C0_entry, 1, conv_fp) != 1){
   fprintf(stdout,"ERROR: in file %s \n", conv_filename); clean_up();
   dtnet->error_func("Could not write to file.\n",TRUE);
 }

 /*----------------------------------------------------*/
 /* write the *.RG file entry  describing the tangible */
 /* marking (number of tokens in each place)           */
 /*----------------------------------------------------*/
 /* bring places back to old order */
 for(p=0; p<num_places; p++)
     tmp_marking[p] = (int)s->m[dtnet->place[p].r_index];
#ifdef MXDEBUG
 printf("marking: (");
 for(p=0; p<num_places; p++)  printf(" %d",tmp_marking[p]);
 printf(" )\n");
#endif
 if (fwrite (tmp_marking, sizeof(int), num_places, rg_fp) != num_places){
   fprintf(stdout,"ERROR: in file %s \n", rg_filename); clean_up();
   dtnet->error_func("Could not write marking to file.\n",TRUE);
 }

 /*-------------------------------*/
 /* write the *.RATES file entry  */
 /*-------------------------------*/
 /*calculate rates */
 for (t=0; t<num_trans; tmp_rates[t++]=0); /*init */

 for (vp=vp_head;   vp != NULL;   vp=vp->next) 
   for(va=vp->mult_arcs; va!=NULL; va=va->next_arc)
     for (t=0; t<num_trans; t++) 
       tmp_rates[t]+= (va->count[t] * va->prob);

 /* determine # of fireable transitions */
 for (t=trans_cnt=0; t<num_trans; t++)
   if (tmp_rates[t] > DEB_accuracy) trans_cnt++;

 /*step_size must be = 1 if any transition fired */
 /* OBSOLETE checking 'cause we correct the thruput :-) 
 if ((trans_cnt > 0) && (step_size > 1.0)){
   fprintf(stdout,"ERROR: step-size(%g), #firing-trans(%d) \n", 
	   step_size, (int)trans_cnt); clean_up();
   dtnet->error_func("Inconsistency: step-size>1 & #firing-trans>0.\n",TRUE);
 } 
 */

 /* write header */
 if (fwrite(&trans_cnt, sizeof(byte), 1, rates_fp) != 1){
   fprintf(stdout,"ERROR: in file %s \n", rates_filename); clean_up();
   dtnet->error_func("Could not write header to file.\n",TRUE);
 }
 
 /* write entries  & sort transitions back to old order */
 for (t=0; t<num_trans; t++)
   if (tmp_rates[dtnet->trans[t].r_index] > DEB_accuracy){
     rate_entry.t_nr = t;
     /* delay expected! despite of "rates"-naming, not my fault */
     /* delay = holdingtime / rate, where rate = #firings/timeunit */
     rate_entry.rate = 
       ((double)step_size) / 
       (tmp_rates[dtnet->trans[t].r_index] * ((double)but_step) );
     if (fwrite(&rate_entry, sizeof(rate_entry), 1, rates_fp) != 1){
       fprintf(stdout,"ERROR: in file %s \n", rates_filename); clean_up();
       dtnet->error_func("Could not write entry to file.\n",TRUE);
     };
   }



 check_S_cnt++; /* one state stored */

 /* discharge ONLY phi-vectors but keep tree-structure and indices */
#ifdef SAVE_MEM
 if (ph_is_byte){ Free(s_ref->ph.b); s_ref->ph.b = NULL; }       
 else           { Free(s_ref->ph.w); s_ref->ph.w = NULL; }
#endif

}/*end extr_res_store*/



void finish_storage()
{
  /* close file handles */
  clean_up();

  /* Anzahl der andauernden Zustaende im File <netname>.AUX ablegen:
  (fuer solve_LGS.c) */
  sprintf (rg_filename, "%s.AUX", dtnet->netname);
  if (!(rg_fp = fopen (rg_filename, "w"))){
    fprintf(stdout,"ERROR: in file %s \n", rg_filename);
    dtnet->error_func("Could not open file for writing.\n",TRUE);
  }
  fprintf (rg_fp, "TANGIBLE MARKINGS %ld\n", S_chain_cnt);

  /* clean up the rest*/
  fclose (rg_fp);
  Free(rg_filename);
  Free(emc_filename);
  Free(conv_filename);
  Free(tmp_marking);
  Free(tmp_rates);

  /* final consistency checking and info */ 
  if (S_cnt != (S_chain_cnt + nxtS_chain_cnt)){
   fprintf(stdout,"ERROR: #S(%ld) !=  #S_chain(%ld) + #nxtS_chain(%ld)\n", 
	  S_cnt, S_chain_cnt, nxtS_chain_cnt);
    dtnet->error_func("Inconsistent state counters.\n",TRUE);
  }
  if (S_cnt != S_chain_cnt){
   fprintf(stdout,"ERROR: %ld states calculated BUT %ld states traversed.\n", 
	  S_cnt, S_chain_cnt);
    dtnet->error_func("Inconsistent state counters.\n",TRUE);
  }
  if (S_cnt != check_S_cnt){
   fprintf(stdout,"ERROR: %ld states calculated BUT %ld states stored.\n", 
	  S_cnt, check_S_cnt);
    dtnet->error_func("Inconsistent state counters.\n",TRUE);
  }

  printf("\nObtained %ld tangible markings and %ld tangible states.\n\n\n", 
	 M_cnt, S_cnt);

}/*finish_storage*/



/*end genrrg_tree.c --------------------------------------------------------- */






