/****************************************************************************/
/* Program:	         solve                                              */
/*                                                                          */
/* File:	         transient_sol.c			            */
/*                                                                          */
/* Purpose:              Transient solution of a timediscrete markov chain. */
/*	                  Pi(n) = Pi(n-1)*P                                 */
/*                                                                          */
/* Programmer:           Andrea Zisowsky                                    */
/*                                                                          */
/* History:              8/96                                               */
/*                                                                          */
/*  10/96                embedding methods                                  */
/****************************************************************************/



#include <stdio.h>
#include <string.h> 
#include <stdlib.h>
#include <malloc.h>
#include <math.h> 
#include <sys/time.h>
#include <signal.h>

#include "netio.h"

#include "matrices_P_C.h"
#include "transient_sol.h"

#define _MATRIXDEBUG 
#define _VEKTORDEBUG

#define false 0
#define true 1
#define MAXPATHLEN  100
#define seconds 0.00000251

/* fuer result measures: */
#define  maxtoken 65535
#define  maxplace 255
 
typedef unsigned char byte;

typedef struct                /* Typ fuer den speichereffizienten Wahrscheinlichkeitsvektor */
{
  int    last;
  double *intermediate;
}m_state;

typedef struct                /* Typ fuer den zeiteffizienten Wahrscheinlichkeitsvektor */
{
  double *intermediate;
}t_state;

/* Import aus derive_disp */
extern FILE *open_solve();    /* oeffnet Backend und initialisiert */
extern void update_solve();   /* neu berechneten Schritt in Backend anzeigen(in Datei schreiben) */
extern void close_solve();    /* schreiben der endgueltigen Ergebnisdatei */

/* Import aus /Shared/src/netio/load_net */
extern NetDescr *load_net();


int     	toptan,           /* Anzahl tangible Markings */
		nonnull,          /* Anzahl Nicht-Null-Elemente*/
                one_step,         /* =1 falls Scrittweite=1, sonst =0 */
                *ia,*ja,          /* Zeilen- bzw. Spaltenindizes der Nicht-Null-Eintraege */  
                drop_no,          /* Anzahl der aus mprob wegen zu geringer Groesse gestrichenen Eintraege */
                embedding,        /* Indikator fuer embedding */
                *c_matrix;        /* C0_matrix (embedding) */

double		*a,               /* Nicht-Null-Eintrage der EMC-Matrix */
		*mprob,           /* Ergebnisvektor  */
                *prod,            /* Hilfsvektor fuer Multiplikation */
                prec,             /* Praezision */
                *y_max,           /* maximale Anzeigengrenze */
                elapsed_time,     /* zur Zeitmessung */
                min_a,min_mprob;  /* Minimaleintraege von a und mprob */

double          n,stepsize,       /* auszufuehrende Schrittzahl, Schrittweite  */
                first_step;       /* erster auszugebener Schritt */
int             num,              /* Anzahl der Ausgaben */
                last=1,
                max_embedded=0;

t_state         *t_e_mprob;       /* Wahrscheinlichkeitsvektor fuer time efficient embedding */
m_state         *m_e_mprob;       /* Wahrscheinlichkeitsvektor fuer memory efficient embedding */
char            emb_opt[3];
t_stat_opts     stat_opts;
int             c_diag_sum=0;
NetDescr*       net;

static char     *netname;



/* fuer result measures: */
char    mpdname[100], tn_name[100], tpdname[100], rgname[100],  staname[100];

FILE    *tn_fp;     	/*  The DSPN description  */
FILE    *sta;     	/*  The file containing output statistics  */
FILE    *rg_fp;   	/*  The Tangible Reachability Set  */
FILE    *mpdf_fp; 	/*  Tangible Markings' Steady-state Probability  */
FILE    *tpdf_fp; 	/*  Tokens' PDF in each place  */

static struct
{	unsigned val;
}  rg;

int	no_places;
int	*notok;
double	tprob[maxtoken + 1];
int	now_marking;
double epsilon = 1.0e-16;
unsigned digits; /* number of digits to display (needed in $model_ERC.c) */

/****************************************************************************
PROZEDUREN
*****************************************************************************/

void read_p_matrix();
void read_c_matrix();
void read_vektor();
void test_read_vektor();
int  get_dimension();

void init_embedding();
void add_intermediate_states();

void vektor_matrix_mult();
void embedded_mult();
void column_sort();

void single_point_solution();
void derive_printed_solution();
void derive_solution();

double add_vektor_elements();
void clean();
void put_vektor();

void display_matrix();
void display_vektor();

/* fuer result measures: */
int act_marking();
void init();
void computeprob();


/****************************************************************************/
void main(argc,argv)
int argc;
char ** argv;
{
  int i,equal;
  double p;
  char buf[6];
  char pathless_netname[MAXPATHLEN];
  String tnethome;

  if (argc < 8)
   {
     fprintf(stderr,"wrong number of arguments");
     exit(1);
   }
  netname = argv[1];                      /* Lesen des Namens des Netzes */
  sscanf(argv[2],"%lf",&n);               /* Lesen der Schrittzahl */
  sscanf(argv[3],"%lf",&stepsize);        /* Lesen der Schrittweite */
  sscanf(argv[4],"%lg",&stat_opts.prec);
  sscanf(argv[4],"%lg",&prec);            /* Lesen der Praezision */
  sscanf(argv[5],"%lf",&first_step);      /* Lesen des ersten auszugebenen Schrittes */
  stat_opts.emb = argv[6];                /* Lesen der Option embedding */ 
  sscanf(argv[6],"%s",&emb_opt);          
  stat_opts.sol = argv[7];                /* Lesen der Loesungsmethode (fuer stationaere Lsg) */
  sscanf(argv[8],"%d",&stat_opts.maxiter);/* Lesen der maximalen Iterationsschritte (fuer stationaere Lsg) */
  sscanf(argv[9],"%d",&stat_opts.scale);

  first_step *= stat_opts.scale;
  n *= stat_opts.scale;
  /*  stepsize *= stat_opts.scale;*/

  if (strcmp(argv[6],"-en")==0)           /* kein embedding */
      embedding = 0;
  else
  { 
    if(strcmp(argv[6],"-et")==0)          /* time efficient embedding */
    {
      strcpy(buf,"time");
      embedding = 1;
    }
    else
      if(strcmp(argv[6],"-em")==0)        /* memory efficient embedding */
      {
	strcpy(buf,"memory");
	embedding = 2;
      }
      else
      {
	printf("wrong embedding option\n");
	exit(1);
      }
  }
  if (embedding)
    printf("Using %s efficient embedding technique\n",buf);

  if (first_step >= n) 
  {
    printf("The first visible timestep has to be smaller than the last\n");fflush(stdout);
    exit(-1);
  }


  ia=ja = (int*)NULL;
  a= (double*)NULL;
  net = load_net(netname,NULL);
  toptan = get_dimension(netname);      /* Lesen der Dimension der P-Matrix aus *.AUX-Datei */
  /* fuer result measures: */
  signal(SIGFPE,SIG_IGN);	
  epsilon = prec;
  for( digits = 0, p = epsilon; p < 1.0; digits++, p *= 10.0 );
  init(argc, argv);


  puts ("\nDeriving sparse representation of state transition probability matrix...\n"); fflush (stdout);
  read_p_matrix(netname);               /* Einlesen der P-Matrix aus *.EMC00-Datei */
  read_c_matrix(netname);               /* Einlesen der C0-Matrix aus *.CONV-Datei */
  read_vektor(netname);                 /* Einlesen bzw Initialisieren des Startvektors */

#ifdef MATRIXDEBUG
  display_matrix();
#endif
  if (embedding)
    init_embedding();
  if (stepsize != 0)
    derive_printed_solution();          /* Loesen des Systems mit oder ohne graphische Ausgabe */
  else single_point_solution();

#ifdef VEKTORDEBUG
 display_vektor();
#endif

  clean();                              /* Entfernen von Elementen, die kleiner als prec sind.*/
 
  (void)free((char *)a);                /* Speicher freigeben */
  (void)free((char *)ia);
  (void)free((char *)ja);
  free(notok);
  fclose(tn_fp);
  fclose(mpdf_fp);
  fclose(rg_fp);

  /*
  if ((tnethome = (String)getenv("TNETHOME")) == NULL)
  {
    printf("Environment variable TNETHOME not set; can't specified model\n");
    exit(-1);
  }
  equal = strlen(tnethome)+5;
  i = 0;
  while (netname[i+equal+1] != NULL)
  { 
    pathless_netname[i] = netname[i + equal+1];
    i++;
  }
  pathless_netname[i] =NULL;
  printf("\nIf you want to use the curves with gnuplot type:\n%s/DTSPN/TransSol/bin/prepare_gnu %s\n\n",tnethome,pathless_netname);
  fflush(stdout);
*/
  exit(0);
}

/*************************************************************************/
void derive_printed_solution()
{  
  int   i,j,k;
  FILE  *ftmp;
  double minX,maxX,t,ka;
  char  command[MAXPATHLEN];
  double waiting = 0.0;

  net = load_net(netname,NULL);
  if(net -> num_rewards == 0)
  {
    printf("No reward measures found. Define them first!\n");
    exit(1);
  }
  if ((stepsize*stat_opts.scale) >n)
  {
    printf("ERROR: stepsize has to be less than or equal to number of steps\n");
    exit(1);
  }
  /*  if ( (stepsize == 1.0) && (stat_opts.scale == 1) )*/
  if ( stepsize  == 1.0)
    one_step = 1;
  else
    one_step = 0;
  puts("Deriving solution...\n");fflush(stdout);
  i = n;
  num = 0;               /* Anzahl "num"  der Ausgaben in .tempcurves Datei berechnen: */
  while(i>first_step)
  {
    if (i>=stepsize)
      i -= stepsize;
    else
      i = 0;
    num++;
  }
  for(i=0;i<first_step;i++)       /* Berechnung der Schritte vor dem ersten auszugebenen Schritt */
  {
    waiting += nonnull*seconds;
    if (waiting > 2)              /* Ausgabe Lebenszeichen ca. alle 3s */
    {
      printf("*");fflush (stdout);
      waiting = 0.0;
    }
    if (!embedding)
      vektor_matrix_mult();
    else
    {
      embedded_mult();
    }
  }
printf("\nFirst visible step reached. Watch graphic now!\n");fflush (stdout);
  minX = (double)(first_step);
  maxX = (double)(n);
  ftmp = open_solve(netname,num,minX,maxX);
  sprintf(command,"/bin/rm -f %s.RESULTS",netname);
  system(command);
  t = (double)(first_step);
  if (embedding)
  {
    if (embedding == 1)
      if((mprob = (double*)malloc(toptan*sizeof(double)))==NULL)  /* Speicher allozieren */
      {
	perror("transient_sol ");
	exit(1);
      }
    add_intermediate_states();

  }
  update_solve(netname,ftmp,mprob,t,minX,maxX,prec,1); 
  if (embedding ==1)
    (void)free((double *)mprob);
  
  for (i=0;i<num;i++)
  {
    for (j = i*stepsize +1+first_step; (j <= (i+1)*stepsize+first_step )&( j<=n ); j++) /* schon j Schritte vorangegangen */
      if (!embedding)           
	vektor_matrix_mult();
      else
	embedded_mult();
    t = (double)(j-1);
    system(command);
    if (embedding)
      add_intermediate_states();
    update_solve(netname,ftmp,mprob,t,minX,maxX,prec,0);
    if (embedding ==1)
      (void)free((double *)mprob);
  }
  free_net(net);
  close_solve(netname,ftmp);
}
/**************************************************************************/
/**************************************************************************
  Name			: get_dimension

  Zweck			: Ermittlung der Dimension der P-matrix

  Eingabeparameter	: Modellname
    
  Rueckgabewert		: Dimension der P-Matrix
---------------------------------------------------------------------------*/
int get_dimension(netname)
char *netname;
{
  register FILE *aux_file;
  char filename[MAXPATHLEN], line[200];
  int dim;

  sprintf (filename, "%s.AUX", netname);      /* zu betrachtender FILE: "netname".AUX */
  if (!(aux_file = fopen (filename, "r")))
    {
      perror (NULL);
      fprintf (stderr, "can't open file '%s' for reading!\n", filename);
      exit (1);
    }
  do
    fgets (line, 198, aux_file);              /* Lese evtl. alles Belanglose, bis der Ausdruck "TANGIBLE MARKINGS" erscheint */
  while (!feof (aux_file) && strncmp (line, "TANGIBLE MARKINGS", 17));

  if (!strncmp (line, "TANGIBLE MARKINGS", 17))
    sscanf (line + 17, "%d", &dim);           /* hier die Dimension einlesen */
  else
    {
      fprintf (stderr, "file '%s' doesn't contain a line 'TANGIBLE MARKINGS <number>' !\n",
	       filename);
      exit (1);
    }
  fclose (aux_file);
  return dim;
}

/**************************************************************************
  Name                  : read_vektor

  Zweck                 : liest den Vektor der Anfangsverteilung ein

  Eingabeparameter      : Modellname

---------------------------------------------------------------------------*/
void read_vektor(netname)
char *netname;
{
  char v_filename[MAXPATHLEN];
  FILE *v_file;
  int i;

  if((mprob = (double*)malloc(toptan*sizeof(double)))==NULL)  /* Speicher allozieren */
    perror("transient_sol ");

  sprintf(v_filename, "%s.DTINI", netname);

  if (!(v_file = fopen(v_filename,"r")))
  { 
    perror (NULL);
    fprintf (stderr, "can't open file '%s' for reading!\n", v_filename);
    exit (1);
  }
  for(i=0;i<toptan;i++)                                       /* Vektor initialisieren */
    mprob[i] = 0.0;

  /* Header lesen */
  if (fread (&P_row_header, SIZEOF_P_row_header, 1, v_file) != 1)
  {
    if (ferror (v_file))
    {
      fprintf (stderr, "can't read file '%s' !\n", v_filename);
      perror (NULL);
      exit (1);
    }
    fclose(v_file);
  }
  else
  {
    for ( i=0; i<P_row_header.entries_in_row ; i++)
    {
      if (fread (&P_row_entry, (int)SIZEOF_P_row_entry,1,v_file) != 1)
      {
	fprintf (stderr, "can't read file '%s' !\n", v_filename);
	perror (NULL);
	exit (1);
      }                          /* Indexverschiebung beachten, array mprob beginnt bei 0 :*/ 
      mprob[(int)P_row_entry.column_index-1] = P_row_entry.probability;
    }
  }

  fclose(v_file);
}
/**************************************************************************
  Name                  : read_c_matrix

  Zweck                 : liest die C-Matrix (embedding) ein

  Eingabeparameter      : Modellname

---------------------------------------------------------------------------*/
/* Lesen der c_matrix (Diagonalmatrix), in der die Aufenhaltsdauer fuer    */
/*  jeden Zustand enthalten ist. Ohne Embedding also nur Einsen.           */
void read_c_matrix(netname)
char *netname;
{
  char c_filename[MAXPATHLEN];
  FILE *c_file;
  int i;

  if((c_matrix = (int*)malloc(toptan*sizeof(int)))==NULL)  /* Speicher allozieren */
    perror("transient_sol ");

  sprintf(c_filename, "%s.CONV00", netname);

  if (!(c_file = fopen(c_filename,"r")))
  { 
    perror (NULL);
    fprintf (stderr, "can't open file '%s' for reading!\n", c_filename);
    exit (1);
  }
  for(i=0;i<toptan;i++)                                       /* Vektor initialisieren */
    c_matrix[i] = 1;

  while(1)
  {
    if (fread((char*)&C0_entry , SIZEOF_C0_entry ,1 , c_file) != 1)
      if (ferror(c_file))
      {
	fprintf(stderr, "can't read file '%s' !\n", c_filename);
	perror (NULL);
	exit (1);
      }
      else
	break;
    c_matrix[C0_entry.state_number-1] = (int)C0_entry.sojourn_time;
  }
  for (i=0;i<toptan;i++)
  {
    c_diag_sum += c_matrix[i];
    if (c_matrix[i] > max_embedded)
      max_embedded = c_matrix[i];
  }

#ifdef MATRIXDEBUG
    printf("\n");
    for (i=0;i<toptan;i++)
      printf("%d ",c_matrix[i]);
    printf("\n");
#endif

  fclose(c_file);
}
/*************************************************************************/
void test_read_vektor(netname)        /* Erstellt fuer Testzwecke den ersten 
                                         Einheitsvektor der Dimension toptan */
char *netname;
{
  int i;

  if((mprob = (double*)malloc(toptan*sizeof(double)))==NULL)
    perror("transient_sol ");
  mprob[0] = 1.0;
  for (i=1; i< toptan; i++)
    mprob[i] = 0.0;
}
/*************************************************************************/
double add_vektor_elements(n,vektor)
int n;
double *vektor;
{
  double sum=0.0;
  int i;

  for (i=0;i<n;i++)
    sum += vektor[i];
  return(sum);
}
/**************************************************************************
  Name			: read_p_matrix

  Zweck			: liest die P-Matrix aus dem File <netname>.EMC00 in ia,ja,a

  Eingabeparameter	: Modellname
    
---------------------------------------------------------------------------*/
void read_p_matrix(netname)
char *netname;
{
  char emc_filename[MAXPATHLEN];
  FILE *emc_fp;
  register unsigned i;
  register unsigned row_nbr = 0;

#ifdef MATRIXDEBUG
 fprintf (stderr, "reading P-matrix\n");
#endif

  /* <netname>.EMC00-File lesen und dabei die Anzahl der Eintraege in jeder Spalte zaehlen: */
  sprintf (emc_filename, "%s.EMC00", netname);
  if (!(emc_fp = fopen (emc_filename,"r")))
  {
    perror (NULL);
    fprintf (stderr, "can't open file '%s' for reading!\n", emc_filename);
    exit (1);
  }

  /* Zunaechst einmal lesen und Eintraege zaehlen, um entsprechenden Speicherplatz zu allozieren.*/
  nonnull = 0;
  while(1)
  {
    /* Header lesen */
    if (fread (&P_row_header, SIZEOF_P_row_header, 1, emc_fp) != 1)   /* lese ersten P_row_header  */  
    {
      if (ferror (emc_fp))
      {
	fprintf (stderr, "can't read file '%s' !\n", emc_filename);
	perror (NULL);
	exit (1);
      }
      fclose(emc_fp);
      break;
    }
    row_nbr++;
    nonnull += P_row_header.entries_in_row;                            /* addiere #Eintraege auf */
    for ( i = nonnull - P_row_header.entries_in_row;                   /* fuer alle Eintraege in dieser Zeile */
	  i < nonnull;
	  i++)
    {
      if (fread (&P_row_entry, (int)SIZEOF_P_row_entry,1,emc_fp) != 1 )/* lese einen Eintrag */
      {
	fprintf (stderr, "can't read file '%s' !\n", emc_filename);
	perror (NULL);
	exit (1);
      }
    }
  }
#ifdef MATRIXDEBUG
    printf("nonnull= %d\n",nonnull);
    printf("toptan= %d\n",toptan);                      
#endif

  /* Speicher allozieren */
  if((a = (double*)malloc(nonnull*sizeof(double)))==NULL)
     {
       perror("transient_sol ");
       exit(1);
     }
  if((ia = (int*)malloc(nonnull*sizeof(int)))==NULL)
     {
       perror("transient_sol ");
       exit(1);
     }
  if((ja = (int*)malloc(nonnull*sizeof(int)))==NULL)
     {
       perror("transient_sol ");
       exit(1);
     }

  if (!(emc_fp = fopen (emc_filename,"r")))
  {
    perror (NULL);
    fprintf (stderr, "can't open file '%s' for reading!\n", emc_filename);
    exit (1);
  }

  /* Erneut lesen und Eintraege speichern. */
  nonnull = 0;
  while(1)
  {
    /* Header lesen */
    if (fread (&P_row_header, SIZEOF_P_row_header, 1, emc_fp) != 1)
    {
      if (ferror (emc_fp))
      {
	fprintf (stderr, "can't read file '%s' !\n", emc_filename);
	perror (NULL);
	exit (1);
      }
      fclose(emc_fp);
      break;
    }
    nonnull += P_row_header.entries_in_row;
    for ( i = nonnull - P_row_header.entries_in_row ; 
	  i < nonnull;
	  i++)
    {
      if (fread (&P_row_entry, (int)SIZEOF_P_row_entry,1,emc_fp) != 1)
      {
	fprintf (stderr, "can't read file '%s' !\n", emc_filename);
	perror (NULL);
	exit (1);
      }
      if (i==0) min_a = P_row_entry.probability;       /* minimales Element initialisieren */
      a[i] = P_row_entry.probability;
      ia[i] = (int)P_row_header.row_index;
      ja[i] = (int)P_row_entry.column_index;
      if (min_a > a[i])   min_a = a[i];                /* Minimum suchen */
    }
  }
}
/*********************************************************************/
void single_point_solution()
{

  derive_solution();
  if (embedding)
    add_intermediate_states();
  put_vektor(netname);
}
/***********************************************************************/
void derive_solution()
{
  int i;
  double waiting=0.0;

  puts("deriving solution(calculating single value)...\n");
  fprintf (stdout, "Estimated duration: %.0f sec\n*",(float)(n*(nonnull)*seconds));fflush (stdout);
  elapsed_time = time(NULL);
  for (i=1; i<= n;i++)
  {
    waiting += nonnull*seconds;
    if (waiting > 2)
    {
      printf("*");fflush (stdout);
      waiting = 0.0;
    }

    if (!embedding)
      vektor_matrix_mult();
    else
      embedded_mult();
  }
  elapsed_time -= time(NULL);
  printf("\nelapsed time: %.0f sec\n",-elapsed_time);
}
/******************************************************************************/
/* Sortiert P-Matrix nach Spalten                                             */
/* Vorgehen: Suche nacheinander fuer alle Array-Plaetze den richtigen Eintrag */
void column_sort()
{
  int i,j,k,l,found;
  double wert,waiting=0.0;

  printf("Transposing state transition probability matrix...\nEstimated duration: %.0f sec\n",nonnull*nonnull*seconds/2);fflush (stdout);
  for (k=0;k<nonnull;k++)
  {
    waiting += (nonnull-k)*seconds;
    if (waiting>2)
    {
      waiting=0.0;
      printf("*");fflush(stdout);
    }
    found = k;                            /* derzeitiges Element kann auf diesem Platz richtig sein */
    for (l=(k+1);l<nonnull;l++)           /* vergleiche mit allen nachstehenden Eintraegen */
      /* falls in gleicher Spalte vorher oder in vorheriger Spalte */
      if (((ja[l] == ja[found])&(ia[l] < ia[found]))|(ja[l] < ja[found]))
	found = l;                        /* ein Element gefunden, das vor das aktuell angenommene gehoert */
    if (k != found)
    {                                     /* Tauschen */
      wert = a[found];
      i = ia[found];
      j = ja[found];
      a[found] = a[k];
      ia[found] = ia[k];
      ja[found] = ja[k];
      a[k] = wert;
      ia[k] = i;
      ja[k] = j;
    }
  }
  puts("\n");
}
/*********************************************************************/
void vektor_matrix_mult()
{ 
  int i;

  if((prod = (double*)malloc(toptan*sizeof(double)))==NULL)  /* Speicher fuer Hilfsvektor allozieren */
  {
    perror("transient_sol ");
    exit(1);
  }
  for (i=0;i<toptan;i++)                             /* Hilfsvektor initialisieren */
    prod[i] = 0.0;


  for (i=0;i<nonnull;i++)                            /* laufe ueber alle Matrix-Elemente */
      prod[ ja[i]-1 ] += a[i]*mprob[ ia[i]-1 ];


  (void)free((double *)mprob);
  mprob = prod;
}
/*********************************************************************/
/*********************************************************************/
void embedded_mult()
{
  int i,ilast;

  if (embedding==1)   /* time efficient embedding method */
  {
    if((mprob = (double*)malloc(toptan*sizeof(double)))==NULL)  /* Speicher allozieren */
    {
      perror("transient_sol ");
      exit(1);
    }
    
    for (i=0;i<toptan;i++)                  /* Vektor initialisieren */
      mprob[i] = 0.0;
    
    for (i=0;i<nonnull;i++)
    {
      ilast = last-c_matrix[ia[i]-1];       /* Bestimmung des intermediate Vektors, der die Wahrscheinlichkeit
                                               enthaelt, aus Zustand ia[i]-1 herauszukommen. */
      if (ilast<0)
	ilast += max_embedded;
      mprob[ ja[i]-1 ] += a[i]*t_e_mprob[ ilast ].intermediate[ ia[i]-1 ];
    }
    
    (void)free((double *)t_e_mprob[last].intermediate);     /* intermediate Vektor last jetzt veraltet */
    t_e_mprob[last].intermediate = mprob;
    if (last < max_embedded-1 )             /* last weitersetzen */
      last ++;
    else
      last = 0;
  }
  else               /* memory efficient embedding method */
  {
    for (i=0;i<toptan;i++)         
      mprob[i] = 0.0;

 /* laufe ueber alle Matrixelemente und addiere in entsprechender Spalte von mprob das Produkt mit entsprechender 
    Spalte von e_mprob last: */   
    for (i=0;i<nonnull;i++)      
      mprob[ ja[i]-1 ] += a[i]*m_e_mprob[ ia[i]-1 ].intermediate[ m_e_mprob[ ia[i]-1 ].last ];
    
    for (i=0;i<toptan;i++)
    {
      m_e_mprob[i].intermediate[m_e_mprob[i].last] = mprob[i]; /* letzte Eintraege setzen */
      if (m_e_mprob[i].last < c_matrix[i]-1)                   /* last weitersetzen */
	m_e_mprob[i].last++;
      else
	m_e_mprob[i].last =0;
    }
  }
}
/************************************************************************/
/* Nur fuer Embedding:
   Falls die Werte aus mprob wirklich gebraucht werden, muessen die einge-
   betteten Zustaende beruecksichtigt werden, denn mprob enthaelt bei Embedding
   nur die Wahrscheinlichkeit, gerade in diesen Zustand gelangt zu sein. 
   Da die Verweildauer in Zustand i c_matrix[i] Einheiten betraegt, muessen 
   die letzten c_matrix[i] intermediate Eintraege aufaddiert werden.*/
void add_intermediate_states()
{
  int i,j,jlast;
  double sum = 0.0;

  if (embedding==1)
  {
    if ((mprob = (double*)malloc(toptan*sizeof(double)))==NULL)
    {
      perror("transient_sol");
      exit(1);
    }
    for (i=0;i<toptan;i++)
    {
      mprob[i] = 0.0;
      for (j=0;j<c_matrix[i];j++)       
      {
	jlast = last -1 -j;
	if (jlast <0)
	  jlast += max_embedded;
	mprob[i] += t_e_mprob[jlast].intermediate[i];
      }
    }
  }
  else
  {
    for (i=0;i<toptan;i++)
      if (c_matrix[i] >1)
	for (j=0;j<c_matrix[i];j++)
	  if (j!=((m_e_mprob[i].last + c_matrix[i] -1)%c_matrix[i]))
	    mprob[i] += m_e_mprob[i].intermediate[j];
  }
#ifdef VEKTORDEBUG
  display_vektor();
#endif
}
/*********************************************************************/
void init_embedding()
{
  int i,j,first;
  
  if (embedding==1)
  {
    if (stat_opts.scale == 1.0)
      last = 0;
    if ((t_e_mprob = (t_state*)malloc(max_embedded*sizeof(t_state)))==NULL)
    {
      perror("transient_sol");
      exit(1);
    }
    for (i=0;i<max_embedded;i++)
    {
      if ((t_e_mprob[i].intermediate = (double*)malloc(toptan*sizeof(double)))==NULL)
      {
	perror("transient_sol");
	exit(1);
      }
      for (j=0;j<toptan;j++)
	t_e_mprob[i].intermediate[j] = 0.0;
    }
    for (i=0;i<toptan;i++)
      t_e_mprob[0].intermediate[i] = mprob[i];
  }
  else
  {
    if ((m_e_mprob = (m_state*)malloc(toptan*sizeof(m_state)))==NULL)
    {
      perror("transient_sol");
      exit(1);
    }
    for (i=0;i<toptan;i++)
    {
      if ((m_e_mprob[i].intermediate = (double*)malloc(c_matrix[i]*sizeof(double)))==NULL)
      {
	perror("transient_sol");
	exit(1);
      }
      for (j=1;j<c_matrix[i];j++)
	m_e_mprob[i].intermediate[j] = 0.0;
      m_e_mprob[i].intermediate[0] = mprob[i];
      if (c_matrix[i] == 1)
	m_e_mprob[i].last = 0;
      else
	m_e_mprob[i].last = 1;
    }
  }
}
/************************************************************************/
/* Raeumt alle Elemente, die kleiner als 1/10-tel Praezision sind, raus.*/
void clean()
{
  int i;
  min_mprob = 1.0;
  drop_no = 0;
  for (i=0;i<toptan;i++)
  {
    if ((mprob[i]!=0.0)&&(min_mprob > mprob[i]))
      min_mprob = mprob[i];
    if (mprob[i] < prec*0.1)
    {
      mprob[i] = 0.0;
      drop_no++;
    }
  }
  fprintf(stdout,"\nSmallest entry in P: %.*g\nSmallest entry in mprob: %.*g\nNumber of dropped elements: %d\n\n",digits,min_a,digits,min_mprob,drop_no);fflush(stdout);
}
/************************************************************************/
/* Ausgabe der Praezision und noch eines doubles vor dem Loesungsvektor */ 
/* in Datei netname.PROBMARK                                            */
void put_vektor(netname)
char *netname;
{
 
  char out_filename[MAXPATHLEN];
  FILE *out_file;
  double buf;

  sprintf(out_filename, "%s.PROBMARK", netname);
  if (!(out_file=fopen(out_filename,"w")))
  {
    fprintf (stderr, "can't open file '%s' !\n", out_filename);
    perror (NULL);
    exit (1);
  }
  buf = -prec;
  if (fwrite(&buf,sizeof(double),1,out_file)!=1)
  {
    fprintf (stderr, "can't write file '%s' !\n", out_filename);
    perror (NULL);
    exit (1);
  }
  buf = -1;
  if (fwrite(&buf,sizeof(double),1,out_file)!=1)
  {
    fprintf (stderr, "can't write file '%s' !\n", out_filename);
    perror (NULL);
    exit (1);
  }

  if (fwrite(mprob,sizeof(double),toptan,out_file)!=toptan)  /* Ausgabe Loesungsvektor */
  {
    fprintf (stderr, "can't write file '%s' !\n", out_filename);
    perror (NULL);
    exit (1);
  }
  fclose(out_file);
}

/************************************************************************/
void display_matrix()
{
    int i;

  printf("\n");
  for (i=0;i<nonnull;i++)
    printf("%d. zeile : %d spalte : %d wert : %.20f\n",i,ia[i],ja[i],a[i]);
  printf("\n");
}
/************************************************************************/
void display_vektor()
{
  int i;

  printf("\n");
  for(i=0;i<toptan;i++)
    printf(" %.10f \n",mprob[i]);
  printf("\n");
}
/************************************************************************/
/* Prozeduren fuer result measures:                                     */
/************************************************************************/
/* The following procedure gives the number of the actual marking       */
int
act_marking(pla)
	int	pla;
{
	return notok[(now_marking-1)*no_places + pla];
}
/************************************************************************/
/*  The following procedure initializes the program data structures     */
void
init(argc, argv)
	int	argc;
	char	*argv[];
{ 
	char	buf[50];
	int	i, akttan;
   
        strcpy(tn_name, argv[1]);      	strcat(tn_name, ".TN");
	strcpy(mpdname, argv[1]);	strcat(mpdname, ".PROBMARK");
	strcpy(tpdname, argv[1]);	strcat(tpdname, ".PROBTOK");
	strcpy(rgname,  argv[1]);	strcat(rgname,  ".RG");
	strcpy(staname, argv[1]);	strcat(staname, ".RESULTS");
	if ((tn_fp = fopen(tn_name, "r")) == NULL)
	{	fprintf(stderr, "Can't open file %s for reading\n", tn_name);
		exit(1);
	}
	no_places = 0;
	while(!no_places)
	{	fscanf(tn_fp, "%s", buf);
		if (strcmp(buf, "PLACES:"))
			continue;
		fscanf(tn_fp, "%s", buf);
		no_places = atoi(buf);
		if (feof(tn_fp))
			exit(1);
	}

	notok = (int *)calloc(toptan*no_places+1, sizeof(int));
	if(notok == NULL)
	{	fprintf(stderr, "\nERROR in gst: not enough memory for notok[]\n");
		exit(1);
	}
	/* read net.rg to fill notok[] */
	if ((rg_fp = fopen(rgname, "r")) == NULL)
	{	fprintf(stderr, "Can't open file %s for reading\n", rgname);
		exit(1);
	}
	akttan = 0;
	while (!feof(rg_fp))
	{	for (i = 0; i <= (no_places-1); i++)
		{	if (fread(&rg, sizeof(rg), 1, rg_fp) != 1)
				if (ferror(rg_fp))
				{	fprintf(stderr, "\nERROR: on reading to file %s !\n",
									 rgname);
					exit(1);
				}
			if (feof(rg_fp))
				break;
			notok[akttan*no_places+i] = rg.val;
		} /*  for  */
		if (feof(rg_fp))
			break;
		akttan++;
	}
} /* init  */
/*************************************************************************/
/*	compute the token probabilities                                  */
/* wird zur Zeit nicht benutzt */
void computeprob(res_vec)
double *res_vec;
{
	int	i,j,k;
	int	maxpt[maxplace+1], minpt[maxplace+1];
        FILE *tpdf_fp;
        double tpdf;

        sprintf(tpdname,"%s.PROBTOK",netname);
        if ((tpdf_fp = fopen(tpdname, "w")) == NULL)
		{	printf("Can't open file %s for writing\n", tpdname);
			exit(1);
		}
	for (i = 1; i <= no_places; i++)
	{	for (j = 0; j <= maxtoken; j++)
			tprob[j] = 0.0;
		maxpt[i] = 0;
		minpt[i] = maxtoken;
		for (j = 1; j <= toptan; j++)
		{	k = notok[(j-1)*no_places+i-1];
			if (k > maxpt[i])
				maxpt[i] = k;
			if (k < minpt[i])
				minpt[i] = k;
			tprob[k] = tprob[k] + mprob[j-1];
		}
		tpdf = (double)minpt[i];
		fwrite((char *)&tpdf, sizeof(double), 1, tpdf_fp);
		tpdf = (double)maxpt[i];
		fwrite((char *)&tpdf, sizeof(double), 1, tpdf_fp);
		for (j = minpt[i]; j <= maxpt[i]; j++)
		{	tpdf = tprob[j];
			fwrite((char *)&tpdf, sizeof(double), 1, tpdf_fp);
		}
	}
	fclose(tpdf_fp);
}

