/*******************************************************************************
Modul   	:	NWAppActions.c

Purpose		: 	Globale Aktionen der Klasse NetWidget, die nicht "uber
                        "Ubersetzungstafel registriert sind

Project Director:	Christoph Lindemann TU Berlin

Programmer	:	Martin M"uller

History		:	release 1: October 1992
*******************************************************************************/


/* NWAppActions.c: Action routines called from Application 
 *
 * NetWidget.c - Manages all objects of a PN.

   Class Hierarchie:
   Core -> Composite -> NetWidget
 */

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <stdio.h>
#include <string.h>

#include "BaseObj.h"
#include "MarkParObj.h"
#include "RateParObj.h"
#include "PlaceObj.h"
#include "TransObj.h"
#include "ArcObj.h"

/* die Abhaengigkeit zur Aussenwelt (global.h) sollte abgeschaft werden
   Communikation ueber CallBack (XtSetSensitive)?? */

#include "NetWidgetP.h"

/*******************************************************************************
Name:
	ClearNet
--------------------------------------------------------------------------------
Funktionstyp:
	globale Aktion
--------------------------------------------------------------------------------
triggert bei event:
	Wird mittels XtCallActionProc aus der Applikation aufgerufen
--------------------------------------------------------------------------------
Zweck:
	Loeschen des Bildfenster und zerstoeren des aktuellen Modelles
*******************************************************************************/

void ClearNet(w,event,params,numparams)
Widget w;
XButtonEvent* event;
String* params;
Cardinal* numparams;

{
NetWidget nw = (NetWidget)w;
int i;

#ifdef debug
printf("Clear \n");
#endif

  for (i=0;i<nw->composite.num_children;i++) 
	if(isArcObject(nw->composite.children[i]))
		XtDestroyWidget(nw->composite.children[i]);
  for (i=0;i<nw->composite.num_children;i++) 
	XtDestroyWidget(nw->composite.children[i]);
  XFillRectangle(XtDisplay(nw),nw->netWidget.picture,nw->netWidget.clear_GC,
		0,0,nw->core.width,nw->core.height);
  SetModifiedFlag(nw,True);
  SetNetModifiedFlag(nw,True);
  nw->netWidget.tagsVisible = False;
  nw->netWidget.ratesVisible = False;
  nw->netWidget.spline = False;
  nw->netWidget.gridsize = Off;
  nw->netWidget.rescale_factor = OneZero;

  nw->netWidget.selectedObject = NULL;
  nw->netWidget.actionTerminated = True;
  nw->netWidget.allowMotionAction = False;
  nw->netWidget.ignoreButtonEvents = False;
  nw->netWidget.results_collected = False;
  nw->netWidget.throughput_computed = False;
  nw->netWidget.selected_intermediate_point = 0;
  nw->netWidget.prev_x = -1;
  nw->netWidget.prev_y = -1;
  nw->netWidget.currTransOrient = ZERO;
  nw->netWidget.rescale_value = 1.0;
  Redisplay(w,NULL);
}

/*******************************************************************************
Name:
	Redraw
--------------------------------------------------------------------------------
Funktionstyp:
	globale Aktion
--------------------------------------------------------------------------------
triggert bei event:
	Wird mittels XtCallActionProc aus der Applikation aufgerufen
--------------------------------------------------------------------------------
Zweck:
	Neuzeichnen des Modelles
*******************************************************************************/

void Redraw(w,event,params,numparams)
Widget w;
XButtonEvent* event;
String* params;
Cardinal* numparams;

{
  NetWidget nw = (NetWidget)w;

  ClearNetPixmap(nw);
  Drawgrid(w,Off,nw->netWidget.gridsize);
  DrawNet(w,NORMAL);
  if (nw->netWidget.tokenGame)
    showEnabledTranss(nw);
  if (nw->netWidget.object == COMPLETE_SELECTION)
      undrawSelectionBorder(nw); /* don't be confused - it DRAWS it */
  
  Redisplay(w,NULL);
}

/*******************************************************************************
Name:
	get_MD_RParamDefinitions
--------------------------------------------------------------------------------
Funktionstyp:
	public function
--------------------------------------------------------------------------------
benutzt in:
	checkSyntax.c
--------------------------------------------------------------------------------
Parameter:
	w: 	Aktuelle Instanz des Modells
    Ausgabe:
	params: Array in denen die vorhanden Parameter gesammelt werden
	numresults: Groesse des Arrays
--------------------------------------------------------------------------------
Zweck:
	Sammeln der Namen und Definition der MD-RateParameter-Definitionen
--------------------------------------------------------------------------------
Bemerkung:
	rparams[i] := name
	rparams[i+1] := definition
	numrparams := 2* Anzahl Parameter Objekte

  	Aufrufende Funktion muss Speicher mit XtFree freigeben.
*******************************************************************************/

#ifdef __STDC__
void get_MD_RParameters(Widget w, String **rparams, Cardinal *numrparams)
#else
void get_MD_RParameters(w, rparams, numrparams)
     Widget w;
     String **rparams;
     Cardinal * numrparams;
#endif
{
  NetWidget nw = (NetWidget)w;
  String *rec;
  Cardinal num;
  String name;
  int i, mds;  /* Marking Defined RateParameterS */
  RateValue *rate;
  
#ifdef debug
  printf("get_MD_RParameters \n");
#endif

  /*
   * first allocate for all
   */
  num = 2*nw->netWidget.numratepar;
  rec = (String*)XtMalloc(num * sizeof(String));

  /*
   * read in 
   */
  for(i=FirstRatePar(nw) , mds = 0;
      i<LastRatePar(nw); 
      i++) {
	XtVaGetValues(nw->composite.children[i],
			XtNrateParName,&name,
			XtNrateParValue,&rate,
			NULL);
	
	if (rate->valueType == Definition) {
	  rec[mds++] = XtNewString(name);
	  rec[mds++] = XtNewString(rate->values.definition);
	}
  }

  *rparams = rec;
  *numrparams = mds; 
}

/*******************************************************************************
Name:
        DeleteRParamByName
--------------------------------------------------------------------------------
Funktionstyp:
	public function
--------------------------------------------------------------------------------
benutzt in:
	checkSyntax.c
--------------------------------------------------------------------------------
Parameter:
	w: 	Aktuelle Instanz des Modells
	name:   Name des zu loeschenden RateParameters
    Ausgabe: -
--------------------------------------------------------------------------------
Zweck:
	Loeschen des angegebenen Raten Parameters
--------------------------------------------------------------------------------
Bemerkung:
*******************************************************************************/
void DeleteRParamByName(w, name)
     Widget w;
     String name;
{
    NetWidget nw = (NetWidget)w;
    Widget rpar = NULL;
    Rate *rate_buf, rate;
    String s;
    int i, x, y;

    if (!name)
    {
	fprintf(stderr, "DeleteRParamByName() : given name is NULL.\n");
	return;
    }

    /* 
     * TRY TO FIND 
     */ 
    for (i=FirstRatePar(nw); i<LastRatePar(nw); i++)
    {
	XtVaGetValues(nw->composite.children[i],
		      XtNrateParName, &s,
		      NULL);
	if (!strcmp(s, name)) 
	{
	    rpar = nw->composite.children[i];
	    break;
	}   
    }

    /* 
     * FOUND ? 
     */
    if (rpar == NULL) 
    {
	fprintf(stderr, "DeleteRParamByName() : No rate named '%s' found.\n",
		name);
	return;
    }
    
    /* 
     * REMOVE THAT RATE FROM ANY TRANSITION 
     */
    for (i=FirstTrans(nw);i<LastTrans(nw);i++)
    {
        XtVaGetValues(nw->composite.children[i],
		      XtNrates,&rate_buf,
		      NULL);

	if (rate_buf->type != RATEPARAM)
	    continue;

	if ((Widget)rate_buf->values.rateparam == rpar)
	{
	    rate.type = RATEVALUE;
	    rate.values.ratevalue = 1.0;
	    if (nw->netWidget.ratesVisible)
		DrawRate(nw->composite.children[i],
			 XtDisplay(nw),
			 nw->netWidget.picture,
			 nw->netWidget.draw_GC);

	    XtVaSetValues(nw->composite.children[i],
			  XtNrates,&rate,
			  NULL);
	    if (nw->netWidget.ratesVisible)
		DrawRate(nw->composite.children[i],
			 XtDisplay(nw),
			 nw->netWidget.picture,
			 nw->netWidget.draw_GC);
	}
    } 
    
    /*
     * DELETE IT
     */
    XtVaGetValues(rpar,
		  XtNxpos, &x,
		  XtNypos, &y,
		  NULL);
    DrawRatePar(rpar, XtDisplay(nw), 
		nw->netWidget.picture, nw->netWidget.draw_GC,
		x,y, nw->netWidget.object_size, 0);
    Redisplay(w,NULL);

    XtDestroyWidget(rpar);

    SetModifiedFlag(nw,True);
    SetNetModifiedFlag(nw,True);
    
}
 
/*******************************************************************************
Name:
        TestRParamName
--------------------------------------------------------------------------------
Funktionstyp:
	public function
--------------------------------------------------------------------------------
benutzt in:
	checkSyntax.c
--------------------------------------------------------------------------------
Parameter:
	w: 	Aktuelle Instanz des Modells
	name:   Name des zu loeschenden RateParameters
    Ausgabe: ok : 0; sonst 1;
--------------------------------------------------------------------------------
Zweck:
	Ueberpruefen, ob der angegebene Name schon vergeben ist
--------------------------------------------------------------------------------
Bemerkung:
*******************************************************************************/
int TestRParamName(w, name)
     Widget w;
     String name;
{
    NetWidget nw = (NetWidget)w;
    int i;
    String RPname;
    char msg[100];

#ifdef debug
    fprintf(stderr, "TestRParamName for \"%s\".\n", name);
#endif

    for (i = FirstRatePar(nw); i<LastRatePar(nw); i++)
    {
	XtVaGetValues(nw->composite.children[i],
		      XtNrateParName,&RPname,
		      NULL);
#ifdef debug
	fprintf(stderr, "TestRParamName whith \"%s\".\n", RPname);
#endif
	if (strcmp(RPname,name) == 0)
	{
	    sprintf(msg, "The Delay/Weightparameter \"%s\" already exists.",
		    name);
	    XtWarning(msg);
	    return 1;
	}
    }
    return 0;
}
