/*******************************************************************************
Modul   	:	NWGrid.c

Purpose		: 	Routinen zum Aufbau und adjustieren der Objekte an 
                        ein Gitternetz

Project Director:	Christoph Lindemann TU Berlin

Programmer	:	Martin M"uller

History		:	release 1: October 1992
*******************************************************************************/

/* NWAppActions.c: Action routines called from Application 
 *
 * NetWidget.c - Manages all objects of a PN.

   Class Hierarchie:
   Core -> Composite -> NetWidget
 */

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <stdio.h>
#include <string.h>

#include "NetWidgetP.h"

static Cardinal GridsizeToDistance();

static GC grid_gc;

/*******************************************************************************
Name:
	Init_Grid
--------------------------------------------------------------------------------
Funktionstyp:
	public function
--------------------------------------------------------------------------------
benutzt in:
	main.c
--------------------------------------------------------------------------------
Parameter:
	w:	aktuelle Instanz des Modelles
--------------------------------------------------------------------------------
Zweck:
	Initalisierung des graphischen Kontextes zum Zeichnen des Gitters
*******************************************************************************/
void Init_Grid(w)
Widget w;
{
NetWidget nw = (NetWidget)w;
XGCValues values;
XtGCMask mask;

  values.line_style = LineOnOffDash;
  values.foreground = 1;
  values.background = 0;
  mask = GCLineStyle | GCForeground | GCBackground ;

  grid_gc = XCreateGC(XtDisplay(nw),nw->netWidget.picture,mask,&values);

  XSetFunction(XtDisplay(nw),grid_gc,GXxor);

}

/*******************************************************************************
Name:
	Free_Grid
--------------------------------------------------------------------------------
Funktionstyp:
	public function
--------------------------------------------------------------------------------
benutzt in:
	Destroy-Methode der Klasse NetWidget
--------------------------------------------------------------------------------
Parameter:
	w:	aktuelle Instanz des Modelles	
--------------------------------------------------------------------------------
Zweck:
	Freigeben des graphischen Kontextes
*******************************************************************************/
void Free_Grid(w)
Widget w;
{
NetWidget nw = (NetWidget)w;

  XFreeGC(XtDisplay(nw),grid_gc);
}

/*******************************************************************************
Name:
	Drawgrid
--------------------------------------------------------------------------------
Funktionstyp:
	public function
--------------------------------------------------------------------------------
benutzt in:
	SetValues-Methode der Klasse NetWidget
--------------------------------------------------------------------------------
Parameter:
	w:	aktuelle Instanz des Modelles
	oldsize: alte Gittergroesse
	newsize: neue Gittergroesse
--------------------------------------------------------------------------------
Zweck:
	Loeschen des alten Gitters, zeichnen des neuen Gitters
*******************************************************************************/

void Drawgrid(w,oldsize,newsize)
Widget w;
Gridsize oldsize,newsize;
{
NetWidget nw = (NetWidget)w;
Cardinal distance ;
int dash_offset = 0;
char griddash[2];
XSegment *segments;
int num_segments,i;
XGCValues values;

  /*undraw old*/
  if (oldsize != Off)
  {
	values.foreground = 0;
	values.background = 1;
/*	XChangeGC(XtDisplay(w),grid_gc,GCForeground | GCBackground,&values);*/
  	distance = GridsizeToDistance(oldsize);
  	griddash[1] = (char)distance - 1;
  	griddash[0] = '\001';
  	XSetDashes(XtDisplay(nw),grid_gc,dash_offset,griddash,2);
  	num_segments = nw->core.height/(int)distance;
  	segments = (XSegment*)XtMalloc(num_segments*sizeof(XSegment));
  	segments[0].x1 = 0;
  	segments[0].x2 = nw->core.width;
  	segments[0].y1 = distance;
  	segments[0].y2 = distance;
  	for (i=1;i<num_segments;i++)
  	{
		segments[i].x1 = segments[i-1].x1;
		segments[i].x2 = segments[i-1].x2;
		segments[i].y1 = segments[i-1].y1 + distance;
		segments[i].y2 = segments[i-1].y2 + distance;
  	}
  	XDrawSegments(XtDisplay(nw),nw->netWidget.picture,grid_gc,
		      segments,num_segments);
	XtFree((char*)segments);
  }
  /*draw new*/
  if (newsize != Off)
  {
	values.foreground = 1;
	values.background = 0;
/*	XChangeGC(XtDisplay(w),grid_gc,GCForeground | GCBackground,&values);*/
  	distance = GridsizeToDistance(newsize);
  	griddash[1] = (char)distance - 1;
  	griddash[0] = '\001';
  	XSetDashes(XtDisplay(nw),grid_gc,dash_offset,griddash,2);
  	num_segments = nw->core.height/(int)distance;
  	segments = (XSegment*)XtMalloc(num_segments*sizeof(XSegment));
  	segments[0].x1 = 0;
  	segments[0].x2 = nw->core.width;
  	segments[0].y1 = distance;
  	segments[0].y2 = distance;
  	for (i=1;i<num_segments;i++)
  	{
		segments[i].x1 = segments[i-1].x1;
		segments[i].x2 = segments[i-1].x2;
		segments[i].y1 = segments[i-1].y1 + distance;
		segments[i].y2 = segments[i-1].y2 + distance;
  	}
  	XDrawSegments(XtDisplay(nw),nw->netWidget.picture,grid_gc,
		      segments,num_segments);
	XtFree((char*)segments);
  }
/* dont do it here - but later */
/*  Redisplay(w,NULL); */ 
}

/*******************************************************************************
Name:
	AdjustEventToGrid
--------------------------------------------------------------------------------
Funktionstyp:
	public function
--------------------------------------------------------------------------------
benutzt in:
	Aktionen des NetWidgets:
	AddArc_2(), AddPlace(), AddImTrans(), AddExpTrans(), AddDetTrans(),
	AddGenTrans(), AddMarkPar(), AddRatePar(); 
 	MoveNode(), MoveNode_2(), MoveTag(), MoveTag_2(), MoveRates(), 
	MoveRates_2(), MoveParam(), MoveParam_2(), MoveArc(), MoveArc_2();
	NodeMotion(), TagMotion(), ParamMotion(), RateMotion(), ArcMotion();
--------------------------------------------------------------------------------
Parameter:
	w:	aktuelle Instanz des Modelles
	e_x,e_y: x,y Koordinaten des erhaltenen Ereignisses
    Ausgabe:
	g_x,g_y: x,y Koordinaten des berechneten Gitterpunktes
--------------------------------------------------------------------------------
Zweck:
        Adjustiert Ereigniskoordianten zum n"achsten Gitterpunkt
*******************************************************************************/
void AdjustEventToGrid(w,e_x,e_y,g_x,g_y)
Widget w;
int e_x,e_y,*g_x,*g_y;
{
NetWidget nw = (NetWidget)w;
Cardinal distance = GridsizeToDistance(nw->netWidget.gridsize);
Cardinal x_factor,y_factor;
int ret_x,ret_y;

  if (distance == 0)
  {
	*g_x = e_x;
	*g_y = e_y;
	return;
  }
  x_factor = (Cardinal)abs(e_x)/distance;
  y_factor = (Cardinal)abs(e_y)/distance;
  
  if (x_factor > 0)
  {
	ret_x = abs(distance*x_factor - e_x) < abs(distance*(x_factor+1) - e_x)
		?
		(int)distance*x_factor
		:
		(int)distance*(x_factor+1);
  }
  else
	ret_x = distance;

  if (y_factor > 0)
  {
	ret_y = abs(distance*y_factor - e_y) < abs(distance*(y_factor+1) - e_y)
		?
		(int)distance*y_factor
		:
		(int)distance*(y_factor+1);
  }
  else
	ret_y = distance;

  *g_x = ret_x;
  *g_y = ret_y;
}

/*******************************************************************************
Name:
	GridsizeToDistance
--------------------------------------------------------------------------------
Funktionstyp:
	private function
--------------------------------------------------------------------------------
benutzt in:
	Drawgrid(), AdjustEventToGrid()
--------------------------------------------------------------------------------
Parameter:
	gridsize: gittergroesse als Wert des Aufzaehlungstypen 'Gridsize'
--------------------------------------------------------------------------------
Rueckgabewert:
	Gittergroesse als Wert in Pixel
--------------------------------------------------------------------------------
Zweck:
	Konvertieren des Datentyps 'Gridsize' zu Pixelanzahl
*******************************************************************************/

static Cardinal GridsizeToDistance(gridsize)
Gridsize gridsize;
{
  switch (gridsize)
  {
    case Off: 	  	
	return  0;break;
    case Five: 	  	
	return  5;break;
    case Ten: 	  
	return 10;break;
    case Fifteen: 
	return 15;break;
    case Twenty:  
	return 20;break;
    case Twentyfive:  
	return 25;break;
    case Thirty:  
	return 30;break;
    case Thirtyfive:  
	return 35;break;
    case Forty:  
	return 40;break;
    default: return 0;
  }
}
