/*******************************************************************************
Modul   	:	ArcObj.c

Purpose		: 	Realisation der Klasse ArcObject

Project Director:	Christoph Lindemann TU Berlin

Programmer	:	Martin M"uller

History		:	release 1: October 1992
*******************************************************************************/

	
/*
 * ArcObj.c 

   Hierarchie:
   Object -> Composite -> ArcObj
 */

#include <math.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <stdio.h>

#include "util.h"
#include "TransObj.h"
#include "PlaceObj.h"

#include "ArcObjP.h"

#define offset(field) XtOffsetOf(ArcObjRec, field)

static XtResource resources[] = {
	{
	XtNfromObj,
	XtCfromObj,
	XtRPointer,
	sizeof(XtPointer),
	offset(arcObj.fromObj),
	XtRPointer,
	NULL
	},
	{
	XtNtoObj,
	XtCtoObj,
	XtRPointer,
	sizeof(XtPointer),
	offset(arcObj.toObj),
	XtRPointer,
	NULL
	},
	{
	XtNarcType,
	XtCarcType,
	XtRarcType,
	sizeof(ArcType),
	offset(arcObj.arcType),
	XtRImmediate,
	(XtPointer) 0
	},
	{
	XtNmultiplicity,
	XtCmultiplicity,
	XtRPointer,
	sizeof(XtPointer),
	offset(arcObj.multiplicity),    
	XtRPointer,
	NULL
	},
	{
	XtNdrawingDirection,
	XtCdrawingDirection,
	XtRdrawingDirection,
	sizeof(DrawingDirection),
	offset(arcObj.drawing_direction),
	XtRImmediate,
	(XtPointer) UP
	},
};

/* Declaration of Core methods */

static void ClassInitialize();
static void Initialize();
static void Destroy();
static Boolean SetValues();


/* these Core methods not needed by ArcObj:
 *
 * static void Realize();
 */

static Cardinal insertPosition();

/* the following are private functions unique to ArcObj */
static Boolean is_under();

/* The following are public functions of ArcObj, declared extern
 * in the public include file: */

CompositeClassExtensionRec arcObjExtension = {
	/* next extension       */ NULL,
	/* record type          */ NULLQUARK,
	/* version              */ XtCompositeExtensionVersion,
	/* record size          */ sizeof(CompositeClassExtensionRec),
	/* accepts_objects      */ True,
	};

CompositeClassExtension arcObjExtensionPtr = &arcObjExtension;


ArcObjClassRec arcObjClassRec = {
    {
    /* object_class fields */
    /* superclass	  	 */ (WidgetClass) &compositeClassRec,
    /* class_name	  	 */ arcClassName,
    /* widget_size	  	 */ sizeof(ArcObjRec),
    /* class_initialize   	 */ ClassInitialize,
    /* class_part_initialize	 */ NULL,
    /* class_inited       	 */ FALSE,
    /* initialize	  	 */ Initialize,
    /* initialize_hook		 */ NULL,
    /* realize		  	 */ NULL,
    /* actions		  	 */ NULL,
    /* num_actions	  	 */ 0,
    /* resources	  	 */ resources,
    /* num_resources	  	 */ XtNumber(resources),
    /* xrm_class	  	 */ NULLQUARK,
    /* compress_motion	  	 */ 0,
    /* compress_exposure  	 */ 0,
    /* compress_enterleave	 */ 0,
    /* visible_interest	  	 */ 0,
    /* destroy		  	 */ Destroy,
    /* resize		  	 */ NULL,
    /* expose		  	 */ NULL,
    /* set_values	  	 */ SetValues,
    /* set_values_hook		 */ NULL,
    /* set_values_almost	 */ NULL,
    /* get_values_hook		 */ NULL,
    /* accept_focus	 	 */ NULL,
    /* version			 */ XtVersion,
    /* callback_private   	 */ NULL,
    /* tm_table		   	 */ NULL,
    /* query_geometry		 */ NULL,
    /* display_accelerator       */ NULL,
    /* extension                 */ NULL
    },
    {
   /* composite fields */
   /* geometry_manager         */ NULL,
   /* change_managed           */ NULL,
   /* insert_child             */ XtInheritInsertChild,
   /* delete_child             */ XtInheritDeleteChild,
   /* extension structure      */ NULL,
    },
    {
    /* dummy               */ 0,
    },
};

WidgetClass arcObjClass = (WidgetClass) & arcObjClassRec;

/*******************************************************************************
Name:
        ClassInitialize
--------------------------------------------------------------------------------
Funktionstyp:
        Methode
*******************************************************************************/

/* ARGSUSED */
static void ClassInitialize( )
{

#ifdef debug
  printf("Arc: ClassInitialize \n");
#endif

  arcObjClassRec.composite_class.extension = (XtPointer)arcObjExtensionPtr;
}

/*******************************************************************************
Name:
        Initialize
--------------------------------------------------------------------------------
Funktionstyp:
        Methode
*******************************************************************************/

/* ARGSUSED */
static void Initialize(treq, tnew, args, num_args)
     Widget treq, tnew;
     ArgList args;
     Cardinal *num_args;
{
  ArcObject new = (ArcObject) tnew;
  Multiplicity* multiplicity;

  void (*func)();

#ifdef debug
  printf("Arc: Initialize \n");
#endif

  /* composite fields */
  new->composite.insert_position = insertPosition;

  /* arcObj fields */
  /* multiplicity */

/********************************************************************************/
  /* checking arc setting , Copying Pointer Reference */
  if ((multiplicity=XtNew(Multiplicity))==NULL) {
	XtError("Arc initialize: Can't allocate mem for new multiplicity");
  }
  if (new->arcObj.multiplicity == NULL) {
	multiplicity->type = ARCVALUE;
	multiplicity->values.arcvalue = 1.0;
  }
  else {
	multiplicity->type = new->arcObj.multiplicity->type;
	switch (multiplicity -> type) { 
	case ARCVALUE:
	    multiplicity->values.arcvalue = 
	      new->arcObj.multiplicity->values.arcvalue;
	    break;
	case MD_DEF:
	    multiplicity->values.md_definition = 
	      XtNewString(new->arcObj.multiplicity->values.md_definition);
	    break;
	default:
	    XtWarning("Arc initialize: no valid (rate?) type, reset to default (1.0)");
	    multiplicity->type = RATEVALUE;
	    multiplicity->values.arcvalue = 1.0;
	}
  }
  new->arcObj.multiplicity = multiplicity;

/*******************************************************************************/
  
  if ((new->arcObj.multiplicity->type == ARCVALUE) && 
      (new->arcObj.multiplicity->values.arcvalue < 1)) {
      XtWarning("Initialize arc: impossible multiplicity setting, reset to 1");
      new->arcObj.multiplicity->values.arcvalue = 1;
  }
  
  /* drawing_direction */
  if (new->arcObj.drawing_direction != UP && new->arcObj.toObj == NULL) {
    XtWarning("Initialize arc: not allowed to set drawing direction");
    new->arcObj.drawing_direction = UP;
  }
  if (new->arcObj.toObj != NULL) {
    if (isTransObject(new->arcObj.toObj))
      new->arcObj.drawing_direction = DOWN;
  }

  /* update connections in transitions */
  if (isTransObject(new->arcObj.toObj)) { 
    XtVaGetValues((Widget)new->arcObj.toObj,
		  XtNaddConnection, &func,
		  NULL);

    (*func)((TransObj)new->arcObj.toObj,
	    new->arcObj.arcType,
	    (PlaceObj)new->arcObj.fromObj,
	    new->arcObj.multiplicity);
  }
  else if (isTransObject(new->arcObj.fromObj)) {
    XtVaGetValues((Widget)new->arcObj.fromObj,
		  XtNaddConnection, &func,
		  NULL);
    (*func)((TransObj)new->arcObj.fromObj,
	    new->arcObj.arcType,
	    (PlaceObj)new->arcObj.toObj,
	    new->arcObj.multiplicity);
  }

  /* private fields */
  new->arcObj.insert_position = 0;
}



/*******************************************************************************
Name:
        SetValues
--------------------------------------------------------------------------------
Funktionstyp:
        Methode
*******************************************************************************/

/* ARGSUSED */
static Boolean SetValues(current, request, new, args, num_args)
     Widget current, request, new;
     ArgList args;
     Cardinal *num_args;
{
  ArcObject curarc = (ArcObject) current;
  ArcObject newarc = (ArcObject) new;
  Multiplicity* multiplicity;
  Boolean do_redisplay = False;
  void (*func)();

/*******************************************************************************/
  /* checking ARC setting Copying Pointer Reference */
/*  if (newarc->arcObj.multiplicity != curarc->arcObj.multiplicity) { bei Transitionen ! */
  if ((multiplicity = XtNew(Multiplicity)) == NULL) {
      XtError("Arc initialize: Can't allocate mem for new Multiplicity");
  }

  if (newarc->arcObj.multiplicity == NULL) {
      XtWarning("Arc setvalues: no Multiplicity found, using old setting");
      multiplicity->type = curarc->arcObj.multiplicity->type;
      switch (multiplicity->type) {
	case ARCVALUE:
	  multiplicity->values.arcvalue = 
	    curarc->arcObj.multiplicity->values.arcvalue;
	  break;
	case MD_DEF:
	  multiplicity->values.md_definition = 
	    XtNewString(curarc->arcObj.multiplicity->values.md_definition);
	  break;
	default: ;
      }
  }
  else {  /* SET MULTIPLICITY */
      multiplicity->actualValue = newarc->arcObj.multiplicity->actualValue;
      multiplicity->type = newarc->arcObj.multiplicity->type;
      switch (multiplicity -> type) { 
	case ARCVALUE:
	  multiplicity->values.arcvalue = 
	    newarc->arcObj.multiplicity->values.arcvalue;
	  break;
	case MD_DEF: 
	  multiplicity->values.md_definition = 
	    XtNewString(newarc->arcObj.multiplicity->values.md_definition);
	  break;
	default:
	  XtWarning(
	  "Transition setvalues: no valid arc type, reset to previous values!");
	  multiplicity->type = curarc->arcObj.multiplicity->type;
	  switch (multiplicity -> type) { 
	    case ARCVALUE:
	      multiplicity->values.arcvalue = 
		curarc->arcObj.multiplicity->values.arcvalue;
	      break;
	    case MD_DEF:
	      multiplicity->values.md_definition = 
		XtNewString(curarc->arcObj.multiplicity->values.md_definition);
	      break;
	    default: ;
	  } /* case 2 */
      } /* case 1 */
  } /* else */
  newarc->arcObj.multiplicity = multiplicity;
  
/*******************************************************************************/

/* ptr Inhalt kopieren ! und Vergleich korregieren "." <=> "j->m" */
  if (newarc->arcObj.multiplicity->type == ARCVALUE) {
      if (curarc->arcObj.multiplicity->type == ARCVALUE) {
	  if ((curarc->arcObj.multiplicity->values.arcvalue 
	       != newarc->arcObj.multiplicity->values.arcvalue) 
	      && (newarc->arcObj.multiplicity->values.arcvalue < 1)) {
	      XtWarning
		("Set Values arc: impossible multiplicity setting; ignored");
           newarc->arcObj.multiplicity->values.arcvalue = curarc->arcObj.multiplicity->values.arcvalue;
	  }
      }
      else /* check new multiplicity.arcvalue --- old MD_DEF */
	if (newarc->arcObj.multiplicity->values.arcvalue < 1) {
	    XtWarning
      ("Set Values arc: impossible multiplicity setting; old mark.-dep. taken");
	    newarc->arcObj.multiplicity->type = MD_DEF;
	    newarc->arcObj.multiplicity->values.md_definition = 
	      XtNewString(curarc->arcObj.multiplicity->values.md_definition);
	}
  } /* type == ARCVALUE */
  else /* check new md_def. */
  {
      if (curarc->arcObj.multiplicity->type == MD_DEF) {
	  if (newarc->arcObj.multiplicity->values.md_definition == "") {
	      XtWarning
		("Set Values arc: impossible mark.-dep setting; ignored");
	      XtFree(newarc->arcObj.multiplicity->values.md_definition);
	      newarc->arcObj.multiplicity->values.md_definition = 
		XtNewString(curarc->arcObj.multiplicity->values.md_definition);
	  }
      }
      else /* ... old multipli.arcvalue */
        if (newarc->arcObj.multiplicity->values.md_definition == "") {
	    XtWarning("Set Values arc: impossible mark.-dep. setting; ignored");
	    newarc->arcObj.multiplicity->type = ARCVALUE;
	    newarc->arcObj.multiplicity->values.arcvalue = 
	      curarc->arcObj.multiplicity->values.arcvalue;
	}
  } /* type == MD_DEF */

  if (curarc->arcObj.drawing_direction != newarc->arcObj.drawing_direction) {
      XtWarning("Set Values arc: not allowed to set drawing direction");
      newarc->arcObj.drawing_direction = curarc->arcObj.drawing_direction;
  }
  if (curarc->arcObj.toObj == NULL && newarc->arcObj.toObj != NULL) {
      if (isTransObject(newarc->arcObj.toObj))
	newarc->arcObj.drawing_direction = DOWN;
  }

  /* update connections in trans */

  if (curarc->arcObj.toObj != newarc->arcObj.toObj ||
      curarc->arcObj.fromObj != newarc->arcObj.fromObj ||
      curarc->arcObj.arcType != newarc->arcObj.arcType ||
      curarc->arcObj.multiplicity != newarc->arcObj.multiplicity) {
      if (isTransObject(curarc->arcObj.toObj)) {
	  XtVaGetValues((Widget)curarc->arcObj.toObj,
			XtNdelConnection, &func,
			NULL);
	  (*func)((TransObj)curarc->arcObj.toObj,
		  curarc->arcObj.arcType,
		  curarc->arcObj.fromObj);
      }
      else if (isTransObject(curarc->arcObj.fromObj)) {
	  XtVaGetValues((Widget)curarc->arcObj.fromObj,
			XtNdelConnection, &func,
			NULL);
	  (*func)((TransObj)curarc->arcObj.fromObj,
		  curarc->arcObj.arcType,
		  curarc->arcObj.toObj);
      }
      if (isTransObject(newarc->arcObj.toObj)) {
	  XtVaGetValues((Widget)newarc->arcObj.toObj,
			XtNaddConnection, &func,
			NULL);
	  (*func)((TransObj)newarc->arcObj.toObj,
		  newarc->arcObj.arcType,
		  newarc->arcObj.fromObj,
		  newarc->arcObj.multiplicity);
      }
      else if (isTransObject(newarc->arcObj.fromObj)) {
	  XtVaGetValues((Widget)newarc->arcObj.fromObj,
			XtNaddConnection, &func,
			NULL);
	  (*func)((TransObj)newarc->arcObj.fromObj,
		  newarc->arcObj.arcType,
		  newarc->arcObj.toObj,
		  newarc->arcObj.multiplicity);
      }
  }

  return do_redisplay;
}

/*******************************************************************************
Name:
Destroy
--------------------------------------------------------------------------------
Funktionstyp:
        Methode
*******************************************************************************/

static void Destroy(w)
     Widget w;
{
  ArcObject arc = (ArcObject) w;
  Multiplicity *multiplicity;
  void (*func)();
  
#ifdef debug
  printf("ARC: Destroy \n");
#endif

  XtVaGetValues((Widget)arc,  
		XtNmultiplicity, &multiplicity,
		NULL); 
  if (multiplicity->type == MD_DEF) 
    XtFree(arc->arcObj.multiplicity->values.md_definition); 

  XtFree((char *)arc->arcObj.multiplicity);   

  if (isTransObject(arc->arcObj.fromObj)) {
      XtVaGetValues((Widget)arc->arcObj.fromObj,
		    XtNdelConnection, &func,
		    NULL);
      (*func)((TransObj)arc->arcObj.fromObj,
	      arc->arcObj.arcType,
	      arc->arcObj.toObj);
  }
  else if (isTransObject(arc->arcObj.toObj)) {
      XtVaGetValues((Widget)arc->arcObj.toObj,
		    XtNdelConnection, &func,
		    NULL);
      (*func)((TransObj)arc->arcObj.toObj,
	      arc->arcObj.arcType,
	      arc->arcObj.fromObj);
  }
  
  arc->arcObj.fromObj = NULL;
  arc->arcObj.toObj = NULL;
}


/*******************************************************************************
Name:
        insertPosition
--------------------------------------------------------------------------------
Funktionstyp:
        Methode
*******************************************************************************/

static Cardinal insertPosition(w)
     Widget w;
{
  ArcObject parent;

  parent = (ArcObject)XtParent(w);

  return parent->arcObj.insert_position ;
}

/*******************************************************************************
Name:
        createImPoint
--------------------------------------------------------------------------------
Funktionstyp:
        "offnetliche funktion
--------------------------------------------------------------------------------
Parameter:
        w: Instanz einer Kante
	x,y: Position des Zwischenpunktes
	insert_position: Einf"ugeposition im Childrenarray
--------------------------------------------------------------------------------
Zweck:
        Erzeugt Zwischenpunkt in angegebener Position im Childrenarray
*******************************************************************************/

void createImPoint(w,x,y,insert_position)
     Widget w;
     int x,y,insert_position;
{
  Widget imp;
  ArcObject arc = (ArcObject)w;
  char name[100],*objname;
  static numObj = 0;

#ifdef debug
  printf("createImPoint \n");
#endif

  if (0 <= insert_position && insert_position <= arc->composite.num_children) {
    arc->arcObj.insert_position = insert_position;
    sprintf(name,"IMP%d",numObj++);
    objname = XtNewString(name);
    imp = XtVaCreateWidget(objname,
			   baseObjClass,
			   w,
			   XtNxpos,x,
			   XtNypos,y,
			   NULL);
  }
}

/*******************************************************************************
Name:
        destroyImPoint
--------------------------------------------------------------------------------
Funktionstyp:
        "offnetliche funktion
--------------------------------------------------------------------------------
Parameter:
        w: Instanz einer Kante
	position: index des Zwischenpunktes
--------------------------------------------------------------------------------
Zweck:
        zerst"oren des Zwischenpunktes in angegbener Position im Childrenarray
*******************************************************************************/

void destroyImPoint(w,position)
     Widget w;
     int position;
{
  ArcObject arc = (ArcObject)w;

#ifdef debug
  printf("DelImPoint \n");
#endif

  if (0<=position && position < arc->composite.num_children) {
      /* ein zerstoertes Widget ist noch weiter vorhanden (bis XtMainLoop),
	 das 'destroyed'-Feld ist nicht oeffentlich, also setze ich die 
	 Koordinaten auf (hoffentlich) unmoegliche Werte */ 
      XtVaSetValues(arc->composite.children[position],
		    XtNxpos,DESTROYED,
		    XtNypos,DESTROYED,
		    NULL);
      XtDestroyWidget(arc->composite.children[position]);
  }
}

/*******************************************************************************
Name:
        resetImPoint
--------------------------------------------------------------------------------
Funktionstyp:
        "offnetliche funktion
--------------------------------------------------------------------------------
Parameter:
        w: Instanz einer Kante
	x,y: neue Position des Zwischenpunktes
	position: Position des Zwischenpunktes im Childrenarray
--------------------------------------------------------------------------------
Zweck:
        Neusetzen der Koordinaten eines Zwischenpunktes
*******************************************************************************/

void resetImPoint(w,x,y,position)
Widget w;
int x,y,position;

{
ArcObject arc = (ArcObject)w;

#ifdef debug
printf("resetImPoint \n");
#endif
  if (0<=position && position < arc->composite.num_children)
  	XtVaSetValues(arc->composite.children[position],
		XtNxpos,x,
		XtNypos,y,
		NULL);
}

/*******************************************************************************
Name:
        rescaleImPoint
--------------------------------------------------------------------------------
Funktionstyp:
        "offnetliche funktion
--------------------------------------------------------------------------------
Parameter:
        w: Instanz einer Kante
	rescale_factor: Faktor, mit dem die Koordiaten ver"andert werden
--------------------------------------------------------------------------------
Zweck:
        Ver"andern der Koordianten aller Zwischenpunkte einer Kante
*******************************************************************************/

void rescaleImPoint(w,rescale_factor)
Widget w;
float rescale_factor;

{
ArcObject arc = (ArcObject)w;
int i,x,y;

#define ImPointClass baseObjClass

#ifdef debug
printf("resetImPoint \n");
#endif
  for (i=0;i<arc->composite.num_children;i++)
  {
  	XtVaGetValues(arc->composite.children[i],
		XtNxpos,&x,
		XtNypos,&y,
		NULL);
  	XtVaSetValues(arc->composite.children[i],
		XtNxpos,(int)((float)x*rescale_factor),
		XtNypos,(int)((float)y*rescale_factor),
		NULL);
  }
}

#define POW_2(X)	((X)*(X))

/*return werte:
	mode SECTION: index des 2. Punktes in IParray, der Section 
			definiert.
  	mode POINT:   index des naecshten Punktes zu event

	NOT_FOUND: keine Kante unter event, oder bei mode Point
			kein IP vorhanden
*/
/*******************************************************************************
Name:
        recievePointOfArc
--------------------------------------------------------------------------------
Funktionstyp:
        "offnetliche funktion
--------------------------------------------------------------------------------
Parameter:
        w: Instanz einer Kante
	x,y: Punkt
	areat: Streubreite
	mode: SECTION Kantenabschnitt in der N"ahe des Punktes finden
	      POINT n"achsten Zwischenpunkt der Kante zum Punkt finden
--------------------------------------------------------------------------------
Rueckgabewert:
        index des n"achsten Zwischenpunktes bei mode POINT
	index des Zwischenpunktes der Kantenabschnitt definiert bei mode SECTION
--------------------------------------------------------------------------------
Zweck:
        Identifizieren eines Zwischenpunktes oder Kantenabschnittes in der 
	N"ahe des angegebenen Punktes
*******************************************************************************/

int recievePointOfArc(w,x,y,area,mode)
Widget w;
int x,y,area,mode;

{
ArcObject arc = (ArcObject)w;
int begin_x,begin_y,end_x,end_y;
int i;
Widget from_obj,to_obj;

#ifdef debug
printf("lookForArc \n");
#endif
    /* recieve from_obj,to_obj */
    XtVaGetValues(w,
		  XtNfromObj,&from_obj,
		  XtNtoObj,&to_obj,
		  NULL);
    /* recive begin coords of arc */
    XtVaGetValues(from_obj,
		  XtNxpos,&begin_x,
		  XtNypos,&begin_y,
		  NULL);
    for (i=0;i<(int)arc->composite.num_children;i++) 
    {
	XtVaGetValues(arc->composite.children[i],
		      XtNxpos,&end_x,
		      XtNypos,&end_y,
		      NULL);
	if (is_under(x,y, begin_x,begin_y,end_x,end_y,area)) 
	{
	    if (mode == POINT )
	    {
		if (sqrt(POW_2(end_x-x)+POW_2(end_y-y))
			>sqrt(POW_2(begin_x-x)+POW_2(begin_y-y)))
			return Max(0,i-1);
	        else		
	    	    return i;
	    }
	    else
		return i;	     
	}
	else 
	{
	    begin_x = end_x;
	    begin_y = end_y;
	}
    }/*for*/
    XtVaGetValues(to_obj,
	          XtNxpos,&end_x,
	          XtNypos,&end_y,
	          NULL);
    if (is_under(x,y, begin_x,begin_y,end_x,end_y,area)) 
    {
	if (mode == POINT)
	{
	    if (arc->composite.num_children > 0)
		return arc->composite.num_children - 1;
	    else
	    	return NOT_FOUND;
	}
	else
	    return (int)arc->composite.num_children;
    }
    else 
	return NOT_FOUND;
}

/*******************************************************************************
Name:
        setMultiplicity
--------------------------------------------------------------------------------
Funktionstyp:
        "offnetliche funktion
--------------------------------------------------------------------------------
Parameter:
        w: Instanz einer Kante
	value: Wert
	mode: INCR Multiplizit"at wird um value erh"oht
	      ABS  Multiplizit"at wird value versehen
--------------------------------------------------------------------------------
Zweck:
        Neusetzen der Multiplizit"at
*******************************************************************************/

void setMultiplicity(w,value,mode)
Widget w;
int value,mode;

{
    Cardinal arcvalue;
    Multiplicity *mult;

    XtVaGetValues(w,
		  XtNmultiplicity,&mult,
		  NULL);
    if (mult->type == ARCVALUE) {
	arcvalue = mult->values.arcvalue;
	
	switch (mode) 
	  {
	    case INCR: 
	      arcvalue += value; 
	      break;
	    case DECR: 
	      if(arcvalue > value)
		arcvalue -= value;
	      break;
	    case ABS: 
	      arcvalue = value;
	      break;
	    default: ;
	  }
	
	mult->values.arcvalue = arcvalue;
	XtVaSetValues(w,
		      XtNmultiplicity,mult,
		      NULL);
    }
    else 
      XtWarning
	("There is already an arc with a marking dependend multiplicity \n\
between this objects. Change the definition (\"old definition\" + 1).");
}

/*******************************************************************************
Name:
       is_under 
--------------------------------------------------------------------------------
Funktionstyp:
        private funktion
--------------------------------------------------------------------------------
Parameter:
        x,y: Punktkoordinaten
	b_x,b_y,e_x,e_y: Anfangs- und Endpunkt eines Streckenabschnitts
	area: Streubreite um den Kantenabschnitt der untersucht wird
--------------------------------------------------------------------------------
Rueckgabewert:
       True, falls Punkt in der N"ahe des Kantenabschnitts, False sonst
--------------------------------------------------------------------------------
Zweck:
       Auffinden eines Kantenabschnitts in der N"ahe eines Punktes
*******************************************************************************/

static Boolean is_under(x,y,b_x,b_y,e_x,e_y,area)
int x,y,b_x,b_y,e_x,e_y,area;

{
double d_x = (double)x;
double d_y = (double)y;
double x1 = (double)b_x;
double y1 = (double)b_y;
double x2 = (double)e_x;
double y2 = (double)e_y;
double D_Max = (double)area;

double alpha,
	delta_x,delta_y,delta_B,
	M,Bg1,Bg2,
	proj_length,Qx,Qy,
	arclaenge,d_q_p1,d_q_p2;
int x_rect,y_rect;

#define Max_M 15.0

#ifdef debug
printf("is_under \n");
#endif
  if ((x2-x1 != 0.0) && (fabs((M=(y2-y1)/(x2-x1))) < Max_M)) 
  {
	/* Steigung */
	/* Schnittpunkt des Kantenabschnitts an Hand P=(x2,y2)*/
	Bg1 = y2 - M*x2;
	/* Schnittpunkt der geraden durch zu Untersuchendeb Punkt
	   mit gleicher Steigung */
	Bg2=d_y-M*d_x;
	/* AbStand der Schnittpkte an Y-achse */
	delta_B = (Bg1-Bg2);

	/* laenge der senkrechten Projektion von Geraden zur Kante */
	alpha = atan(M);
	proj_length=(delta_B*cos(alpha));
	/* Projektion ueberpruefen */
	if (fabs(proj_length) <= D_Max/2.0) 
	{
	    /* Streckenabschnitt ueberpruefen */
	    /*Projektionspunkt auf Geraden bestimmen */
	    delta_x = proj_length*sin(alpha);
	    delta_y = proj_length*cos(alpha);
	    Qx = d_x + delta_x;
	    Qy = d_y + delta_y;
	    arclaenge = sqrt(POW_2(x1-x2)+POW_2(y1-y2));
	    /* Abstand von Q zu P1 */
	    d_q_p1=sqrt(POW_2(x1-Qx)+POW_2(y1-Qy));
	    if (d_q_p1 > arclaenge) 
	    {
		return False;
	    }
	    /* Abstand von Q zu P2 */
	    d_q_p2=sqrt(POW_2(x2-Qx)+POW_2(y2-Qy));
	    if (d_q_p2 > arclaenge) 
	    {
		return False;
	    }
	    /* if you have reached this point every thing will be allright */
	    return True;
	}
	else /* Projektionslaenge stimmt nicht */
	    return False;
  }
  else 
  {
  	/*Steile Kante*/
	/* rechteck untersuchen ausgehen vom oberen linken Punkt */
	x_rect = Max(Min(b_x,e_x)-area/2,0);
	y_rect = Min(b_y,e_y);
	if (x_rect<x && x<x_rect+area && y_rect<y && y< y_rect+abs(b_y - e_y))
		return True;
	else
		return False;
  }
}


/*******************************************************************************
Name:
        getImPCoords
--------------------------------------------------------------------------------
Funktionstyp:
        "offnetliche funktion
--------------------------------------------------------------------------------
Parameter:
        w: Instanz einer Kante
	position: Position des Zwischenpunktes im Childrenarray
    Bei R"uckkehr:
        x,y: Koordinaten des Zwischenpunktes
--------------------------------------------------------------------------------
Zweck:
Erhalten der Koordianten eines Zwischenpunktes
*******************************************************************************/

void getImPCoords(w,position,x,y)
Widget w;
int *x,*y,position;
{
ArcObject arc = (ArcObject)w;
Widget toObj,fromObj;
int numChild;

  numChild = (int)arc->composite.num_children;
  if (position == -1) 
  {
	XtVaGetValues((Widget)arc,
		      XtNfromObj,&fromObj,
		      NULL);
	XtVaGetValues((Widget)fromObj,
		      XtNxpos,x,
		      XtNypos,y,
		      NULL);
		      return;
  }
  if (numChild > 0 && position < numChild) 
  	XtVaGetValues(arc->composite.children[position],
			XtNxpos,x,
			XtNypos,y,
			NULL);
  else 
  {
	XtVaGetValues((Widget)arc,
		      XtNtoObj,&toObj,
		      NULL);
	XtVaGetValues(toObj,
		      XtNxpos,x,
		      XtNypos,y,
		      NULL);
  }
}

