/******************************************************************************
 File:		expoly.c

 Purpose:	routines for manipulating expolynomial distributions
		and for computing the factors needed by the
		randomization procedure

 Programmer:	Reinhard German,
                Ulrich Hertlein

 History:	Release 1: June 2, 1993	
                Release 2: September 17, 1993
                Release 3: March 15, 1994
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include <math.h>
#include "expoly.h"
#include "weighter.h"

#include "portab.h"

#define ALLOC_SIZE 128        /* array factors enlargement */
#define FRESH      (ALLOC_SIZE * sizeof(double))


/*
 * Malloc: malloc with error checking
 */
char *Malloc(size)
unsigned size;
{
   char *p;

   if( !(p = (char *)malloc(size)) )
   {
      fprintf(stderr, "ERROR: can't malloc %d bytes!\n",size);
      exit(1);
   }

   return( p );
}

/*
 * MyRealloc: realloc with NULL-pointer handling and error checking;
 *            Realloc clears the freshly allocated memory, but this only
 *            works in this special case since we always know by how many
 *            bytes the allocated memory is increased
 */
char *MyRealloc(ptr, size)
register char *ptr;
register unsigned size;
{
   register char *np;

   /* alloc initial memory if necessary */
   if( !ptr )
   {
      np = (char *)Malloc(size);
      memset(np,0,size);
   }
   else
   {
      if( !(np = (char *)realloc(ptr,size)) )
      {
         fprintf(stderr, "ERROR: can't realloc %d bytes!\n",size);
         exit(1);
      }

      /* clear 'fresh' memory */
      ptr = np + size - FRESH;
      memset(ptr,0,FRESH);
   }

   return( np );
}

#ifdef TEST_ONLY

/****************************************************************************

function:	scan_expoly

 input:		fname: name of file which contains expoly distribution
		return_expoly: pointer to expoly distribution
 
 output:	return_expoly points to the expoly distrinution 

 purpose:	scans an expolynomial distribution from an ascii file
		and converts it to the internal data structure
		The expected file format is the format of the output of
		print_expoly.

***************************************************************************/
int scan_expoly(fname, return_expoly)
char *fname;
time_l **return_expoly;
{
  FILE *fp;
  time_l *expoly;
  power_m *power;
  expo_n *expo;
  char line[80];
  double x1=0.0, x2=0.0, p=0.0, lambda=0.0, b=0.0, c=0.0;
  unsigned l, m;
  enum {initial, read_int, read_expo} status = initial;

  if (!(fp=fopen(fname,"r"))) {
    fprintf(stderr, "ERROR: can't open file %s!\n", fname);
    return 0; 
  }; /* if */
  
  *return_expoly = (time_l *) Malloc(sizeof(time_l));
  expoly = *return_expoly;

  /* read lines of input file and scan them: */
  while (fgets(line, 80, fp)) {
    if (sscanf(line, "interval %d:\tx1=%lf, x2=%lf, p=%lf", &l, &x1, &x2, &p)) {
      if (status==read_int)
	return 0;
      else if (status==read_expo) {
        expoly->next = (time_l *) Malloc(sizeof(time_l));
	expoly = expoly->next;
      }; /* else if status */
      expoly->x1 = (double) x1;
      expoly->x2 = (double) x2;
      expoly->p = (double) p;
      expoly->next = NULL;
      status = read_int;
    } /* if scanf */
    else if (sscanf(line, "power m=%d:\tlambda=%lf, b=%lf, c=%lf", 
                    &m, &lambda, &b, &c)==4) {
      if (status==read_int) { 
        expoly->power = (power_m *) Malloc(sizeof(power_m));
        power = expoly->power;
      } /* if status */
      else if (status==read_expo) {
        power->next = (power_m *) Malloc(sizeof(power_m));
        power = power->next;
      } /* else if status */
      else return 0;
      power->m = m;
      power->expo = (expo_n *) Malloc(sizeof(expo_n));
      expo = power->expo;
      expo->lambda = (double) lambda;
      expo->b = (double) b;
      expo->c = (double) c;
      expo->next = NULL;
      power->next = NULL;
      status = read_expo;
    } /* else if scanf */
    else if (sscanf(line, "\tlambda=%lf, b=%lf, c=%lf", &lambda, &b, &c)) {
      if (status==read_expo) {
        expo->next = (expo_n *) Malloc(sizeof(expo_n));
        expo = expo->next;
      }
      else return 0;
      expo->lambda = (double) lambda;
      expo->b = (double) b;
      expo->c = (double) c;
      expo->next = NULL;
    }; /* else if scanf */
  }; /* while */
  
  fclose(fp);
  if (status!=read_expo)
    return 0;
  else
    return 1;

}; /* scan_expoly */

/* print the computed factors: */
void print_factors(omega_factors, psi_factors, R, omega_error, psi_error)
  double *omega_factors,
	 *psi_factors,
	 omega_error,
	 psi_error;
  unsigned R;
{
  double x=0,y=0;
  unsigned k;
  printf("\n");
  printf("---------------------------------------------------------------\n"); 
  printf("*** print omega and psi factors, R=%d: ***\n", R);
  printf("omega_error=%.16f, psi_error=%.16f\n", omega_error, psi_error);
  printf("---------------------------------------------------------------\n"); 
  for(k=0; k <= R; k++) {
    x+=omega_factors[k];
    y+=psi_factors[k];
    printf("k=%d, omega_factor=%.16f, psi_factor=%.16f\n", 
	   k, omega_factors[k], psi_factors[k]);
  };
  printf("---------------------------------------------------------------\n"); 
  printf("omega_approx=%.16f, psi_approx=%.16f\n",x,y);
  printf("---------------------------------------------------------------\n"); 
}; /* print_factors */

#endif 


/* set deterministic distribution */
time_l *set_det_expoly(tau)
double tau;
{
   time_l *expoly, *return_expoly;
   power_m *power;
   expo_n *expo;

   return_expoly = (time_l *)Malloc(sizeof(time_l));
   expoly = return_expoly;
   expoly->x1 = 0.0;
   expoly->x2 = tau;
   expoly->p = 0.0;
   expoly->power = (power_m *) Malloc(sizeof(power_m));
      power = expoly->power;
      power->m = 0;
      power->expo = (expo_n *) Malloc(sizeof(expo_n));
         expo = power->expo;
         expo->lambda = 0.0;
         expo->b = 1.0;
         expo->c = 0.0;
         expo->next = NULL;
      power->next = NULL;
   expoly->next = (time_l *) Malloc(sizeof(time_l));
   expoly = expoly->next;
   expoly->x1 = tau;
   expoly->p = 1.0;
   expoly->power = (power_m *) Malloc(sizeof(power_m));
      power = expoly->power;
      power->m = 0;
      power->expo = (expo_n *) Malloc(sizeof(expo_n));
         expo = power->expo;
         expo->lambda = 0.0;
         expo->b = 0.0;
         expo->c = 0.0;
         expo->next = NULL;
      power->next = NULL;
   expoly->next = NULL;

   return( return_expoly );
} /* set_det_expoly */

/* set uniform distribution */
time_l *set_uniform_expoly(t1, t2)
double t1, t2;
{
   time_l *expoly, *return_expoly;
   power_m *power;
   expo_n *expo;

   return_expoly=(time_l *) Malloc(sizeof(time_l));
   expoly = return_expoly;
   expoly->x1 = 0.0; 
   expoly->x2 = t1; 
   expoly->p = 0.0;
   expoly->power = (power_m *) Malloc(sizeof(power_m));
      power = expoly->power;
      power->m = 0;
      power->expo = (expo_n *) Malloc(sizeof(expo_n));
         expo = power->expo; 
         expo->lambda = 0.0;
         expo->b = 1.0;
         expo->c = 0.0;
         expo->next = NULL;
      power->next = NULL;
   expoly->next = (time_l *) Malloc(sizeof(time_l));
   expoly = expoly->next;
   expoly->x1 = t1;
   expoly->x2 = t2;
   expoly->p = 0.0;
   expoly->power = (power_m *) Malloc(sizeof(power_m));
      power = expoly->power;
      power->m = 0;
      power->expo = (expo_n *) Malloc(sizeof(expo_n));
         expo = power->expo;
         expo->lambda = 0.0;
         expo->b = 1+t1/(t2-t1); 
         expo->c = 1/(t2-t1);
         expo->next = NULL;
      power->next = (power_m *) Malloc(sizeof(power_m));
      power = power->next;
      power->m = 1;
      power->expo = (expo_n *) Malloc(sizeof(expo_n));
         expo = power->expo;
         expo->lambda = 0.0;
         expo->b = -1/(t2-t1);
         expo->c = 0.0;
         expo->next = NULL;
      power->next = NULL;
   expoly->next = (time_l *) Malloc(sizeof(time_l));
   expoly = expoly->next;
   expoly->x1 = t2;
   expoly->p = 0.0;
   expoly->power = (power_m *) Malloc(sizeof(power_m));
      power = expoly->power;
      power->m = 0;
      power->expo = (expo_n *) Malloc(sizeof(expo_n));
         expo = power->expo;
         expo->lambda = 0.0;
         expo->b = 0.0;
         expo->c = 0.0;
         expo->next = NULL;
      power->next = NULL;
   expoly->next = NULL;

   return( return_expoly );
} /* set_test_expoly */

/* print an expolynomial distribution: */
void print_expoly(expoly)
  time_l *expoly;
{
  power_m *power;
  expo_n *expo;
  unsigned l;

  fprintf(stderr,"\t\t\t\t Expolynomial distribution:\n");
  for( l = 0; expoly; expoly = expoly->next, l++ ) 
  {
    fprintf(stderr,"\t\t\t\t ----------------------------------------------------------\n");
    fprintf(stderr,"\t\t\t\t Interval %d: ",l);
    fprintf(stderr,"\t\t\t\t x1=%.16g, ", expoly->x1);
    if (expoly->next)
      fprintf(stderr,"x2=%.16g, ", expoly->x2);
    else 
      fprintf(stderr,"x2=infinity, ");
    fprintf(stderr,"p=%.16g\n", expoly->p);
    for (power = expoly->power; power; power = power->next)
    {
      fprintf(stderr,"\t\t\t\t power m=%d:", power->m); 
      for (expo = power->expo; expo; expo = expo->next)
      {
	fprintf(stderr,"\t\t\t\t lambda=%.16g, b=%.16g, c=%.16g\n",
	       expo->lambda, expo->b, expo->c);
      }
    }
  }
  fprintf(stderr,"\t\t\t\t ----------------------------------------------------------\n");
} /* print_expoly */



/******************************************************************************
 Function:     poisson_prob

 Input:

 Output:

 Purpose:
******************************************************************************/
void poisson_prob(k, epsilon, qx, beta, beta_sum, boundary, run, raising_probs, prev_weight, weight_sum, W, w)
unsigned k;
double   epsilon;
double   qx, 
         **beta, 
         *beta_sum;
unsigned *boundary, 
         run;
int      *raising_probs;
double   *prev_weight, 
         *weight_sum,
         *W, 
         **w;
{ 
   /* decide for an approach */
/*
   if( qx < 100.0 )
   {
      *boundary=1;
      if( !k )
      {
         *beta = (double *) Malloc(sizeof(double));
         **beta = exp(-qx);
      }
      else
         **beta *= qx/k;
   }
   else
*/
   {
      int           l,
                    r;
      double        left,
                    right;
      double        *weights;
      char          ok;

      if( qx == (double )0.0 )
      {
         if( !k )
         {
            *beta = (double *)Malloc(sizeof(double));
            **beta = (double )1.0;
	    *boundary = 0;
         }
         else
            **beta = (double )0.0;
      *beta_sum = (double) 1.0;
      }
      else
      { 
         if( !k )
         {
            /* call weighter */
            weighter(qx,epsilon,MINVALUE,MAXVALUE,&left,&right,&weights,W,&ok,run);

            /* check execution */
            if( !ok )
            {
               fprintf(stderr,"poisson_prob: weighter() returned with error\n");
               exit( -1 );
            }

            *boundary = (unsigned )nint(right);
            *beta = weights;

            if((*w = (double *)calloc((size_t)(*boundary + 1), sizeof(double))) == NULL)
            {
               fprintf(stderr,"poisson_prob: cannot allocate enough memory for w\n");
               exit( -1 );
            }
            *prev_weight = 0.0;
            *weight_sum = 0.0;
            *raising_probs = 1;
         }
         else 
            if( k <= *boundary )
               (*beta)++;
            else
	    {
               fprintf(stderr,"poisson_prob: k is bigger than number of computed poisson_probs\n");
               exit( -1 );
            }

         (*w)[k] = **beta;
         **beta /= *W;

         if(*raising_probs && ((*w)[k] < *prev_weight))
            *raising_probs = 0;

         if(*raising_probs)
         { 
            *weight_sum += (*w)[k];
            *prev_weight = (*w)[k];
         } 
         else
         {
            *weight_sum = 0;
            l = 0;
            r = k;
            while (l < r )
               if ((*w)[l] < (*w)[r])
                 *weight_sum += (*w)[l++];
              else
                 *weight_sum += (*w)[r--];
            *weight_sum += (*w)[l];
         }

         if(k == *boundary)
            free(*w);

         *beta_sum = *weight_sum / *W;
      }
   }
}

/******************************************************************************
 Function:     calculate_factors

 Input:	       expoly: internal representation of expolynomial distribution
	       q: 1.02* max. abs. diag. entry of generator matrix
	       omega_factors: vector of factors for transient probs 
	       psi_factors: vector of factors for sojourn times 
    	       R : right truncation point

 Output:       values in omega_factors, psi_factors, and R 

 Purpose:      calculation of the factors needed for each iteration of
	       the randomization procedure
******************************************************************************/
void calculate_factors(expoly, epsilon, q, omega_factors, psi_factors, R,
                       omega_error, psi_error)
time_l *expoly;
double epsilon,q;
double **omega_factors;
double **psi_factors;
unsigned *R;
double *omega_error;
double *psi_error;
{
   power_m *power;
   expo_n *expo;

   double x1, x2, p, lambda, b, c,
          qx1, qpl, qplx1, qplx2,
          alpha1, alpha2,
          *beta1, *beta2,
          beta_sum1, beta_sum2,
          W1, W2,
          *w1, *w2,
          prev_weight1, prev_weight2,
          weight_sum1, weight_sum2,
          gamma,
          kappa1, kappa2,
          temp, diff,
          exact1, exact2,
          error1, error2,
          psi_exact = 0.0;
   int    raising_probs1, raising_probs2;
   unsigned h, k, m;

   unsigned lsize = ALLOC_SIZE;
   unsigned bound1, bound2;
#ifdef BOUNDS
   unsigned max_bounds, old_bounds;
#endif

   int interval = 0;
   unsigned run, tmp_size;
   double *omega_tmp, *psi_tmp;

   int prec_reached, poisson1_pos, poisson2_pos, bounds_reached;

   /* check for Q underflow */
   if (q < 1.0e-34)
      q = 0.0;

   /* initialize factor pointers to default size */
   *omega_factors = (double *)MyRealloc(NULL,lsize * sizeof(double));
   *psi_factors = (double *)MyRealloc(NULL,lsize * sizeof(double));

   *R = 0;
   bound1 = bound2 = 0;

   /* for all time intervals of the expolynomial distribution ... */
   for(; expoly; expoly=expoly->next, interval++ )
   {
      fprintf(stderr,"\t\t\t\t computing factors for interval %d\n",interval);
      x1 = expoly->x1;
      if (expoly->next)
         x2 = expoly->x2;

      p = expoly->p;

      /* calculate factors corresponding to point mass */
      if( p > 0.0 )
      {
         qx1 = q * x1;
         kappa1 = 1.0;

         for( k = 0; (kappa1 > epsilon) && (k <= MAXINDEX); k++ )
         {
            poisson_prob(k, epsilon, qx1, &beta1, &beta_sum1, &bound1, 0, 
                         &raising_probs1, &prev_weight1, &weight_sum1, &W1, &w1);

            kappa1 = 1 - beta_sum1;

            /* enlarge list if necessary */
            if( k >= lsize )
	    {
               lsize += ALLOC_SIZE;
               *omega_factors = (double *)MyRealloc(*omega_factors,lsize * sizeof(double));
               *psi_factors = (double *)MyRealloc(*psi_factors,lsize * sizeof(double));
	    }
            (*omega_factors)[k] += p * (*beta1);
         }

         if( k-1 > *R )
            *R = k-1;
      }
      if( !expoly->next )
         break;
     
      /* for all power terms ... */
      for( power = expoly->power; power; power=power->next )
      {
         m = power->m;

         /* for all exponential terms */
         for( expo = power->expo; expo; expo=expo->next )
         {
            lambda = expo->lambda;
            b = expo->b;
            c = expo->c;

            /* compute exact sums of alpha-factors */
            if( lambda == 0.0 )
            {
               exact1 = pow(x1, (double) (m+1)) / (m+1);
	       exact2 = pow(x2, (double) (m+1)) / (m+1);
            }
            else
            {
               kappa1 = 1.0; 
               kappa2 = 1.0;
               temp = 1.0/lambda;

               for( h = 0; h <= m; h++ )
               {
                  poisson_prob(h, epsilon, lambda*x1, &beta1, &beta_sum1, &bound1, 0,
                               &raising_probs1, &prev_weight1, &weight_sum1, &W1, &w1);
                  poisson_prob(h, epsilon, lambda*x2, &beta2, &beta_sum2, &bound2, 0,
                               &raising_probs2, &prev_weight2, &weight_sum2, &W2, &w2);

                  kappa1 = 1 - beta_sum1;
                  kappa2 = 1 - beta_sum2;

/* printf("kappa1:%le kappa2:%le\n", kappa1, kappa2); */

                  if( h )
                     temp *= h/lambda;
               }

               exact1 = temp*kappa1;
               exact2 = temp*kappa2;
            }

	    temp = b*exact2 - b*exact1;
            psi_exact += temp; 
	    
	    if (q < 1.0e-34)
	    {
	       (*omega_factors)[0] += c;
	       (*psi_factors)[0] += temp;
	    }
	    else
	    {
               /* allocate memory for vector copies */
               tmp_size = lsize;
               omega_tmp = (double *)Malloc(tmp_size * sizeof(double));
               psi_tmp = (double *)Malloc(tmp_size * sizeof(double));

               /* compute alpha-factors */
               qpl = q+lambda;
               qplx1 = qpl*x1;
               qplx2 = qpl*x2;

#ifdef BOUNDS
               max_bounds = old_bounds = 0;
#endif
               run = 0;
               do
	       {
                  /**********************************************/
                  /*     	changed for agnes		*/
                  /**********************************************/
                  /* fprintf(stderr,"Performing run %d\n",run); */
                  /**********************************************/
                  gamma = 1.0/qpl;

                  kappa1 = 1.0;
                  kappa2 = 1.0;

                  /* clear vector */
                  for( k = 0; k < tmp_size; k++ )
	          {
                     omega_tmp[k] = (double )0.0;
                     psi_tmp[k] = (double )0.0;
                  }

                  for( h = 0; h <= m; h++ )
                  {
	             poisson_prob(h, epsilon, qplx1, &beta1, &beta_sum1, &bound1,run,
                                  &raising_probs1, &prev_weight1, &weight_sum1, &W1, &w1);
                     poisson_prob(h, epsilon, qplx2, &beta2, &beta_sum2, &bound2,run,
                                  &raising_probs2, &prev_weight2, &weight_sum2, &W2, &w2);

                     if( h )
                        gamma *= h/qpl;

                     kappa1 = 1 - beta_sum1;
                     kappa2 = 1 - beta_sum2;

/* printf("kappa1:%le kappa2:%le\n", kappa1, kappa2); */

                  }

                  alpha1 = gamma*kappa1;
                  alpha2 = gamma*kappa2;
                  diff = alpha2-alpha1;
                  /* printf("diff:%le\n",diff); */

                  omega_tmp[0] += c*diff;
                  psi_tmp[0] += b*diff;

                  error1 = exact1-alpha1;
                  error2 = exact2-alpha2;

                  /* reset flags */
                  prec_reached = 0;
                  poisson1_pos = poisson2_pos = 1;
                  bounds_reached = 0;
#ifdef BOUNDS
                  old_bounds = max_bounds;
#endif

                  for( k = 1; !prec_reached && poisson2_pos && !bounds_reached &&
				(k < MAXINDEX); k++ )
                  {
                     if( (k+m) > bound1 && qplx1 > 0.0 )
                        bounds_reached = 1;
                     if( (k+m) > bound2 && qplx2 > 0.0 )
                        bounds_reached = 1;
                     if( bounds_reached )
                     {
                        fprintf(stderr,"\t\t\t\t\t warning: bounds reached, k=%d\n", k);
                        continue;
                     }

                     poisson_prob(k+m, epsilon, qplx1, &beta1, &beta_sum1, &bound1, run, 
                                  &raising_probs1, &prev_weight1, &weight_sum1, &W1, &w1);
                     poisson_prob(k+m, epsilon, qplx2, &beta2, &beta_sum2, &bound2, run,
                                  &raising_probs2, &prev_weight2, &weight_sum2, &W2, &w2);

                     gamma *= (((double) (k+m))/k) * (q/qpl);
                     kappa1 = 1 - beta_sum1;
                     kappa2 = 1 - beta_sum2;
                 
/*      printf("kappa1:%le kappa2:%le\n", kappa1, kappa2); */

               if( kappa1 < 0.0 && poisson1_pos)
                     {
                        fprintf(stderr,"Warning: 1-Poisson-probs negative on left side\n");
                        poisson1_pos = 0;
                     }
                     if( kappa2 < 0.0 )
		     {
		        fprintf(stderr,"Warning: 1-Poisson-probs negative on right side\n");
		        poisson2_pos = 0;
		        continue;
		     }

                     alpha1 = (poisson1_pos ? gamma*kappa1 : 0.0);
                     alpha2 = gamma*kappa2;
                     diff = alpha2-alpha1;

                     /* printf("diff:%le k:%u beta1:%le beta2:%le\n",diff,k,*beta1,*beta2); */                    

                     /* enlarge lists if necessary */
                     if( k >= tmp_size )
                     {
                        tmp_size += ALLOC_SIZE;
                        omega_tmp = (double *)MyRealloc(omega_tmp,tmp_size * sizeof(double));
                        psi_tmp = (double *)MyRealloc(psi_tmp,tmp_size * sizeof(double));
                     }
                     omega_tmp[k] += c*diff;
                     psi_tmp[k] += b*diff;

                     /* calculate error */
                     error1 = (poisson1_pos ? error1-alpha1 : 0.0);
                     error2 -= alpha2;
                     prec_reached = (error1 < epsilon) && (error2 < epsilon);
                  }

                  if( k-1 > *R )
                     *R = k-1;

#ifdef BOUNDS
                  max_bounds = (bound1 > bound2 ? bound1 : bound2);
#endif
                  run++;
               } while( !prec_reached && run < 10 ); 

               /* enlarge array if necessary */
               if( lsize < tmp_size )
               {
                  lsize = tmp_size;
                  *omega_factors = (double *)MyRealloc(*omega_factors,lsize * sizeof(double));
                  *psi_factors = (double *)MyRealloc(*psi_factors,lsize * sizeof(double));
	       }

               /* add temp. vector to factors */
               for( k = 0; k < lsize; k++ )
	       {
                  /* printf("k:%u omega:%le psi:%le\n",k,omega_tmp[k],psi_tmp[k]); */
                  (*omega_factors)[k] += omega_tmp[k];
                  (*psi_factors)[k] += psi_tmp[k];
               }
               free(omega_tmp);
               free(psi_tmp);
	    } /* if q < 1e-34 */
	 } /* for ... expo */
      } /* for ... power */
   } /* for ... expoly */

   /* compute resulting error bounds */
   *omega_error = 1.0;
   *psi_error = psi_exact;

   for( k = 0; k <= *R; k++ )
   {
      *omega_error -= (*omega_factors)[k];
      *psi_error -= (*psi_factors)[k];
   }

#ifdef DEBUG_FACTORS
   {
      char fname[256];
      FILE *fp;
      double o_sum = 0.0, p_sum = 0.0;

      sprintf(fname,"Factors.%u", getpid());
      if( fp = fopen(fname,"w") )
      {
         fprintf(fp,"R = %u, omega_error = %g, psi_error = %g\n", *R,*omega_error,*psi_error);
         for( k = 0; k <= *R; k++ )
	 {
            o_sum += (*omega_factors)[k];
            p_sum += (*psi_factors)[k];

            fprintf(fp,"k = %d: omega = %g, o_sum = %g, psi = %g, p_sum = %g\n", k,(*omega_factors)[k],o_sum,(*psi_factors)[k],p_sum);
	 }

         fclose(fp);
      }
   }
#endif

   /*
    * The End.
    */
}

