 
#include "module.h"
#include "main.h"
#include "editor.h"
#include "net.h"
#include "page.h"
#include "netcl.h"
#include "node.h"
#include "int.h"
#include "list.h"

#include "relation.h"
#include "error.h"
#include "netcls.h"
#include "includer.h"

//-----------------------------------------------------------------------------
// Das Objekt NodeMeltInfo beschreibt eine bereits vorgenommene Verschmelzung
// es wird der Name des verschmolzenen und des verbliebnenen Knotens
// gespeichert.
//
class NodeMeltInfo 
{
public:
  Ident melted, real;
  bool operator == (const NodeMeltInfo& m) { return m.melted == melted; }
  bool operator != (const NodeMeltInfo& m) { return m.melted != melted; }
  NodeMeltInfo(Ident& m, Ident& r): melted(m), real(r) {}
};

#define NoExplicitInstanciation
#include "linked.cc"
OINSTANCIATE(NodeMeltInfo)

typedef Map<Node, Node> Node2NodeMap;
typedef LinkedList<NodeMeltInfo> MeltInfos;

//-----------------------------------------------------------------------------

class SymbolTranslator: public Module
{
public:
  SymbolTranslator(Net& n): Module(n, "Translate into AHCPN") {}
  bool Init() { return net.Prime()->Class()->Name() == "APSYS_Prime"; }
  bool Open(Editor&,Page&) { return true; }
  bool Change(Editor&,Page&,Page&) { return true; }
  void Call(Editor&, Page&);

  static void TranslateAPSYS(Net* source, Net* target, Error& error);
  
  static void TranslateAPSYSPage(Page* source, Page* target, Page* struc,
    Page* nstruc, Node2NodeMap& instances, MeltInfos& melted, Error& error);

  static void TranslateSymbol(Node* symbol, Page* target, Page* struc,
    Page* wpl, Node2NodeMap& instances, Error& error);

  static Node* FindStructNode(Node* node, Page* struc, Error& error);

  static void MeltNodes(Node* symbol1, Node* symbol2, Connector* conn,
    Position pos, Page* struc, Page* nstruc, Node2NodeMap& instances,
    MeltInfos& melted, Error& error);

  static int CheckConnection(Node* symbol1, Node* symbol2, Connector* conn,
    Page* struc, Error& error);

  static void LayoutPlaces(Page* page);
};

MODULE_NOTIFY(SymbolTranslator)

//-----------------------------------------------------------------------------
// Der Aufruf
//
void SymbolTranslator::Call(Editor& editor, Page&)
{
  Error error(editor);
  Net* target = new Net(NetClass::GetNetClass("AHCPN_Prime")); // Netz erzeugen
  TranslateAPSYS(&net, target, error);            // bersetzer aufrufen
  if (error) delete target;
  else target->OpenEditor();
}

//-----------------------------------------------------------------------------
// TranslateAPSYS bersetzt ein APSYS-Netz in ein AHCPN.
//
void SymbolTranslator::TranslateAPSYS(Net* source, Net* target, Error& error)
{
  Node* struc = 0;
  Node2NodeMap instances;                        
  MeltInfos melted;                            

  Page* prime = source->Prime();
  Iterator<Node*> nodes = prime->Nodes();
  Node* node;

  for (node = nodes; node; node=++nodes)          // Strukturseite suchen
    if (node->Class()->Name()=="Structure"
        && *node->Attr("active")) {
      if (struc) {                                
        error(Error::TooManyStructures, node); return; } // zu viele?
      else struc = node;
  }
  if (!struc) {                                   // keine Struktur gefunden?
    error(Error::NoStructure, prime); return; }

  Page* strucp = struc->SubPage();
  Node* new_struc = new Node(target->Class()->GetNodeClass("Structure"),
                             target->Prime(), struc->Name(), struc->Mid());

  TranslateAPSYSPage(strucp, new_struc->SubPage(),// Strukturseite bersetzen
                     strucp, new_struc->SubPage(), instances, melted, error);
  if (error) return;
  
  for (node = nodes; node; node=++nodes)          // Alle Arbeitsplanseiten
    if (node->Class()->Name()=="Workplan"         // bersetzen
        && *node->Attr("active")) {
      Node* new_wpl = new Node(target->Class()->GetNodeClass("Workplan"),
                               target->Prime(), node->Name(), node->Mid());

      TranslateAPSYSPage(node->SubPage(), new_wpl->SubPage(), strucp,
                         new_struc->SubPage(), instances, melted, error);
      if (error) return;
  }
  LayoutPlaces(new_struc->SubPage());
}

//-----------------------------------------------------------------------------
// TranslateAPSYSPage bersetzt eine Seite (Struktur- oder Arbeitsplan) des
// APSYS-Netzes in eine Seite des AHCPNs.
//
void SymbolTranslator::TranslateAPSYSPage(Page* source, Page* target, 
       Page* struc, Page* nstruc, Node2NodeMap& instances, MeltInfos& melted,
       Error& error)
{
  bool wpl = source!=struc;
  
  Iterator<Node*> nodes = source->Nodes();
  Node* node;
    
  for (node=nodes; node; node=++nodes) {          // Fr alle Knoten
    if (node->Class()->GroupName() == "Defs") {   // Definitionen kopieren
      NodeClass* nodecl = target->Class()->GetNodeClass(node->Class()->Name());
      node->Copy(target)->SetClass(nodecl);
    }                                             // Verzweigungsknoten kommen
    else if (node->Class()->Name() != "Branch") {                    // spter
      TranslateSymbol(node, target, struc, wpl?source:0, // Fr alle Symbole 
                      instances, error);          // Templates instanziieren
      if (error) return;
    }        
  }       

  if (!wpl) return;                               // Im Arbeitsplan...

  Iterator<Connector*> conns = source->Connectors(); // Fr alle Kanten
  for (Connector* conn=conns; conn; conn=++conns)    // zwischen 2 Symbolen
    if (conn->GetInput()->Class()->Name() != "Branch"
        && conn->GetOutput()->Class()->Name() != "Branch") {
      MeltNodes(conn->GetInput(), conn->GetOutput(), // Verschmelzung vornehmen
                conn, conn->Mid(), struc, nstruc, instances, melted, error);
      if (error) return;
  }

  for (node=nodes; node; node=++nodes) 
    if (node->Class()->Name() == "Branch") {      // Fr alle Verzweigungsknotn
      Iterator<Connector*> ins = node->Inputs();
      Iterator<Connector*> outs = node->Outputs();
      
      for (Connector* in=ins; in; in=++ins)       // Alle beteiligten Stellen
        for (Connector* out=outs; out; out=++outs) { // verschmelzen
          MeltNodes(in->GetInput(), out->GetOutput(), out, node->Mid(),
                    struc, nstruc, instances, melted, error);
          if (error) return;
      }
  }
  LayoutPlaces(target);
}

//-----------------------------------------------------------------------------
// TranslateSymbol instanziiert fr ein Symbol aus dem APSYS-Netz das
// ensprechende Template im AHCPN.
//
void SymbolTranslator::TranslateSymbol(Node* symbol, Page* target, Page* struc,
       Page* wpl, Node2NodeMap& instances, Error& error)
{
  Text in, out, brcond, brweight;
  Connector* inconn = symbol->Inputs();
  Connector* outconn = symbol->Outputs();
  Struct* attr=0;
  
  if (wpl) {                                      // Im Arbeitsplan...
    if (inconn) {                                 // Die Attribute der ein- 
      in = charP(*inconn->Attr("weight"));        // und ausgehenden Kanten   
      if (in == "") { error(Error::ArcWithoutWeight, inconn); return; }// lesen

      brcond = charP(*inconn->Attr("brcond"));
      brweight = charP(*inconn->Attr("brweight"));
    }
    else in = "0'e";

    if (outconn) { 
      out = charP(*outconn->Attr("weight"));
      if (out == "") { error(Error::ArcWithoutWeight, outconn); return; }
    }
    else out = "0'e";
    
    attr = symbol->Attribute();
  }
                                                  // Assoziierten Knoten finden
  Node* struct_node = FindStructNode(symbol, struc, error);
  if (error) return;
                                                  // Template des Symbols findn
  Page* tplnet = NetClassCreator::FindTemplateNet(symbol->Class()->Name());
  if (!tplnet) return;

  error.SetContext(symbol);
  Node* subtrans = TemplateIncluder::InstanciateTemplate( // Template instanz.
          target, tplnet, symbol->Name(), symbol->Mid(), 
          struct_node->Attribute(), attr, struc, wpl, &in, &out, error);
  if (error) return;
  error.SetContext(0);
  
  instances.Add(symbol, subtrans);                // erzeugte Vertretertrans.
                                                  // merken
  if (wpl) {
    if (subtrans->NumInputs() > 1) {              // Anschlubed. n. eindeutig
      error(Error::AmbigousConnConditons, symbol, in); return; } 
      
    if (!subtrans->NumInputs() && inconn) {       // Kante nicht erlaubt
      error(Error::ArcNotInTemplate, inconn, in); return; }

    if (subtrans->NumOutputs() > 1) {
      error(Error::AmbigousConnConditons, symbol, 0); return; } 

    if (!subtrans->NumOutputs() && outconn) {
      error(Error::ArcNotInTemplate, outconn, 0); return; }

    if (inconn) *inconn->Attr("weight") = charP(in); // Attribute in die Kanten
    if (brcond) *inconn->Attr("brcond") = charP(brcond); // der Vertretertrans
    if (brweight) *inconn->Attr("brweight") = charP(brweight); // eintragen

    if (outconn) *outconn->Attr("weight") = charP(out);
  }
}

//-----------------------------------------------------------------------------
// FindStructNode sucht zu einem Symbol im Arbeitsplan das assoziierte
// Symbol in der Struktur heraus.
//
Node* SymbolTranslator::FindStructNode(Node* node, Page* struc, Error& error)
{
  Iterator<Node*> snodes = struc->Nodes();
  for (Node* snode=snodes; snode; snode=++snodes) 
    if (snode->Name()==node->Name()) {
      if (snode->Class()->Name() != node->Class()->Name()) 
        error(Error::IllegalStructRef2, node, snode);// In der Struktur falsch?
      return snode;
    }
  error(Error::NoStructRef2, node);               // oder nicht vorhanden?
  return 0;
}

//-----------------------------------------------------------------------------
// MeltNodes verschmilzt jeweils zwei Knoten im erzeugten AHCPN-Arbeitsplan.
// Diese Verschmelzung wird dann auch in der Struktur vorgenommen.
//
void SymbolTranslator::MeltNodes(Node* symbol1, Node* symbol2, Connector* conn,
       Position /*pos*/, Page* struc, Page* nstruc, Node2NodeMap& instances, 
       MeltInfos& melted, Error& error)
{                                                 
  int real = CheckConnection(symbol1, symbol2,    // Verschmelzung berprfen
                             conn, struc, error); // und echten Knoten finden
  if (error) return;

  Node* sub1 = instances[symbol1];                // Vertretertransitionen
  Node* sub2 = instances[symbol2];                // heraussuchen
  if (sub1->Name() == sub2->Name()) return;
  
  Node* node1 = ConnectorP(sub1->Outputs())->GetOutput(); // Verbindungsstellen
  Node* node2 = ConnectorP(sub2->Inputs())->GetInput(); // heraussuchen
  
  Node* rnode = (real==1) ? node1 : node2;        
  Node* mnode = (real==2) ? node1 : node2;
  Ident rname = rnode->Name();                    // Namen merken
  Ident mname = mnode->Name();

  //rnode->SetPos(pos); 
  rnode->Melt(mnode);                             // Stellen im APL verschmelzn

  Ident r = rname, m = mname;

  OpaqueIterator<NodeMeltInfo> miter = melted;    // Sind diese Stellen schon
  NodeMeltInfo* inf;                              // mal verschmolzen worden?
  for (inf=miter; inf; inf=++miter) {             // -> echte Stellen finden
    if (r == inf->melted) { r = inf->real; inf=miter; continue; }
    if (m == inf->melted)   m = inf->real;
  }

  if (r==m) return;
  if (m!=mname) {                                 // Verschmelzungskonflikt?
    error(Error::MeltConflict, conn); return; }

  rnode = nstruc->GetNode(charP(r));              // Strukturelle Knoten finden
  mnode = nstruc->GetNode(charP(m));
  
  rnode->Melt(mnode);                             // und ebenfalls verschmelzen

  melted.Append(NodeMeltInfo(mname, r));          // und merken
}

//-----------------------------------------------------------------------------
// CheckConnection sucht fr eine Verbindung zwischen 2 Symbolen die
// entsprechende "Verschmelzungskante" aus der Library heraus und stellt
// so fest, auf welcher Seite der Kante der echte Knoten zu finden ist.
//
int SymbolTranslator::CheckConnection(Node* symbol1, Node* symbol2, 
      Connector* conn, Page* struc, Error& error)
{
  Iterator<Connector*> conns = struc->Connectors();
  Connector* sconn;                               

  for (sconn=conns; sconn; sconn=++conns)         // Assoziierte Kante in der 
    if (sconn->GetInput()->Name() == symbol1->Name() // Struktur suchen
        && sconn->GetOutput()->Name() == symbol2->Name()) break;
  
  if (!sconn) {                                   // Keine gefunden? Fehler!
    error(Error::NoMatchingArc, conn); return 0; }

  Connector* lconn;
  conns = popp_library->Prime()->Connectors();    // Kante in der Library 
  for (lconn=conns; lconn; lconn=++conns) {       // suchen
    if (lconn->GetInput()->Name() == symbol1->Class()->GroupName()
        && lconn->GetOutput()->Name() == symbol2->Class()->GroupName()) break;
  }
  if (!lconn) {                                   // Keine gefunden? Fehler!
    error(Error::IllegalArc, conn); return 0; } 

  return (lconn->Class()->Name()=="FromReal") ? 1 : 2; // KantenTyp auswerten
}

//-----------------------------------------------------------------------------
// LayoutPlaces ordnet die Stellen in dem erzeugten Netz so an, da die in der
// Mitte zwischen den mit ihnen verbundenen Vertretertransitionen liegen.
//
void SymbolTranslator::LayoutPlaces(Page* page)
{
  Iterator<Node*> nodes = page->Nodes();
  Node* node;

  for (node = nodes; node; node=++nodes)          // Fr alle Stellen
    if (node->Class()->Name()=="PPlace" || node->Class()->Name()=="EPlace") {
      
      LinkedList<Node*> trans;

      Iterator<Connector*> ins = node->Inputs();  // verbundene Transitionen
      for (Connector* in=ins; in; in=++ins)       // heraussuchen
        if (!trans.IsIn(in->GetInput()))
          trans.Append(in->GetInput());

      Iterator<Connector*> outs = node->Outputs();
      for (Connector* out=outs; out; out=++outs)
        if (!trans.IsIn(out->GetOutput()))
          trans.Append(out->GetOutput());

      Position mid(0,0);

      Iterator<Node*> titer = trans;
      for (Node* t=titer; t; t=++titer)           // Mittelpunkt bestimmen
        mid = mid + t->Mid();

      if (trans.Length()==2) {
        node->SetPos(mid/trans.Length());         // Position zuweisen
      }
    }
}
