#include <ctype.h>
#include <iostream.h>

#include "module.h"
#include "main.h"
#include "editor.h"
#include "net.h"
#include "page.h"
#include "netcl.h"
#include "node.h"
#include "int.h"
#include "list.h"
#include "relation.h"
#include "error.h"

typedef char* charP;

typedef ValueMap<Ident, Node> Ident2Node;
typedef ValueMap<Ident, Ident> Ident2Ident;
typedef ComplexValueMap<Ident, LinkedList<Node*> > Ident2NodeList;
typedef ComplexMap<Page, ValueMap<Ident, Ident> > Page2Ident2Ident;


class Translator: public Module
{
  static void TranslateNet(Net* source, Page* target, Error& error);

  static void TranslateStructAndWPLs(Page* structure, Iterator<Node*> WPLs,
    Page* page, Error& error);

  static void SearchStructureAndWPLs(
    Page* structure, Iterator<Node*> WPLs, Ident2Node& StructNodes,
    Ident2NodeList& WPLNodes, LinkedList<Node*>& Defs, Error& error);

  static void CopyStructure(Page* structure, Iterator<Node*> nodes, Page* page,
    Ident2Node& TransNodes, Error& error);

  static void TranslateDefinitions(Page* page, Page* structure,
    Iterator<Node*> defs,Page2Ident2Ident& NewDefNames, Error& error);

  static Node* CopyDefinition(Node* def, Page* page, int version);

  static void TranslateArc(Connector* newarc, Connector* arc, Error& error);

  static void TranslatePlace(Node* newplace, Node* place, bool used);

  static void TranslateSubstTrans(Node* subtrans, Node* struc,
    Iterator<Node*> WPLs, Error& error);

  static void TranslateBranchConditions(Node* place, Node* struc,
    Iterator<Node*> WPLs, Page2Ident2Ident& VarSubsts, Error& error);
  
  static void TranslateTransition(Node* trans, Node* struc,
    Iterator<Node*> WPLs, Page2Ident2Ident& VarSubsts, Error& error);

  static Text TranslateDistribution(Node* struc, Node* rtrans,
    Ident2Ident* varsubst);

  static void TranslateProductArcs(bool input, Iterator<Connector*> arcs,
    Iterator<Connector*> WPLarcs, Ident2Ident* varsubst, Text& binds,
    Error& error);

  static void TranslateElemArcs(bool input, Iterator<Connector*> arcs,
    Text& binds);

  static void SubstVars(Text& expr, Ident2Ident* subst);

public:
  Translator(Net& n): Module(n, "Translate into HCPN") {}
  bool Init() { return net.Prime()->Class()->Name() == "AHCPN_Prime"; }
  bool Open(Editor&,Page&) { return true; }
  bool Change(Editor&,Page&,Page&) { return true; }
  void Call(Editor&, Page&);
};

MODULE_NOTIFY(Translator)

//-----------------------------------------------------------------------------
// Call wird vom Editor aufgerufen, wenn der entsprechende Meneintrag
// angewhlt wurde.
//
void Translator::Call(Editor& editor, Page&)
{
  Error error(editor);
  Net* hcpn = new Net(NetClass::GetNetClass("HCPN")); // Netz erzeugen
  TranslateNet(&net, hcpn->Prime(), error);           // bersetzer aufrufen
  if (error) delete hcpn;
  else hcpn->OpenEditor();
}

//-----------------------------------------------------------------------------
// TranslatePrime bersetzt das bergebene AHCPN und schreibt das Ergebins
// auf die gewnschte Seite.
//
void Translator::TranslateNet(Net* source, Page* target, Error& error)
{
  Page* prime = source->Prime();
  Iterator<Node*> nodes = prime->Nodes();
  Page* structure = 0;                            // Strukturseite
  LinkedList<Node*> WPLs;                         // Arbeitsplan-Knoten

  for (Node* node=nodes; node; node= ++nodes) {   // Alle Knoten in der
    if (node->Class()->Name()=="Structure"        // Primepage durchsuchen
        && *node->Attr("active")) {     
      if (structure) {                            // Mehr als eine Struktur?
        error(Error::TooManyStructures, node); return; }
      else structure = node->SubPage();
    }
    else if (node->Class()->Name()=="Workplan"    // Alle APLs in die Liste
             && *node->Attr("active"))
           WPLs.Append(node);
  }

  if (!structure) {                               // Keine Struktur?
    error(Error::NoStructure, prime); return; }

  TranslateStructAndWPLs(structure, WPLs, target, error);// bersetzer aufrufen
}

//-----------------------------------------------------------------------------
// TranslateStructAndWPLs bersetzt eine Struktur und die zugehrigen
// Arbeitsplne in ein HCPN. Unterseiten werden durch einen rekursiven Aufruf
// dieser Funktion mit bersetzt. Um diesen Rekursiven Aufruf zu erleichtern,
// ist WPLs eine Liste von Knoten, deren Unterseiten die Arbeitsplne enthalten
//
void Translator::TranslateStructAndWPLs(Page* structure, Iterator<Node*> WPLs,
                                        Page* page, Error& error)
{
  Ident2Node StructNodes;
  Ident2NodeList WPLNodes;
  LinkedList<Node*> Defs;
                                                  // Zuerst werden alle Knoten
  SearchStructureAndWPLs(structure, WPLs,         // auf der Struktur und in d.
     StructNodes, WPLNodes, Defs, error);         // Arbeitsplnen in diesen
  if (error) return;                              // Datenstrukturen erfat

  Ident2Node TransNodes;                          // Dann wird die Seite
  CopyStructure(structure, StructNodes.Vals(),    // kopiert und die Netzklasse
                page, TransNodes, error);         // angepat
  if (error) return;

  Page2Ident2Ident NewDefNames;                   // Jetzt werden die Defs
  TranslateDefinitions(page, structure, Defs,     // hinzugefgt und ggf.
                       NewDefNames, error);       // umbenannt
  if (error) return;

  Ident* name;
  Node* node;
  
  MapIterator<Ident, Node> tniter = TransNodes;   // und dann fr alle Knoten:
  for (tniter.Start(name, node); node; tniter.Step(name, node)) {
    Ident nodeclass = node->Class()->Name();
    
    if (nodeclass=="EPlace" || nodeclass=="PPlace") { // Stelle bersetzen
      TranslatePlace(node, StructNodes[name], *WPLNodes[name]);
      if (nodeclass=="PPlace")                    // Verzweigungsbed. bersetzn
        TranslateBranchConditions(node, StructNodes[name], *WPLNodes[name],
                                  NewDefNames, error);
    } 

    else if (nodeclass=="TimedTrans" || nodeclass=="ImmTrans") {
      TranslateTransition(node, StructNodes[name], // Transition bersetzen
                          *WPLNodes[name], NewDefNames, error);
      if (error) return;
    }

    else if (nodeclass=="SubTrans") {            // Vertretertrans. bers.
      TranslateSubstTrans(node, StructNodes[name], *WPLNodes[name], error);
      if (error) return;
    }
  }                                               // fertig !
  page->Deselect();
}

//-----------------------------------------------------------------------------
// SearchStructureAndWPLs durchsucht Struktur und Arbeitsplne nach
// korrespondierenden Knoten und trgt diese in die bergebenen Tabellen ein.
// Dabei werden auch erste Checks auf Fehler durchgefhrt.
//
void Translator::SearchStructureAndWPLs(
       Page* structure, Iterator<Node*> apls,     // Struktur und Arbeitsplne
       Ident2Node& StructNodes,                   // Name -> Struktureller
                                                  // Knoten mit diesem Namen
       Ident2NodeList& WPLNodes,                  // Name -> Liste aller
                                                  // Knoten mit diesem Namen
                                                  // in den Arbeitsplnen
       LinkedList<Node*>& Defs,                   // Alle Definitionen auf der
       Error& error )                             // Struktur und den APlnen
{
  Node* node;
  Iterator<Node*> nodes = structure->Nodes();
  
  for (node=nodes; node; node=++nodes)            // Fr alle Struktur-Knoten
    if (node->Class()->Name()=="Def")             // Definitionen nach Defs
      Defs.Append(node);
    else if (node->Class()->Name()=="Comment") {} // Kommentare raus
    else {                                        // Alle anderen Knoten werden
      StructNodes.Add(&node->Name(), node);       // in StructNodes gespeichert
      WPLNodes.AddSection(&node->Name());         // WPLNodes initialisieren
    }

  for (Node* apl=apls; apl; apl=++apls) {         // Fr alle Knoten
    nodes = apl->SubPage()->Nodes();              // in allen Arbeitsplnen
    for (node=nodes; node; node=++nodes) {
      if (node->Class()->Name()=="Def")           // Definitionen nach Defs
        Defs.Append(node);
      else if (node->Class()->Name()=="Comment"){}// Kommentare raus
      else {                                      // Alle anderen Knoten mssen
        Node* related = StructNodes[&node->Name()]; // einen entsprechenden
        if (!related) {                           // Knoten in der Struktur
          error(Error::NoStructRef, node);        // besitzen
          return;
        }
        else if (node->Class()->Name() !=         // Dieser mu vom gleichen
                 related->Class()->Name()) {      // Typ sein
          error(Error::IllegalStructRef, node, related);
          return;
        }
        else WPLNodes[&node->Name()]->Append(node); // Knoten merken
      }
    }
  }
}

//-----------------------------------------------------------------------------
// CopyStructure kopiert die Strukturseite und wandelt es mit Hilfe von
// Node::SetClass und Connector::SetClass in ein HCPN um.
//
void Translator::CopyStructure(Page* structure, Iterator<Node*> nodes,
                               Page* page, Ident2Node& TransNodes,Error& error)
{
  NetClass* HCPN = NetClass::GetNetClass("HCPN");

  for (Node* node=nodes; node; node=++nodes) {    // zuerst alle Knoten
    Node* copy = node->Copy(page);                // kopieren und dabei die
    copy->SetClass(HCPN->GetNodeClass(node->Class()->Name())); // Knotenklasse
    TransNodes.Add(&node->Name(), copy);          // anpassen
  }

  Iterator<Connector*> citer = structure->Connectors();
  for(Connector* arc=citer; arc; arc=++citer) {   // und dann dasselbe fr 
    Connector* copy = arc->Copy(page);            // alle Kanten
    copy->SetClass(HCPN->GetConnectorClass(arc->Class()->Name()));
    TranslateArc(copy, arc, error);               // auerdem gleich bersetzen
    if (error) return;
  }
}

//-----------------------------------------------------------------------------
// TranslateDefinitions kopiert alle Definitionen aus Struktur und Arbeits-
// plnen in die HCPN-Seite. Um Namenskonflikte zu vermeiden, werden dabei die
// Definitionen aus den Arbeitsplnen umbenannt. Diese Namensnderungen werden
// in NewDefNames gespeichert.
// Annahme: die Defs sind in der gleichen Reihenfolge, wie sie von
// SearchStructureAndWPLs gefunden wurden.
//
void Translator::TranslateDefinitions(Page* page, Page* structure,
       Iterator<Node*> defs, Page2Ident2Ident& NewDefNames, Error& error)
{
  Node* def;
  Node* copy;
  Ident2Ident* structdefs = NewDefNames.AddSection(structure);
  Ident2Ident* WPLdefs = NULL;

  for (def=defs; def && def->GetPage() == structure; def=++defs) { 
     copy=CopyDefinition(def, page, 0);           // Alle Defs aus der Struktur
     if (!copy) { error(Error::DefExists, def); return; } // kopieren
     structdefs->Add(&def->Name(), &copy->Name()); // und merken
  }

  for (int WPL_nr = 0; def; def=++defs) {         // Fr alle Defs aus den APLs
    WPLdefs = NewDefNames[def->GetPage()];        
    if (!WPLdefs) {                               // Anderer Arbeitsplan?
      WPLdefs = NewDefNames.AddSection(def->GetPage()); WPL_nr++; }

    if ((*structdefs)[&def->Name()]) {            // Gleiche Def in Struktur 
      error(Error::DefExists, def); return; }     // und APL?

    copy = CopyDefinition(def, page, WPL_nr);     // Definition kopieren
    if (!copy) {
      error(Error::DefExists, def); return; }     // Unerwarteter Nameskonflikt
    WPLdefs->Add(&def->Name(), &copy->Name());    // Umbenennung merken
 }
}

//-----------------------------------------------------------------------------
// CopyDefinition kopiert eine Definition in die HCPN-Seite und benennt sie um,
// wenn sie aus einem Arbeistplan kam (version>0)
//
Node* Translator::CopyDefinition(Node* def, Page* page, int version)
{
  Node* copy = def->Copy(page);                   // Def kopieren und Klasse
  copy->SetClass(page->Class()->GetNodeClass(def->Class()->Name())); // ndern

  if (version>0) {                                // ggf. Namen ndern
    copy->SetName(def->Name()+"_"+Int(version));
  }
  if (page->NameExists(copy)) {                   // Trotzdem Namenskonflikt?
    delete def; return NULL; }                   
  else return copy;

}

//-----------------------------------------------------------------------------
// TranslateArc bersetzt eine Kante
//
void Translator::TranslateArc(Connector* newarc, Connector* arc, Error& error)
{
  if (!arc->Virtual()) {
    if (arc->Class()->Name()=="EArc")
      if (*arc->Attr("weight")=="")               // Eine Elementarkante muss 
        error(Error::ArcWithoutWeight, arc);      // ein Gewicht haben

    if (arc->Class()->Name()=="PArc")             // Bei Produktkanten wird
      *newarc->Attr("unused") = true;             // erst spter ber die
  }                                               // Benutzung entschieden
}

//-----------------------------------------------------------------------------
// TranslatePlace bersetzt eine Stelle
//
void Translator::TranslatePlace(Node* newplace, Node* place, bool used)
{
  if (!place->Virtual()) {                        // Attribute kopieren
    if (!used && place->Class()->Name()=="PPlace") // Hat eine Produktstelle
      *newplace->Attr("unused") = true;           // keine korresponierenden
  }                                               // Stellen, so ist sie
}                                                 // unbenutzt

//-----------------------------------------------------------------------------
// TranslateSubstTrans bersetzt eine Vertretertransition. Dabei wird
// TranslateStructandWPLs rekursiv aufgerufen.
//
void Translator::TranslateSubstTrans(Node* subtrans, Node* struc,
                                     Iterator<Node*> WPLs, Error& error)
{
  TranslateStructAndWPLs(struc->SubPage(), WPLs,  // Zuerst die Unterseite
                         subtrans->SubPage(), error); // bersetzen
  if (!bool(WPLs)) {                              // Keine Arbeitsplne?
    *subtrans->Attr("unused") = true;             //  -> unbenutzt
    return;
  }
  Iterator<Node*> nodes = subtrans->SubPage()->Nodes(); // Fr alle virtuellen
  for (Node* node=nodes; node; node=++nodes)      // Knoten auf der Unterseite
    if (node->Virtual()) {
      Node* real = node->Real();
      if (node->Class()->Name()=="EPlace" ||      // Wenn der Knoten auf der
          !*node->Attr("unused"))                 // Unterseite benutzt ist,

        for (int from=0; from<2; from++)          // und dort aus(ein-)gehende
          if (bool(from ? node->Outputs()         // Kanten hat, dann finde
                        : node->Inputs()))        // die entsprechende (virt.)
          {                                       // Kante zwischen dem realen
            Iterator<Connector*> conns            // Knoten und der Vertreter-
              = from ? subtrans->Inputs() : subtrans->Outputs(); // Transition
            for (Connector* conn=conns; conn; conn=++conns) {
              Node* node2 = from ? conn->GetInput() : conn->GetOutput();
              if (node2==real || node2->Virtual() && node2->Real()==real)
                *conn->Attr("unused") = false;    // und markiere diese benutzt
            }
          }
    }
}

//-----------------------------------------------------------------------------
// TranslateBranchConditions erzeugt aus den Verzweigungsbedingungen der
// Produktkanten an einer Stelle die entsprechende Entscheidungstransition
//
void Translator::TranslateBranchConditions(Node* place, Node* /*struc*/,
       Iterator<Node*> WPLs, Page2Ident2Ident& VarSubsts, Error& /*error*/)
{
  List bindings;
  
  for (Node* rplace=WPLs; rplace; rplace=++WPLs) {// Fr alle korresp. Stellen
    Ident2Ident* varsubst = VarSubsts[rplace->GetPage()];

    Iterator<Connector*> inputs = rplace->Inputs();
    Iterator<Connector*> outputs = rplace->Outputs();
    LinkedList<Connector*> ins, outs;

    if (int(outputs)<=1) continue;
    
    for (Connector* in=inputs; in; in=++inputs) {
      Text from = charP(*inputs.First()->Attr("weight"));
      for (Connector* out=outputs; out; out=++outputs) {
        Text to = charP(*out->Attr("weight"));
        Text guard = charP(*out->Attr("brcond"));
        Text weight = charP(*out->Attr("brweight"));
        if (to==from) continue;
        
        Text* entry = new Text;
        if (guard!="") SubstVars(guard, varsubst);
        else guard="-";
        
        *entry += charP(Text("(") + guard + Text(") ")
                      + Text("in#") + place->Name()
                      + Text("=") + from + Text(", out#")
                      + place->Name() + Text("=")
                      + to + (weight=="" ? Text("") : Text("; ") + weight)
                      + "; Immediate(1)");
        bindings.Append(entry);
      }
    }
  }
  
  if (!bindings.Entries()) return;

  NetClass* HCPN = NetClass::GetNetClass("HCPN"); 
  NodeClass* nodecl = HCPN->GetNodeClass("ImmTrans");
  ConnectorClass* conncl = HCPN->GetConnectorClass("PArc");
  
  Page* page = place->GetPage();
  Position pos = place->Mid() + Position(-40,40); // layouten
  
  Node* trans = new Node(nodecl, page, 0, pos);   // Entscheidungstransition
                                                  // erzeugen
  *ListP(trans->Attr("bindings")) = bindings;     // bergangstabelle kopieren

  Connector* conn1 = new Connector(conncl, page); // Kanten erzeugen
  Connector* conn2 = new Connector(conncl, page);
 
  conn1->SetConnections(place, trans);            // Kanten verbinden
  conn2->SetConnections(trans, place);
  
  trans->RearrangeConnections();
}

//-----------------------------------------------------------------------------
// TranslateTransition bersetzt eine Transition. Dabei wird fr jede
// korrespondierende Transition aus den Arbeitsplnen jewals ein bergang in
// der bergangstabelle erzeugt. Dabei werden die Variablenumbennennungen
// bercksichtigt.
//
void Translator::TranslateTransition(Node* trans, Node* struc,
       Iterator<Node*> WPLs, Page2Ident2Ident& VarSubsts, Error& error)
{
  Connector* conn;

  Text EArcBinds = "";
  TranslateElemArcs(true, struc->Inputs(), EArcBinds); // Kantenbelegungen der
  TranslateElemArcs(false, struc->Outputs(), EArcBinds); // Elementarkanten

  if (!bool(WPLs)) {
    bool unused = false;                          // Keine WPLs? Wenn
    Iterator<Connector*> conns = struc->Inputs(); // mindestens eine Produkt-
    for (conn=conns; conn; conn=++conns)          // kante angeschlossen ist,
      if (conn->Class()->Name()=="PArc")          // dann Transition als
        unused = true;                            // unbenutzt markieren

    conns = struc->Outputs();
    for (conn=conns; conn; conn=++conns)
      if (conn->Class()->Name()=="PArc")
        unused = true;
                                                  // Ist gar keine Kante
    if (!bool(struc->Inputs()) && !bool(struc->Outputs())) // angeschlossen?
      unused = true;                              // -> unbenutzt

    if (unused) *trans->Attr("unused") = true;    // Unbenutzte Transitionen
    else {                                        // kriegen keinen Eintrag
      Text* entry = new Text(EArcBinds);          // der bergangstabelle
      Text distr = TranslateDistribution(struc, 0, 0);
      *entry += distr;
      ((List*)trans->Attr("bindings"))->Append(entry); // Eintrag erzeugen
    }
  }

  for (Node* rtrans=WPLs; rtrans; rtrans=++WPLs) {// Fr alle korresp. Trans.
    Ident2Ident* varsubst = VarSubsts[rtrans->GetPage()];
                                                  // Eintrag in der bergangs-
    Text* entry = new Text;                       // tabelle erzeugen:
    Text guard = charP(*rtrans->Attr("guard"));
    if (guard!="") {                              // Zuerst die lokale Schalt-
      SubstVars(guard, varsubst);                 // bedingung
      *entry += charP(Text("(") + guard + ") "); }
    else *entry += "(-) ";
                                                  // Dann die Kantenbelegungen:
    Text ArcBinds = EArcBinds;                    //  Elementarkanten

    TranslateProductArcs(true, trans->Inputs(),   //  eingehende Produktkanten
      rtrans->Inputs(), varsubst, ArcBinds, error);
    if (error) { delete entry; return; }

    TranslateProductArcs(false, trans->Outputs(), //  ausgehende Produktkanten
      rtrans->Outputs(), varsubst, ArcBinds, error);
    if (error) { delete entry; return; }

    if (!bool(ArcBinds)) { delete entry; continue; } // Keine Kantenbelegungen?

    Text distr = TranslateDistribution(struc, rtrans, varsubst);
    *entry += ArcBinds;                           // Eintrag erzeugen
    *entry += distr;

    ((List*)trans->Attr("bindings"))->Append(entry); // Eintrag speichern
  }
}

//-----------------------------------------------------------------------------
// TranslateDistribution erzeugt fr die bergangstabelle eine textuelle
// Reprsentation der Schaltzeit und Schaltzeitverteilung einer Transition.
//
Text Translator::TranslateDistribution(Node* struc, Node* rtrans,
       Ident2Ident* varsubst)
{
  Text distr = "Immediate", delay = "", weight = "";
  Text result;

  bool timed = struc->Class()->Name()=="TimedTrans";

  if (rtrans) {
    if (timed) {                                  // Zeitbehaftete Transitionen
      distr = charP(*rtrans->Attr("distr"));      // haben eine Schaltzeit und
      delay = charP(*rtrans->Attr("delay"));      // eine Schaltzeitverteilung
    }
    else                                          // Zeitlose Transitionen
      delay = charP(*rtrans->Attr("priority"));   // haben eine Prioritt
                                                  // (auch nach delay)
    weight = charP(*rtrans->Attr("weight"));
  }
  result = "";

  if (weight!="") {
    SubstVars(weight, varsubst);
    result += charP(Text("; ") + weight);         // Gewicht anhngen
  }

  if (delay=="") {                                // kein Delay angegeben
    if (timed) {                                  // -> Default-Werte aus der
      distr = charP(*struc->Attr("distr"));       // Struktur benutzen
      delay = charP(*struc->Attr("delay"));
    }
    else delay = charP(*struc->Attr("priority"));
  }
  else SubstVars(delay, varsubst);                // Variablen ersetzen

  result += charP(Text("; ") + distr              // Delay und Verteilung
            + Text("(") + delay + Text(")"));     // anhngen

  return result;
}

//-----------------------------------------------------------------------------
// TranslateProductArcs erzeugt aus den Beschriftungen der Kanten im Arbeits-
// plan eine Liste von textuellen Reprsentationen dieser Kantenbelegungen
// fr die bergangstabelle. Etwaige Variablenumbennenungen werden
// bercksichtigt.
//
void Translator::TranslateProductArcs(bool input, Iterator<Connector*> arcs,
       Iterator<Connector*> WPLarcs, Ident2Ident* varsubst,
       Text& binds, Error& error)
{
  Connector* arc, *wparc;

  for (wparc=WPLarcs; wparc; wparc=++WPLarcs) {   // Fr jede APL-Kante
    for (arc=arcs; arc; arc=++arcs)               // die Korresponierende Kante
      if (arc->GetInput()->Name() == wparc->GetInput()->Name() // suchen
      && arc->GetOutput()->Name() == wparc->GetOutput()->Name()) break;

    if (!arc) { error(Error::NoMatchingArc, wparc); return; } // keine gefunden
    *arc->Attr("unused") = false;

    if (binds!="") binds += ", ";
    if (input) binds += charP(Text("in#") + arc->GetInput()->Name());
    else       binds += charP(Text("out#") + arc->GetOutput()->Name());

    Text weight = charP(*wparc->Attr("weight"));
    if (weight=="") { error(Error::ArcWithoutWeight, wparc); return; }
    SubstVars(weight, varsubst);
    binds += charP(Text("=") + weight);
  }
}

//-----------------------------------------------------------------------------
// TranslateElemArcs erzeugt aus den Beschriftungen der Elementarkanten in der
// Struktur eine Liste von textuellen Reprsentationen dieser Kantenbelegungen
//
void Translator::TranslateElemArcs(bool input, Iterator<Connector*> arcs,
   Text& binds)
{
  Connector* arc;

  for (arc=arcs; arc; arc=++arcs)
    if (arc->Class()->Name()=="EArc") {
      if (binds!="") binds += ", ";
      if (input) binds += charP( Text("in#") + arc->GetInput()->Name() );
      else       binds += charP( Text("out#") + arc->GetOutput()->Name() );

      Text weight = charP(*arc->Attr("weight"));
      if (weight=="") binds += "=1'e";
      else binds += charP(Text("=(") + weight + Text(")*1'e"));
    }
}

//-----------------------------------------------------------------------------
// SubstVars benennt in dem bergebenen Ausdruck alle Variablen nach der
// bergebenenen Namensabbildung um.
//
void Translator::SubstVars(Text& expr, Ident2Ident* subst)
{
  if (subst) {
    Text newexpr = "";
    Ident ident = "";

    for (char* c=expr; ; c++) {                   // Den String durchgehen
      if (isalpha(*c) || *c=='_' || (ident!="" && isdigit(*c)))
        ident+= *c;                               // Idents rausfiltern
      else {
        if (ident!="") {                          // Ident gefunden?, dann
          if ((*subst)[&ident]) ident = *(*subst)[&ident]; // ggf. ersetzen
          newexpr+=charP(ident); ident="";        
        }
        if (*c) newexpr+= *c; else break;
      }
    }
    expr = newexpr;                               // Neuen Ausdruck bergeben
  }
}


