/******************************************************************************

  filename :	simu_service.c

  purpose  :	service functions for the simulation engine

  author   :    Christian Kelling
  		Technical University Berlin
		Computer Science Department

  date     :	June 1993

  exported functions:
	void	compute_connections
	void	define_measures
	char	*er_malloc
	void	compute_connections
	void	print_net
	void 	read_net

 local functions
	void	searc_place
	int	next_line
	void 	check_and_add_aff
	void	check_and_add_measure
        void    all_affected

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../../../Shared/src/netio/netio.h"
#include "sim_master_types.h"

extern	int	no_places;
extern	int	no_trans;
extern  int	no_ena_fcts;
extern  int	no_md_fcts;
extern  int	no_del_par;
extern  int	no_mark_par;
extern  int	no_md_arcs;
extern	int	no_measures;
extern	double  current_time;
extern	place *places;
extern	transition *transitions;
extern  FILE	*net_file;
extern  FILE	*inf_file;
extern  FILE	*erc_file;

extern	char	model_name[]; 

extern int *probMeasureId;
extern int numberProbMeasures;

#ifdef CVR
  extern meas_trans_list *meas_trans;
#endif
#ifndef RST
  extern int only_basic_components;
#endif

#define TRUE 1
#define FALSE 0
#define INPUT_TRANS  0
#define OUTPUT_TRANS 1
#define SINGLE_PROB_ESTIMATOR   1
#define JOINED_PROB_ESTIMATOR   2
#define COND_PROB_ESTIMATOR     3
#define PROB_TRANS_NUMBER  4 
#define MEAN_TRANS_NUMBER  3 
#define DELAY_DIFF_FACTOR 10 

/*************************************************************************

  function:	er_malloc

  parameters:	number of bytes

  purpose:	allocation of memory, exit the program if memory is 
		unsufficient

  return:	pointer to the allocated memory block

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

char *er_malloc (bytes)
     int bytes;
{
  char *pointer;
  if (!bytes)
    return (NULL);
  pointer = (char *)malloc(bytes);
  if (pointer == NULL)
    {
      printf ("Not enough memory for %d bytes\n", bytes);
      exit (FAIL);
    }
  return pointer;
}

/*************************************************************************

  function:	next_line

  parameters:	in_file: file to read
		buffer: pointer wher to store the rest

  purpose:	skips to the beginning of the next data line of the 
		description file and and identifies the token, the rest
		of the line is stored in the buffer

  return	identifier of the token; 0 = EOF 

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

#define NO_TOKENS	14
char	tokens[NO_TOKENS][30] = {"eof", "FCT", "CONNECTED",
			"MDA", "/*mean", "/*prob", "case", "if", 
			"update_measure(", "break", "MD_ARCS", "ENA_FCTS",
			"WEIGHTS", "MEASURE"};
int next_line(in_file, buffer)
     FILE *in_file;
     char *buffer;
{	
  int	beg, tok, copy;
  char	in_token[50];

  while(1)
    {	if (fgets(buffer, NEXT_LINE_BUF, in_file) == NULL)
	  return 0;		/* EOF reached */
	for (beg = 0; beg < NEXT_LINE_BUF; beg++)
	  if (!isspace(buffer[beg]))
	    break;
	if (buffer[beg] && (beg < NEXT_LINE_BUF) &&
	    strncmp(buffer,"--",2))
	  break;          /* line is not empty nor a comment */
      }
  for (copy = 0; buffer[beg] && !isspace(buffer[beg]); )
    in_token[copy++] = buffer[beg++];
  in_token[copy] = 0;
  for (tok = 1; tok < NO_TOKENS; tok++)
    if (!strcmp(in_token, tokens[tok]))
      {	for (copy = 0; buffer[beg]; )
	  buffer[copy++] = buffer[beg++];
	buffer[copy] = 0;
	return tok;
      }
  return -1;			/* not found */
}


/*************************************************************************

  function:	check_and_add_aff 

  parameters:	tra_nr: transition that has the activity
		tra_aff: transition that is to check

  purpose:	check wether this transition aff_trans is already recorded in 
		the list of affected_trans, if not add to list 

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

void check_and_add_aff (tra_nr, aff_trans)
     int tra_nr, aff_trans;
{
  int found_flag = FALSE;
  aff_list *new_node;
  
  new_node = transitions[tra_nr].affected_trans;
  while (new_node != NULL)
    {
      if (new_node->number == aff_trans)
	{
	  found_flag = TRUE;
	  break; /* already recorded */
	}
      new_node = new_node->next;
    }
  if (!found_flag)
    {
      if (transitions[tra_nr].affected_trans != NULL)
	{
	  new_node = (aff_list*)er_malloc(sizeof(aff_list));
	  new_node->next = transitions[tra_nr].affected_trans;
	  transitions[tra_nr].affected_trans = new_node;
	}
      else
	{
	  transitions[tra_nr].affected_trans
	    = (aff_list*)er_malloc(sizeof(aff_list));
	  transitions[tra_nr].affected_trans->next = NULL;
	}
      transitions[tra_nr].affected_trans->number = aff_trans;
    }
} /* end check_and_add_aff */


/*************************************************************************

  function:	check_and_add_measure

  parameters:	place number and measure id

  purpose:	check wether this place already record this measure
		if not add to list 

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

void check_and_add_measure (place_nr, meas_id)
     int place_nr, meas_id;
{
  int found_flag = FALSE;
  meas_list *new_node;
  
  new_node = places[place_nr].measure_list;
  while (new_node != NULL)
    {
      if (new_node->id == meas_id)
	{
	  found_flag = TRUE;
	  break; /* already recorded */
	}
      new_node = new_node->next;
    }
  if (!found_flag)
    {
      new_node = (meas_list *)er_malloc(sizeof(meas_list));
      if (places[place_nr].measure_list != NULL)
	new_node->next = places[place_nr].measure_list;
      else
        new_node->next = NULL;
      places[place_nr].measure_list = new_node;
      places[place_nr].measure_list->id = meas_id;
      places[place_nr].no_measures ++;
    }
} /* end check_and_add_measure */


/*************************************************************************

  function:	search_place

  parameters:	place number, transition number, arc type

  purpose:	search for the place in all other transition connections 
		but not in the own one , if found call check_and_add_aff

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

void search_place(trans_nr, place_nr, arc_type)
     int trans_nr, place_nr, arc_type;
{
 int i_trans, i;
 for (i_trans=0; i_trans<no_trans; i_trans++)
   {
     if (i_trans == trans_nr)
       continue;
     switch (arc_type)
       {
       case INP_OR_INH:
	 for (i=0; i<transitions[i_trans].no_inputs; i++)
	   if (transitions[i_trans].in_from_place[i].element == place_nr)
	     check_and_add_aff(trans_nr, i_trans);
	 /* no break !!!, goes trough for both types */
       case INH_ARC:
	 for (i=0; i<transitions[i_trans].no_inh; i++)
	   if (transitions[i_trans].inh_from_place[i].element == place_nr)
	     check_and_add_aff(trans_nr, i_trans);
	 break;
       case ENA_CONN:
       case MDA_CONN:
	 for (i=0; i<transitions[i_trans].no_inputs; i++)
           if (transitions[i_trans].in_from_place[i].element == place_nr)
	     check_and_add_aff(i_trans, trans_nr);  /*reverse to above , cause
		   affected trans. is known, but not the affecting transition*/
	 for (i=0; i<transitions[i_trans].no_outputs; i++)
           if (transitions[i_trans].out_to_place[i].element == place_nr)
	     check_and_add_aff(i_trans, trans_nr);  /*reverse to above , cause
		   affected trans. is known, but not the affecting transition*/
	 break;
       }
   }
} /* end search_place */


/*************************************************************************

  function:	compute_connections

  parameters:	-

  purpose:	compute the connections between transitions and places
		form the internal data structure for representation of 
		the DSPN

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/
void compute_connections ()
{
  int i_place, i_trans, i_arc, place_nr, no_rel_places, code, id, type;
  int i, j;
  char buf[NEXT_LINE_BUF];
  aff_list *aff_node;
  void all_affected();  

  for (i_trans=0; i_trans<no_trans; i_trans++)
    {
      for (i_arc=0; i_arc<transitions[i_trans].no_outputs; i_arc++)
	{
	  place_nr = transitions[i_trans].out_to_place[i_arc].element;
	  search_place(i_trans, place_nr, INP_OR_INH); 
	}
      for (i_arc=0; i_arc<transitions[i_trans].no_inputs; i_arc++)
	{
	  place_nr = transitions[i_trans].in_from_place[i_arc].element;
	  search_place(i_trans, place_nr, INH_ARC);
	}
    } /* end for all transitions */
  rewind(inf_file);
  next_line(inf_file, buf);
  while (no_md_arcs && (code = next_line(inf_file, buf)))
    {
      if (code == 11) break; /* list of enabling functions reached */
      switch (code) 
	{
	case 3:
	  sscanf(buf, "%d %d %d %d", &id, &type, &i_trans, &place_nr);
	  break;
	case 2: 
	  sscanf(buf, "%d", &no_rel_places);
	  for (i_place=0; i_place<no_rel_places; i_place++)
	    {
	      next_line(inf_file, buf);
	      if ((type == INP_ARC) || (type == INH_ARC))
		{
		  sscanf(buf, "%d", &place_nr);
		  search_place(i_trans, place_nr, MDA_CONN);
		}
	    }
	  break;
	default:
	  printf("ERROR: no list of relevant places for m-d arcs\n");
	  exit (FAIL);
	} /* end switch */ 
    } /* end while */
  if (code != 11) 
    next_line(inf_file, buf); 
  while (no_ena_fcts && (code = next_line(inf_file, buf)))
    {
      if (code == 12) break; /* list of md weights reached */
      switch (code)
	{
	case 1:
	  sscanf(buf, "%d %d", &id, &i_trans);
	  if (transitions[i_trans].ena_id)  
	    {
	      printf
		("ERROR: enabling function for trans. %s already defined\n",
		 transitions[i_trans].name);
	      exit(FAIL);
	    }
	  else
	    transitions[i_trans].ena_id = id;
	  break;
	case 2:
	  sscanf(buf, "%d", &no_rel_places);
	  for (i_place=0; i_place<no_rel_places; i_place++)
	    {
	      next_line(inf_file, buf);
	      sscanf(buf, "%d", &place_nr);
	      search_place(i_trans, place_nr, ENA_CONN);
	    }
	  break;
	default:
	  printf("ERROR: no list of relevant places for enabling fcts\n");
	  exit (FAIL);
	} /* end switch */ 
    } /* end while */
  all_affected();
  for (i=0; i<no_trans; i++)  /* compute the number of affected transitions */
    {
      aff_node = transitions[i].affected_trans;
      j = 0;
      while (aff_node != NULL)
	{
	  aff_node = aff_node->next;
	  j++;
	}
      transitions[i].no_aff = j;
    }  /* ||| hier kann man die liste noch in ein durchgehendes array konvertieren */
}  /* end compute_connections */


/*************************************************************************

  function:	print_net

  parameters:	-

  purpose:	print the internal net representation, for debugging purpose 
		only

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/
void print_net()
{
  int i, j;
  aff_list *new_node;

  for (i=0; i<no_trans; i++)
    {
      printf("Transition: %s Number: %d, Type: %d\n", 
	     transitions[i].name, i, transitions[i].type);
      printf("\tDelay: %f, Priority: %d\n", 
	     transitions[i].delay, transitions[i].prio);
      printf("\tmd-Delay-id: %d\n", transitions[i].mark_dep_id);
      printf("\tena_fctn-id: %d\n", transitions[i].ena_id);
      
      printf("\tOutput Arcs to Places\n");
      if (transitions[i].no_outputs == 0)
	printf ("No output arcs to places from transition %s\n", 
		transitions[i].name);
      for (j=0; j<transitions[i].no_outputs; j++)
	printf("\t\tPlace: %d Multiplicity: %d\n",
	       transitions[i].out_to_place[j].element,
	       transitions[i].out_to_place[j].multiplicity);
      
      printf("\tInput Arcs from Places\n");
      if (transitions[i].no_inputs == 0)
	printf ("No input arcs from places to transition %s\n", 
		transitions[i].name);
      for (j=0; j<transitions[i].no_inputs; j++)
	printf("\t\tPlace: %d Multiplicity: %d\n",
	       transitions[i].in_from_place[j].element,
	       transitions[i].in_from_place[j].multiplicity);
      
      printf("\tInhibitor Arcs from Places\n");
      if (transitions[i].no_inh == 0)
	printf ("No inhibitor arcs from places to transition %s\n", 
		transitions[i].name);
      for (j=0; j<transitions[i].no_inh; j++)
	printf("\t\tPlace: %d Multiplicity: %d\n",
	       transitions[i].inh_from_place[j].element,
	       transitions[i].inh_from_place[j].multiplicity);
      new_node = transitions[i].affected_trans;
      printf("Transitions affected by transition %d (name: %s)\n",
	    i, transitions[i].name);
      if (new_node == NULL)
	printf("No affected transitions in list\n");
      while (new_node != NULL)
	{
	  printf("Transition: %d\n", new_node->number);
	  new_node = new_node->next;
	}
    }
  for (i=0; i<no_places; i++)
    {
      printf("Place: %s, Number %d, Initial Marking: %d\n",
	     places[i].name, i, places[i].token);
    }
} /* end print_net */


/*************************************************************************

  function:	read_net

  parameters:	filepointer for <name>.NET and <name>.DEFINFO files

  purpose:	reads the DSPN description from the .DEFINFO and the .NET file

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

void read_net (net, inf)
     FILE  *net, *inf;
{
  static char buf[NEXT_LINE_BUF];
  int code, i_mpar, i_place, i_trans;
  int  i, j, no_arcs;
  par_name  *mpar;
  NetDescr *description;

  
  no_del_par = no_mark_par = no_md_arcs = 0;
  i_mpar = i_place = 0;
  i_trans = -1;

  description = load_net(model_name, NULL);

  no_places = description->num_places;
  places = (place *)er_malloc(no_places*sizeof(place));
  no_trans = description->num_trans;
  transitions = (transition *)er_malloc(no_trans * sizeof(transition));
  no_del_par = description->num_delaypars;
  no_mark_par = description->num_markpars;
  mpar = (par_name *)er_malloc(no_mark_par * sizeof(par_name));
  for (i_mpar=0; i_mpar<no_mark_par; i_mpar++)
    {
      mpar[i_mpar].name = description->markpar[i_mpar].name;
      mpar[i_mpar].value = (double)description->markpar[i_mpar].value;
    }
  for (i_place=0; i_place<no_places; i_place++)
    {
      places[i_place].name = description->place[i_place].name;
      if (!(description->place[i_place].is_value))
	places[i_place].token = 
	  description->place[i_place].marking.markpar->value;
      else
	places[i_place].token = description->place[i_place].marking.value;
      places[i_place].no_measures = NO_MEASURE;
      places[i_place].measure_list = NULL;
    }
  for (i_trans=0; i_trans<no_trans; i_trans++)
    {
      transitions[i_trans].name = description->trans[i_trans].name;
      transitions[i_trans].affected_trans = NULL;
      transitions[i_trans].prio = description->trans[i_trans].priority;
      transitions[i_trans].type = description->trans[i_trans].type;
      /* ||| definition von type auf TransType aus netio.h aendern */
      transitions[i_trans].fire_policy = description->trans[i_trans].f_policy;
      /* ||| definition von fire_policy auf FiringPolicy aus netio.h aendern */
      if (description->trans[i_trans].inf_server)
	{
	  if (transitions[i_trans].fire_policy == RA) 
	    {
	      printf("ERROR: infinite server for RA-trans.\n");
	      printf("not allowed (transition %s)\n",
		     transitions[i_trans].name);
              exit( FAIL );
	    }
	  transitions[i_trans].serv_type = INF_SERVER;
	  }
      else
	transitions[i_trans].serv_type = SING_SERVER;
      if ((description->trans[i_trans].delay_type == MARKDEP) || 
	  (description->trans[i_trans].delay_type == PMFDEF) ||
	  ((description->trans[i_trans].delay_type == DELAYPAR) &&
	   !(description->trans[i_trans].delay.delay_par->is_value)))
	transitions[i_trans].mark_dep_id = i_trans+1;
      else 
	{
	  transitions[i_trans].mark_dep_id = 0;
	  if (description->trans[i_trans].delay_type == DELAYPAR)
	    transitions[i_trans].delay = 
	      description->trans[i_trans].delay.delay_par->delay.value;
	  else
	    transitions[i_trans].delay = 
	      description->trans[i_trans].delay.value;
	}
      transitions[i_trans].ena_id = 0;
      transitions[i_trans].no_cv_meas = 0;

      no_arcs = description->trans[i_trans].num_inarcs;
      if (no_arcs)
        transitions[i_trans].in_from_place =
          (conn_list*)er_malloc(no_arcs*sizeof(conn_list));
      else
        transitions[i_trans].in_from_place = NULL;
      transitions[i_trans].no_inputs = no_arcs;
      for (i=0; i<no_arcs; i++)
	{
	  for (j=0; j<no_places; j++)
	    if (!strcmp(description->trans[i_trans].in_arc[i].place->name, 
			places[j].name))
	      break;
	  if (j >= no_places)
	    {
	      printf
		("ERROR: unknown place name in arc definition, place\n%s\n",
		 description->trans[i_trans].in_arc[i].place->name);
	      exit (FAIL);
	    }
	  transitions[i_trans].in_from_place[i].element = j;
          if (!(description->trans[i_trans].in_arc[i].is_value))
	      transitions[i_trans].in_from_place[i].id = ++no_md_arcs;
	  else
	    {
	      transitions[i_trans].in_from_place[i].multiplicity =
                description->trans[i_trans].in_arc[i].multipl.value;
              transitions[i_trans].in_from_place[i].id = 0;
            }
	}

      no_arcs = description->trans[i_trans].num_outarcs;
      if (no_arcs)
        transitions[i_trans].out_to_place =
          (conn_list*)er_malloc(no_arcs*sizeof(conn_list));
      else
        transitions[i_trans].out_to_place = NULL;
      transitions[i_trans].no_outputs = no_arcs;
      for (i=0; i<no_arcs; i++)
	{
	  for (j=0; j<no_places; j++)
	    if (!strcmp(description->trans[i_trans].out_arc[i].place->name, 
			places[j].name))
	      break;
	  if (j >= no_places)
	    {
	      printf
		("ERROR: unknown place name in arc definition, place\n%s\n",
		 description->trans[i_trans].out_arc[i].place->name);
	      exit (FAIL);
	    }
	  transitions[i_trans].out_to_place[i].element = j;
          if (!(description->trans[i_trans].out_arc[i].is_value))
	      transitions[i_trans].out_to_place[i].id = ++no_md_arcs;
	  else
	    {
	      transitions[i_trans].out_to_place[i].multiplicity =
                description->trans[i_trans].out_arc[i].multipl.value;
              transitions[i_trans].out_to_place[i].id = 0;
            }
	}

      no_arcs = description->trans[i_trans].num_inharcs;
      if (no_arcs)
        transitions[i_trans].inh_from_place =
          (conn_list*)er_malloc(no_arcs*sizeof(conn_list));
      else
        transitions[i_trans].inh_from_place = NULL;
      transitions[i_trans].no_inh = no_arcs;
      for (i=0; i<no_arcs; i++)
	{
	  for (j=0; j<no_places; j++)
	    if (!strcmp(description->trans[i_trans].inh_arc[i].place->name, 
			places[j].name))
	      break;
	  if (j >= no_places)
	    {
	      printf
		("ERROR: unknown place name in arc definition, place\n%s\n",
		 description->trans[i_trans].inh_arc[i].place->name);
	      exit (FAIL);
	    }
	  transitions[i_trans].inh_from_place[i].element = j;
          if (!(description->trans[i_trans].inh_arc[i].is_value))
	      transitions[i_trans].inh_from_place[i].id = ++no_md_arcs;
	  else
	    {
	      transitions[i_trans].inh_from_place[i].multiplicity =
                description->trans[i_trans].inh_arc[i].multipl.value;
              transitions[i_trans].inh_from_place[i].id = 0;
            }
	}
    } /* for all trans */
  
  rewind (inf);
  while (code = next_line(inf, buf))
    {
      switch (code)
	{
	case 11: 
	  sscanf(buf, "%d", &no_ena_fcts);
	  break;
	}
    } /* end while info */
} /* end read net */


#ifdef CVR


/***************************************************************************

 function:        insert_cvs

 parameters:      meas_id: measure id
		  trans  : list of transitions

 purpose:         insert a measure and belongig list of transition into
		  the global stucture meas_trans
		 
 author:          Ralf Putz

 date:            July 1994

****************************************************************************/

void insert_cvs(meas_id,trans)
     int meas_id;
     trans_list *trans;
{
  meas_trans_list *temp_meas_trans;

  temp_meas_trans = (meas_trans_list *)er_malloc(sizeof(meas_trans_list));
  temp_meas_trans->meas_id = meas_id;
  temp_meas_trans->trans = trans;
  temp_meas_trans->next = meas_trans;
  meas_trans = temp_meas_trans;
}

/***************************************************************************

 function:        read_cvs

 parameters:      

 purpose:         read the global structure meas_trans, init the cv_lists
		  of all transitions and init the measures
		 
 author:          Ralf Putz

 date:            July 1994

****************************************************************************/

void read_cvs()
{
  extern void Init_CVR_Measure();
  int i_control,i_trans;
  cv_list_elem *new_cv_list;
  trans_list *help_trans;
  meas_trans_list *help_meas_trans;

  help_meas_trans = meas_trans;
  while (meas_trans != NULL)
  {
    i_control = 0;
    help_trans = meas_trans->trans;
    while (help_trans != NULL)
      {
	for(i_trans = 0; i_trans < no_trans; i_trans++)
	  if( !strcmp(transitions[i_trans].name,help_trans->name))
	    break;
	new_cv_list = (cv_list_elem *) er_malloc(sizeof(cv_list_elem));
	new_cv_list->meas_number = meas_trans->meas_id;
	new_cv_list->cv_number = i_control ++;
	new_cv_list->next = transitions[i_trans].cv_list;
	transitions[i_trans].cv_list = new_cv_list;
	transitions[i_trans].no_cv_meas ++;
	help_trans = help_trans->next;
      } /* end while */
    Init_CVR_Measure (meas_trans->meas_id,i_control); 
    meas_trans = meas_trans->next;
  }
  meas_trans = help_meas_trans;
}



/***************************************************************************

 function:        free_trans_delay_list

 parameters:      l: pointer to a trans_delay_list

 purpose:         free allocated memory of a trans_delay_list
		 
 author:          Ralf Putz

 date:            July 1994

****************************************************************************/

static void free_trans_delay_list(l)
	    trans_delay_list *l;
{
  trans_delay_list *temp;

  if( l != NULL )
  {
    while( l->next != NULL )
    {
      temp = l;
      l = l->next;
      free(temp);
    }
    free(l);
  }
}

/**************************************************************************/


/***************************************************************************

 function:        sort_trans_candidates

 parameters:      place_nr   : place number
		  in_out_flag: specifies wether input- or output-transitions
			       should be sorted

 purpose:         sort transitions specified by place_nr and in_out_flag and 
		  return a list of them

 author:          Ralf Putz

 date:            July 1994

****************************************************************************/

trans_delay_list *sort_trans_candidates(place_nr,in_out_flag)
		 int place_nr,in_out_flag;
{
  int i,j,number;
  conn_list *helplist;
  trans_delay_list *head,*tail,*help,*temp;

  head = NULL;
  for(i = 0;i < no_trans; i++)
  {  
    if((transitions[i].type != IMM) && (transitions[i].type != DET)
       && (transitions[i].type != GEN) 
       && (transitions[i].serv_type != INF_SERVER))
    {
      if( in_out_flag == INPUT_TRANS )
      { 
	helplist = transitions[i].out_to_place;
	number = transitions[i].no_outputs;
      }
      else
      {
	helplist = transitions[i].in_from_place;
	number = transitions[i].no_inputs;
      }
      for( j = 0;j < number;j++ )
      {
	if( helplist[j].element == place_nr ) 
	{
	  temp = (trans_delay_list *)er_malloc(sizeof(trans_delay_list));
	  if ( head == NULL )
	  {
	    strcpy(temp->name,transitions[i].name);
	    temp->delay = transitions[i].delay;
	    temp->next = NULL;
	    head = temp;
	    tail = temp;
	  }
	  else          /* in_out_flag == OUTPUT_TRANS  */
	  {
	    help = head;
	    while( (help != NULL) && (help->delay < transitions[i].delay) )
	      help = help->next;
	    if( help == NULL )   /* hinten einfuegen */
	    {
	      strcpy(temp->name,transitions[i].name);
	      temp->delay = transitions[i].delay;
	      temp->next = NULL;
	      tail->next = temp;
	      tail = tail->next;
	    }
	    else
	    {
	      if( help == head )  /* vorne einfuegen */
	      {
		strcpy(temp->name,transitions[i].name); 
		temp->delay = transitions[i].delay; 
		temp->next = head;
		head = temp;
	      }
	      else   /* irgendwo in der Mitte einfuegen */
	      {
		strcpy(temp->name,help->name); 
		temp->delay = help->delay;
		temp->next = help->next;
		strcpy(help->name,transitions[i].name);
		help->delay = transitions[i].delay;
		help->next = temp;
	      }
	    }
	  }
	}
      } /* end for */          
    } /* end if (transition != IMM) ... */
  } /* end for */  
  return head;
} /* end  */ 

/***************************************************************************

 function:        second_sort_trans_candidates

 parameters:      place_nr   : place number

 purpose:         sort transitions specified by rule 4.

 author:          Ralf Putz

 date:            Februar 1995

****************************************************************************/

trans_delay_list *second_sort_trans_candidates(place_nr)
		 int place_nr;
{
  int i,j,k,number1,number2;
  conn_list *helplist1,*helplist2;
  trans_delay_list *head,*tail,*help,*temp,*candidates;

  head = NULL;
  for( i = 0;i < no_trans; i++ )
    if( transitions[i].type == IMM )
    {
      helplist1 = transitions[i].in_from_place;
      number1 = transitions[i].no_inputs;
      for( j = 0; j < number1; j++ )
        if( helplist1[j].element == place_nr )
        {  
           helplist2 = transitions[i].out_to_place;
           number2 = transitions[i].no_outputs; 
           for( k = 0; k < number2; k++ )
           {
             candidates = sort_trans_candidates( helplist2[k].element,
                                                         OUTPUT_TRANS );
             if( candidates != NULL )
             {
               temp = (trans_delay_list *)er_malloc(sizeof (trans_delay_list));
 	       if ( head == NULL )
	        {
	         strcpy(temp->name,candidates->name);
	         temp->delay = candidates->delay;
	         temp->next = NULL;
	         head = temp;
	         tail = temp;
	       }
	       else          /* in_out_flag == OUTPUT_TRANS  */
	       {
	         help = head;
	         while( (help != NULL) && (help->delay < candidates->delay) )
	           help = help->next;
	         if( help == NULL )   /* hinten einfuegen */
	         {
	           strcpy(temp->name,candidates->name);
	           temp->delay = candidates->delay;
	           temp->next = NULL;
	           tail->next = temp;
	           tail = tail->next;
	         }
	         else
	         {
	           if( help == head )  /* vorne einfuegen */
	           {
		     strcpy(temp->name,candidates->name); 
		     temp->delay = candidates->delay; 
		     temp->next = head;
		     head = temp;
	           }
	           else   /* irgendwo in der Mitte einfuegen */
	           {
		     strcpy(temp->name,help->name); 
		     temp->delay = help->delay;
		     temp->next = help->next;
		     strcpy(help->name,candidates->name);
		     help->delay = candidates->delay;
		     help->next = temp;
	           }
	         }
               }
             }
           }
         }
     }
     return head;
} 


/***************************************************************************

 function:        get_short_delayed_trans

 parameters:      candidates: list of transitions sorted by belonging delays

 purpose:         if a transition exists which has a DELAY_DIFF_FACTOR smaller
		  delay than all the others this one is returned, else NULL
		 
 author:          Ralf Putz

 date:            July 1994

****************************************************************************/


trans_delay_list *get_short_delayed_trans(candidates)
		 trans_delay_list *candidates;
{
  if( (candidates != NULL) && (candidates->next != NULL) )
    return(((candidates->delay * DELAY_DIFF_FACTOR) <= candidates->next->delay) 
	   ? candidates: NULL);
  else
    return NULL;
}

/***************************************************************************

 function:        handle_prob_estimator

 parameters:      place_nr:    place number
		  no_measures: measure number
		

 purpose:         apply cv-rules to probability-estimator
		 
 author:          Ralf Putz

 date:            July 1994

****************************************************************************/


void handle_prob_estimator(place_nr,meas_id,flag)
     int place_nr,meas_id,flag;
{
  trans_delay_list *candidates,*temp_candidates;
  trans_list *temp_trans, *trans = NULL;
  int counter = 0;

  switch(flag) 
  {
    case 1:   /* SINGLE_PROB_ESTIMATOR */
      candidates = sort_trans_candidates(place_nr,OUTPUT_TRANS);
      temp_candidates = candidates;
      while(temp_candidates != NULL && counter < PROB_TRANS_NUMBER)
      {
	temp_trans =  (trans_list *)er_malloc(sizeof(trans_list));
	strcpy(temp_trans->name,temp_candidates->name);
	temp_trans->next = trans;
	trans = temp_trans;
	counter++;
	temp_candidates = temp_candidates->next;
      }
      free_trans_delay_list(candidates);
      if(counter < PROB_TRANS_NUMBER)
      {
	candidates = sort_trans_candidates(place_nr,INPUT_TRANS);
	temp_candidates = candidates;
	while(temp_candidates != NULL && counter < PROB_TRANS_NUMBER)
	{
	  temp_trans =  (trans_list *)er_malloc(sizeof(trans_list));
	  strcpy(temp_trans->name,temp_candidates->name);
	  temp_trans->next = trans;
	  trans = temp_trans;
	  counter++;
	  temp_candidates = temp_candidates->next;
	}
	free_trans_delay_list(candidates);
      }
      break;
    case 2:   /* JOINED_PROB_ESIMATOR */
    case 3:   /* COND_PROB_ESTIMAROR  */
      break;       
  } 
  insert_cvs(meas_id,trans);
}

/***************************************************************************

 function:        handle_mean_estimator

 parameters:      place_nr:    place number 
		  no_measures: measure number
		

 purpose:         apply cv-rules to mean-estimator
		 
 author:          Ralf Putz

 date:            July 1994

****************************************************************************/

void handle_mean_estimator(place_nr,meas_id)
     int place_nr,meas_id;
{
  trans_delay_list *candidates,*one_candidate,*temp_candidates;
  trans_list *temp_trans, *trans = NULL;
  int counter = 0;

  if( (candidates = sort_trans_candidates(place_nr,OUTPUT_TRANS)) == NULL)
      candidates = second_sort_trans_candidates(place_nr);  
  if ( candidates != NULL)
  {  
    if ( (one_candidate = get_short_delayed_trans(candidates)) != NULL)
      {
	 temp_trans =  (trans_list *)er_malloc(sizeof(trans_list));
	 strcpy(temp_trans->name,one_candidate->name);
	 temp_trans->next = trans;
	 trans = temp_trans;
       }
    else
    {
      temp_candidates = candidates; 
      while(temp_candidates != NULL && counter < MEAN_TRANS_NUMBER)
      {
	temp_trans =  (trans_list *)er_malloc(sizeof(trans_list));
	strcpy(temp_trans->name,temp_candidates->name);
	temp_trans->next = trans;
	trans = temp_trans;
	counter++;
	temp_candidates = temp_candidates->next;
      }
    }
    free_trans_delay_list(candidates);
  }
  insert_cvs(meas_id,trans);    
}

#endif


/*************************************************************************

  function:	define_measures

  parameters:	-

  purpose:	reads the measure definition to determine the places for which 
		measures should be recorded, uses the <name>_ERC.c file 
		marks the places related top measures

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/
void define_measures ()
{
  static char buf[NEXT_LINE_BUF];
  
  int  code, place_nr, meas_id, i_dummy,prob_counter;
  char s_dummy[30], end_string[30];

  no_measures = 0;
  while (code = next_line(erc_file, buf))
    switch (code)
      {
      case 4:  /* mean estimator */
	no_measures ++;
	next_line (erc_file, buf);
	sscanf (buf, "%s %d", s_dummy, &place_nr);
	if ((code = next_line(erc_file, buf)) != 6) /* no case follows */
	  {
	    printf("ERROR: no \"case\" in measure definition %s\n", buf);
	    exit (FAIL);
	  }
	sscanf(buf, "%d:", &meas_id);
	check_and_add_measure (place_nr, meas_id);
#ifdef CVR
	handle_mean_estimator(place_nr,meas_id);
#endif
	break;
      case 5: /* probablity estimator */
        if(probMeasureId==NULL)
        {
         numberProbMeasures=1;
         probMeasureId = (int*)malloc(sizeof(int));
        }
        else 
        {
         numberProbMeasures++;
         probMeasureId =(int*)realloc(probMeasureId,numberProbMeasures*sizeof(int));
        }
	no_measures ++;
	if ((code = next_line(erc_file, buf)) != 6) /* no case follows */
	  {
	    printf("ERROR: no \"case\" in measure definition %s\n", buf);
	    exit (FAIL);
	  }
	sscanf(buf, "%d:", &meas_id);
        probMeasureId[numberProbMeasures-1]=meas_id;
#ifdef debug
	{
         int i;
         for (i=0;i<numberProbMeasures;i++)
          printf("probMeasureId[%d]=%d\n",i,probMeasureId[i]);
        }
#endif
	next_line(erc_file, buf);
	sscanf(buf, "%s", end_string);
        prob_counter = 0;
	while ((strstr(end_string, "{") == NULL) 
	       && (strstr(end_string, "endif") == NULL))
	  {
	    next_line(erc_file, buf);
	    sscanf(buf, "%s %d %s %s", 
		   s_dummy, &place_nr, s_dummy, end_string);
	    check_and_add_measure (place_nr, meas_id);
	    prob_counter ++;
	  }
#ifdef CVR
	if(prob_counter == 1) 
	  handle_prob_estimator(place_nr,meas_id,SINGLE_PROB_ESTIMATOR);
	if(prob_counter > 1)
	  handle_prob_estimator(place_nr,meas_id,JOINED_PROB_ESTIMATOR);   
#endif
	if (strstr(end_string, "{") != NULL) /* cond. prob. measure */
	  {
	    next_line(erc_file, buf); /* skip first update */
	    no_measures ++;
#ifdef STR
      printf("\nERROR: in transient simulation only basic components are allowed for measures\n");
      exit (FAIL);
#endif
	    if (next_line(erc_file, buf) != 7) /* no if-expression */
	      {
		printf
		  ("ERROR: no if-expression in cond. prob. def.\n");
		exit (FAIL);
	      }
	    sscanf(buf, "%s", end_string);
	    while (strstr(end_string, "endif") == NULL)
	      {
		next_line(erc_file, buf);
		sscanf(buf, "%s %d %s %s", 
		       s_dummy, &place_nr, s_dummy, end_string);
		check_and_add_measure (place_nr, meas_id);
#ifdef CVR               
		handle_prob_estimator(place_nr,meas_id,COND_PROB_ESTIMATOR);
#endif 
	      }
	    if (next_line(erc_file, buf) != 8) /* no update fctn */
	      {
		printf
		  ("ERROR: no update fctn. in cond. prob. def.\n");
		exit (FAIL);
	      }
	    else
	      {
		meas_id ++;
		sscanf(buf, "%d", &i_dummy);
		if (i_dummy != meas_id)
		  {
		    printf("ERROR: wrong measure id in line %s\n", buf);
		    exit (FAIL);
		  }
	      }
	  } /* end if end string contains )) */
	break;
      case 7:
      case 8:
      case 9:
	break;
      }  /* end switch */
  if (no_measures == 0)
    {
      printf("ERROR: non sense simulation, no measures defined\n");
      exit (FAIL);
    }
  if (no_measures != (meas_id+1))
    {
      printf("ERROR: not all or more then defined measures found\n");
      printf("expected: %d, found %d\n", meas_id+1, no_measures);
      exit (FAIL);
    }
#ifdef STR
  if(!only_basic_components)
    {
      printf("\nERROR: in transient simulation only basic components are allowed for measures\n");
      exit (FAIL);
    }
#endif
} /*end define_measures */



  

/*************************************************************************

  function:	all_affected

  parameters:	-

  purpose:	adds some affected Transitions to the affected list
                for the new simu_engine
                This function is called at the end of
                compute_connections

  author:	Matthias Hoffmann

  date:		March 1995

*************************************************************************/
void all_affected()
{
  int i, j, k, l;
  
  for( i=0; i<no_trans; i++ )
    for( k=0 ; k<transitions[i].no_inputs ; k++ )
      for( j=0; j<no_trans; j++ )      
        for( l=0 ; l<transitions[j].no_inputs ; l++ )
          if( transitions[i].in_from_place[k].element 
              == transitions[j].in_from_place[l].element )
            check_and_add_aff( i, j );
  for( i=0; i<no_trans; i++ )
    if( transitions[i].ena_id )
      for( j=0; j<no_trans; j++ )      
        check_and_add_aff( i, j );
  for( i=0; i<no_trans; i++ )
  { for( k=0 ; k<transitions[i].no_inputs ; k++ )
      if( transitions[i].in_from_place[k].id )
        for( j=0; j<no_trans; j++ )      
          check_and_add_aff( i, j );
    for( k=0 ; k<transitions[i].no_inh ; k++ )
      if( transitions[i].inh_from_place[k].id )
        for( j=0; j<no_trans; j++ )      
          check_and_add_aff( i, j );
  }     
  for( i=0; i<no_trans; i++ )
    if( transitions[i].mark_dep_id &&
        ((transitions[i].fire_policy==RE) || (transitions[i].fire_policy==RA)))
      { /*if( transitions[i].type == GEN )
        { printf( "The generalized <md>-Transitions %d with fire_policy RA or RE is not allowed.\n", i );
          exit( FAIL );
        }*/
        for( j=0; j<no_trans; j++ )      
          check_and_add_aff( i, j );
      }  
  for( i=0; i<no_trans; i++ )
    if( transitions[i].no_inputs == 0 )
      for( j=0; j<no_trans; j++ )      
        check_and_add_aff( j, i );
}

#ifndef RST
/*************************************************************************

  function:	get_measure_names

  parameters:	net_file

  purpose:	create a list of all measure names
                - look wheater all measures are basic measures (#STR)
                - if Measure non basic measure it will get <name>_Part%d (#else)

  author:	Ralf Putz

  date:		December 1995
  
  release:      March 1997 

*************************************************************************/
meas_names_list *get_measure_names(file)
FILE *file;
{
  int code, i;
  static char buffer[NEXT_LINE_BUF]; 
  char measure_name[30];
  meas_names_list *measure_list, *temp_list1, *tail;

  rewind(file);
  measure_list = NULL;
  tail = NULL;
  while( code = next_line(file, buffer) )
    switch(code)
      {
        case 13:
          sscanf(buffer,"%s",measure_name);
          temp_list1 = (meas_names_list *)er_malloc(
      					   sizeof(meas_names_list));
          if(measure_list == NULL)
	    {
              strcpy(temp_list1->name,measure_name); 
              temp_list1->next = NULL;
              measure_list = temp_list1;
              tail = temp_list1;
            }             
          else
	    {
               strcpy(temp_list1->name,measure_name); 
               temp_list1->next = NULL;
               tail->next  = temp_list1;
               tail = tail->next;
	     }
          next_line(file,buffer);
#ifdef STR
          for(i=0; i < strlen(buffer); i++)
            if(buffer[i] == '+' || buffer[i] == '-' || buffer[i] == '*' ||
               buffer[i] == '/' || buffer[i] == '^' )
              only_basic_components = false;  
#else 
          {
           int noParts=1;
           char rest[6];
           for(i=0; i < strlen(buffer); i++)
            if(buffer[i] == '+' || buffer[i] == '-' || buffer[i] == '*' ||
               buffer[i] == '/' || buffer[i] == '^' )
            {
              noParts++;
              temp_list1 = (meas_names_list *)er_malloc(
      		  			       sizeof(meas_names_list)+7);
              strcpy(temp_list1->name,measure_name); 
              sprintf(rest,"_Part%d",noParts);
              strcat(temp_list1->name,rest);
              temp_list1->next = NULL;
              tail->next  = temp_list1;
              tail = tail->next;
	    }
	  }
#endif              
          break;
        default:
          break;
      }
    return measure_list;
}
    
#endif
