/*************************************************************************

  filename:	spec_var.c

  purpose:	It implements functions to perform an accurate statistical
		analysis of simulation output data.
		It evaluates the final simulation results for a required 
		level of accuracy.
		The function for testing the required precision uses the
		Method of Spectral Variance Analysis.

		See [K. Pawlikowsky]: "Steady-State Simulation of Queueing
		  Processes: Problems ans Solutions",
		  ACM Computing Surveys 1990, p. 123.

  author:       Christian Kelling
                Technical University of Berlin
		Computer Science Department
		Institute for Technical Computer Science

  date:		May 1991, November 1992


  exported functions:
  	void	SPEC_Analysis_Init %  % 
	void	SPEC_RewardMeasure_Init
	int	SPEC_Analysis
	void	SPEC_RewardMeasure_data_out
	void	SPEC_Globals_out
	double  SPEC_Mean_out
	double  SPEC_Lower_out
	double  SPEC_Upper_out

  local functions:
	int	SPEC_Estimation

*************************************************************************/


#include <math.h>
#include <stdio.h>

#include "num_utils.h"
#include "stat_utils.h"
#include "sim_master_types.h"
#include "spec_var.h"


/* external functions */
#ifdef __STDC__
extern double uniPdfAreaErr(char   *measureName,
                            double lowerBorder,
                            double upperBorder,
                            int n,
                            double meanValue,
                            double variance,
                            double alpha);      /* Irrtumswahrscheinlichkeit */ 
#else
extern double uniPdfAreaErr();
#endif
/* external variables */
extern  double  current_time;
extern	double	eps_max;
extern  double  errPercent;
extern  meas_names_list *measure_names;
extern  int     StatSema;
extern  int     NoPrecCount;
extern 	int	conf_level;
extern 	int	maxsamples;
extern  int     time_exceeded;
extern  bool    min_flag;
extern  bool    trans_detection;
extern  FILE    *stat_file;

/*****************************/
/* variables used by RESTART */
/*****************************/

#ifdef RST

#define END_OF_SEQUENCE 0.0 

extern double ret_product;

#endif

/*****************************/

/* global variables used by the modul */

static long 		SPEC_n_max;


static int		paw_stattotal	= 0;

char *er_malloc ();


/*************************************************************************

  functions:	SPEC_Mean_out SPEC_Lower_out SPEC_Upper_out SPEC_prec_out

  parameters:   MeasureStat -> pointer to a reward measure statistic

  return:	the mean of the measure and the bounds of the conf. interval

  author:       Christian Kelling
                Technical University of Berlin

  date:         May 1993

*************************************************************************/

double SPEC_Mean_out(MeasureStat)
     SpecStat *MeasureStat;
{
#ifdef RST
  return MeasureStat->MeanValue /*/ret_product*/ ;
#else
  return MeasureStat->SampleCount*LOCAL_BATCH_SIZE_DEF*
      MeasureStat->MeanValue/MeasureStat->EndTime;
#endif
}

double SPEC_Lower_out(MeasureStat)
SpecStat *MeasureStat;
{
#ifdef RST
  return (MeasureStat->MeanValue - MeasureStat->DeltaX) /*/ret_product*/;
#else
  return MeasureStat->SampleCount*LOCAL_BATCH_SIZE_DEF*
    (MeasureStat->MeanValue - MeasureStat->DeltaX)/MeasureStat->EndTime;
#endif
}

double SPEC_Upper_out(MeasureStat)
SpecStat *MeasureStat;
{
#ifdef RST
  return (MeasureStat->MeanValue + MeasureStat->DeltaX)/*/ret_product*/;
#else
  return MeasureStat->SampleCount*LOCAL_BATCH_SIZE_DEF*
    (MeasureStat->MeanValue + MeasureStat->DeltaX)/MeasureStat->EndTime;
#endif
}

int SPEC_prec_out(MeasureStat)
SpecStat *MeasureStat;
{

  if (MeasureStat->NoPrecision == true)
    return 1;
  else
    return 0;
}


/*************************************************************************

  function:	SPEC_RewardMeasure_data_out

  parameters:	MeasureStat -> pointer to a reward measure statistic

  purpose:	performs the output of results of a statistic

  author:	Christian Kelling
		Technical University of Berlin

  date:		August 93

*************************************************************************/

void SPEC_RewardMeasure_data_out(MeasureStat)
SpecStat *MeasureStat;
{
  FILE *stat;
  stat = stat_file;
  if (trans_detection)
    {
      fprintf(stat, "\nTransient Analysis at slaves (maximum values):\n");
      fprintf(stat, "==============================================\n\n");
      fprintf(stat, "Length: %d samples\t",MeasureStat->InitLength);
      fprintf(stat, "TransTime: %f\n",MeasureStat->TransTime);
    }
  else
      fprintf(stat, "\nTransient Analysis was turned off.\n");
  fprintf(stat, "\nSteady State Analysis:\n");
  fprintf(stat, "======================\n\n");
  fprintf(stat, "Epsilon: %f\t",MeasureStat->Epsilon);
  fprintf(stat, "Samples: %lu\t",LOCAL_BATCH_SIZE_DEF*MeasureStat->SampleCount);
  fprintf(stat, "VirtTime: %f\n",MeasureStat->EndTime);
  fprintf(stat, "MeanValue: %e\t",MeasureStat->MeanValue);
  fprintf(stat, "LowInterv: %e\t",
         (MeasureStat->MeanValue - MeasureStat->DeltaX));
  fprintf(stat, "UpInterv: %e\n",
         (MeasureStat->MeanValue + MeasureStat->DeltaX));
  fprintf(stat, "\nBatchSize: %d\t", MeasureStat->BatchSize);
  fprintf(stat, "NumbOfMeans: %lu\t", MeasureStat->BatchMeanNumber);
  fprintf(stat, "Checkpoints: %lu", MeasureStat->CheckNumber);
  fprintf(stat, "\nLastCheckpoint: %lu\t", LOCAL_BATCH_SIZE_DEF*MeasureStat->CheckLocLast);
  fprintf(stat, "NextCheckpoint: %lu\n", LOCAL_BATCH_SIZE_DEF*MeasureStat->CheckLocNext);
  fprintf(stat, "StopSimul: %d\t", MeasureStat->StopSimulation);
  fprintf(stat, "NoPrecision: %d\n\n", MeasureStat->NoPrecision);
} /* SPEC_RewardMeasure_data_out */


/*************************************************************************

  function:	SPEC_Globals_out

  parameters:	

  purpose:	performs the output of global results of the statistical
		analysis for a simulation run.

  author:	Christian Kelling
		Technical University of Berlin

  date:		August 93

*************************************************************************/

void SPEC_Globals_out()
{
  FILE *stat;
  stat = stat_file;
  
  fprintf(stat, "\n\n\nSpectral Variance Analysis:\n");
  fprintf(stat, "===========================\n\n");
  fprintf(stat, "ConfidenceLevel: %d\n",conf_level);
  fprintf(stat, "Statistics: %d\t",paw_stattotal);
  fprintf(stat, "NoPrecision: %d\n",NoPrecCount);

} /* SPEC_Globals_out */


/*************************************************************************

  function:	SPEC_Ananlysis_Init

  parameters:	maxsamples -> the maximum allowed length of the simulation
			run measured in the number of recorded observations.
			The calling function uses the number of messages
			to transmit.

  purpose:	init local data of this modul and global variables used
                by all statistics

  author:	Christian Kelling
		Technical University of Berlin

  date:		Nov 1992

*************************************************************************/

void SPEC_Analysis_Init (maxsamples)
long maxsamples;

{

  SPEC_n_max = maxsamples;

} /* SPEC_Analysis_Init */


/*************************************************************************

  function:	SPEC_RewardMeasure_Init

  parameters:	MeasureStat -> Pointer to the new defined statistic

  purpose:	Init data of an environment to begin spectralvariance analysis

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1991, Nov 1992

*************************************************************************/

SpecStat *SPEC_RewardMeasure_Init (index)
int index;
{
  SpecStat *MeasureStat;
  meas_names_list *help_measure_names;
  int  counter;

  MeasureStat = (SpecStat *) er_malloc (sizeof (SpecStat));
  StatSema ++;
  NoPrecCount ++;

  MeasureStat->kindOfMeasure=OTHER_MEAS;
  MeasureStat->StopSimulation = false;
  MeasureStat->StopSendedToBackend=false;
  MeasureStat->NoPrecision = true;
  MeasureStat->Epsilon = 1.0 - eps_max;
  MeasureStat->StatStatus = STEADY_STATE;
  MeasureStat->SampleCount = 1;
  MeasureStat->SendFactor = 1;
  MeasureStat->sum = 0.0;
  MeasureStat->BatchSize = 1;
  MeasureStat->BatchMeanNumber = 1;
  MeasureStat->CheckNumber = 1;
  MeasureStat->CheckLocLast = 0; 
  MeasureStat->BlockAdminSum = 0;
  MeasureStat->BlockAdminFirst = 0; 
  	/*relative 0 to MeasureStat->InitLength in difference to Paw*/
  MeasureStat->InitLength = 2*M0; 
  /* hier wird ggf die groesste aller von slaves 
     gesendeten Laengen eingetragen*/
  MeasureStat->TransTime = 0.0;
  MeasureStat->CheckLocNext = 2*M0;   /* 2*M0 waere korrekt */
  MeasureStat->MeanValue = 0.0;
  
  paw_stattotal++;

  help_measure_names = measure_names;
  counter = 0;
  while( counter < index )
    {
      help_measure_names = help_measure_names->next;
      counter++;
    }
  if (help_measure_names->name!=NULL)
   strcpy(MeasureStat->name,help_measure_names->name);
  else strcpy(MeasureStat->name,"NoName");

  return (MeasureStat);
} /*RewardMeasure_Init*/


/*************************************************************************

  function:	Batching

  parameters:	MeasureStat -> Pointer the the environment of a STATISTICS

  returns:	m  -> new batch size

  purpose:	modifies the old batch size into the new batch size whenever
                ...

  author:	Christian Kelling
		Technical University of Berlin

  date:		Nov 1992

*************************************************************************/

int Batching (MeasureStat)

SpecStat *MeasureStat;
{
	int m,s;
/* a_seq:  ein makro fuer MeasureStat->AnalysedSequence*/

	m = MeasureStat->BatchSize;
	a_seq[MeasureStat->BatchMeanNumber] = MeasureStat->sum/m;
	if(MeasureStat->BatchMeanNumber==2*M0)
	{
		for(s = 1;s<=M0;s++)
		  a_seq[s] = ((a_seq[2*s-1]+a_seq[2*s])/2);
		m+=m;
		MeasureStat->BatchMeanNumber = M0;
	}
	MeasureStat->BatchMeanNumber++;
	MeasureStat->sum = 0.0;
	return(m);
}/*Batching*/


/*************************************************************************

  function:	SPEC_Estimation

  parameters:	MeasureStat -> Pointer the a reward measure statistic

  returns:	true  -> the required precision has been reached.
		false -> the required precision has not been reached; too short
			 simulation run is assumed.

  purpose:	calculates estimates and tests their precision until a required
		precision is reached.
		The boolean result will be assigned to the environment flag
		StopSimulation too.

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1991, Nov 1992, Aug 93

*************************************************************************/

static bool SPEC_Estimation(MeasureStat)
SpecStat *MeasureStat;
{
	static double safecoeff = SAFE_DEF_STEADY;
	static int varseqlen = VARSEQLEN_DEF;
	static int kappa = FREEDOM_DEF_STEADY;

	int s;
	double alpha = ALPHAT_DEF/2;
	double EstMean, student_distr(), Variance, *testsamplesarray;

	testsamplesarray = &a_seq[MeasureStat->BatchMeanNumber-varseqlen];

	Variance = SpectralVarAnalysis(testsamplesarray,varseqlen);
	EstMean = 0.0;
	for(s = 1;s < MeasureStat->BatchMeanNumber;s++)
		EstMean += a_seq[s];
	EstMean /= (MeasureStat->BatchMeanNumber-1);
	MeasureStat->Epsilon = student_distr(kappa,1.0-alpha);
	MeasureStat->DeltaX = MeasureStat->Epsilon;
	MeasureStat->DeltaX *= sqrt(Variance);
	MeasureStat->Epsilon *= (sqrt(Variance)/EstMean);
	MeasureStat->MeanValue = EstMean;
	if((MeasureStat->Epsilon < eps_max) && min_flag &&
           ((MeasureStat->kindOfMeasure==PROB_MEAS &&   /* it's a probabilty */
             (uniPdfAreaErr(MeasureStat->name,
                            SPEC_Lower_out(MeasureStat),
                            SPEC_Upper_out(MeasureStat),
                            STEPS,
                            SPEC_Mean_out(MeasureStat),
                            varseqlen*Variance/MeasureStat->EndTime,
                            ALPHAT_DEF) <= errPercent)
            ) ||
            MeasureStat->kindOfMeasure==OTHER_MEAS      /* is not a p. */
           )
          )
	  {
	    MeasureStat->StopSimulation = true;
	    MeasureStat->NoPrecision = false;
          }
	else
	  {
	    MeasureStat->CheckNumber++;
	    MeasureStat->CheckLocLast = MeasureStat->CheckLocNext;
	    MeasureStat->CheckLocNext=
	      (long)(MeasureStat->CheckLocLast * safecoeff);
	  }
	return (MeasureStat->StopSimulation);
}/*SPEC_Estimation*/

/*************************************************************************/

#ifdef RST
bool VARIANCE_Analysis(MeasureStat, sample)

SpecStat *MeasureStat;
double sample;
{
static double x[50];
static int n=0;
int i;
double alpha=ALPHAT_DEF/2,
       student_distr(), 
       variance,
       est_mean,
       first_sum=0.0,
       second_sum=0.0;

  if (MeasureStat == NULL)
    {
      printf("Variance Analysis: Null-Pointer for env.\n");
      exit(FAIL);
    }
  if (MeasureStat->StopSimulation)
    return(!StatSema);
  else
    {
      if (sample!=END_OF_SEQUENCE)
	{
        x[n]=sample;
        ++n; 
#ifdef DEBUG_RST
      printf("sample: %le sample count: %d\n", 
             sample,MeasureStat->SampleCount);
#endif
	}
      else
	{
        for (i=0;i<n;++i)
	  {
          first_sum+=x[i]*x[i];
	  second_sum+=x[i];
#ifdef DEBUG_RST
          printf("sample: %le sample count: %d\n",sample,MeasureStat->SampleCount);
#endif
	  }
        if (n!=1)
	  {
	  est_mean=second_sum/(double)n;
          variance=(first_sum-second_sum*second_sum/(double)n)/((double)n-1.0);
          MeasureStat->Epsilon=student_distr(n-1,1.0-alpha);
	  MeasureStat->DeltaX = MeasureStat->Epsilon;
	  MeasureStat->DeltaX *= sqrt(variance)/sqrt((double)n);
          MeasureStat->Epsilon *= sqrt(variance)/(est_mean*sqrt((double)n));
	  MeasureStat->MeanValue = est_mean;
	  printf("eps:%le  mean:%le\n",MeasureStat->Epsilon,MeasureStat->MeanValue);
	  }
        else
          fprintf(stdout,"skipped sample\n");
        n=0;
	if((MeasureStat->Epsilon < eps_max)&&min_flag)
	  {
	    MeasureStat->StopSimulation = true;
	    MeasureStat->NoPrecision = false;
	  }
	}
/*        if (!min_flag&&(n==0))
	printf ("no min flag\n"); */
      if (MeasureStat->StopSimulation != true)
	{
	  if (n!=0)
	    MeasureStat->SampleCount++;
	}
      else
	{
	  StatSema--;
	  NoPrecCount--;
	}
    }/* else */
  if (MeasureStat->StopSimulation != true)
    {
      if (MeasureStat->SampleCount > SPEC_n_max)
	{
	  printf(
	       "\nNOTE: A measure has exceeded the maximum number of samples\n");
	  printf(
	 "      Simulation stops without reaching the predefined accuracy!\n");
	  MeasureStat->StopSimulation = true;
	  MeasureStat->NoPrecision = true;
	  StatSema--;
	}
      if (time_exceeded)
	{
          printf(
	"\nNOTE: The simulation has exceeded the specified model time limit.\n");
          printf(
         "      Simulation stops without reaching the predefined accuracy!\n");
          MeasureStat->StopSimulation = true;
          MeasureStat->NoPrecision = true;
	  StatSema--;
        }
    }
  return (MeasureStat->StopSimulation);
}/*VARIANCE_Analysis*/
#endif

/*************************************************************************

  function:	SPEC_Analysis
  
  parameters:	MeasureStat -> Pointer the the environment of a STATISTICS
  		sample -> new sampled observation

  returns:	true  -> All created STATISTICS have been reached the required
			 precision. This marks the maximum length of simulation
			 run. 
		false -> There are already STATISTICS with an unacceptable precision.

  purpose:	performs the main function of the spectral analysis.
		
		This function returns immediatly when the environment of the
		STATISTICS denotes an accurate precision. It will no further
		observations analyse.
		The return value true means all created STATISTICS have get the
		required precision. At this point the simulation run could be
		terminated by the calling function or an upper function level.


  author:	Christian Kelling
		Technical University of Berlin

  date:		Nov 1992

*************************************************************************/

bool SPEC_Analysis(MeasureStat, sample)

SpecStat *MeasureStat;
double sample;
{
  if (MeasureStat == NULL)
    {
      printf("Spectral Analysis: Null-Pointer for env.\n");
      exit(FAIL);
    }
  if (MeasureStat->StopSimulation)
    return(!StatSema);
  else
    {
      MeasureStat->sum += sample;
      if((MeasureStat->SampleCount % MeasureStat->BatchSize) == 0)
	MeasureStat->BatchSize = Batching(MeasureStat);
/*      printf("sample: %e sample count: %d next check: %d\n", 
             sample,
             MeasureStat->SampleCount,
	     MeasureStat->CheckLocNext);
      if (!min_flag)
	printf ("no min flag\n"); */
      if((MeasureStat->SampleCount >= MeasureStat->CheckLocNext))
	SPEC_Estimation(MeasureStat); 
      if (MeasureStat->StopSimulation != true)
	  MeasureStat->SampleCount++;
      else
	{
	  StatSema--;
	  NoPrecCount--;
	}
    }/* else */
  if (MeasureStat->StopSimulation != true)
    {
      if (MeasureStat->SampleCount > SPEC_n_max)
	{
	  printf(
	       "\nNOTE: A measure has exceeded the maximum number of samples\n");
	  printf(
	 "      Simulation stops without reaching the predefined accuracy!\n");
	  MeasureStat->StopSimulation = true;
	  MeasureStat->NoPrecision = true;
	  StatSema--;
	}
      if (time_exceeded)
	{
          printf(
	"\nNOTE: The simulation has exceeded the specified model time limit.\n");
          printf(
         "      Simulation stops without reaching the predefined accuracy!\n");
          MeasureStat->StopSimulation = true;
          MeasureStat->NoPrecision = true;
	  StatSema--;
        }
    }
  return (MeasureStat->StopSimulation);
}/*SPEC_Analysis*/


