/******************************************************************************

Programm	:	graph

Komponente	:	csmc_csmcs.c

Aufgabe		:	Prozeduren fuer die Untersuchung der zugeordneten
			Markov - Ketten (SMC), die sich mit den Komponenten der
			SMC (CSMC) beschaeftigen.
			Exportiert die Prozedur part_smc_into_csmcs(),
			die eine SMC in die CSMCs teilt und die CSMCs
			weiterverarbeitet; insbesondere wird auf Isomorphie mit
			anderen CSMCs testet.		

Projektleiter	:	Christoph Lindemann TU Berlin

Programmierer	:	Armin Zimmermann

Datum		:	Juni 1993

Aenderung       :       Juni 1995 (Patrick Drescher)

******************************************************************************/

#include <stdio.h>
#include "graph_types.h"
#include "csmc_types.h"
#include "portab.h"

/* forward definitions */
Iso_Class *any_isomorphism();

void
part_smc_into_csmcs(p_smc)
	SMC	*p_smc;		/* SMC to part */
{
	graph	*csmc_new, *csmc_ok;
	node	*n_walk, *n_now;
	edge	*e_walk;
	char	preds;

	p_smc->states = NULL;
	csmc_new = new_graph();
	while ((n_walk = cut_node(p_smc->graph, NULL)) != NULL)
	{	csmc_ok = new_graph();
		n_walk->bits |= VISITED;
		paste_node(csmc_new, n_walk);
		while ((n_now = cut_node(csmc_new, NULL)) != NULL)
		{	paste_node(csmc_ok, n_now);
			e_walk = n_now->successors;
			preds = 1;
			while ((e_walk != NULL) || preds)  /* scan successors and predececcors */
			{	while(e_walk != NULL)
				{	n_walk = (node *)e_walk->to_node;
					if (!(e_walk->bits & PREEMPTING) && 
								!(n_walk->bits & VISITED))
					{	cut_node(p_smc->graph, n_walk);
						n_walk->bits |= VISITED;
						paste_node(csmc_new, n_walk);
					}
					e_walk = e_walk->next;
				}
				e_walk = preds ? n_now->predecessors : NULL;
				preds = 0;
			}
		}
		/* a new CSMC has been found */
		found_new_csmc(p_smc, csmc_ok);
		no_csmcs++;
	}
	delete_graph(&csmc_new);
	delete_graph(&(p_smc->graph));
}

/* a new CSMC has been found */
/* csmc_ok wird uebergeben */
found_new_csmc(p_smc, csmc_gr)
	SMC		*p_smc;
	graph		*csmc_gr;
{
	int		no_preempting = 0, no_initial = 0;
	char		iso_bits = 0, csmc_bits = 0;
	node		*n_walk, **initial_states;
	edge		*e_walk;
	Iso_Class	*iso_basic, *new_class;
	Tau_CSMC	*new_csmc, *tau_walk;
	Iso_CSMC	*new_iso;

#ifdef DEBUG 
	printf("==================\n");
	printf("\nfound new CSMC:\n");
	show_graph(csmc_gr, 0xff);
	printf("=================\n");
#endif

	/* get and count initial states, count preempting edges, mark states unvisited */
	initial_states = (node **)emalloc(csmc_gr->no_states * sizeof(node *));
	scan_list(n_walk, csmc_gr->first_node)
	{	if (n_walk->bits & INITIAL)
			initial_states[no_initial++] = n_walk;
		scan_list(e_walk, n_walk->successors)
			no_preempting += ((e_walk->bits & PREEMPTING) > 0);
		n_walk->bits &= ~VISITED;
	}

#ifdef DEBUG
	printf("no_preempting edges: %d\n", no_preempting);
#endif

	/* is the deterministic transition marking dependent ? */
	if (p_smc->trans->mark_ind)
		csmc_bits |= MARK_DEP_TAU;

#ifdef DEBUG
	printf("transition MD: %d\n", csmc_bits & MARK_DEP_TAU);
#endif

	/* obtain covering initial states (CIS) and find cycles */
	if (find_cov_ini_states(csmc_gr, initial_states, no_initial))
		iso_bits |= CYCLIC;
	efree(initial_states);

#ifdef DEBUG
	printf("cycles: %d\n", iso_bits & CYCLIC);
#endif

	/* prepare informations about new CSMC */
	new_class = (Iso_Class *)emalloc(sizeof(Iso_Class));
	new_class->no_preempt_arcs = no_preempting;
	new_class->bits = iso_bits;
	new_class->graph = csmc_gr;
	new_class->degrees = NULL;
	new_csmc = (Tau_CSMC *)emalloc(sizeof(Tau_CSMC));
	new_class->Tau_CSMCs = new_csmc;
	new_csmc->graph = csmc_gr;
	new_csmc->bits = csmc_bits;
	new_csmc->next = NULL;
	new_csmc->no_isom_csmcs = 0;
	new_csmc->Iso_CSMCs = 0;

	/* test for isomorphism with isomorphism classes */
	iso_basic = any_isomorphism(new_class);

	if (iso_basic != NULL)		/* add to found isomorphism class */
	{	iso_basic->no_total_csmcs++;
		scan_list(tau_walk, iso_basic->Tau_CSMCs)
			if (d_equal(tau_walk->tau, new_csmc->tau))
				break;
		if (tau_walk != NULL)		/* equal tau */
		{	new_iso = (Iso_CSMC *)emalloc(sizeof(Iso_CSMC));
			new_iso->bits = new_csmc->bits;
			new_iso->graph = new_csmc->graph;
			new_iso->next = tau_walk->Iso_CSMCs;
			tau_walk->Iso_CSMCs = new_iso;
			tau_walk->no_isom_csmcs++;
			efree(new_csmc);
		} else				/* different tau */
		{	if (iso_basic->Tau_CSMCs != NULL)
				for (tau_walk = iso_basic->Tau_CSMCs; tau_walk->next != NULL; 
						tau_walk = tau_walk->next);
			if (tau_walk != NULL)
				tau_walk->next = new_csmc;
			else
			{	new_csmc->next = iso_basic->Tau_CSMCs;
				iso_basic->Tau_CSMCs = new_csmc;
			}
			iso_basic->no_tau_csmcs++;
		}
		/* delete new_class */
		efree(new_class->sort_degrees);
		efree(new_class->degrees);
		efree(new_class);
	} else				/* create new isomorphism class */
		new_isomorphism_class(new_class);
}

/* no_pred_compare compares the number of predecessors of an initial state (for qsort) */
int
no_pred_compare(a, b)
	node	**a, **b;
{
	return ((*a)->no_pred - (*b)->no_pred);
}

/* obtain covering initial states (CIS) and test the CSMC for cycles
   (all states must be unvisited) */
int
find_cov_ini_states(gra, ini_states, no_ini_states)
	graph		*gra;
	node		**ini_states;
	int		no_ini_states;
{
	int		act_ini, i, cycles = 0;
	node		*n_walk, *h_walk;

	/* mark all initial states as CIS */
	scan_array(act_ini, no_ini_states)
	{	ini_states[act_ini]->bits |= COV_INITIAL;
	}

	/* sort initial states; root states come first */
	qsort((char *)ini_states, no_ini_states, sizeof(node *), no_pred_compare);

	/* scan all ancestors of each initial state */
	scan_array(act_ini, no_ini_states)
	{	n_walk = ini_states[act_ini];
		if (!(n_walk->bits & COV_INITIAL))
			continue;
		/* set all states to unvisited */
		scan_list(h_walk, gra->first_node)
			h_walk->bits &= ~VISITED;
		depth_first_search_CIS(n_walk, n_walk->number, &cycles);
	}
	gra->no_cov_states = i = 0;
	scan_array(act_ini, no_ini_states)
		gra->no_cov_states += ((ini_states[act_ini]->bits & COV_INITIAL) > 0);
	gra->cov_states = (node **)emalloc(gra->no_cov_states * sizeof(node *));
	scan_array(act_ini, no_ini_states)
		if (ini_states[act_ini]->bits & COV_INITIAL)
			gra->cov_states[i++] = ini_states[act_ini];
#ifdef DEBUG
	printf("covering initial states:");
	scan_array(i, gra->no_cov_states)
		show_node(gra->cov_states[i], NODE_BITS);
#endif

	return cycles;
}

/* procedure for find_cov_ini_states to scan all ancestors of a node */
depth_first_search_CIS(d_node, start_num, cycle_found)
	node		*d_node;
	int		start_num;
	int		*cycle_found;
{
	edge		*e_walk;
	node		*n_walk;

	d_node->bits |= MARKED | VISITED;
	if (d_node->number != start_num)	/* we are not at the start node */
		d_node->bits &= ~COV_INITIAL;	/* reset cov. initial bit */
	scan_list(e_walk, d_node->successors)
	{	if (e_walk->bits & PREEMPTING)
			continue;
		n_walk = (node *)e_walk->to_node;
		if (n_walk->bits & MARKED)
			*cycle_found = 1;
		if (!(n_walk->bits & VISITED))
			depth_first_search_CIS(n_walk, start_num, cycle_found);
	}
	d_node->bits &= ~MARKED;
}

/* is there an isomorphism with a Basic_CSMC ? NULL: no */
Iso_Class *
any_isomorphism(t_class)
	Iso_Class	*t_class;
{
	Iso_Class	*iso_walk;
	Tau_CSMC	*t_csmc = t_class->Tau_CSMCs;
	graph		*gra = t_class->graph, iso_graph;
	node		*n_walk;
	edge		*e_walk;
	char		tau_isomorphic;
	double		sum_of_all_Q_entries = 0, new_tau;
	int		num, bit, edge_count = 0, node_count = 0;

	/* calculate the sum of all Q-entries */
	scan_list(n_walk, gra->first_node)
		scan_list(e_walk, n_walk->successors)
			sum_of_all_Q_entries += e_walk->rate;
	t_class->sum_of_Qi = sum_of_all_Q_entries;

	/* count edges and number nodes */
	scan_list(n_walk, gra->first_node)
	{	n_walk->number = node_count++;
		edge_count += n_walk->no_succ;
	}
	t_class->no_edges = edge_count + gra->no_cov_states;

	scan_list(iso_walk, Iso_Classes)
	{	/* test, if an isomorphism is possible */
		if (t_class->no_preempt_arcs != iso_walk->no_preempt_arcs)
			continue;
		if ((t_class->bits & CYCLIC) != (iso_walk->bits & CYCLIC))
			continue;
		if (gra->no_states != iso_walk->graph->no_states)
			continue;
		if (gra->no_cov_states != iso_walk->graph->no_cov_states)
			continue;
		if (t_class->no_edges != iso_walk->no_edges)
			continue;

		/* rescale the new CSMC, if necessary */
		new_tau = 1.0;
		if (iso_walk->sum_of_Qi > 0.0)
			new_tau = t_class->sum_of_Qi / iso_walk->sum_of_Qi;
		if (d_equal(new_tau, 1.0))
			tau_isomorphic = 0;
		else	tau_isomorphic = 1;
	
		/* calculate the degrees, if necessary */
		if ((t_class->degrees == NULL) || !d_equal(new_tau, t_csmc->tau))
		{	if (t_class->degrees != NULL)
				efree(t_class->degrees);
			calc_degrees(t_class);
			/* test, if the initial partitions are equal */
			bit = 0;
			scan_array(num, gra->no_states)
				if (t_class->sort_degrees[num] != iso_walk->sort_degrees[num])
				{	bit = 1;
					break;	/* not isomorphic */
				}
			if (bit)
				continue;	/* not isomorphic */
			t_csmc->tau = new_tau;
		}
		
		/* test graphs for isomorphism */
		if (test_for_isomorphism(iso_walk, t_class, tau_isomorphic))
			return iso_walk;	/* isomorphism found */
	}
	return NULL;				/* no isomorphism found */
}

/* variables for the dfs-numbering of nodes and edges */
int		get_edge_node_num, get_edge_edge_num;
iso_edge	*iso_edges;

/* dfs for get_edge_list */
dfs_edge_list(n_from)
	node		*n_from;
{
	edge		*e_walk;
	node		*n_to;
	int		act_edge_num;

	n_from->bits |= VISITED;
	n_from->number = get_edge_node_num++;
	scan_list(e_walk, n_from->successors)
	{	act_edge_num = ++get_edge_edge_num;
		iso_edges[act_edge_num].node_from = n_from;
		iso_edges[act_edge_num].rate = e_walk->rate;
		if (e_walk->bits & PREEMPTING)
			iso_edges[act_edge_num].node_to = NULL;
		else
		{	n_to = (node *)e_walk->to_node;
			if (!(n_to->bits & VISITED))
				dfs_edge_list(n_to);
			iso_edges[act_edge_num].node_to = n_to;
		}
	}
}

/* DFS-number the nodes and get an edge list */
get_edge_list(t_class)
	Iso_Class	*t_class;
{
	graph		*gra = t_class->graph;
	node		*n_walk;
	iso_edge	*i_walk, *j_walk, i_mem;
	int		num, from, to, max;

	/* mark states unvisited */
	scan_list(n_walk, gra->first_node)
		n_walk->bits &= ~(VISITED | MARKED);

	iso_edges = (iso_edge *)emalloc(t_class->no_edges * sizeof(iso_edge));
	t_class->edges = iso_edges;

	/* depth-first search from the cov. initial states */
	get_edge_node_num = 0;
	get_edge_edge_num = -1;
	scan_array(num, gra->no_cov_states)
	{	get_edge_edge_num++;
		n_walk = gra->cov_states[num];
		iso_edges[get_edge_edge_num].node_from = NULL;
		iso_edges[get_edge_edge_num].node_to = n_walk;
		dfs_edge_list(n_walk);
	}

	/* calculate the orders of the subgraphs for all edges */
	iso_edges[0].order = 0;
	for (num = 1; num < t_class->no_edges; num++)
	{	i_walk = &(iso_edges[num-1]);
		from = (i_walk->node_from == NULL) ? 0 : i_walk->node_from->number;
		to   = (i_walk->node_to == NULL) ? 0 : i_walk->node_to->number;
		iso_edges[num].order = MAX(i_walk->order, MAX(from, to));
	}

	/* reorder edges if necessary */
	for (num = 1; num < t_class->no_edges; num++)
	{	i_walk = &(iso_edges[num]);
		from = (i_walk->node_from == NULL) ? 0 : i_walk->node_from->number;
		to   = (i_walk->node_to == NULL) ? 0 : i_walk->node_to->number;
		max = MAX(from, to);
		if (max >= i_walk->order)
			continue;
		for (to = num; to > 1; to--)
		{	j_walk = &(iso_edges[to-1]);
			if (max > j_walk->order)
				break;
			/* exchange i_walk and j_walk */
			i_walk = &(iso_edges[to]);
			memmove(&i_mem, i_walk, sizeof(iso_edge));
			memmove(i_walk, j_walk, sizeof(iso_edge));
			memmove(j_walk, &i_mem, sizeof(iso_edge));
			j_walk->order = i_walk->order;
		}
	}
}


/* procedure for qsort to compare two node degrees */
int
degree_comp(deg1, deg2)
	int	*deg1, *deg2;
{
	return (*deg1 - *deg2);
}

/* calculate the state degrees of a new isomorphism class */
calc_degrees(test_class)
	Iso_Class	*test_class;
{
	graph		*test_graph = test_class->graph;
	node		*n_walk;
	edge		*e_walk;
	int		M, from, to;
	double		*row_sum, *col_sum, max_row_sum = 0.0, max_col_sum = 0.0;
	int		*in_degree, *out_degree, max_degree = 0;

	/* calculate informations about degrees */
	row_sum = (double *)emalloc(test_graph->no_states * sizeof(double));
	col_sum = (double *)emalloc(test_graph->no_states * sizeof(double));
	in_degree  = (int *)emalloc(test_graph->no_states * sizeof(int));
	out_degree = (int *)emalloc(test_graph->no_states * sizeof(int));
	scan_array(from, test_graph->no_states)
	{	row_sum[from] = col_sum[from] = 0.0;
		in_degree[from] = out_degree[from] = 0;
	}
	scan_list(n_walk, test_graph->first_node)
	{	from = n_walk->number;
		scan_list(e_walk, n_walk->successors)
		{	out_degree[from]++;
			max_degree = MAX(max_degree, out_degree[from]);
			row_sum[from] += e_walk->rate;
			max_row_sum = MAX(max_row_sum, row_sum[from]);
			if (!(e_walk->bits & PREEMPTING))
			{	to = ((node *)e_walk->to_node)->number;
				in_degree[to]++;
				max_degree = MAX(max_degree, in_degree[to]);
				col_sum[to] += e_walk->rate;
				max_col_sum = MAX(max_col_sum, col_sum[to]);
			}
		}
	}
	if (max_degree > MAX_M)
		printf("M is too large (%d>%d) for optimal initial partition!\n",
								max_degree, MAX_M);
	M = MAX_M;

	/* calculate the state degrees */
	test_class->degrees = (int *)emalloc(test_graph->no_states * sizeof(int));
	test_class->sort_degrees = (int *)emalloc(test_graph->no_states * sizeof(int));
	scan_array(from, test_graph->no_states)
	{	to = in_degree[from] + M * out_degree[from]
			    + M * M * (unsigned)(row_sum[from] / max_row_sum * (float)(M-1))
			+ M * M * M * (unsigned)(col_sum[from] / max_col_sum * (float)(M-1));
		test_class->degrees[from] = to;
		test_class->sort_degrees[from] = to;
	}

	/* obtain initial partition, sorted degrees */
	qsort((char *)test_class->sort_degrees, test_graph->no_states, sizeof(int), degree_comp);
}

/* variables for isomorphism_test and match(edge) */
char		isomorphism, diff_tau;
int		max_no_edges, max_no_nodes;
int		edge_count;
double		tau_factor;
iso_edge	*edges1;
int		*degrees1, *degrees2;
node		**G1_to_G2, **G2_to_G1;
node		*G2_nodes;

/* to1 is an unassigned vertex; verify that the number of edges
   in the corresponding induced subgraph in G2 does not exceed the
   number of edges in G1(order[edge]) */
edge_count_test(act_edge)
	int		act_edge;
{
	node		*n_walk, *t_walk;
	edge		*e_walk;
	int		i;

	n_walk = G1_to_G2[edges1[act_edge].order];
	if (n_walk == NULL)
		return 0;
	scan_list(e_walk, n_walk->successors)
	{	if (e_walk->bits & PREEMPTING)
		{	edge_count--;
			continue;
		}
		t_walk = (node *)(e_walk->to_node);
		if (G2_to_G1[t_walk->number] != NULL)
			edge_count--;
	}
	scan_array(i, edges1[act_edge].order)
		scan_list(e_walk, G1_to_G2[i]->successors)
		{	if (e_walk->bits & PREEMPTING)
				continue;
			t_walk = (node *)(e_walk->to_node);
			if (t_walk == n_walk)
				edge_count--;
		}
	return edge_count;
}
			

/* backtracking algorithm to find an isomorphism */
match(edge_num)
	int		edge_num;
{
	double		weight, t_weight;
	node		*from1, *to1, *to2;
	edge		*e_walk;
	int		i;

	/* test, if the search was successful */
	if (edge_num >= max_no_edges)
	{	/* isomorphism found ! */
		isomorphism = 1;
		return;
	}
	/* actual edge */
	weight = edges1[edge_num].rate;
	from1 = edges1[edge_num].node_from;
	to1 = edges1[edge_num].node_to;
	if (from1 == NULL)
	{	if (edge_count_test(edge_num))
			return;		/* unequal number of edges -> backtrack */
		/* additional edge; 'to' is a cov. initial state */
		scan_list(to2, G2_nodes)
		{	if ((degrees1[to1->number] != degrees2[to2->number])
						|| G2_to_G1[to2->number] != NULL)
				continue;
			/* possible node found */
			G1_to_G2[to1->number] = to2; G2_to_G1[to2->number] = to1;
			edge_count = 0;
			match(edge_num + 1);
			if (isomorphism)
				return;
			G1_to_G2[to1->number] = NULL; G2_to_G1[to2->number] = NULL;
		}
		return;
	}
	/* normal edge */
	if ((to1 != NULL) && (G1_to_G2[to1->number] == NULL))
	{	/* to1 is an unassigned vertex; test number of edges */
		if (edge_count_test(edge_num))
			return;		/* unequal number of edges -> backtrack */
		/* assign to to1 an unassigned vertex to2 in G2 */
		scan_list(e_walk, G1_to_G2[from1->number]->successors)
		{	/* test, if nodes can correspond */
			if (e_walk->bits & PREEMPTING)
				continue;
			to2 = (node *)(e_walk->to_node);
			if (G2_to_G1[to2->number] != NULL)
				continue;		/* already assigned node */
			if (degrees1[to1->number] != degrees2[to2->number])
				continue;		/* degrees must match */
			t_weight = diff_tau ? e_walk->rate / tau_factor : e_walk->rate;
			if (!d_equal(weight, t_weight))
				continue;		/* rates must match */
			G1_to_G2[to1->number] = to2; G2_to_G1[to2->number] = to1;
			edge_count = 1;
			match(edge_num + 1);
			if (isomorphism)
				return;
			G1_to_G2[to1->number] = NULL; G2_to_G1[to2->number] = NULL;	
		}
		return;
	}
	/* to1 is an assigned node or the edge is preempting; verify
	   that there is an edge in G2 corresponding to from1->to1 */
	i = 0;
	scan_list(e_walk, G1_to_G2[from1->number]->successors)
	{	t_weight = diff_tau ? e_walk->rate / tau_factor : e_walk->rate;
		if (!d_equal(weight, t_weight))	/* different rate */
			continue;
		if ((to1 == NULL) || (e_walk->bits & PREEMPTING))
		{	if (!(e_walk->bits & PREEMPTING) || (to1 != NULL))
				continue;
			i = 1;			/* found preempting edge */
			break;
		}
		to2 = (node *)(e_walk->to_node);
		if (to2 == G1_to_G2[to1->number])
		{	i = 1;			/* found normal edge */
			break;
		}
	}
	if (!i)
		return;				/* no edge found; backtrack */
	edge_count++;
	match(edge_num + 1);
}

/* test two graphs for isomorphism */
int
test_for_isomorphism(basic_class, test_class, different_tau)
	Iso_Class	*basic_class, *test_class;
	int		different_tau;
{
	graph		*test_graph = test_class->graph;
	node		*n_walk, *t_walk, *o_walk;
	int		i;

	/* some preparations for the match procedure */
	isomorphism = 0; diff_tau = (char)different_tau;
	max_no_edges = basic_class->no_edges;
	max_no_nodes = test_graph->no_states;
	tau_factor = different_tau ? test_class->Tau_CSMCs->tau : 1.0;
	edges1 = basic_class->edges;
	degrees1 = basic_class->degrees; degrees2 = test_class->degrees;
	G1_to_G2 = (node **)emalloc(test_graph->no_states * sizeof(node *));
	G2_to_G1 = (node **)emalloc(test_graph->no_states * sizeof(node *));
	G2_nodes = test_graph->first_node;
	scan_array(i, max_no_nodes)
		G1_to_G2[i] = G2_to_G1[i] = NULL;

#ifdef DEBUG
	printf("==============\niso-test:\n");
	printf("Graph-alt:\n");
	show_graph(basic_class->graph, 0x3b);
	printf("Graph-test:\n");
	show_graph(test_class->graph, 0x3b);
#endif

	/* start backtracking algorithm to match edges and nodes */
	match(0);

#ifdef DEBUG
	printf("isomorphism: %d\n", isomorphism);
#endif

	if (!isomorphism)
		return 0;

	/* sort the linear list of nodes of G2 according to G1 */
	o_walk = NULL;
	n_walk = basic_class->graph->first_node;
	while (n_walk != NULL)
	{	t_walk = G1_to_G2[n_walk->number];
		if (o_walk == NULL)
			test_graph->first_node = t_walk;
		else	o_walk->next = t_walk;
		n_walk = n_walk->next;
		o_walk = t_walk;
	}
	o_walk->next = NULL;

	efree(G1_to_G2);
	efree(G2_to_G1);
	return 1;
}

/* create new isomorphism class */
new_isomorphism_class(n_class)
	Iso_Class	*n_class;
{
	Iso_Class	*iso_walk;
	graph		*test_graph = n_class->graph;
	node		*n_walk;

	/* DFS-number the nodes and get an edge list */
	get_edge_list(n_class);

	/* calculate the state degrees (again, but the states are dfs-ordered) */
	calc_degrees(n_class);

	/* sort CSMC-states topological */
#ifdef DEBUG
	printf("new iso-class:\n");
	printf("before topological sorting:\n");
	show_graph(n_class->graph, 0x3b);
#endif

	top_sort_cyclic_graph(n_class->graph, n_class->bits & CYCLIC);

#ifdef DEBUG
	printf("after topological sorting:\n");
	show_graph(n_class->graph, 0x3b);
#endif

	/* test CSMC for special topology and  resort the states
	by ascending tree levels, if a BRANCH-topology is detected */
	test_csmc_topology(n_class);

	if (((n_class->bits & (DIRECT_TOP | ERLANG_TOP | HYPO_TOP | BRANCH_TOP))
	     && !(n_class->bits & NO_TOP_CALC)) || (Iso_Classes == NULL))
	{	/* enter the new class at the beginning of the list */
		n_class->next = Iso_Classes;
		Iso_Classes = n_class;
	} else
	{	/* enter the new class at the end of the list */
		for (iso_walk=Iso_Classes; iso_walk->next!=NULL; iso_walk=iso_walk->next);
		iso_walk->next = n_class;
		n_class->next = NULL;
	}
	n_class->Tau_CSMCs->tau = 1.0;
	n_class->no_tau_csmcs = 1;
	n_class->no_total_csmcs = 1;
	no_iso_classes++;
}

