/******************************************************************************

Programm	:	graph

Komponente	:	csmc_spec.c

Aufgabe		:	Dieser Programmteil enthaelt Prozeduren zur
			Erkennung, Numerierung und direkten Berechnung 
			der transienten Loesung von Komponenten von
			zugeordneten Markov-Ketten (CSMC), die spezielle
			Topologien aufweisen.
			Folgende Topologien werden verarbeitet:
			DIRECT_TOP =	direkte Topologie; die CSMC
					besteht nur aus einem Zustand
			ERLANG_TOP = 	Erlang-Topologie; die CSMC besteht
					aus einer linearen Kette von
					Zustaenden; die Raten sind gleich
			HYPO_TOP =	Hypoexponential-Topologie; die CSMC besteht
					aus einer linearen Kette von
					Zustaenden; alle Raten sind verschieden
			BRANCH_TOP =	N-Branch-Topologie; die Zustaende
					sind baumfoermig angeordnet, die
					Anzahl der Nachfolger ist gleich,
					die Raten entsprechen sich auf einer
					Ebene und der Baum ist ausgewogen

Projektleiter	:	Christoph Lindemann TU Berlin

Programmierer	:	Armin Zimmermann

Datum		:	Juni 1993

******************************************************************************/


#include <stdio.h>
#include <string.h>
#include <math.h>
#include "graph_types.h"
#include "csmc_types.h"
#include "matrices_P_C.h"
#include "rand_types.h"

/* rand_procs ueberlisten */
int no_csmc_states, no_total_states, first_state_no;
void my_exit(){}
char *visited;
int branch_n, branch_k, *branch_level, **state_succs;
sp_matrix *A_matrix;
sp_vector **A_matrix_rows;
char *leaf_state;

/* procedure for a qsort */
int	*levels;
level_comp(a, b)
	node		**a, **b;
{
	return	(levels[(*a)->number] - levels[(*b)->number]);
}

/* another procedure for qsort */
rate_comp(a, b)
	double		*a, *b;
{
	if (*a > *b)
		return  1;
	if (*a < *b)
		return -1;
	return 0;
}

/* test CSMC for special topology
   (the CSMC states are already topological sorted) */
void
test_csmc_topology(t_class)
	Iso_Class	*t_class;
{
	graph		*gra = t_class->graph;
	node		*n_walk, *n_to, **level_nodes, **num_to_node;
	edge		*e_walk, *te_walk;
	int		no_state_successors, bit = 1;
	int		no_levels, i;
	double		comp_rate, **level_rates, *hypo_rates, test_val;

	/* test for DIRECT topology */
	if (gra->no_states == 1)
	{	t_class->bits |= DIRECT_TOP;
		return;
	}

	if ((t_class->bits & CYCLIC) || (t_class->no_preempt_arcs))
		return;			/* no topology possible */
	if (gra->no_cov_states > 1)
		return;			/* only one cov. initial state possible */
	if (gra->cov_states[0]->no_pred)
		return;			/* the CIS must be a root state */

	/* test, if the number of state successors is the same for all nodes */
	no_state_successors = gra->first_node->no_succ;
	comp_rate = gra->first_node->successors->rate;	/* must exist */
	scan_list(n_walk, gra->first_node)
	{	if (n_walk->no_succ && (no_state_successors != n_walk->no_succ))
			return;
		scan_list(e_walk, n_walk->successors)
			if (!d_equal(comp_rate, e_walk->rate))
				bit = 0;		/* different rates */
	}

	/* test for ERLANG and HYPO topology */
	if (no_state_successors == 1)
	{	if (bit)
		{	t_class->bits |= ERLANG_TOP;
			return;
		}
		/* test for HYPO topology */
		i = bit = 0;
		hypo_rates = (double *)emalloc((gra->no_states - 1) * sizeof(double));
		scan_list(n_walk, gra->first_node)
			if (n_walk->successors != NULL)
				hypo_rates[i++] = n_walk->successors->rate;
		qsort((char *)hypo_rates, gra->no_states - 1, sizeof(double), rate_comp);
		scan_array(i, gra->no_states - 2)
			if (hypo_rates[i] > hypo_rates[i+1] - MAX_LAMBDA_DIFF_HYPO)
			{	bit = 1;		/* same rates */
				break;
			}
		if (!bit)	
		{	t_class->bits |= HYPO_TOP;	/* HYPO topology */
			/* due to numerical problems: dont use hypo topology */
			t_class->bits |= NO_TOP_CALC;
			return;
		}
	}

	/* calculate level-information for tree */
	levels = (int *)emalloc(gra->no_states * sizeof(int));
	scan_array(no_levels, gra->no_states)
		levels[no_levels] = 0;
	levels[gra->first_node->number] = 0;
	no_levels = 0;
	scan_list(n_walk, gra->first_node)
	{	bit = levels[n_walk->number] + 1;
		if ((bit == 1) && (n_walk != gra->first_node))
		{	printf("ERROR in test_csmc_top: csmc isnt topological ordered!\n");
			exit(1);
		}
		scan_list(e_walk, n_walk->successors)
		{	n_to = (node *)e_walk->to_node;	/* cant be a preempting edge */
			levels[n_to->number] = bit;
			no_levels = MAX(no_levels, bit);
		}
	}

	/* test, if the tree is balanced (each leaf has the same level) */
	scan_list(n_walk, gra->first_node)
		if ((!n_walk->no_succ) && (levels[n_walk->number] != no_levels))
		{	efree(levels);
			return;			/* no branching topology */
		}
	no_levels++;	/* number of rates is 1 bigger than leaf level */

	/* test, if the rates are the same on each level */
	level_nodes = (node **)emalloc(no_levels * sizeof(node *));
	scan_array(bit, no_levels)
		level_nodes[bit] = NULL;	
	scan_list(n_walk, gra->first_node)
	{	if (!n_walk->no_succ)	/* leaf state */
			continue;
		bit = levels[n_walk->number];
		if (level_nodes[bit] == NULL)
		{	level_nodes[bit] = n_walk;
			continue;
		}
		e_walk = level_nodes[bit]->successors;
		te_walk = n_walk->successors;
		while (e_walk != NULL)
		{	if (!d_equal(e_walk->rate, te_walk->rate))
			{	efree(level_nodes);
				efree(levels);
				return;		/* no branching topology */
			}
			e_walk = e_walk->next;
			te_walk = te_walk->next;
		}
	}

	/* CSMC has a branching topology */
	t_class->bits |= BRANCH_TOP;
	/* due to numerical problems: dont solve branch topology analytical */
	t_class->bits |= NO_TOP_CALC;
	t_class->branch_n = no_state_successors;
	t_class->branch_k = no_levels;

	/* resort the tree by ascending levels */
	num_to_node = (node **)emalloc(gra->no_states * sizeof(node *));
	scan_list(n_walk, gra->first_node)
		num_to_node[n_walk->number] = n_walk;

	qsort((char *)num_to_node, gra->no_states, sizeof(node *), level_comp);
	n_walk = NULL;
	scan_array(bit, gra->no_states)
		if (n_walk == NULL)
			gra->first_node = n_walk = num_to_node[bit];	
		else
		{	n_walk->next = num_to_node[bit];
			n_walk = n_walk->next;
			n_walk->next = NULL;
		}

	efree(num_to_node);
	efree(level_nodes);
	efree(levels);
}

/* number nodes of a CSMC with special topology */
void
number_csmc_topology(n_class, state_count)
	Iso_Class	*n_class;
	int		*state_count;
{
	Tau_CSMC	*tau_walk;
	Iso_CSMC	*iso_walk;
	node		*n_walk;

	scan_list(tau_walk, n_class->Tau_CSMCs)
	{	scan_list(n_walk, tau_walk->graph->first_node)
		{	n_walk->orig_mark->number = (*state_count)++;
			write_rg_and_rates_file(n_walk->orig_mark);
		}
		scan_list(iso_walk, tau_walk->Iso_CSMCs)
			scan_list(n_walk, iso_walk->graph->first_node)
			{	n_walk->orig_mark->number = (*state_count)++;
				write_rg_and_rates_file(n_walk->orig_mark);
			}
	}
}

/* solve direct topology */
void
solve_direct_top(class_id, tau, r_graph, p_fp, c_fp)
	int		class_id;
	double		tau;
	graph		*r_graph;
	FILE		*p_fp, *c_fp;
{
	node		*r_node = r_graph->first_node;
	tang_marking	*t_walk = r_node->orig_mark;
	tang_path	*p_walk;
	sp_vector	*p_vector = new_vector();
	sp_entry	*entry_walk;
	double		value, sum_of_rates = t_walk->sum_of_rates;
	double		exp_sum_rates, one_m_exp_sum_rates;
	int		e_num, index, no_edges = r_node->no_succ;

	exp_sum_rates = exp( -sum_of_rates );
	one_m_exp_sum_rates = 1.0 - exp_sum_rates;

	/* assemble P vector entries from deterministic firing paths */
	scan_array(e_num, t_walk->no_det_paths)
	{	p_walk = &(t_walk->det_paths[e_num]);
		index = ((tang_marking *)(p_walk->marking))->number;
		value =  no_edges ? exp_sum_rates * p_walk->prob : p_walk->prob;
		add_to_vector(p_vector, NULL, index, value);
	}

	/* assemble P vector entries from exponential firing paths */
	scan_array(e_num, t_walk->no_exp_paths)
	{	p_walk = &(t_walk->exp_paths[e_num]);
		index = ((tang_marking *)(p_walk->marking))->number;
		value = one_m_exp_sum_rates * p_walk->prob / sum_of_rates;
		add_to_vector(p_vector, NULL, index, value);
	}

	/* write header to EMC file */
	P_row_header.row_index = t_walk->number;
	P_row_header.entries_in_row = p_vector->no_entries;
	if (fwrite (&P_row_header, SIZEOF_P_row_header, 1, p_fp) != 1)
	{	printf("ERROR while writing to EMC%02u.\n", class_id);
		exit(1);
	}

	/* write firing path probabilities */
	scan_list(entry_walk, p_vector->entries)
	{	/* write P matrix entry to EMC */
		P_row_entry.column_index = entry_walk->num;
		P_row_entry.probability = entry_walk->val;
		if (fwrite (&P_row_entry, SIZEOF_P_row_entry, 1, p_fp) != 1)
		{	printf("ERROR while writing to EMC%02u.\n", class_id);
			exit(1);
		}
	}

	/* write conversion factor */
	/* write header to CONV */
	C_row_header.row_index = t_walk->number;
	C_row_header.entries_in_row = 1;
	if (fwrite (&C_row_header, SIZEOF_C_row_header, 1, c_fp) != 1)
	{	printf("ERROR while writing to CONV%02u.\n", class_id);
		exit(1);
	}

	/* write C matrix entry */
	C_row_entry.column_index = t_walk->number;
	C_row_entry.conversion_factor = no_edges ? (one_m_exp_sum_rates) / sum_of_rates : tau;
	C_row_entry.conversion_factor *= det_transition_delay(t_walk);
	if (fwrite (&C_row_entry, SIZEOF_C_row_entry, 1, c_fp) != 1)
	{	printf("ERROR while writing to CONV%02u.\n", class_id);
		exit(1);
	}
	del_vector(&p_vector);
}


/* solve erlang topology */
int		erl_no_states, max_probs;
double		*c_i, *c_n, *p_i, *p_n;

/* function to return the p- and c-matrix entries for erlang topologies */
void
erlang_p_c_value(from, to, p, c)
	int		from, to;
	double		*p, *c;
{
	register int	diff = to - from;

	if (diff >= max_probs)
	{	*p = 0.0;		/* solve numerical problems */
		*c = 0.0;
		return;
	}

	if (to == erl_no_states - 1)	/* last state */
	{	*p = p_n[diff];
		*c = c_n[diff];
		return;
	}
	*p = p_i[diff];			/* normal state */
	*c = c_i[diff];
}

void
solve_erlang_top(class_id, tau, r_graph, p_fp, c_fp)
	int		class_id;
	double		tau;
	graph		*r_graph;
	FILE		*p_fp, *c_fp;
{
	node		*n_walk = r_graph->first_node, *nh_walk;
	tang_marking	*t_walk, *th_walk;
	tang_path	*p_walk;
	sp_vector	*p_vector = new_vector(), *c_vector = new_vector();
	sp_entry	*entry_walk;
	double		lambda = n_walk->successors->rate;
	double		lambda_tau = lambda * 1.0, value, c_value;
	int		i, index, from, to, no_states = r_graph->no_states;

	/* initialize */
	erl_no_states = r_graph->no_states;
	c_i = (double *)emalloc(no_states * sizeof(double));
	c_n = (double *)emalloc(no_states * sizeof(double));
	p_i = (double *)emalloc(no_states * sizeof(double));
	p_n = (double *)emalloc(no_states * sizeof(double));

	/* start values */
	c_i[0] = ( 1.0 - exp( -lambda_tau )) / lambda;
	c_n[0] = 1.0;
	p_i[0] = exp( -lambda_tau );
	p_n[0] = 1.0;

	/* calculate necessary values */
	max_probs = no_states;
	scan_array(i, no_states - 1)
	{	p_i[i+1] = p_i[i] * lambda_tau / (double)(i+1);
		p_n[i+1] = p_n[i] - p_i[i];
		if (p_n[i+1] < 0.0)
			p_n[i+1] = 0.0;
		c_i[i+1] = c_i[i] - p_i[i+1] / lambda;
		if (c_i[i+1] < 0.0)
			c_i[i+1] = 0.0;
		c_n[i+1] = c_n[i] - c_i[i];
		if (c_n[i+1] < 0.0)
			c_n[i+1] = 0.0;
		if (is_zero(p_i[i+1]) && (p_n[i+1] < 0.5))
		{	max_probs = i+1;
			break;
		}
	}
	
	/* scan all markings */
	scan_list(n_walk, r_graph->first_node)
	{	t_walk = n_walk->orig_mark;
		from = t_walk->number - r_graph->first_node->orig_mark->number;

		/* write informations for non-initial state */
		if (!(n_walk->bits & INITIAL))
		{	/* write empty header to CONV file */
			C_row_header.row_index = t_walk->number;
			C_row_header.entries_in_row = 0;
			if (fwrite (&C_row_header, SIZEOF_C_row_header, 1, c_fp) != 1)
			{	printf("ERROR while writing to CONV%02u.\n", class_id);
				exit(1);
			}

			/* write empty header to EMC file */
			P_row_header.row_index = t_walk->number;
			P_row_header.entries_in_row = 0;
			if (fwrite (&P_row_header, SIZEOF_P_row_header, 1, p_fp) != 1)
			{	printf("ERROR while writing to EMC%02u.\n", class_id);
				exit(1);
			}
			continue;
		}

		scan_list(nh_walk, n_walk)
		{	th_walk = nh_walk->orig_mark;
			to = th_walk->number - r_graph->first_node->orig_mark->number;
			erlang_p_c_value(from, to, &value, &c_value);

			/* break, if no more values */
			if ((value<1e-10) && (c_value<1e-10))
				break;

			/* assemble P vector entries from deterministic firing paths */
			if (value>1e-10)
				scan_array(i, th_walk->no_det_paths)
				{   p_walk = &(th_walk->det_paths[i]);
				    index = ((tang_marking *)(p_walk->marking))->number;
				    add_to_vector(p_vector, NULL, index, value * p_walk->prob);
				}

			/* obtain C vector entry */
			if (c_value>1e-10)
			{	index = th_walk->number;
				c_value *= det_transition_delay(th_walk);
				add_to_vector(c_vector, NULL, index, c_value);
			}
		}

		/* write header to CONV file */
		t_walk = n_walk->orig_mark;
		C_row_header.row_index = t_walk->number;
		C_row_header.entries_in_row = c_vector->no_entries;
		if (fwrite (&C_row_header, SIZEOF_C_row_header, 1, c_fp) != 1)
		{	printf("ERROR while writing to CONV%02u.\n", class_id);
			exit(1);
		}

		/* write C vector */
		scan_list(entry_walk, c_vector->entries)
		{	C_row_entry.column_index = entry_walk->num;
			C_row_entry.conversion_factor = entry_walk->val;
			if (fwrite (&C_row_entry, SIZEOF_C_row_entry, 1, c_fp) != 1)
			{	printf("ERROR while writing to CONV%02u.\n", class_id);
				exit(1);
			}
		}

		/* write header to EMC file */
		P_row_header.row_index = t_walk->number;
		P_row_header.entries_in_row = p_vector->no_entries;
		if (fwrite (&P_row_header, SIZEOF_P_row_header, 1, p_fp) != 1)
		{	printf("ERROR while writing to EMC%02u.\n", class_id);
			exit(1);
		}

		/* write firing path probabilities */
		scan_list(entry_walk, p_vector->entries)
		{	/* write P matrix entry to EMC */
			P_row_entry.column_index = entry_walk->num;
			P_row_entry.probability = entry_walk->val;
			if (fwrite (&P_row_entry, SIZEOF_P_row_entry, 1, p_fp) != 1)
			{	printf("ERROR while writing to EMC%02u.\n", class_id);
				exit(1);
			}
		}
		clear_vector(p_vector);
		clear_vector(c_vector);
	}

	/* clean up */
	del_vector(&p_vector);
	del_vector(&c_vector);
	efree(c_i);
	efree(c_n);
	efree(p_i);
	efree(p_n);
}

/* solve hypo topology */
void
solve_hypo_top(class_id, tau, r_graph, p_fp, c_fp)
	int		class_id;
	double		tau;
	graph		*r_graph;
	FILE		*p_fp, *c_fp;
{
	node		*from_walk, *to_walk;
	tang_marking	*from_mark, *to_mark;
	tang_path	*p_walk;
	sp_vector	*p_vector = new_vector(), *c_vector = new_vector();
	sp_entry	*entry_walk;
	double		*lambda, *p_factor, *c_factor, *a;
	double		p_i, p_n, c_i, c_n;
	int		i, from, to, no_states = r_graph->no_states;
	int		first_num = r_graph->first_node->orig_mark->number;

	/* initialize */
	a        = (double *)emalloc(no_states * sizeof(double));
	lambda   = (double *)emalloc(no_states * sizeof(double));
	p_factor = (double *)emalloc(no_states * sizeof(double));
	c_factor = (double *)emalloc(no_states * sizeof(double));
	for (from_walk = r_graph->first_node; from_walk->next != NULL; from_walk = from_walk->next)
	{	i = from_walk->orig_mark->number - first_num;
		lambda[i] = from_walk->successors->rate;
		p_factor[i] = exp(- lambda[i] * tau);
		c_factor[i] = (1.0 - p_factor[i]) / lambda[i];
	}

	/* scan all initial states */
	scan_list(from_walk, r_graph->first_node)
	{	from_mark = from_walk->orig_mark;
		from = from_mark->number - first_num;

		/* write informations for non-initial state */
		if (!(from_walk->bits & INITIAL))
		{	/* write empty header to CONV file */
			C_row_header.row_index = from_mark->number;
			C_row_header.entries_in_row = 0;
			if (fwrite (&C_row_header, SIZEOF_C_row_header, 1, c_fp) != 1)
			{	printf("ERROR while writing to CONV%02u.\n", class_id);
				exit(1);
			}

			/* write empty header to EMC file */
			P_row_header.row_index = from_mark->number;
			P_row_header.entries_in_row = 0;
			if (fwrite (&P_row_header, SIZEOF_P_row_header, 1, p_fp) != 1)
			{	printf("ERROR while writing to EMC%02u.\n", class_id);
				exit(1);
			}
			continue;
		}

		/* initializations for from_state */
		p_n = 1.0;
		c_n = tau;
		for (i = from; i < no_states; i++)
			a[i] = 1.0;

		/* scan to - markings and calculate p- and c-values */
		scan_list(to_walk, from_walk)
		{	/* initializations */
			p_i = c_i = 0.0;
			to_mark = to_walk->orig_mark;
			to = to_mark->number - first_num;

			/* derive new values */
			if (to == no_states-1)
			{	p_i = p_n;		/* last state */
				c_i = c_n;
			} else
			{	for (i = from; i <= to; i++)
				{	p_i += a[i] * p_factor[i];
					c_i += a[i] * c_factor[i];
				}
				p_n -= p_i;
				c_n -= c_i;
			}

			/* assemble P vector entries from deterministic firing paths */
			if (!is_zero(p_i))
				scan_array(i, to_mark->no_det_paths)
				{	p_walk = &(to_mark->det_paths[i]);
					i = ((tang_marking *)(p_walk->marking))->number;
					add_to_vector(p_vector, NULL, i, p_i * p_walk->prob);
				}

			/* obtain C vector entry */
			if (!is_zero(c_i))
			{	c_i *= det_transition_delay(to_mark);
				add_to_vector(c_vector, NULL, to_mark->number, c_i);
			}

			/* test for truncation possibility */
			if (is_zero(p_n))
				break;		/* no more values */
			if (to >= no_states - 2)
				continue;	/* dont calculate a[n] */

			/* obtain new a - values */
			for (i = from; i < no_states-1; i++)
				a[i] *= (i == to+1) ? lambda[to] / (lambda[from] - lambda[i])
						    : lambda[to] / (lambda[to+1] - lambda[i]);
		}

		/* write header to CONV file */
		C_row_header.row_index = from_mark->number;
		C_row_header.entries_in_row = c_vector->no_entries;
		if (fwrite (&C_row_header, SIZEOF_C_row_header, 1, c_fp) != 1)
		{	printf("ERROR while writing to CONV%02u.\n", class_id);
			exit(1);
		}

		/* write C vector */
		scan_list(entry_walk, c_vector->entries)
		{	C_row_entry.column_index = entry_walk->num;
			C_row_entry.conversion_factor = entry_walk->val;
			if (fwrite (&C_row_entry, SIZEOF_C_row_entry, 1, c_fp) != 1)
			{	printf("ERROR while writing to CONV%02u.\n", class_id);
				exit(1);
			}
		}

		/* write header to EMC file */
		P_row_header.row_index = from_mark->number;
		P_row_header.entries_in_row = p_vector->no_entries;
		if (fwrite (&P_row_header, SIZEOF_P_row_header, 1, p_fp) != 1)
		{	printf("ERROR while writing to EMC%02u.\n", class_id);
			exit(1);
		}

		/* write firing path probabilities */
		scan_list(entry_walk, p_vector->entries)
		{	/* write P matrix entry to EMC */
			P_row_entry.column_index = entry_walk->num;
			P_row_entry.probability = entry_walk->val;
			if (fwrite (&P_row_entry, SIZEOF_P_row_entry, 1, p_fp) != 1)
			{	printf("ERROR while writing to EMC%02u.\n", class_id);
				exit(1);
			}
		}
		clear_vector(p_vector);
		clear_vector(c_vector);
	}

	/* clear */
	del_vector(&p_vector);
	del_vector(&c_vector);
	efree(lambda);
	efree(p_factor);
	efree(c_factor);
	efree(a);		
}

/* --- solve branch topology */
void
solve_branch_top(class_id, tau, r_graph, p_fp, c_fp)
	int		class_id;
	double		tau;
	graph		*r_graph;
	FILE		*p_fp, *c_fp;
{
	

	printf("geht noch nicht\n");
}

/* solve special topologies direct */
void
solve_csmc_topology(m_name, s_class)
	char		*m_name;
	Iso_Class	*s_class;
{
	Tau_CSMC	*tau_walk;
	Iso_CSMC	*iso_walk;
	void		(*solve_procedure)();
	int		m_len = strlen(m_name) + 20;
	char		*conv_filename, *emc_filename;
	FILE		*emc_fp, *conv_fp;

	/* prepare file handles */
	conv_filename = (char *)emalloc(m_len * sizeof(char));
	emc_filename = (char *)emalloc(m_len * sizeof(char));
	sprintf (conv_filename, "%s.CONV%02u", m_name, s_class->class_id);
	sprintf (emc_filename, "%s.EMC%02u", m_name, s_class->class_id);
	if ((emc_fp = fopen(emc_filename, "w")) == NULL)
	{	printf("ERROR: could not open %s for writing.\n", emc_filename);
		exit(1);
	}
	if ((conv_fp = fopen(conv_filename, "w")) == NULL)
	{	printf("ERROR: could not open %s for writing.\n", conv_filename);
		exit(1);
	}

	/* set the right solve procedure */
	if (s_class->bits & DIRECT_TOP)
		solve_procedure = solve_direct_top;
	else if (s_class->bits & ERLANG_TOP)
		solve_procedure = solve_erlang_top;
	else if (s_class->bits & HYPO_TOP)
		solve_procedure = solve_hypo_top;
	else if (s_class->bits & BRANCH_TOP)
		solve_procedure = solve_branch_top;

	/* scan all sets of CSMCs with equal tau */
	scan_list(tau_walk, s_class->Tau_CSMCs)
	{	(*solve_procedure)(s_class->class_id, tau_walk->tau, 
						tau_walk->graph, emc_fp, conv_fp);
		scan_list(iso_walk, tau_walk->Iso_CSMCs)
			(*solve_procedure)(s_class->class_id, tau_walk->tau, 
						iso_walk->graph, emc_fp, conv_fp);
	}

	/* clean up */
	fclose(conv_fp);
	fclose(emc_fp);
	efree(conv_filename);
	efree(emc_filename);
}
