/******************************************************************************

Programm	:	graph

Komponente	:	csmc_types.h

Aufgabe		:	Typdefinitionen fuer die Untersuchung der
			zugeordneten Markov - Ketten

Bemerkung	:	graph_types.h muss bereits includiert worden sein

Projektleiter	:	Christoph Lindemann TU Berlin

Programmierer	:	Armin Zimmermann

Datum		:	Juni 1993

Aenderung       :       Juni 1995 (Patrick Drescher)

******************************************************************************/

#define	COMP_DOUBLE_EPS		1e-10		/* epsilon for comparing doubles */
#define	MAX_LAMBDA_DIFF_HYPO	1e-10		/* if the difference between two
						   rates of the hypoexponential
						   topology is less than this,
						   randomization is applied */
#define	MAX_HYPO_A_VALUE	1e8		/* maximal a value for hypo
						   topology to avoid numerical
						   problems */

		/* macros for short scanning of linear lists and arrays */
#define scan_list(walk,list)	for (walk = list; walk != NULL; walk = walk->next)
#define scan_array(count,num)	for (count = 0; count < num; count++)


/* types for abstract graph implementation */

	/* edge - bits */
#define	PREEMPTING	0x01

typedef struct gr_arc
{	double		rate;		/* state transition rate (succ only) */
	char		bits;		/* preempting ? .. */
	void		*to_node;	/* pointer to succ/pred node, type gr_node* */
					/* preempting case: original tang_marking * */
	struct gr_arc	*next;		/* linear list */
}  edge;

	/* node - bits */
#define	INITIAL		0x01		/* node is an initial state */
#define	COV_INITIAL	0x02		/* node is an covering initial state */
#define	VISITED		0x04		/* node is visited (DFS) */
#define	MARKED		0x08		/* node is marked */
#define	LAST_OF_SCC	0x10		/* node is the last one (in the linear order)
					   of a strongly connected component */

typedef struct gr_node
{	int		number;		/* identification number of node */
	char		bits;		/* initial state, visited .. */
	char		no_succ;	/* number of successors */
	char		no_pred;	/* number of predecessors */
	edge		*successors;	/* list of successors */
	edge		*predecessors;	/* list of predecessors */
	tang_marking 	*orig_mark;	/* pointer to original marking in RRG */
	struct gr_node	*next;		/* linear list of nodes */
}  node;

typedef struct graph
{	node		*first_node;	/* pointer to first node */
	int		no_states;	/* number of states in graph */
	int		no_cov_states;	/* number of covering states */
	node		**cov_states;	/* set of covering states for DFS algor. */
/*	struct graph	*next;		   linear list for sets of graphs */
}  graph;

/* type for subordinated markov chains */

typedef struct SMC
{	timed_t		*trans;		/* description of deterministic transition */
	int		no_states;	/* number of states */
	int		first_state_no;	/* number of first state */
	graph		*graph;		/* abstract graph */
	tang_marking	*states;	/* linear list of SMC states */
}  SMC;

/* types for isomorphism test */

#define	MAX_M		100		/* maximal 255 */

typedef struct iso_edge
{	double	rate;			/* the edge's rate */
	node	*node_from;		/* predecessing node */
	node	*node_to;		/* successing node */
	int	order;			/* order of the subgraph */
}  iso_edge;

/* types for components of subordinated markov chains */

	/* CSMC - bits */
#define	MARK_DEP_TAU	0x01		/* corresp. det. transition is marking dep. */

	/* isomorphic csmc with equal tau*/
typedef struct Iso_CSMC
{	char		bits;		/* some properties */
	graph		*graph;		/* abstract graph */
	struct Iso_CSMC *next;		/* linear list */
}  Iso_CSMC;

	/* isomorphic csmc with different tau */
typedef struct Tau_CSMC
{	double		tau;		/* firing delay of det. trans. of CSMC */
	char		bits;		/* MD det. trans. */
	int		no_isom_csmcs;	/* number of total isomorphic CSMCs */
	Iso_CSMC	*Iso_CSMCs;	/* list of isomorphic CSMCs with the same tau */
	graph		*graph;		/* abstract graph */
	struct Tau_CSMC *next;		/* linear list */
}  Tau_CSMC;

	/* isomorphism class bits */	
#define	CYCLIC		0x01		/* class is cyclic */
#define	DIRECT_TOP	0x02		/* special topology 'direct' (one state) */
#define	ERLANG_TOP	0x04		/* special topology 'erlang' */
#define	HYPO_TOP	0x08		/* special topology 'hypo' */
#define	BRANCH_TOP	0x10		/* special topology 'branching' */
#define	NO_TOP_CALC	0x20		/* dont use topology */

	/* isomorphism class of CSMCs */
typedef struct Iso_Class
{	int		class_id;	/* id number of isom. class (= EMC, CONV-number) */
	int		first_state_no;	/* number of first state */
	int		no_preempt_arcs;/* number of preempting arcs */	
	char		bits;		/* some properties */
	int		no_total_csmcs;	/* number of all CSMCs in this class */
	int		no_tau_csmcs;	/* number of CSMC-sets with diff. tau */
	Tau_CSMC	*Tau_CSMCs;	/* linear list of isomorphic CSMCs */

	int		branch_n, branch_k; /* info for branching topology */

	/* informations (kept to save time) for the isomorphism test */
	int		no_edges;	/* number of all CSMC-edges + add. edges */
	iso_edge	*edges;		/* ordered array of edges */
	double		sum_of_Qi;	/* sum of all Qi - entries for tau = 1 */
	graph		*graph;		/* abstract graph for isom. tests for tau = 1 */
	int		*sort_degrees;	/* initial partition (sorted degrees) */
	int		*degrees;	/* degrees of nodes */	

	struct Iso_Class *next;		/* linear list */
} Iso_Class;

/* type for .RATES file */
typedef struct rates_type
{	byte	t_nr;
	double	rate;
}  rates_type;

/* external definitions */

	/* from graph_main.c */
extern void 		*emalloc();	
extern tang_marking 	**det_markings, *exp_markings;
extern int              last_class_id;
extern int              no_gen_trans;
extern int		no_det_trans, no_tang_markings, no_exp_markings;
extern int		*no_nodes, no_places;
extern double		exp_transition_rate();
extern double		det_transition_delay();
extern FILE		*rates_file;
extern FILE		*rg_file;
extern t_sub		timed;
extern char		*net_name;

	/* from graph_timed.c */
extern void		set();		
extern double		det_transition_delay();

	/* from csmc_main.c */
extern	int		no_iso_classes;
extern	int		no_normal_classes;
extern	int		no_csmcs;
extern	Iso_Class	*normal_classes;
extern	Iso_Class	*Iso_Classes;
extern	SMC		**SMCs;
extern	void		write_rg_and_rates_file();
extern void		send_data_to_rand();

	/* from csmc_csmcs.c */
extern void		part_smc_into_csmcs();

	/* from csmc_spec.c */
extern void		test_csmc_topology();
extern void		number_csmc_topology();
extern void		solve_csmc_topology();

	/* from csmc_graph.c */
/* option bits */
#define	NODE_BITS	0x01
#define	SUCCESSORS	0x02
#define	PREDECESSORS	0x04
#define	RATES		0x08
#define	ORIG_MARKS	0x10
#define	VERTEX_BITS	0x20
#define DET_TRANS	0x40

extern	graph		*new_graph();
extern	void		delete_graph();
extern	void		show_graph();
extern	node 		*new_node();
extern	node		*cut_node();
extern	void		paste_node();
extern	void		show_node();
extern	void		add_edge();
extern	void		show_edge();
extern	void		top_sort_acyclic_graph();
