/*****************************************************************************

	Programm	:	graph

	Komponente	:	graph_subnet.c

	Aufgabe		:	Unterprogramme, die die 
				Erreichbarkeitsgraphen der
				Subnetze berechnen

	Programmierer	:	Armin Zimmermann

	Datum		:	Juni 1992

******************************************************************************/

#include <stdio.h>
#include <memory.h>
#include "portab.h"


#include "graph_types.h"			/* definitions and types */
#include "graph_extern.h"

extern char	*emalloc();			/* from graph_main */
extern void	efree();
extern int	*current_mark;			/* from graph_timed.c */
extern double	md_rate();			/* from <model>_MDF.c */

/*	local_to_global_mark obtains the (partial) global marking from a local one */
void
local_to_global_mark(sub, l_mark, g_mark)
	subnet	*sub;		/* corresponding subnet */
	int	*l_mark;	/* local marking */
	int	*g_mark;	/* resulting global marking, space must be provided */
{
	int	pl;

	for (pl = 0; pl < no_places; pl++)
		g_mark[pl] = -1;			/* mark as unknown */
	for (pl = 0; pl < sub->no_places; pl++)
		g_mark[sub->places[pl].number] = l_mark[pl];	/* copy subnet places */
}

/*	imm_transition_weight returns the (possibly marking dependent) 
	firing weight of the immediate transition			*/
int	*weight_mark = NULL;

double
imm_transition_weight(sub, l_mark, trans)
	subnet	*sub;		/* corresponding subnet */
	int	*l_mark;	/* local marking */
	int	trans;	/* local transition number */
{
	double	ret;

	if (!sub->transitions[trans].wgt_id)
		return sub->transitions[trans].prob;	/* independent transition */
	if (weight_mark == NULL)			/* get memory space */
		weight_mark = (int *)emalloc(no_places * sizeof(int));
	local_to_global_mark(sub, l_mark, weight_mark);
	current_mark = weight_mark;			/* parameter for act_marking() */
	ret = md_rate(sub->transitions[trans].wgt_id);	/* marking dependent rate */
	if (ret > 0.0)
		return ret;
	printf("Error: marking dependent firing weight of transition");
	printf(" %s", sub->transitions[trans].name);
	printf(" evaluates to 0\n");
	exit(1);
}

/*	is_in_packed tests a transition for being a member of a packed set */
int
is_in_packed(tra, set)
	int	tra;
	byte	*set;
{
	int	i;

	for(i = 1; i <= set[0]; i++)
		if (set[i] == tra)
			return 1;
	return 0;
}

/*	bitmasks for marking coding / decoding */
word	bit_mask[17] = { 0x000,
	  0x0001, 0x0003, 0x0007, 0x000F, 0x001F, 0x003F, 0x007F, 0x00FF,
	  0x01FF, 0x03FF, 0x07FF, 0x0FFF, 0x1FFF, 0x3FFF, 0x7FFF, 0xFFFF };

/*	code_submarking codes a submarking into a packed format using only
	as many bits as needed and returns a pointer to the resulting memory area */
word *
code_submarking(sub_mem, sub_no_pla, sub_pla, mark)
	int	sub_mem;	/* memory needed to store one marking */
	int	sub_no_pla;	/* number of places in the subnet */	
	place	*sub_pla;	/* places of the subnet */
	int	*mark;		/* an array with the numbers of tokens in places */
{
	word	*pack;
	int	i, bits, words, count = 0;

	pack = (word *)emalloc(sub_mem * sizeof(word), 15);
	for(i = 0; i < sub_mem; i++)
		pack[i] = 0;			/* clear */
	for(i = 0; i < sub_no_pla; i++)
	{	bits = sub_pla[i].bits;
		if (mark[i] > sub_pla[i].bound)
		{	printf("ERROR: number of tokens in %s exceeded upper bound %d !\n", 
				sub_pla[i].name, sub_pla[i].bound);
			exit(1);
		}
		words = count / 16;
		pack[words] |= mark[i] << (count % 16);
		if ((count+bits) / 16 != words)
			pack[words+1] |= mark[i] >> (16 - count % 16);
		count += bits;
	}
	return pack;
}

/*	decode_submarking decodes a submarking from packed format and returns
	a pointer to the resulting array with number of tokens in places */
void
decode_submarking(sub_no_pla, sub_pla, code, to)
	int	sub_no_pla;	/* number of places in the subnet */	
	place	*sub_pla;	/* places of the subnet */
	word	*code;		/* memory space with coded marking */
	int	*to;		/* result place */
{
	int	i, bits, words, count = 0;

	for(i = 0; i < sub_no_pla; i++)
	{	bits = sub_pla[i].bits;
		words = count / 16;
		to[i] = (code[words] >> (count % 16)) & bit_mask[bits];
		if ((count+bits) / 16 != words)
			to[i] |= (bit_mask[bits-16+count%16] & code[words+1])
					<< (16 - count % 16);
		count += bits;
	}
}

/*	enter_disconnected enters a disconnected places marking into the tree
	and returns a pointer to the memory area, where the marking is stored	*/
word *
enter_disconnected(sub, g_token)
	t_sub	*sub;		/* subnet containing disconnected places */
	int	*g_token;	/* number of tokens in places (global marking) */
{
	sub_node   *node = sub->tree, **insert;
	sp_pointer *lookup;
	int	pla_nr = 0, i, left, right, mid, tokens, old_bound;
	int	*token;

	if (!sub->no_places)		/* no disconnected places */
		return NULL;
	token = (int *)emalloc(sub->no_places * sizeof(int), 17);
	for (i = 0; i < sub->no_places; i++)		/* local numbers */
		token[i] = g_token[sub->places[i].number];
#ifdef DEBUG
	printf("Enter new disconnected marking:\n");
	for (i=0; i<sub->no_places; i++)
		printf("%s:%d  ", sub->places[i].name, token[i]);
	printf("\n");
#endif
	if (node == NULL)		/* tree is empty */
	{	node = sub->tree = (sub_node *)emalloc(sizeof(sub_node), 18);
		if (sub->places[0].bound > ARRAY_LIMIT)
		{	node->bound = SPARSE_START;	/* sparse array */
			node->number = 1;
			node->next = (sub_node **)emalloc(SPARSE_START * sizeof(sp_pointer), 19);
			lookup = (sp_pointer *)node->next;
			lookup[0].index = token[0];
			insert = (sub_node **)&(lookup[0].next);
			node = NULL;		/* skip search process */
		} else
		{	node->bound = 0;		/* array */
			node->next = (sub_node **)emalloc((sub->places[0].bound + 1) *
								sizeof(sub_node *), 20);
			for(i = 0; i <= sub->places[0].bound; i++)
				node->next[i] = NULL;	/* clear */
		}
	}
	while((node != NULL) && (pla_nr < sub->no_places))
	{	tokens = token[pla_nr];
		if (tokens > sub->places[pla_nr].bound)
		{	printf("ERROR: number of tokens in %s(%d) exceeded upper bound %d !\n", 
				sub->places[pla_nr].name, tokens, sub->places[pla_nr].bound);
			exit(1);
		}
		if (node->bound)
		{	lookup = (sp_pointer *)node->next;	/* sparse array */
			left = 0; right = node->number - 1;
			mid = (left + right) / 2;
			if (tokens <= lookup[left].index)
				mid = left;
			if (tokens >= lookup[right].index)
				mid = left = right;
			while((tokens != lookup[mid].index) && (mid != left))
			{	if (tokens > lookup[mid].index)
					left = mid;
				else	right = mid;
				mid = (left + right) / 2;
			}
			if (tokens == lookup[mid].index)	/* found */
				node = (sub_node *)lookup[mid].next;
			else
			{	mid += (tokens > lookup[mid].index);
				if (++(node->number) > node->bound)
				{	old_bound = node->bound;
					node->bound = SPARSE_NEXT(node->bound);
					if (node->bound <= old_bound)
						node->bound = SPARSE_MAX;
					node->next = (sub_node **)emalloc(node->bound
								 * sizeof(sp_pointer), 21);
					memmove(node->next,lookup,  
							old_bound * sizeof(sp_pointer));
					efree((char *)(lookup));
					lookup = (sp_pointer *)node->next;
				}
				memmove( &(lookup[mid+1]),&(lookup[mid]),
					sizeof(sp_pointer) * (node->number - mid - 1));
				lookup[mid].index = tokens;
				insert = (sub_node **)&(lookup[mid].next);
				break;
			}
		} else
		{	if (node->next[tokens] == NULL)	/* array */
			{	insert = (sub_node **)&(node->next[tokens]);
				break;
			}
			node = node->next[tokens];
		}
		pla_nr++;
	}
	if (pla_nr >= sub->no_places)			/* marking already exists */
	{	efree((char *)(token));
		return (word *)node;
	}
	/* insert new nodes into tree */
	for(pla_nr++ ; pla_nr < sub->no_places; pla_nr++)
	{	tokens = token[pla_nr];
		if (tokens > sub->places[pla_nr].bound)
		{	printf("ERROR: number of tokens in %s(%d) exceeded upper bound %d !\n", 
				sub->places[pla_nr].name, tokens, sub->places[pla_nr].bound);
			exit(1);
		}
		*insert = node = (sub_node *)emalloc(sizeof(sub_node), 22);
		if (sub->places[pla_nr].bound > ARRAY_LIMIT)
		{	node->bound = SPARSE_START;	/* sparse array */
			node->number = 1;
			node->next = (sub_node **)emalloc(SPARSE_START * sizeof(sp_pointer), 23);
			lookup = (sp_pointer *)node->next;
			lookup->index = tokens;
			insert = (sub_node **)&(lookup->next);
		} else
		{	node->bound = 0;		/* array */
			node->next = (sub_node **)emalloc((sub->places[pla_nr].bound + 1) *
							sizeof(sub_node *), 24);
			for(i = 0; i <= sub->places[pla_nr].bound; i++)
				node->next[i] = NULL;	/* clear */
			insert = (sub_node **)&(node->next[tokens]);
		}
	}
	*insert = (sub_node *)code_submarking(sub->memory, sub->no_places, sub->places, token);
	efree((char *)(token));
	sub->no_markings++;
	return (word *)*insert;
}

/*	enter_sub_marking enters a subnet-marking into the reachability tree,
	tests for immediate loops and returns a pointer to the marking		*/
sub_marking *
enter_submarking(sub, token)
	subnet	*sub;		/* corresponding subnet */
	int	*token;		/* number of tokens in places */
{
	sub_marking *marking;
	mem_sub_marking *mem_marking;
	sub_node   *node = sub->tree, **insert;
	sp_pointer *lookup;
	int	pla_nr = 0, i, left, right, mid, tokens, old_bound;

#ifdef DEBUG
	printf("Enter new submarking:\n");
	for (i=0; i<sub->no_places; i++)
		printf("%s:%d  ", sub->places[i].name, token[i]);
	printf("\n");
#endif
	if (node == NULL)		/* tree is empty */
	{	node = sub->tree = (sub_node *)emalloc(sizeof(sub_node), 25);
		if (sub->places[0].bound > ARRAY_LIMIT)
		{	node->bound = SPARSE_START;	/* sparse array */
			node->number = 1;
			node->next = (sub_node **)emalloc(SPARSE_START * sizeof(sp_pointer),
26);
			lookup = (sp_pointer *)node->next;
			lookup[0].index = token[0];
			insert = (sub_node **)&(lookup[0].next);
			node = NULL;		/* skip search process */
		} else
		{	node->bound = 0;		/* array */
			node->number = sub->places[0].bound;
			node->next = (sub_node **)emalloc((sub->places[0].bound + 1) *
								sizeof(sub_node *), 27);
			for(i = 0; i <= sub->places[0].bound; i++)
				node->next[i] = NULL;	/* clear */
		}
	}
	while((node != NULL) && (pla_nr < sub->no_places))
	{	tokens = token[pla_nr];
		if (tokens > sub->places[pla_nr].bound)
		{	printf("ERROR: number of tokens in %s(%d) exceeded upper bound %d !\n", 
				sub->places[pla_nr].name, tokens, sub->places[pla_nr].bound);
			exit(1);
		}
		if (node->bound)
		{	lookup = (sp_pointer *)node->next;	/* sparse array */
			left = 0; right = node->number - 1;
			mid = (left + right) / 2;
			if (tokens <= lookup[left].index)
				mid = left;
			if (tokens >= lookup[right].index)
				mid = left = right;
			while((tokens != lookup[mid].index) && (mid != left))
			{	if (tokens > lookup[mid].index)
					left = mid;
				else	right = mid;
				mid = (left + right) / 2;
			}
			if (tokens == lookup[mid].index)	/* found */
				node = (sub_node *)lookup[mid].next;
			else
			{	mid += (tokens > lookup[mid].index);
				if (++(node->number) > node->bound)
				{	old_bound = node->bound;
					node->bound = SPARSE_NEXT(node->bound);
					if (node->bound <= old_bound)
						node->bound = SPARSE_MAX;
					node->next = (sub_node **)emalloc(node->bound
								 * sizeof(sp_pointer), 28);
					memmove(node->next,lookup, 
							old_bound * sizeof(sp_pointer));
					efree((char *)(lookup));
					lookup = (sp_pointer *)node->next;
				}
				memmove(&(lookup[mid+1]),&(lookup[mid]),
					sizeof(sp_pointer) * (node->number - mid - 1));
				lookup[mid].index = tokens;
				insert = (sub_node **)&(lookup[mid].next);
				break;
			}
		} else
		{	if (node->next[tokens] == NULL)	/* array */
			{	insert = (sub_node **)&(node->next[tokens]);
				break;
			}
			node = node->next[tokens];
		}
		pla_nr++;
	}
	if (pla_nr >= sub->no_places)			/* marking already exists */
	{	marking = (sub_marking *)node;
		if (mem_bit || marking->paths != NULL)
			return marking;
		printf("ERROR: immediate loop detected, marking:\n");
		for(i=0; i<sub->no_places; i++)
			printf("%s:%d  ", sub->places[i].name, token[i]);
		printf("\n");
		exit(1);
	}
	/* insert new nodes into tree */
{	int i;
	for (i=0; i<no_subnets; i++)
		if (&(subnets[i]) == sub)
			break;
	no_van_submarks[i]++;
}
	for(pla_nr++ ; pla_nr < sub->no_places; pla_nr++)
	{	tokens = token[pla_nr];
		if (tokens > sub->places[pla_nr].bound)
		{	printf("ERROR: number of tokens in %s(%d) exceeded upper bound %d !\n", 
				sub->places[pla_nr].name, tokens, sub->places[pla_nr].bound);
			exit(1);
		}
		*insert = node = (sub_node *)emalloc(sizeof(sub_node), 29);
		if (sub->places[pla_nr].bound > ARRAY_LIMIT)
		{	node->bound = SPARSE_START;	/* sparse array */
			node->number = 1;
			node->next = (sub_node **)emalloc(SPARSE_START * sizeof(sp_pointer), 30);
			lookup = (sp_pointer *)node->next;
			lookup->index = tokens;
			insert = (sub_node **)&(lookup->next);
		} else
		{	node->bound = 0;		/* array */
			node->number = sub->places[pla_nr].bound;
			node->next = (sub_node **)emalloc((sub->places[pla_nr].bound + 1) *
							sizeof(sub_node *), 31);
			for(i = 0; i <= sub->places[pla_nr].bound; i++)
				node->next[i] = NULL;	/* clear */
			insert = (sub_node **)&(node->next[tokens]);
		}
	}
	sub->no_markings++;
	if (mem_bit)
	{	mem_marking = (mem_sub_marking *)emalloc(sizeof(mem_sub_marking), 80);
		*insert = (sub_node *)mem_marking;
		mem_marking->prob = 0.0;
		mem_marking->mark = code_submarking(sub->memory, 
				sub->no_places, sub->places, token);
		mem_marking->forward = NULL;
		mem_marking->backward = NULL;
		mem_marking->no_pred = 0;
		return (sub_marking *)mem_marking;
	}
	marking = (sub_marking *)emalloc(sizeof(sub_marking), 32);
	*insert = (sub_node *)marking;
	marking->no_paths = 0;
	marking->paths = NULL;
	return marking;
}

/*	release_tree frees the entire submarking tree of a subnet (save memory) */
void
release_tree(tree, no_pla)
	sub_node	*tree;
	int		no_pla;
{
	int		i;
	fire_path	*f_walk;
	sp_pointer	*sp_walk;
	mem_sub_marking	*m_sub_m;

	if (tree == NULL)
		return;		/* nothing to do */
	if (!no_pla)		/* bottom of tree */
	{	m_sub_m = (mem_sub_marking *)tree;
		efree(m_sub_m->mark);
		efree(m_sub_m);
		return;
	}
	sp_walk = (sp_pointer *)tree->next;
	for (i=0; i<tree->number; i++)
		if (tree->bound)	/* sparse */
			release_tree(sp_walk[i].next, no_pla-1);
		else	release_tree(tree->next[i], no_pla-1);
	efree(tree->next);
	efree(tree);
}

/*	test_global_concession tests if a global transition is enabled (without priority);
	the variable current_mark (graph_timed.c) must be set to the current global marking */
int
test_global_conc(mark, i_mat, h_mat)
	int	 *mark;		/* local marking */
	sp_place *i_mat;	/* input matrix row */
	sp_place *h_mat;	/* inhibit matrix row */
{
	int	pla, value;

	for(pla = 1; pla <= i_mat[0].num; pla++)
	{	value = i_mat[pla].plu_id ? md_arc_mult(i_mat[pla].plu_id) : i_mat[pla].value;
		if (value && (mark[i_mat[pla].num] < value))
			return 0;		/* not enough input token */
	}
	for(pla = 1; pla <= h_mat[0].num; pla++)
	{	value = h_mat[pla].plu_id ? md_arc_mult(h_mat[pla].plu_id) : h_mat[pla].value;
		if (value && (mark[h_mat[pla].num] >= value))
			return 0;		/* too many inhibitor token */
	}
	return 1;
}

/*	test_concession tests if a subnet - transition is enabled (without priority) */
int	*gh_mark = NULL;	/* space for global marking */
int
test_concession(sub, mark, trans)
	subnet	*sub;		/* corresponding subnet */
	int	*mark;		/* marking */
	int	trans;		/* transition number */
{
	int	pla, value;
	imm_t	*tra = &(sub->transitions[trans]);
	sp_place *i_mat = tra->i_matrix;
	sp_place *h_mat = tra->h_matrix;

	if (gh_mark == NULL)
		gh_mark = (int *)emalloc(no_places * sizeof(int));
	local_to_global_mark(sub, mark, gh_mark);
	current_mark = gh_mark;		/* for marking dependencies */
	if (tra->ena_id)		/* marking dep. enabling function */
		if (!md_ena_function(tra->ena_id))
			return 0;
	return test_global_conc(mark, i_mat, h_mat);
}

/*	fire_global_trans fires a global transition; the variable current_mark (graph_timed.c)
	must be set to the current global marking */
void
fire_global_trans(mark, mark_next, num, c_mat)
	int	*mark;		/* marking */
	int	*mark_next;	/* new marking */
	int	num;		/* number of places in this (sub-)net */
	sp_place *c_mat;	/* C - matrix of the transition */
{
	int	pla, value;

	memmove(mark_next, mark, num * sizeof(int));	/* copy old marking */
	for(pla = 1; pla <= c_mat[0].num; pla++)
	{	value = c_mat[pla].value;		/* 0 if marking dependent */
		if (c_mat[pla].plu_id)
			value += md_arc_mult(c_mat[pla].plu_id);
		if (c_mat[pla].min_id)
			value -= md_arc_mult(c_mat[pla].min_id);
		mark_next[c_mat[pla].num] += value;
	}
}

/*	fire_transition fires a local transition */
void
fire_transition(sub, mark, mark_next, num, c_mat)
	subnet	*sub;		/* corresponding subnet */
	int	*mark;		/* local marking */
	int	*mark_next;	/* new marking */
	int	num;		/* number of places in this (sub-)net */
	sp_place *c_mat;	/* C - matrix of the transition */
{

	if (gh_mark == NULL)
		gh_mark = (int *)emalloc(no_places * sizeof(int));
	local_to_global_mark(sub, mark, gh_mark);
	current_mark = gh_mark;
	fire_global_trans(mark, mark_next, num, c_mat);
}

/*	mix_paths mixes the firing paths of two transitions */
void
mix_paths(mix_mark, new_mark, factor)
	sub_marking	*mix_mark;	/* set of old paths */
	sub_marking	*new_mark;	/* new paths */
	double		factor;		/* probability factor of new paths */
{
	int		old, new, enter;
	sub_marking	*old_mark;

	old_mark = (sub_marking *)emalloc(sizeof(sub_marking), 33);
	memmove(old_mark,mix_mark, sizeof(sub_marking));	/* copy data */
	mix_mark->paths = (fire_path *)emalloc((old_mark->no_paths + new_mark->no_paths)
							* sizeof(fire_path), 34);
	old = new = enter = 0;
	while((old < old_mark->no_paths) || (new < new_mark->no_paths))
	{	if ((old < old_mark->no_paths) && (new < new_mark->no_paths) &&
		    (old_mark->paths[old].marking == new_mark->paths[new].marking))
		{	mix_mark->paths[enter].marking = old_mark->paths[old].marking;
			mix_mark->paths[enter++].prob = old_mark->paths[old++].prob +
						new_mark->paths[new++].prob * factor;
			continue;
		}
		if ((old >= old_mark->no_paths) || (new < new_mark->no_paths) &&
		    (old_mark->paths[old].marking > new_mark->paths[new].marking))
		{	mix_mark->paths[enter].marking = new_mark->paths[new].marking;
			mix_mark->paths[enter++].prob = new_mark->paths[new++].prob * factor;
		} else
		{	mix_mark->paths[enter].marking = old_mark->paths[old].marking;
			mix_mark->paths[enter++].prob = old_mark->paths[old++].prob;
		}
	}
	mix_mark->no_paths = enter;
	efree((char *)(old_mark->paths));
	efree((char *)(old_mark));
}

/*	sub_end_markings calculates the set of submarkings reachable from one initial
	subnet - marking and returns a pointer to the sub_marking entry in the subnet-tree
	(containing the set of firing paths with their total probabilities) */
sub_marking *
sub_end_markings(sub, mark, enabled)
	subnet	*sub;		/* the corresponding subnet */
	int	*mark;		/* initial submarking */
	byte	*enabled;	/* set of transitions having concession in this marking */
{
	byte		*enas = enabled, *enas_next, *fire = NULL;
	sub_marking	*marking, *fire_marking, mix_marking;
	int		tra, flag = 0, ecs, i, no_fire = 0, each;
	int		*marking_next;
	double		prob_sum = 0.0, now_prob;

	marking = enter_submarking(sub, mark);		/* look for marking */
	if (marking->paths != NULL)			/* old marking */
		return marking;
	if (enas == NULL)				/* first recursion level */
	{	enas = (byte *)emalloc(sub->no_trans * sizeof(byte), 35);
		for(tra = 0; tra < sub->no_trans; tra++)
			enas[tra] = test_concession(sub, mark, tra);
	}
	for(ecs = 0; (ecs < sub->no_ecs) && !flag; ecs++)
	{	for(i = 1; i <= sub->ecs[ecs][0]; i++)
		{	tra = sub->ecs[ecs][i];		/* transition number */
			if (enas[tra])			/* transition is enabled */
			{	if (!flag)
					fire = (byte *)emalloc(sub->ecs[ecs][0] * sizeof(byte), 36);
				fire[no_fire++] = tra;
				flag = 1;
			}	
		}
	}
	if (!flag)					/* final submarking */
	{	marking->no_paths = 1;
		marking->paths = (fire_path *)emalloc(sizeof(fire_path), 37);
		marking->paths[0].prob = 1.0;		/* probability 1 */
		marking->paths[0].marking = code_submarking(sub->memory,
							sub->no_places, sub->places, mark);
		return marking;
	}
	marking_next = (int  *)emalloc(sub->no_places * sizeof(int),  38);
	enas_next    = (byte *)emalloc(sub->no_trans  * sizeof(byte), 39);
	mix_marking.no_paths = 0;
	mix_marking.paths = NULL;
	for(i = 0; i < no_fire; i++)
	{	tra = fire[i];
		now_prob = imm_transition_weight(sub, mark, tra);
		prob_sum += now_prob;		/* sum up probabilities */
		fire_transition(sub, mark, marking_next, sub->no_places, 
						sub->transitions[tra].c_matrix);
		/* calculate new enas */
		memmove(enas_next,enas, sub->no_trans * sizeof(byte));	/* copy data */
		for(each = 0; each < sub->no_trans; each++)		/* get new enas */
		{	flag = 0;
			if (each == tra)
				flag = 1;
			else	if (enas[each])
				{	if (is_in_packed(each, sub->transitions[tra].diss))
						flag = 1;
				} else
					if (is_in_packed(each, sub->transitions[tra].enas))
						flag = 1;
			if (flag)
			enas_next[each] = test_concession(sub, marking_next, each);
		}
		fire_marking = sub_end_markings(sub, marking_next, enas_next);
		mix_paths(&mix_marking, fire_marking, now_prob);
	}
	efree((char *)(marking_next));
	efree((char *)(enas_next));
	efree((char *)(fire));
	if (enabled == NULL)
		efree((char *)(enas));
	marking->no_paths = mix_marking.no_paths;
	marking->paths = (fire_path *)emalloc(marking->no_paths * sizeof(fire_path), 40);
	memmove(marking->paths, mix_marking.paths, marking->no_paths * sizeof(fire_path));
	for(i = 0; i < marking->no_paths; i++)
		marking->paths[i].prob /= prob_sum;		/* normalize probabilities */
	efree((char *)(mix_marking.paths));
	return marking;
}

/*	list_move moves one entry of a twice concatenated list into another */
void
list_move(pointer, s_list, d_list)
	mem_sub_marking	*pointer;	/* entry to move */
	mem_sub_marking	**s_list;		/* source list, may be NULL */
	mem_sub_marking	**d_list;		/* destination list, may be NULL */
{

	if (s_list != NULL)
	{	/* take the entry out of the source list */
		if (pointer->backward == NULL)		/* 1st entry */
			*s_list = pointer->forward;
		else	pointer->backward->forward = pointer->forward;
		if (pointer->forward != NULL)		/* not the last entry */
			pointer->forward->backward = pointer->backward;
	}
	if (d_list == NULL)
	{	pointer->forward = pointer->backward = NULL;
		return;
	}
	/* put the entry in the destination list */
	pointer->backward = NULL;
	pointer->forward = *d_list;
	*d_list = pointer;
	if (pointer->forward != NULL)		/* not the last entry */
		pointer->forward->backward = pointer;
}

/*	mem_sub_end_markings does the same as sub_end_markings 
	with less memory space in a slower, straightforward way	 */
static int 	*m_next = NULL;
static int	*m_now = NULL;
static byte 	*fire;
sub_marking *
mem_sub_end_markings(sub, mark)
	subnet	*sub;		/* the corresponding subnet */
	int	*mark;		/* initial submarking */
{
	sub_marking 	*return_val;
	mem_sub_marking	*now, *next, *first;
	imm_t		*sn_tr	= sub->transitions;
	fire_path	*f_walk;
	end_mind	*e_walk;
	double		factor;
	int		sn_no_pl = sub->no_places;
	int		sn_no_tr = sub->no_trans;
	int		no_end_marks = 0, no_marks = 1;
	int		ecs, tra, i, no_fire;

	if (m_next == NULL)
	{	m_next = (int *)emalloc(no_places * sizeof(int), 81);
		m_now = (int *)emalloc(no_places * sizeof(int), 86);
		fire = (byte *)emalloc(no_trans * sizeof(byte), 82);
	}
	sub->list1 = sub->list2 = sub->e_list = NULL;
	sub->tree = NULL;
	now = (mem_sub_marking *)enter_submarking(sub, mark);	/* enter 1st marking */
	first = now;
	list_move(now, NULL, &(sub->list1));
	/* generate the set of vanishing submarkings */
	while ((now = sub->list1) != NULL)
	{	decode_submarking(sn_no_pl, sub->places, now->mark, m_now);
		list_move(now, &(sub->list1), &(sub->list2));
		/* calculate enabled transitions (one ecs only) */
		no_fire = 0;
		for(ecs = 0; (ecs < sub->no_ecs) && !no_fire; ecs++)
		{	for(i = 1; i <= sub->ecs[ecs][0]; i++)
			{	tra = sub->ecs[ecs][i];		/* transition number */
				if (test_concession(sub, m_now, tra))
					fire[no_fire++] = tra;
			}
		}
		if (!no_fire)				/* final submarking */
			continue;			/* go on */
		for (i=0; i<no_fire; i++)
		{	tra = fire[i];
			fire_transition(sub, m_now, m_next, sub->no_places, sn_tr[tra].c_matrix);
			next = (mem_sub_marking *)enter_submarking(sub, m_next);
			if (next->no_pred++)
				continue;
			list_move(next, NULL, &(sub->list1));
		}
	}
	/* ordered scan of vanishing submarkings */
	list_move(first, &(sub->list2), &(sub->list1));
	first->prob = 1.0;
	while ((now = sub->list1) != NULL)
	{	decode_submarking(sn_no_pl, sub->places, now->mark, m_now);
		no_fire = 0; factor = 0.0;
		for(ecs = 0; (ecs < sub->no_ecs) && !no_fire; ecs++)
		{	for(i = 1; i <= sub->ecs[ecs][0]; i++)
			{	tra = sub->ecs[ecs][i];		/* transition number */
				if (test_concession(sub, m_now, tra))
				{	fire[no_fire++] = tra;
					factor += imm_transition_weight(sub, m_now, tra);
				}	
			}
		}
		if (!no_fire)				/* final submarking */
		{	list_move(now, &(sub->list1), &(sub->e_list));
			no_end_marks++;
			continue;			/* go on */
		}
		factor = now->prob / factor;		/* mark-prob / sum of all en. trans. */
		for (i=0; i<no_fire; i++)
		{	tra = fire[i];
			fire_transition(sub, m_now, m_next, sub->no_places, sn_tr[tra].c_matrix);
			next = (mem_sub_marking *)enter_submarking(sub, m_next);
			if (!--(next->no_pred))
				list_move(next, &(sub->list2), &(sub->list1));
			next->prob += imm_transition_weight(sub, m_now, tra) * factor;
		} /* fire transitions */
		list_move(now, &(sub->list1), NULL); 	/* delete */
	}
	if (sub->list2 != NULL)
	{	printf("ERROR: immediate loop detected, marking:\n");
		for(i=0; i<sub->no_places; i++)
			printf("%s:%d  ", sub->places[i].name, m_next[i]);
		printf("\n");
		exit(1);
	}
	/* prepare return values */
	return_val = (sub_marking *)emalloc(sizeof(sub_marking), 83);
	return_val->no_paths = no_end_marks;
	return_val->paths = (fire_path *)emalloc(no_end_marks * sizeof(fire_path), 84);
	now = sub->e_list;		/* set of end markings */
	for (i = 0; (i < no_end_marks) && (now != NULL); i++)
	{	f_walk = &(return_val->paths[i]);
		f_walk->prob = now->prob;
		e_walk = sub->end_mind;
		while ((e_walk != NULL) && memcmp(e_walk->mark, now->mark, 
						sub->memory * sizeof(word)))
			e_walk = e_walk->next;
		if (e_walk == NULL)	/* not found */	
		{	e_walk = (end_mind *)emalloc(sizeof(end_mind), 87);
			e_walk->mark = (word *)emalloc(sub->memory * sizeof(word), 88);
			memmove(e_walk->mark, now->mark, sub->memory * sizeof(word));
			e_walk->next = sub->end_mind;
			sub->end_mind = e_walk;
		}
		f_walk->marking = e_walk->mark;	/* copy results */
		now = now->forward;
	}
	if ((i < no_end_marks) || (now != NULL))
	{	printf("mem_sub_markings: internal error 2\n");
		exit(1);		/* unequal number of end markings */
	}
	/* clean up */
	release_tree(sub->tree, sub->no_places);
	return return_val;
}

/*	net_end_markings calculates the set of subnet - final markings reachable from
	one marking and returns a pointer to an array which contains a pointer to another
	array and a number of entries for each subnet. Each of this arrays describes
	the set of reachable markings of the corresponding subnet and the probability 
	for this firing path. */
sub_marking *
net_end_markings(subs, tsub, marking, disc_result)
	subnet	*subs;		/* the net's subnets */
	t_sub	*tsub;		/* the "timed transitions" subnet */
	int	*marking;	/* the marking (NULL -> take M0) */
	word	**disc_result;	/* the disconnected places resulting marking (return par.) */
{
	subnet	*sub;
	int	net, *sub_mark, i;
	sub_marking *result, *inter;

	result = (sub_marking *)emalloc(no_subnets * sizeof(sub_marking), 41);
	for (net = 0; net < no_subnets; net++)
	{	sub = &(subs[net]);
		sub_mark = (int *)emalloc(sub->no_places * sizeof(int), 42);
		if (marking == NULL)
		{	for(i = 0; i < sub->no_places; i++)
				sub_mark[i] = sub->places[i].m0;
		} else
		{	for(i = 0; i < sub->no_places; i++)
				sub_mark[i] = marking[sub->places[i].number];
		}
		if (mem_bit)		/* save memory option */
			inter = mem_sub_end_markings(sub, sub_mark);
		else	inter = sub_end_markings(sub, sub_mark, NULL);
		memmove(&(result[net]),inter, sizeof(sub_marking));
		efree((char *)(sub_mark));
	}
	if ((sub_mark = marking) == NULL)
	{	sub_mark = (int *)emalloc(no_places * sizeof(int), 43);
		for(i = 0; i < tsub->no_places; i++)
			sub_mark[tsub->places[i].number] = tsub->places[i].m0;
	}
	*disc_result = enter_disconnected(tsub, sub_mark);
	if (marking == NULL)
		efree((char *)(sub_mark));
	return result;
}
