/*****************************************************************************

	Programm	:	graph

	Komponente	:	graph_timed.c

	Aufgabe		:	Unterprogramme, die den
				Erreichbarkeitsgraphen des
				DSPNs berechnen
				Abspeicherung der Ergebnisse

	Programmierer	:	Armin Zimmermann

	Datum		:	Juni 1992

        Aenderung       :       Juni 1995 (Patrick Drescher)

******************************************************************************/

#include <stdio.h>

#include "portab.h"

#include "graph_types.h"			/* definitions and types */
#include "graph_extern.h"

extern double md_rate();			/* from <net>_MDF.c; marking-dependent rates */
extern int    md_ena_function();		/* marking dependent enabling functions */
extern int    md_arc_mult();			/* marking dependent arc multiplicities */

extern char *emalloc();				/* from graph_main */
extern void efree();
extern int *decode_submarking();		/* from graph_subnet */
extern sub_marking *net_end_markings();
extern int test_global_conc();
extern void fire_global_trans();

int	*table;
int	*current_mark;

byte	bit_select[8] = {1, 2, 4, 8, 16, 32, 64, 128};

/*	bit tests, if a bit is set (in a byte) */
int
bit(test, nr)
	byte	test;
	int	nr;
{
	if (nr < 0 || nr > 7)
		return 0;
	return test & bit_select[nr];
}

/*	set sets a bit in a byte */
void
set(test, nr)
	byte	*test;
	int	nr;
{
	if (nr < 0 || nr > 7)
		return;
	*test = *test | bit_select[nr];
}

/*	global_marking obtains a global marking from a coded tangible marking	*/
static int *help_mark = NULL;
void
global_marking(sub_marks, disc_mark, marking)
	word	**sub_marks;	/* the submarkings for each subnet */
	word	*disc_mark;	/* the disconnected-places-marking */
	int	*marking;	/* marking; return parameter */
{
	int	i, s_nr;
	t_sub	*t_subnet = &timed;
	subnet 	*sub = subnets;

	if (help_mark == NULL)
		help_mark = (int *)emalloc(no_places * sizeof(int), 85);
	if (t_subnet->no_places)
	{	decode_submarking(t_subnet->no_places, t_subnet->places, disc_mark, help_mark);
		for(i=0; i<t_subnet->no_places; i++)
			marking[t_subnet->places[i].number] = help_mark[i];
	}
	for(s_nr=0; s_nr<no_subnets; s_nr++)
	{	decode_submarking(sub[s_nr].no_places, sub[s_nr].places,
							sub_marks[s_nr], help_mark);
		for(i=0; i<sub[s_nr].no_places; i++)
			marking[sub[s_nr].places[i].number] = help_mark[i];
	}
}

/*	print_global_marking shows a global marking (for error messages) */
void
print_global_marking(mark)
	int	*mark;		/* marking */
{
	int pla, tok, no, sb, bit=0;
	subnet	*sub;

	for(pla=0; pla<timed.no_places; pla++)
	{	no = timed.places[pla].number;
		tok = mark[no];
		if (tok)
		{	bit = 1;
			if (tok>1)
				printf("%d*", tok);
			printf("%s  ", timed.places[pla].name);
		}
	}
	for(sb = 0; sb < no_subnets; sb++)
	{	sub = &subnets[sb];
		for(pla=0; pla<sub->no_places; pla++)
		{	no = sub->places[pla].number;
			tok = mark[no];
			if (tok)
			{	bit = 1;
				if (tok>1)
					printf("%d*", tok);
				printf("%s  ", sub->places[pla].name);
			}
		}
	}
	if (!bit)
		printf("No Token in places.");
	printf("\n");
}

/*	place_mark returns the number of tokens for a place in a given marking
	(for use with marking-dependent firing rates)		*/
int
act_marking(place_no)
	int	place_no;
{
	int	ret;
	ret = current_mark[place_no];
	if (ret >= 0)
		return ret;
	printf("Marking dependency error\n");
	exit(1);
}

/*	det_transition_delay returns the (deterministic) transition firing delay */
double
det_transition_delay(mark)
	tang_marking	*mark;		/* the current marking */
{
	int	tra = mark->det_num;
	int	*marking;
	double	delay;

	marking = (int *)emalloc(no_places * sizeof(int), 50);
	global_marking(mark->mark, mark->disc, marking);
	if (!mark->det)
	{	printf("ERROR (in det_transition_delay): no det. transition is enabled in\n");
		print_global_marking(marking);
		exit(1);
	}
	if (!timed.transitions[tra].type)
	{	printf("ERROR (in det_transition_delay): transition %s is exponential.\n",
					timed.transitions[tra].name);
		exit(1);
	}
	if (timed.transitions[tra].infinite)	
	{	printf("ERROR (in det_transition_delay): transition %s is infinite-server.\n",
					timed.transitions[tra].name);
		exit(1);
	}
	if (!timed.transitions[tra].mark_ind)		/* normal type */
	{	efree(marking);
		return timed.transitions[tra].value;
	}
	current_mark = marking;				/* marking-dependent */
	delay = md_rate(timed.transitions[tra].mark_ind);
	if (delay <= 1.0e-8)
	{	printf("ERROR: marking-dependent firing rate of transition ");
		printf("%s evaluates to 0 in marking\n", timed.transitions[tra].name);
		print_global_marking(mark);
		exit(1);
	}
	efree(marking);
	return delay;
}

/*	exp_transition_rate returns the (exponential) transition firing rate */
double
exp_transition_rate(tsub, tra, mark)
	t_sub	*tsub;		/* timed transition's subnet */
	int	tra;		/* number of transition */
	int	*mark;		/* the current marking */
{
	int	i, pla, new, min, value, id;
	double	rate;

	if (tsub->transitions[tra].mark_ind)			/* marking-dependent firing rate */
	{	current_mark = mark;
		rate = md_rate(tsub->transitions[tra].mark_ind);
		if (rate > 1.0e-8)
			return rate;
		printf("ERROR: marking-dependent firing rate of transition ");
		printf("%s evaluates to 0 in marking\n", tsub->transitions[tra].name);
		print_global_marking(mark);
		exit(1);
	}
	if (!tsub->transitions[tra].infinite)	
		return tsub->transitions[tra].value;		/* normal type */
	min = mark[tsub->transitions[tra].i_matrix[1].num];
	for(i = 1; i <= tsub->transitions[tra].i_matrix[0].num; i++)
	{	pla = tsub->transitions[tra].i_matrix[i].num;
		id  =  tsub->transitions[tra].i_matrix[i].plu_id;
		value = id ? md_arc_mult(id) : tsub->transitions[tra].i_matrix[i].value;
		if (!value)
			continue;
		new = mark[pla] / value;
		min = min > new ? new : min;			/* look for minimum */
	}
	if (!min)
	{	printf("ERROR: firing rate of infinite-server transition ");
		printf("%s evaluates to 0 in marking\n", tsub->transitions[tra].name);
		print_global_marking(mark);
		exit(1);
	}
	return tsub->transitions[tra].value / (double)min;	/* infinite server */
}

/*	calc_prob calculates the probability of a path to a tangible marking */
word **
calc_prob(in, pro)
	sub_marking *in;
	double	*pro;
{
	word	**back;
	int	i;

	*pro = 1.0;
	back = (word **)emalloc(no_subnets * sizeof(word *), 51);
	for(i=0; i<no_subnets; i++)
	{	back[i] = in[i].paths[table[i]].marking;
		*pro *= in[i].paths[table[i]].prob;
	}
	return back;
}

/*	first_tang_marking returns the first tangible marking from a set of
	reachable ones together with the probability and initializes the
	process of walking through all markings	*/
word **
first_tang_marking(input, prob)
	sub_marking *input;		/* set of reachable markings for each subnet */
	double	*prob;			/* probability (return parameter) */
{
	int	i;

	for(i=0; i<no_subnets; i++)
		table[i] = 0;		/* initialize table of counters */
	return calc_prob(input, prob);
}

/*	next_tang_marking returns the next tangible marking from a set of
	reachable ones together with the probability */
word **
next_tang_marking(sub, input, prob)
	subnet	*sub;			/* the net's subnets */
	sub_marking *input;		/* set of reachable markings for each subnet */
	double	*prob;			/* probability (return parameter) */
{
	int	i;

	for(i=0; i<no_subnets; i++)
	{	if (++(table[i]) < input[i].no_paths)
			return calc_prob(input, prob);
		table[i] = 0;
	}
	*prob = 0.0;
	return NULL;			/* no more combinations */
}

/*	max_paths returns the maximum number of possible firing paths */
int
max_paths(input)
	sub_marking *input;
{
	int i, mult = 1;

	for(i=0; i<no_subnets; i++)
		mult *= input[i].no_paths;
	return mult;
}

/*	merge_paths adds a new firing path */
void
merge_paths(marking, index, rate)
	tang_marking *marking, *index;
	double rate;
{
	static int max_entry;
	int	left, right, mid, old;
	tang_path *lookup;

	if (index == marking)				/* dont care about self-loops */
	{	marking->sum_of_rates -= rate;		/* correct rate */
		return;
	}
	if (!marking->no_exp_paths)
	{	marking->no_exp_paths = 1;
		marking->exp_paths = (tang_path *)emalloc(SPARSE_START * sizeof(tang_path), 52);
		marking->exp_paths[0].prob = rate;
		marking->exp_paths[0].marking = (void *)index;
		max_entry = SPARSE_START;
		return;
	}
	lookup = marking->exp_paths;
	left = 0; right = marking->no_exp_paths - 1;
	mid = (left + right) / 2;
	if ((void *)index <= lookup[left].marking)
		mid = left;
	if ((void *)index >= lookup[right].marking)
		mid = left = right;
	while(((void *)index != lookup[mid].marking) && (mid != left))
	{	if ((void *)index > lookup[mid].marking)
			left = mid;
		else	right = mid;
		mid = (left + right) / 2;
	}
	if ((void *)index == lookup[mid].marking)		/* found */
		lookup[mid].prob += rate;
	else
	{	mid += ((void *)index > lookup[mid].marking);
		if (++(marking->no_exp_paths) > max_entry)
		{	old = max_entry;
			max_entry = SPARSE_NEXT(max_entry);
			marking->exp_paths = (tang_path *)emalloc(max_entry * sizeof(tang_path), 53);
			memmove(marking->exp_paths,lookup, old * sizeof(tang_path));
			efree((char *)(lookup));
			lookup = marking->exp_paths;
		}
		memmove(&(lookup[mid+1]),&(lookup[mid]),
			sizeof(tang_path) * (marking->no_exp_paths - mid - 1));
		lookup[mid].marking = index;
		lookup[mid].prob = rate;
	}
}

/*	enter_tang_marking enters a tangible marking into the reduced reachability tree
	and returns a pointer to the found or created marking		*/
tang_marking *
enter_tangmarking(sub_marks, disc_mark)
	word	**sub_marks;	/* pointer to subnet - markings (malloc'd!) */
	word	*disc_mark;	/* pointer to disconnected places marking */
{
	static int mark_num = 0;
	tang_marking *marking;
	tang_node   *node = r_graph, **insert;
	tsp_pointer *lookup;
	int	sn_nr = 0, i, left, right, mid, old_bound;
	word	*index;

#ifdef DEBUG
	{	extern subnet *subnets;
		extern t_sub timed;
		int	*marking;
		marking = (int *)emalloc(no_places * sizeof(int), 54);
		global_marking(sub_marks, disc_mark, marking);
		printf("\nEnter tangible marking:\n");
		print_global_marking(marking);
		efree((char *)(marking));
	}
#endif
	if (node == NULL)				/* tree is empty */
	{	node = r_graph = (tang_node *)emalloc(sizeof(tang_node), 55);
		if (disc_mark == NULL)			/* no disc. places */
			node->bound = 1;
		else	node->bound = SPARSE_START;	/* sparse array */
		node->number = 1;
		node->next = (tsp_pointer *)emalloc(node->bound * sizeof(tsp_pointer), 56);
		node->next[0].index = disc_mark;
		insert = (tang_node **)&(node->next[0].next);
		node = NULL;				/* don't go into while */
	}
	while((node != NULL) && (sn_nr < no_subnets+1))
	{	index = sn_nr ? sub_marks[sn_nr-1] : disc_mark;
		lookup = node->next;
		left = 0; right = node->number - 1;
		mid = (left + right) / 2;
		if ((void *)index <= (void *)lookup[left].index)
			mid = left;
		if ((void *)index >= (void *)lookup[right].index)
			mid = left = right;
		while(((void *)index != lookup[mid].index) && (mid != left))
		{	if ((void *)index > (void *)lookup[mid].index)
				left = mid;
			else	right = mid;
			mid = (left + right) / 2;
		}
		if ((void *)index == lookup[mid].index)		/* found */
			node = (tang_node *)lookup[mid].next;
		else
		{	mid += ((void *)index > (void *)lookup[mid].index);
			if (++(node->number) > node->bound)
			{	old_bound = node->bound;
				node->bound = SPARSE_NEXT(node->bound);
				if (node->bound <= old_bound)
					node->bound = SPARSE_MAX;
				node->next = (tsp_pointer *)emalloc(node->bound 
								* sizeof(tsp_pointer), 57);
				memmove(node->next,lookup, 
						old_bound * sizeof(tsp_pointer));
				efree((char *)(lookup));
				lookup = (tsp_pointer *)node->next;
			}
			memmove(&(lookup[mid+1]),&(lookup[mid]), 
				sizeof(tsp_pointer) * (node->number - mid - 1));
			lookup[mid].index = index;
			insert = (tang_node **)&(lookup[mid].next);
			break;
		}
		sn_nr++;
	}
	if (sn_nr > no_subnets)			/* marking already exists */
	{	efree((char *)(sub_marks));
		marking = (tang_marking *)node;
		return marking;
	}
	/* insert new nodes into tree */
	for(sn_nr++ ; sn_nr <= no_subnets; sn_nr++)
	{	index = sn_nr ? sub_marks[sn_nr-1] : disc_mark;
		*insert = node = (tang_node *)emalloc(sizeof(tang_node), 58);
		node->bound = SPARSE_START;	/* sparse array */
		node->number = 1;
		node->next = (tsp_pointer *)emalloc(SPARSE_START * sizeof(tsp_pointer), 59);
		lookup = node->next;
		lookup->index = index;
		insert = (tang_node **)&(lookup->next);
	}
	marking = (tang_marking *)emalloc(sizeof(tang_marking), 60);
	*insert = (tang_node *)marking;
	marking->number = mark_num++;
	marking->disc = disc_mark;
	marking->mark = sub_marks;
	marking->no_det_paths = 0;
	marking->no_exp_paths = 0;
	marking->det_paths = NULL;
	marking->exp_paths = NULL;
	marking->det = marking->det_num = 0;
	marking->sum_of_rates = 0.0;
	marking->next = new_markings;
	new_markings = marking;			/* new list entry */
	return marking;
}

/*	release_paths releases the space allocated by net_end_markings */
void
release_paths(e_mark)
	sub_marking	*e_mark;
{
	int	i;

	if (mem_bit)
		for (i=0; i<no_subnets; i++)
			efree(e_mark[i].paths);
	efree(e_mark);
}

/*	generate_rg calculates the reduced reachability graph */
void
generate_rg(subs, tsub)
	subnet	*subs;
	t_sub	*tsub;
{
	tang_marking	*marking, *to_marking;
	sub_marking 	*end_mark;
	word		*disc, **path;
	double		prob, rate;
	int		i, tra, fire, first, *i_mark, *n_mark, *h_mark;
	tang_path	*h_path;
	byte		*h_tnums, h_fire;

	h_tnums = (byte *)emalloc(no_places * sizeof(byte), 61);
	table = (int *)emalloc(no_subnets * sizeof(int), 62);
	no_absorb_states = 0;
	i_mark = (int *)emalloc(no_places * sizeof(int), 63);
	n_mark = (int *)emalloc(no_places * sizeof(int), 64);
	end_mark = net_end_markings(subs, tsub, NULL, &disc);	/* evaluate M0 */
	path = first_tang_marking(end_mark, &prob);		/* get one tang. marking */
	enter_tangmarking(path, disc);	     /* enter as first marking */
	if (next_tang_marking(subs, end_mark, &prob) != NULL)
		printf("Warning: vanishing initial marking with different following markings.\n");
	release_paths(end_mark);
	while(new_markings != NULL)
	{	marking = new_markings;
		new_markings = new_markings->next;
		global_marking(marking->mark, marking->disc, i_mark);
#ifdef DEBUG
		printf("\nevaluate marking (#%d): ", marking->number);
		print_global_marking(i_mark);
		h_mark = (int *)emalloc(no_places * sizeof(int), 65);
#endif
		fire = h_fire = 0;
		for(tra = 0; tra < tsub->no_trans; tra++)	/* for each timed transition */
		{	current_mark = i_mark;
			if (!test_global_conc(i_mark, tsub->transitions[tra].i_matrix,
						      tsub->transitions[tra].h_matrix))
				continue;			/* not enabled */
			fire = 1;
			h_tnums[h_fire++] = tra;
			current_mark = i_mark;
			fire_global_trans(i_mark, n_mark, no_places,
						tsub->transitions[tra].c_matrix);
#ifdef DEBUG
			printf("=> firing %s\n   to (van) marking ", tsub->transitions[tra].name);
			print_global_marking(n_mark);
			printf("  paths:\n");
#endif
			if (tsub->transitions[tra].type)	/* deterministic */
			{	if (marking->det)
				{	printf("ERROR: two deterministic transitions ");
					printf("(%s ", tsub->transitions[marking->det_num].name);
					printf("and %s) ", tsub->transitions[tra].name);
					printf("are enabled in marking:\n");
					print_global_marking(i_mark);
					exit(1);
				}
				marking->det_num = tra;
				marking->det = 1;
			}
			end_mark = net_end_markings(subs, tsub, n_mark, &disc);
			path = first_tang_marking(end_mark, &prob);
			first = 1;
			if (tsub->transitions[tra].type)	/* deterministic */
			{	i = max_paths(end_mark);
				marking->det_paths = (tang_path *)emalloc(i * sizeof(tang_path), 66);
				marking->no_det_paths = i;
				i = 0;
				while(path != NULL || first)
				{	first = 0;
					to_marking = enter_tangmarking(path, disc);
#ifdef DEBUG
					global_marking(to_marking->mark, to_marking->disc, h_mark);
					printf("      with %f to (#%d) ", prob, to_marking->number);
					print_global_marking(h_mark);
#endif
					marking->det_paths[i].prob = prob;
					marking->det_paths[i++].marking = (void *)to_marking;
					path = next_tang_marking(subs, end_mark, &prob);
				}
			} else					/* exponential */
			{	rate = 1.0 / exp_transition_rate(tsub, tra, i_mark);
				marking->sum_of_rates += rate;
				while(path != NULL || first)
				{	first = 0;
					to_marking = enter_tangmarking(path, disc);
#ifdef DEBUG
					global_marking(to_marking->mark, to_marking->disc, h_mark);
					printf("      with %f to (#%d) ", prob*rate
									, to_marking->number);
					print_global_marking(h_mark);
#endif
					merge_paths(marking, to_marking, prob * rate);
					path = next_tang_marking(subs, end_mark, &prob);
				}
			}
			efree(end_mark);
		}
		if (!fire)
		{	no_absorb_states++;
			printf("No transition enabled in marking:\n");
			print_global_marking(i_mark);
			printf("ERROR: absorbing state detected,");
			printf(" no steady-state solution possible.\n");
			exit(1);
		}
		h_path = marking->exp_paths;		/* fit space */
		marking->exp_paths = (tang_path *)emalloc(marking->no_exp_paths 
						* sizeof(tang_path), 67);
		memmove(marking->exp_paths,h_path, marking->no_exp_paths * sizeof(tang_path));
		efree((char *)(h_path));
		marking->fire_tra_num = h_fire;
		marking->fire_nums = (byte *)emalloc(h_fire * sizeof(byte), 68);
		for (h_fire = 0; h_fire < marking->fire_tra_num; h_fire++)
			marking->fire_nums[h_fire] = h_tnums[h_fire];	/* copy rates info */
		if (marking->det)	/* deterministic */
		{	i = tsub->transitions[marking->det_num].det_no;
			marking->next = det_markings[i];
			det_markings[i] = marking;
			no_nodes[i]++;
		} else
		{	marking->next = exp_markings;
			exp_markings = marking;	
			no_exp_markings++;
		}
		no_tang_markings++;
	}
	efree(h_tnums);
#ifdef DEBUG
	efree((char *)(h_mark));
#endif
}
