/******************************************************************************

Programm	:	rand

Komponente	:	rand_main.c

Parameter	:	argv[1]: Host, on which the exec_rands-Process
				 is running
			argv[2]: socket portnumber of exec_rands-Process
			argv[3]: Name of the DSPN
			argv[4]: Epsilon (optional, sonst default-Wert)

Aufgabe		:	Hauptprogramm fuer die Bestimmung der 
			transienten Loesung von Isomorphieklassen
			zugeordneter Markov-Ketten mit der
			Randomization - Technik

Projektleiter	:	Christoph Lindemann TU Berlin

Programmierer	:	Interprozesskommunikation: Tobias Bading
			Randomization: Enrik Baumann, Armin Zimmermann

Datum		:	Juni 1993

******************************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "rand_types.h"
#include "matrices_Q_delta.h"
#include "matrices_P_C.h"

#include "portab.h"

#define	WRITE_EMPTY_ENTRIES 		/* write empty CONV and EMC file entries ? */

static void sigill_handler (), sigfpe_handler (), sigbus_handler (),
     sigsegv_handler (), sigpipe_handler ();

static int socket_fd;			/* Filedescriptor fuer die Socket-Verbindung zum Server */
static FILE *socket_fp;			/* Filepointer fuer die Socket */
static char *conv_filename;		/* Filenamen der Files <argv[3]>.CONV<MC_number> und */
static char *emc_filename;		/* <argv[3]>.EMC<MC_number> */
static FILE *conv_fp, *emc_fp;		/* Filepointer fuer obige Files */
static double max_qii;			/* maximal diagonal entry */
static int eta_k;			/* Anzahl der Uebergaenge */

#define DEFAULT_EPSILON  1e-14
static double epsilon;      /* Wird ggf. aus argv[4] mit einem Wert versorgt */

/*** Eigene exit-Routine: ***/
void my_exit (exit_code)
int exit_code;
{
  write (socket_fd, "ready.\n", 7);	/* An diesem String erkennt der Server, */
  exit (exit_code);			/* dass wir fertig sind. */
}

/********************************************/
/*** Routinen zur Behandlung von Fehlern: ***/
/********************************************/
/* Da rand auf anderen Rechnern gestartet werden kann,
   wird durch diese signal-handler ein evtl. auftretender
   Fehler ueber die Socket an den exec_rands-Prozess gesendet. */
static void sigill_handler ()
{
  fprintf (stderr, "error in %s: Illegal Instruction\n", __FILE__);
  my_exit (1);
}
static void sigfpe_handler ()
{
  fprintf (stderr, "error in %s: Floatingpoint exception\n", __FILE__);
  my_exit (1);
}
static void sigbus_handler ()
{
  fprintf (stderr, "error in %s: Bus error\n", __FILE__);
  my_exit (1);
}
static void sigsegv_handler ()
{
  fprintf (stderr, "error in %s: Segmentation fault\n", __FILE__);
  my_exit (1);
}
static void sigpipe_handler ()
{
  /* Socketverbindung zum Server futsch, also kein my_exit() */
  exit (1);
}

/* global variables */
int		class_id;		/* identification number of rand process */
int             no_csmcs;		/* total number of CSMCs */
int             no_tau_sets;		/* number of sets of CSMCs with equal tau */
tau_set		*tau_sets;		/* array of sets of CSMCs with equal tau */
int		no_total_states;	/* no_csmc_states + no_preempt_arcs */
int		no_csmc_states;		/* number of states of the csmc */
int		no_preempt_arcs;	/* number of preempting edges */
int		first_state_no;		/* global number of the first csmc state */
char		acyclic;		/* the class is acyclic */

char		*last_of_scc;		/* bit for each local state; last state of a
					   strongly connected component of the CSMC */
char		*calc;			/* bit for each state; any initial state 
					   with this local number in an isomorphism set ? */
char		*leaf_state;		/* is the state a leaf node of the graph ? */
double		*q_ii;			/* diagonal entries of the Q matrix */
sp_matrix	*Q_matrix;		/* the local Q-matrix of the isomorphic CSMCs */
sp_matrix	*Delta_matrix;		/* Delta - matrix with probabilities of immediate
					   firing paths after a firing of a det. transition */

/* variables for the randomization algorithm */
sp_vector	*Phi_row, *old_Phi_row;
sp_matrix	*A_matrix;
sp_vector	**A_matrix_rows;
double		act_q_value;		/* the actual q value */
int		L_min, R_max;		/* minimal left and maximal right truncation points */
int		first_row;		/* var for cutting off useless rows of Q and A */
double		q_factor = 1.02;	/* factor for calculating q */
int		step_size;		/* test: old Phi == new Phi ? */

/* variables for the efficient solution of the branching topology */
char		branch;			/* the class has a branching topology */
int		branch_n, branch_k;	/* parameters of branching topology */
int		**state_succs;		/* successors of states */
char		*visited;		/* bits for the depth first search */
int		*branch_level;		/* level of each state */
char		*level_obtained;	/* level solution obtained ? */
int		level_start_state_no;	/* state number of first calculated state in level */

/*	emalloc tries to allocate required memory space and
	tests if it was successfully allocated; if not => exit */
char *
emalloc(bytes)
	int bytes;
{
	int *point;

	if (!bytes)
		return NULL;
	point = (int *)malloc((unsigned)(bytes));
	if (point == NULL)
	{	printf("ERROR: not enough memory in rand.");
		exit(3);
	}
	return (char *)point;	
}

/*	efree frees a memory pointer */
void
efree(point)
	char	*point;
{
	if (point == NULL)
		return;
	free(point);
}




/******************************************************************************

 Funktion:	init

 Parameter:	argv aus main()

 Rueckgabewert:	---

 Zweck:		- Aufbauen der Verbindung zum Server-Prozess 'exec-rands'
		- Umlenken von stdout und stderr in die Socket

******************************************************************************/
static void init (argv)
char *argv[];
{
  struct sockaddr_in server;
  register struct hostent *hp;
  unsigned tries;

  /* Verbindung zum Server (exec_rands.c) aufnehmen: */
  if ((socket_fd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
    exit (1);
  if (!(hp = gethostbyname (argv[1])))
    exit (1);

  memcpy (&server.sin_addr,hp->h_addr, hp->h_length);
  server.sin_family = AF_INET;
  server.sin_port = htons (atoi (argv[2]));
  for (tries = 1; tries <= 60; ++tries)
    if (connect (socket_fd, (struct sockaddr *) &server, sizeof (server)) != -1)
      break;
    else
      sleep (3);
  if (tries > 60)
  exit (1);

  /* Einen neuen Filepointer fuer die Socket erstellen: */
  if (!(socket_fp = fdopen (socket_fd, "r+")))
    exit (1);

  /* Standardausgabe und Standardfehler in die Socket umlenken: */
  if (dup2 (socket_fd, 1) < 0)
    exit (1);
  if (dup2 (socket_fd, 2) < 0)
    exit (1);

  /* Ausgabe-Zeilen sofort ausgeben und nicht buffern: */
  setvbuf(stdout,NULL,_IOLBF,BUFSIZ);
  setvbuf(stderr,NULL,_IOLBF,BUFSIZ);

  /* Eigene Fehlerbehandlungsroutinen anmelden: */
  signal (SIGILL, sigill_handler);
  signal (SIGFPE, sigfpe_handler);
  signal (SIGBUS, sigbus_handler);
  signal (SIGSEGV, sigsegv_handler);
  signal (SIGPIPE, sigpipe_handler);

  conv_filename = emalloc (strlen (argv[3]) + 8);
  emc_filename = emalloc (strlen (argv[3]) + 7);

  /* read id number of the isom. class */
  if (fread(&class_id, sizeof(int), 1, socket_fp) != 1)
	SOCKET_READ_ERROR;

  sprintf (conv_filename, "%s.CONV%02u", argv[3], class_id);
  sprintf (emc_filename, "%s.EMC%02u", argv[3], class_id);

  /*** .CONV00 und .EMC00 bzw. zum Schreiben oeffnen: ***/
  if (!(conv_fp = fopen (conv_filename, "w")))
  {   fprintf (stderr, "error in %s: can't open '%s' for writing !\n",
			__FILE__, conv_filename);
      my_exit (1);
  }
  if (!(emc_fp = fopen (emc_filename, "w")))
  {   fprintf (stderr, "error in %s: can't open '%s' for writing !\n",
			__FILE__, emc_filename);
      my_exit (1);
  }
}


/* main procedure of the randomization process */
main(argc, argv)
	int		argc;
	char		**argv;
{
	time_t		endtime, starttime;
	char		hostname[30];

	clock(); /* Prozessuhr starten.. ( fuer show_rusage() )*/
 	
	fclose (stdin);
	close (1);
	close (2);

	if (!gethostname (hostname, 29))
		if (strcmp (argv[1], hostname))
		/* rand auf anderem Rechner als exec_rands, also rsh austricksen: */
			switch (fork ())
			{	case -1: exit (1);
				case 0:  break;
				default: exit (0);
			}

	/* Kommandozeilenparameter ueberpruefen: */
	if (argc < 4)
		exit (1);

	/* Epsilon initialisieren */
	if(argc > 4)
	   sscanf(argv[4],"%lg",&epsilon);
      	else
	   epsilon=DEFAULT_EPSILON;
    	
	/* initialize sockets */
	init (argv);
	/* get the start time */
	starttime = time (NULL);

	/* read necessary data from exec_rands */
	read_data(socket_fp);

	starttime = time (NULL);
	/* truncated randomization */
	randomization ();

	endtime = time (NULL);

	if (branch)
	{	printf("Efficient solving of %d-branch topoloy with ", branch_n);
		printf("%d levels.\n", branch_k);
	}

	printf ("delta = %lg  eta = %u\n", max_qii, eta_k);
	show_rusage();
	fclose (stdout);
	fclose (stderr);
	fclose (emc_fp);
	fclose (conv_fp);

	my_exit (0);
}

/* truncated randomization algorithm */
randomization()
{
	tau_set		*act_tau_set;
	int		act_tau_set_no, act_iso_csmc_no, global_num, lev;
	iso_csmc	*act_iso_csmc;
	state		*act_state;
	sp_vector	*out_P_row, *out_C_row;
	sp_entry	*e_walk;
	double		act_val;
	int		start_state_no;		/* number of actual state to start the rand. */

	/* initializations */
	first_row = 0;
	act_q_value = 0.0;
	Phi_row = new_vector();
	A_matrix = new_matrix();
	out_P_row = new_vector();
	out_C_row = new_vector();

	/* for each possible start state */
	scan_array(start_state_no, no_csmc_states)
	{	if (!calc[start_state_no])
		{	/* empty entries */
#ifdef WRITE_EMPTY_ENTRIES
			scan_array(act_iso_csmc_no, no_csmcs)
				write_empty_entries(start_state_no, act_iso_csmc_no);
#endif
			continue;		/* no initial state here */
		}

		/* initialize tau - sets */
		scan_array(act_tau_set_no, no_tau_sets)
		{	act_tau_set = &(tau_sets[act_tau_set_no]);
			act_tau_set->rem_beta = 1.0;
			if (!act_tau_set->calc[start_state_no])
			{	
#ifdef WRITE_EMPTY_ENTRIES
				scan_array(act_iso_csmc_no, act_tau_set->no_iso_csmcs)
				write_empty_entries(start_state_no, 
					act_tau_set->iso_csmcs[act_iso_csmc_no].csmc_no);
#endif
				continue;		/* no initial state here */
			}

			/* test, if the solution can be analytically calculated */
			if (leaf_state[start_state_no])
			{	add_to_vector(act_tau_set->P_row, NULL, start_state_no, 1.0);
				add_to_vector(act_tau_set->C_row, NULL, start_state_no,
							act_tau_set->tau * act_q_value);
			}
		}

		/* solution calculated */
		if (leaf_state[start_state_no])
			goto derive_results;

		if (!branch)
		{	/* truncated randomization procedure */
			trunc_random(start_state_no);
		} else
		{	/* efficient solving of branching topology */
			lev = branch_level[start_state_no];
			if (level_obtained[lev])
			    scan_array(act_tau_set_no, no_tau_sets)
			    {	act_tau_set = &(tau_sets[act_tau_set_no]);
				if (!act_tau_set->calc[start_state_no])
					continue;
				map_branch(act_tau_set, level_start_state_no, start_state_no);
			    }
			else
			{   level_obtained[lev] = 1;
			    trunc_random(start_state_no);
			    scan_array(act_tau_set_no, no_tau_sets)
			    {	/* copy the values for later use */
				act_tau_set = &(tau_sets[act_tau_set_no]);
				if (!act_tau_set->lev_calc[branch_level[start_state_no]])
					continue;
				memset(act_tau_set->P_row_copy,0, no_csmc_states * sizeof(double));
				memset(act_tau_set->C_row_copy,0, no_csmc_states * sizeof(double));
				scan_list(e_walk, act_tau_set->P_row->entries)
					act_tau_set->P_row_copy[e_walk->num] = e_walk->val;
				scan_list(e_walk, act_tau_set->C_row->entries)
					act_tau_set->C_row_copy[e_walk->num] = e_walk->val;
				level_start_state_no = start_state_no;
			    }
			}
		}

		/* derive results */
derive_results:	scan_array(act_tau_set_no, no_tau_sets)
		{	act_tau_set = &(tau_sets[act_tau_set_no]);
			if (!act_tau_set->calc[start_state_no])
				continue;
			scan_array(act_iso_csmc_no, act_tau_set->no_iso_csmcs)
			{	/* was this state initial ? */
				act_iso_csmc = &(act_tau_set->iso_csmcs[act_iso_csmc_no]);
				act_state = &(act_iso_csmc->states[start_state_no]);
				if (!act_state->initial)
				{
#ifdef WRITE_EMPTY_ENTRIES
				write_empty_entries(start_state_no, act_iso_csmc->csmc_no);
#endif
					continue;
				}

				/* get the P row with global state numbers */
				scan_list(e_walk, act_tau_set->P_row->entries)
				{   global_num = global_state_num(e_walk->num, act_iso_csmc);
				    add_to_vector(out_P_row, NULL, global_num, e_walk->val);
				}

				/* calculate the P row with delta probabilities */
				mult_vector_matrix(out_P_row, Delta_matrix);

				/* get the C row with global state numbers and tau_i */
				scan_list(e_walk, act_tau_set->C_row->entries)
				{	/* dont take preempted states */
					if (e_walk->num >= no_csmc_states)
						continue;

					global_num = global_state_num(e_walk->num, act_iso_csmc);
					act_val = act_iso_csmc->states[e_walk->num].tau
						    / act_tau_set->tau * e_walk->val / act_q_value;
					add_to_vector(out_C_row, NULL, global_num, act_val);
				}

				/* write the results */
				global_num = global_state_num(start_state_no, act_iso_csmc);
				
				/* write header to CONV file */
				C_row_header.row_index = global_num;
				C_row_header.entries_in_row = out_C_row->no_entries;
				if (fwrite (&C_row_header, SIZEOF_C_row_header, 1, conv_fp) != 1)
				{	printf("ERROR while writing to CONV%02u.\n", class_id);
					my_exit(1);
				}

				/* write C vector */
				scan_list(e_walk, out_C_row->entries)
				{	C_row_entry.column_index = e_walk->num;
					C_row_entry.conversion_factor = e_walk->val;
					if (fwrite (&C_row_entry, SIZEOF_C_row_entry,
									1, conv_fp) != 1)
					{	printf("ERROR writing to CONV%02u.\n", class_id);
						my_exit(1);
					}
				}

				/* write header to EMC file */
				P_row_header.row_index = global_num;
				P_row_header.entries_in_row = out_P_row->no_entries;
				if (fwrite (&P_row_header, SIZEOF_P_row_header, 1, emc_fp) != 1)
				{	printf("ERROR while writing to EMC%02u.\n", class_id);
					my_exit(1);
				}

				/* write P vector */
				scan_list(e_walk, out_P_row->entries)
				{	/* write P matrix entry to EMC */
					P_row_entry.column_index = e_walk->num;
					P_row_entry.probability = e_walk->val;
					if (fwrite (&P_row_entry, SIZEOF_P_row_entry,
									1, emc_fp) != 1)
					{	printf("ERROR writing to EMC%02u.\n", class_id);
						my_exit(1);
					}
				}

				clear_vector(out_P_row);
				clear_vector(out_C_row);
			}
			clear_vector(act_tau_set->P_row);
			clear_vector(act_tau_set->C_row);
		}
	}

}

/* truncated randomization */
trunc_random(i_state)
{
	tau_set		*act_tau_set;
	int		step_no, act_tau_set_no, i;
	double		act_beta, C_factor;

	/* get the new A and Q matrix */
	calc_Q_and_A_matrix(i_state);

	/* set initial state entry in P row */
	add_to_vector(Phi_row, NULL, i_state, 1.0);

	old_Phi_row = copy_vector(Phi_row);	
	scan_array(step_no, R_max + 1)
	{	/* test, if the steady state is reached */
		if ((step_no > L_min) && !(step_no % step_size))
			if (comp_vector(Phi_row, old_Phi_row))
			{	
#ifdef DEBUG
			    fprintf(stderr, "break: %d of %d..%d\n", step_no, L_min, R_max);
#endif
			    /* steady state reached, calculate results */
			    scan_array(act_tau_set_no, no_tau_sets)
			    {   act_tau_set = &(tau_sets[act_tau_set_no]);
				if ((!branch && !act_tau_set->calc[i_state]) || 
				    ( branch && !act_tau_set->lev_calc[branch_level[i_state]]) ||
				    (step_no > act_tau_set->R))
					continue;	/* nothin to be done */

				/* calculate resulting P vector */
				vector_add_mult(act_tau_set->P_row, Phi_row, act_tau_set->rem_beta);

				/* calculate values */
				C_factor = 0.0;
				for (i = step_no; i <= act_tau_set->R; i++)
				{   act_beta = (i >= act_tau_set->L) ? 
					act_tau_set->beta[i - act_tau_set->L] : 0.0;
				    act_tau_set->rem_beta -= act_beta;
				    C_factor += act_tau_set->rem_beta;
				}

				/* calculate resulting C vector */
				vector_add_mult(act_tau_set->C_row, Phi_row, C_factor);
			    }
			    break;
			} else
			{	del_vector(&old_Phi_row);
				old_Phi_row = copy_vector(Phi_row);
			}

		/* for each set with equal tau */
		scan_array(act_tau_set_no, no_tau_sets)
		{	act_tau_set = &(tau_sets[act_tau_set_no]);
			if ((!branch && !act_tau_set->calc[i_state]) ||
			    ( branch && !act_tau_set->lev_calc[branch_level[i_state]]) ||
			    ( step_no > act_tau_set->R))
				continue;	/* nothin to be done */

			/* calculate P if L <= i_state <= R */
			if (step_no >= act_tau_set->L)
			{	act_beta = act_tau_set->beta[step_no - act_tau_set->L];
				act_tau_set->rem_beta -= act_beta;
				vector_add_mult(act_tau_set->P_row, Phi_row, act_beta);
			}

			/* calculate C if i_state <= R */
			vector_add_mult(act_tau_set->C_row, Phi_row, act_tau_set->rem_beta);
		}
		/* get the new Phi_row */
		mult_vector_matrix(Phi_row, A_matrix);
	}
	clear_vector(Phi_row);
	del_vector(&old_Phi_row);
}

/* write empty C and P matrix entry */
write_empty_entries(state_no, csmc_no)
	int		state_no, csmc_no;
{
	int		global_num;

	global_num = no_csmcs * state_no + csmc_no + first_state_no;
	C_row_header.row_index = global_num;
	C_row_header.entries_in_row = 0;
	if (fwrite (&C_row_header, SIZEOF_C_row_header, 1, conv_fp) != 1)
	{	printf("ERROR while writing to CONV%02u.\n", class_id);
		my_exit(1);
	}

	/* write empty P matrix entry */
	P_row_header.row_index = global_num;
	P_row_header.entries_in_row = 0;
	if (fwrite (&P_row_header, SIZEOF_P_row_header, 1, emc_fp) != 1)
	{	printf("ERROR while writing to EMC%02u.\n", class_id);
		my_exit(1);
	}
}

char	first_call_to_calc = 1;

/* get the new A and Q matrix */
calc_Q_and_A_matrix(state_no)
	int		state_no;		/* number of the actual initial state */
{
	tau_set		*act_tau_set;
	sp_vector	*v_walk;
	sp_entry	*e_walk;
	char		calc_bit = 0;
	int		i, act_tau_set_no;
	double		new_q = 0.0, L_d, R_d, W;

	/* can i cut something ? */
	if (!first_call_to_calc && state_no && !acyclic && !last_of_scc[state_no-1])
		return;

	if (first_call_to_calc)		/* first call to calc_Q_.. */
	{	calc_bit = 1;
		first_call_to_calc = 0;
	}

	else
	{	del_rows_from_matrix(Q_matrix, state_no - 1);	
		del_rows_from_matrix(A_matrix, state_no - 1);
		for (i = first_row; i < state_no; i++)
			if (act_q_value <= q_ii[i])
			{	calc_bit = 1;
				break;
			}
	}
	first_row = state_no;

	/* do we have to calculate q new ? */
	if (!calc_bit)
		return;

	/* calculate the new q */
	for (i = first_row; i < no_csmc_states; i++)
		if (q_ii[i] > new_q)
			new_q = q_ii[i];
	new_q *= q_factor;

	/* do we have to calculate A new ? */
	if (d_comp(new_q, act_q_value))
		return;
	act_q_value = new_q;

	/* calculate poisson probabilities for tau sets */
	L_min = 100000;
	R_max = 0;
	scan_array(act_tau_set_no, no_tau_sets)
	{	act_tau_set = &(tau_sets[act_tau_set_no]);
		weighter(act_q_value * act_tau_set->tau, 1e-14, 1.0e-300, 1.0e300, &L_d, &R_d,
					&(act_tau_set->beta), &W, &calc_bit);
		if (!calc_bit)
		{	printf("weighter fails.\n");
			my_exit(1);
		}
		act_tau_set->L = (int)(L_d + 0.5);
		act_tau_set->R = (int)(R_d + 0.5);
		if (act_tau_set->L < L_min)
			L_min = act_tau_set->L;
		if (act_tau_set->R > R_max)
			R_max = act_tau_set->R;

		scan_array(i, act_tau_set->R - act_tau_set->L + 1)
			act_tau_set->beta[i] /= W;
	}

	/* get the new step_size */
	i = R_max - L_min;
	if (i <= 100)
		step_size = 5;
	else if (i <= 1000)
		step_size = 10;
	else	step_size = 20;

	/* calculate the new A matrix */
	del_matrix(&A_matrix);
	A_matrix = copy_matrix(Q_matrix);
	scan_list(v_walk, A_matrix->rows)
		scan_list(e_walk, v_walk->entries)
			e_walk->val /= act_q_value;

	/* add unity matrix; side effect: A matrix is square */
	scan_array(i, no_total_states)
		if (!leaf_state[i])
			add_entry_to_matrix(A_matrix, i, i, 1.0);

	/* efficient A matrix */
	if (!A_matrix_rows)
		A_matrix_rows = (sp_vector **)emalloc(no_total_states * sizeof(sp_vector *));
	memset(A_matrix_rows,0, no_total_states * sizeof(sp_vector *));
	scan_list(v_walk, A_matrix->rows)
		A_matrix_rows[v_walk->num] = v_walk;
}
		
/* read data for an isomorphic csmc */
read_iso_data(i_csmc, tau_calc, tau_lev_calc, fp)
	iso_csmc	*i_csmc;
	char		*tau_calc;
	char		*tau_lev_calc;
	FILE		*fp;
{
	sp_vector	*v_walk;
	sp_entry	*e_walk;
	state		*act_state;
	double		tau_i, prob;
	char		md_tau;
	int		act_state_no, state_no, global_state_no;
	int		act_path_no, path_nos;
	static sp_vector *vec_walk, *old_vec_walk, *start_vec;
	sp_entry	*entry_walk;

	i_csmc->states = (state *)emalloc(no_csmc_states * sizeof(state));
	i_csmc->orig_no = (int *)emalloc(no_preempt_arcs * sizeof(int));

	/* tau marking dependent ? */
	if (fread(&md_tau, sizeof(char), 1, fp) != 1)
		SOCKET_READ_ERROR;

	/* read the absolute tau, if not marking dependent */
	if (!md_tau)
		if (fread(&tau_i, sizeof(double), 1, fp) != 1)
			SOCKET_READ_ERROR;

	/* get csmc states */
	scan_array(act_state_no, no_csmc_states)
	{	act_state = &(i_csmc->states[act_state_no]);
		global_state_no = global_state_num(act_state_no, i_csmc);

		/* get tau */
		if (md_tau)
			if (fread(&tau_i, sizeof(double), 1, fp) != 1)
				SOCKET_READ_ERROR;
		act_state->tau = tau_i;

		/* initial state ? */
		if (fread(&(act_state->initial), sizeof(char), 1, fp) != 1)
			SOCKET_READ_ERROR;
		if (act_state->initial)
		{	tau_calc[act_state_no] = 1;
			calc[act_state_no] = 1;
			if (branch)
				tau_lev_calc[branch_level[act_state_no]] = 1;
		}

		/* delta matrix row */
		if (!Delta_matrix->rows || (vec_walk->num > global_state_no))
			start_vec = Delta_matrix->rows;
		else	start_vec = vec_walk;
		scan_list(vec_walk, start_vec)
		{	if (vec_walk->num > global_state_no)
				break;
			old_vec_walk = vec_walk;
		}
		if (old_vec_walk == NULL)
		{	vec_walk = new_vector();
			vec_walk->next = Delta_matrix->rows;
			Delta_matrix->rows = vec_walk;
		} else
		{	vec_walk = new_vector();
			vec_walk->next = old_vec_walk->next;
			old_vec_walk->next = vec_walk;
		}
		vec_walk->num = global_state_no;
		entry_walk = NULL;

		if (fread(&path_nos, sizeof(int), 1, fp) != 1)
			SOCKET_READ_ERROR;
		vec_walk->no_entries = path_nos;
		scan_array(act_path_no, path_nos)
		{	if (entry_walk == NULL)
				vec_walk->entries = entry_walk = new_entry();
			else	entry_walk = entry_walk->next = new_entry();
			/* state number */
			if (fread(&(entry_walk->num), sizeof(int), 1, fp) != 1)
				SOCKET_READ_ERROR;
			/* probability */
			if (fread(&(entry_walk->val), sizeof(double), 1, fp) != 1)
				SOCKET_READ_ERROR;
		}
	}

	/* read informations about preempted states */
	scan_array(act_state_no, no_preempt_arcs)
	{	/* read marking number */
		if (fread(&global_state_no, sizeof(int), 1, fp) != 1)
			SOCKET_READ_ERROR;
		i_csmc->orig_no[act_state_no] = global_state_no;

		/* make the Delta matrix square */
		scan_list(v_walk, Delta_matrix->rows)
			if (v_walk->num >= global_state_no)
				break;
		if ((v_walk == NULL) || (v_walk->num > global_state_no))
		{	add_entry_to_matrix(Delta_matrix, global_state_no, global_state_no, 1.0);
			continue;
		}
		scan_list(e_walk, v_walk->entries)
			if (e_walk->num >= global_state_no)
				break;
		if ((e_walk == NULL) || (e_walk->num > global_state_no))
			add_entry_to_matrix(Delta_matrix, global_state_no, global_state_no, 1.0);
	}
}


/* read data from exec_rand */
read_data(fp)
	FILE		*fp;
{
	tau_set		*tau_p;
	int		act_preempt_no, act_state_no, i, j;
	int		act_edge_no, no_p_edges, no_n_edges, state_num;
	int		act_tau_set, act_iso_csmc, act_csmc_no = 0;
	double		act_rate;
	char		d_char;
	sp_vector	*vec_walk = NULL;
	sp_entry	*entry_walk;

	/* number of states */
	if (fread(&no_csmc_states, sizeof(int), 1, fp) != 1)
		SOCKET_READ_ERROR;

	/* global number of first state */
	if (fread(&first_state_no, sizeof(int), 1, fp) != 1)
		SOCKET_READ_ERROR;

	/* number of preempting edges */
	if (fread(&no_preempt_arcs, sizeof(int), 1, fp) != 1)
		SOCKET_READ_ERROR;

	/* number of states with preempted */
	no_total_states = no_csmc_states + no_preempt_arcs;

	/* number of isomorphic csmcs */
	if (fread(&no_csmcs, sizeof(int), 1, fp) != 1)
		SOCKET_READ_ERROR;

	/* is the class acyclic ? */
	if (fread(&acyclic, sizeof(char), 1, fp) != 1)
		SOCKET_READ_ERROR;

	/* has the class branching topology ? */
	if (fread(&branch, sizeof(char), 1, fp) != 1)
		SOCKET_READ_ERROR;

	/* informations about branching topology */
	if (branch)
	{	/* number of edges */
		if (fread(&branch_n, sizeof(int), 1, fp) != 1)
			SOCKET_READ_ERROR;
		/* number of levels */
		if (fread(&branch_k, sizeof(int), 1, fp) != 1)
			SOCKET_READ_ERROR;

		/* some initializations for the branching topology */
		state_succs = (int **)emalloc(no_csmc_states * sizeof(int *));
		visited = (char *)emalloc(no_csmc_states * sizeof(char));
		branch_level = (int *)emalloc(no_csmc_states * sizeof(int));
		act_state_no = 0;
		scan_array(j, branch_k)
		{	act_edge_no = no_states_in_level(branch_n, j);
			scan_array(i, act_edge_no)
				branch_level[act_state_no++] = j;
		}

		level_obtained = (char *)emalloc(branch_k * sizeof(char));
		memset(level_obtained,0, branch_k * sizeof(char));
	}

	/* read the Q matrix */
	act_preempt_no = no_csmc_states;
	Q_matrix = new_matrix();
	Delta_matrix = new_matrix();
	q_ii = (double *)emalloc(no_csmc_states * sizeof(double));
	max_qii = 0.0;
	eta_k = 0;
	last_of_scc = (char *)emalloc(no_csmc_states * sizeof(char));
	memset(last_of_scc,0, no_csmc_states * sizeof(char));

	/* mark leaf states */
	leaf_state = (char *)emalloc(no_total_states * sizeof(char));
	scan_array(act_state_no, no_total_states)
		leaf_state[act_state_no] = (act_state_no >= no_csmc_states);

	scan_array(act_state_no, no_csmc_states)
	{	if (vec_walk == NULL)
			Q_matrix->rows = vec_walk = new_vector();
		else	vec_walk = vec_walk->next = new_vector();
		vec_walk->num = act_state_no;
		entry_walk = NULL;

		if (branch)
			state_succs[act_state_no] = (int *)emalloc(branch_n * sizeof(int));

		if (!acyclic)
		{	/* last state of a strongly connected component ? */
			if (fread(&d_char, sizeof(char), 1, fp) != 1)
				SOCKET_READ_ERROR;
			if (d_char)
				last_of_scc[act_state_no] = 1;
			else	last_of_scc[act_state_no] = 0;
		}

		/* read successor info for branch */
		if (branch && branch_level[act_state_no] < branch_k-1)
			scan_array(act_edge_no, branch_n)
			{	if (fread(&state_num, sizeof(int), 1, fp) != 1)
					SOCKET_READ_ERROR;
				state_succs[act_state_no][act_edge_no] = 
							(state_num - first_state_no) / no_csmcs;
			}

		/* number of normal edges */
		if (fread(&no_n_edges, sizeof(int), 1, fp) != 1)
			SOCKET_READ_ERROR;
		eta_k += no_n_edges;
		vec_walk->no_entries = no_n_edges;
		scan_array(act_edge_no, no_n_edges)
		{	if (entry_walk)
				entry_walk = entry_walk->next = new_entry();
			else	vec_walk->entries = entry_walk = new_entry();

			/* global number of state */
			if (fread(&state_num, sizeof(int), 1, fp) != 1)
				SOCKET_READ_ERROR;
			entry_walk->num = (state_num - first_state_no) / no_csmcs;

			/* rate */
			if (fread(&(entry_walk->val), sizeof(double), 1, fp) != 1)
				SOCKET_READ_ERROR;

			if (entry_walk->num == act_state_no)
			{	/* diagonal entry */
				q_ii[act_state_no] = -entry_walk->val;
				if (q_ii[act_state_no] > max_qii)
					max_qii = q_ii[act_state_no];
			}
		}

		/* number of preempting edges */
		if (fread(&no_p_edges, sizeof(int), 1, fp) != 1)
			SOCKET_READ_ERROR;

		/* test for being a leaf state */
		if (!no_p_edges && (no_n_edges == 1))
			leaf_state[act_state_no] = 1;

		eta_k += no_p_edges;
		vec_walk->no_entries += no_p_edges;
		scan_array(act_edge_no, no_p_edges)
		{	if (entry_walk)
				entry_walk = entry_walk->next = new_entry();
			else	vec_walk->entries = entry_walk = new_entry();

			entry_walk->num = act_preempt_no++;

			/* rate */
			if (fread(&(entry_walk->val), sizeof(double), 1, fp) != 1)
				SOCKET_READ_ERROR;
		}
	}

	/* data structure initializations */
	if (fread(&no_tau_sets, sizeof(int), 1, fp) != 1)
		SOCKET_READ_ERROR;
	tau_sets = (tau_set *)emalloc(no_tau_sets * sizeof(tau_set));
	calc = (char *)emalloc(no_csmc_states * sizeof(int));
	memset(calc,0, no_csmc_states * sizeof(char));

	scan_array(act_tau_set, no_tau_sets)
	{	tau_p = &(tau_sets[act_tau_set]);
		tau_p->P_row_copy = (double *)emalloc(no_csmc_states * sizeof(double));
		tau_p->C_row_copy = (double *)emalloc(no_csmc_states * sizeof(double));
		
		/* number of csmcs with the same tau */
		if (fread(&(tau_p->no_iso_csmcs), sizeof(int), 1, fp) != 1)
			SOCKET_READ_ERROR;

		/* relative tau value for this set */
		if (fread(&(tau_p->tau), sizeof(double), 1, fp) != 1)
			SOCKET_READ_ERROR;

		/* some initializations */
		tau_p->P_row = new_vector();
		tau_p->C_row = new_vector();
		tau_p->iso_csmcs = (iso_csmc *)emalloc(tau_p->no_iso_csmcs * sizeof(iso_csmc));
		tau_p->calc = (char *)emalloc(no_csmc_states * sizeof(int));
		memset(tau_p->calc,0, no_csmc_states * sizeof(char));
		if (branch)
		{	tau_p->lev_calc = (char *)emalloc(branch_k * sizeof(char));
			memset(tau_p->lev_calc,0, branch_k * sizeof(char));
		}

		/* read data about isomorphic csmcs with the same tau */
		scan_array(act_iso_csmc, tau_p->no_iso_csmcs)
		{	tau_p->iso_csmcs[act_iso_csmc].csmc_no = act_csmc_no++;
			read_iso_data(&(tau_p->iso_csmcs[act_iso_csmc]), tau_p->calc, 
				tau_p->lev_calc, fp);
		}
	}
}
