/****************************************************************************
PROGRAMM	:	solve_LGS						     

PARAMETER	:	Name des DSPN-Netzes (argv[1])                       

Exitstatus	:	0 falls erfolgreich beendet			     
			1 falls Fehler waehrend der Programmausfuerung       

ZWECK		:	Loesen des Gleichgewichtszustandes der eingebetteten 
			Markoffkette eines DSPN. 	     
			Anschliessend Konvertieren der Stationaeren Losung   
			in DSPN-Zustandswahrscheinlichkeiten fuer die Markenver-     
			teilungen.					     

Datum		:	Okt. 93

Programmierer  :	Martin Mueller
	
Projektleiter  : 	Christoph Lindemann				     
*                                                                           *
*  Modifiziert	  :	Joachim Schneider				    *
*									    *
*  Datum	  :	July 1994 					    *
*****************************************************************************
*                                                                           * 
*  Modifiziert	  :	Patrick Drescher				    * 
*									    * 
*  Datum	  :	April 1996 					    * 
******************************************************************************/ 
/*
#define DEBUG_IT
*/
/*
#define SHOW_MAT
*/

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <malloc.h>
#include <math.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <memory.h> /* for memcpy */

#include "portab.h"
#ifdef SOL2
#include <siginfo.h>
#include <ucontext.h>
#endif

#include "matrices_P_C.h"
#include "sor.h"

#define USAGE "usage: solve_LGS [-a|-i] <model> <maxiter> <epsilon>"
#define TNET_RC ".TimeNET"

#define ITERATIVE_OPTION "-i"
#define AUTOMATED_OPTION "-a"

#define OK 0

#define DONE -1

#define UNSTRUCTURED 0
#define BLOCKDIAGONAL 1

#define ITERATIVE 0
#define AUTOMATED 1

/* constants used for direct solution */
#define EXACT 0
#define APPROX 1

#define USE_ITERATIVE 1
#define SOLUTION_OBTAINED 0

/* a large matrix has more than L_MATRIX tangible markings */
#define L_MATRIX 10000

#define MAX_DROP_TOL 0.01

/* if we have a large matrix we drop elements lower than LM_DROP*/
#define LM_DROP 1.0e-15

/* constants used for iterative solution */
#define MAXITER 1000
#define EPSILON 1.0e-7

#define false 0
#define true 1

int     	toptan, 	/* Anzahl tangible Markings */
		nonnull,	/* Anzahl Nicht-Null-Elemente*/
    		*ia,		/* Pointer auf die jeweiligen Zeilenanfaenge
				   der EMC-Matrix in a,ja*/
                *ja,	        /* Spaltenindizes der Nicht-Null-Eintraege */
                *r_ind,
                *min_ci,
                structure,
                method,
                num_drop;

unsigned        maxiter;

double		*a,		/* Nicht-Null-Eintrage der EMC-Matrix */
		*nprob,		/* Loesungsvektor des Gleichungssystems */
		*mprob,		/* Ergebnisvektor nach Normalisierung */
		prec,		/* Precision; es wird nirgendwo ein Wert zugewiesen,
                                   jedoch bei renormal() benutzt */

                *direct_nprob,  /* L"osungsvektor des Gleichungssystems nach 
                                   direkter L"osungsmethode */
                direct_prec;    /* Pr"azision des L"osungsvektors */


double 		row_min = 100.0;
double 		row_max = 0.0;
double 		abs_min = 100.0;
double 		abs_max = 0.0;
double 		rmaxdiff = 0.0;

long    	exectime; /*Hilfsgroessen zur Laufzeitberechnung*/

static time_t start_time;
static char *netname;
static double epsilon;
static double drop_tolerance;
static  int LU_mode;


/****************************************************************************
PROZEDUREN
*****************************************************************************/

double accuracy(); 
void read_p_matrix();
double get_drop_tolerance();
int get_dimension();
void mat2rpfm();
void renormal();
void derive_solution();
int derive_direct_solution();
void sta();
void print_sta();
void display_rp_matrix();
void display_matrix();
void display_pattern();
void check_rp_matrix();
void check_matrix();
int get_nrnbr();

#ifdef sun 
#ifdef SOL2
void sample_handler(int sig, siginfo_t *sip, ucontext_t *uap) {
     char *label;

     switch (sip->si_code) {
     case FPE_FLTINV: label = "INVALID-OP"; break;
     case FPE_FLTRES: label = "INEXACT"; break;
     case FPE_FLTDIV: label = "DIVISION-BY-ZERO"; break;
     case FPE_FLTUND: label = "UNDERFLOW"; break;
     case FPE_FLTOVF: label = "OVERFLOW"; break;
     default: label = "???"; break;
     }

     fprintf(stderr, "FP exception %s (0x%x) occurred at address %p.\n",
     	label, sip->si_code, (void *) sip->si_addr);
     abort();
}
#else
void sample_handler( sig, code, scp, addr )
int sig ;      /* sig == SIGFPE always */
int code ;
struct sigcontext *scp ;
char *addr ;
{    char *label;

     switch (code) {
     case FPE_FLTOPERR_TRAP: label = "INVALID-OP"; break;
     case FPE_FLTINEX_TRAP: label = "INEXACT"; break;
     case FPE_FLTDIV_TRAP: label = "DIVISION-BY-ZERO"; break;
     case FPE_FLTUND_TRAP: label = "UNDERFLOW"; break;
     case FPE_FLTOVF_TRAP: label = "OVERFLOW"; break;
     default: label = "???"; break;
     }
	printf("ieee exception %s (0x%x) occurred at pc %X \n",label,code,scp->sc_pc);
}
#endif
#endif

/**************************************************************************
  Name			: main

  Zweck			: Hauptprogramm
			  - Argumente ueberpruefen und analysieren
			  - Datenstrukturen initialisieren
			  - Gewuenschtes Losungsverfahren anwenden
			  - Konvertierung zu DSPN Zustandswahrscheinlichkeiten

  Eingabeparameter	: argc	-> Anzahl der Argumente
			  argv	-> Pointer auf Argumentvektoren
---------------------------------------------------------------------------*/
void main(argc,argv)
int argc;
char ** argv;
{
#ifdef sun
  sigfpe_handler_type oh2,oh3,oh4,oh5;
#endif

  clock(); /* Prozessuhr starten.. ( fuer show_rusage() )*/
 
  puts ("\n\nSOLVING THE LINEAR SYSTEM OF GLOBAL BALANCE EQUATIONS ...\n");
  fflush (stdout);

  start_time = time (NULL);

/* set up ieee_handler with own handling */
#ifdef SIGNAL_HANDLER
#ifdef sun
#ifndef SOL2 /* Scheint, als gaebe es ieee_handler() unter SunOS5.x nicht mehr.. */
  oh2 = (sigfpe_handler_type)ieee_handler("get","devision", oh2); 
  oh3 = (sigfpe_handler_type)ieee_handler("get","underflow", oh3); 
  oh4 = (sigfpe_handler_type)ieee_handler("get","overflow", oh4); 
  oh5 = (sigfpe_handler_type)ieee_handler("get","invalid", oh5);
 
  if(ieee_handler("set","devision" ,(sigfpe_handler_type)sample_handler ) != 0)
    {
      printf("ieee_handler can't set exception handling\n");
      fflush (stdout);
    }

  if(ieee_handler("set","underflow" ,(sigfpe_handler_type)sample_handler ) != 0)
    {
      printf("ieee_handler can't set exception handling\n");
      fflush (stdout);
    }

  if(ieee_handler("set","overflow" ,(sigfpe_handler_type)sample_handler ) != 0)
    {
      printf("ieee_handler can't set exception handling\n");
      fflush (stdout);
    }
  if(ieee_handler("set","invalid" ,(sigfpe_handler_type)sample_handler ) != 0)
    {
      printf("ieee_handler can't set exception handling\n");
      fflush (stdout);
    }

/*  set up ieee_handler with abort-handling 
  if(ieee_handler("set","devision" ,SIGFPE_ABORT ) != 0)
    {
      printf("ieee_handler can't set exception handling\n");
      fflush (stdout);
    }

  if(ieee_handler("set","underflow" ,SIGFPE_ABORT ) != 0)
    {
      printf("ieee_handler can't set exception handling\n");
      fflush (stdout);
    }

  if(ieee_handler("set","overflow" ,SIGFPE_ABORT ) != 0)
    {
      printf("ieee_handler can't set exception handling\n");
      fflush (stdout);
    }

  if(ieee_handler("set","invalid" ,SIGFPE_ABORT ) != 0)
    {
      printf("ieee_handler can't set exception handling\n");
      fflush (stdout);
    }
*/
#endif
#else
  signal (SIGFPE,SIG_IGN);
#endif
#endif

  start_time = time (NULL);

  if (argc < 5)
    {
      fprintf(stderr,USAGE);
      exit(1);
    }

  if (!strcmp(argv[1],ITERATIVE_OPTION))
    method = ITERATIVE;
  else
  if (!strcmp(argv[1],AUTOMATED_OPTION))
    method = AUTOMATED;
  else
    {
      fprintf(stderr,USAGE);
      exit(1);
    }
     
  netname = argv[2];
  sscanf(argv[3],"%u",&maxiter);
  sscanf(argv[4],"%lg",&epsilon);
  
  ia=ja=r_ind=min_ci = (int*)NULL;
  a=nprob=mprob = (double*)NULL;

  if (method == AUTOMATED)
    drop_tolerance = get_drop_tolerance(epsilon);

  if (drop_tolerance > MAX_DROP_TOL)
    {
      drop_tolerance = 0.0;
      LU_mode = EXACT;
    }
  else
    LU_mode = APPROX;

  toptan = get_dimension(netname);

  puts ("\nDeriving sparse representation of transition probability matrix in column-wise format.\n"); fflush (stdout);

  read_p_matrix(netname,&a,&ia,&ja);

/*
  check_matrix(toptan,nonnull,a,ia,ja);
*/

#ifdef SHOW_MAT
  display_matrix(toptan,nonnull,a,ia,ja);
  fflush(stdout);
#endif

  mat2rpfm(&a,&ia,&ja);
/*
  check_rp_matrix(toptan,nonnull,a,ia,ja);
*/
#ifdef SHOW_MAT
  display_rp_matrix(toptan,a,ia,ja);
  fflush(stdout);
#endif

  { /* examin matrix structure */
    register int min_index,cmi,cnt,i;

    min_index = min_ci[0];
    cnt = 0;

    for (i=1;i<toptan;i++)
      {
	cmi = min_ci[i];
	if (cmi < min_index)
	  cnt++;
	else
	  min_index = cmi;
      }
    if ((double)cnt/(double)toptan < 0.05)
      structure = BLOCKDIAGONAL;
    else
      structure = UNSTRUCTURED;
  }
  free((char*)min_ci);

  print_sta();
  
  derive_solution();

  renormal();

  (void)free((char *)nprob);
  (void)free((char *)mprob);
  (void)free((char *)a);
  (void)free((char *)ia);
  (void)free((char *)ja);

  show_rusage();

#ifdef SIGNAL_HANDLER  
#ifdef sun
#ifndef SOL2
  ieee_handler("set","devision", oh2); 
  ieee_handler("set","underflow", oh3); 
  ieee_handler("set","overflow", oh4); 
  ieee_handler("set","invalid", oh5);
#endif
#endif
#endif

  exit (0);
}


/**************************************************************************
  Name			: get_drop_tolerance

  Zweck			: drop-tolerance aus dem Environmentfile lesen:

  Rueckgabewert		: drop-tolerance aus dem Environmentfile
---------------------------------------------------------------------------*/
double get_drop_tolerance(eps)
double eps;
{
  char filename[MAXPATHLEN], line[200], value[100], *home;
  FILE *fp;
  double drop_tol;

  if (!(home = getenv ("HOME")))
  {
    fprintf (stderr, "environment variable 'HOME' not set!\n");
    exit (1);
  }
  sprintf (filename, "%s/%s", home, TNET_RC);

  if (!(fp = fopen (filename, "r")))
  {
    fprintf (stderr, "can't open file '%s' for reading!\n", filename);
    exit (1);
  }

  do
    fgets (line, 198, fp);
  while (!feof (fp) && strncasecmp (line, "DROPTOLERANCE", 13));
  if (!strncasecmp (line, "DROPTOLERANCE", 13))
    {
      sscanf (line, "%*s %s", value);
      if (!strncasecmp (value, "AUTO", 4))
	drop_tol = eps*0.1;
      else
      if (!strncasecmp (value, "NONE", 4))
	drop_tol = 0.0;
      else
	{
	  if(sscanf(value,"%lf",&drop_tol) != 1)
	    {
	      drop_tol = eps*0.1;
	      fprintf(stderr,"Missing value for DROPTOLERANCE in file ~/%s",TNET_RC);
	      fprintf(stderr,"Setting DROPTOLERANCE to default (EPSILON*0.01): %g",drop_tol);
	      fflush(stderr);
	    }
	}
    }
  else
    drop_tol = eps*0.1;

  return drop_tol;
}


/**************************************************************************
  Name			: accuracy

  Zweck		        : Bestimmung der bislang erreichten Praezision

  Eingabeparameter	: 
                          n  -> Dimension der Matrix
			  ja -> Zeilenpointer der Matrix
			  ia -> Spaltenindizes
			  a  -> Nichtnullelemente der matrix
                          x  -> Loesungsvektor
    
---------------------------------------------------------------------------*/
double accuracy( n, a, ia, ja,  x)
int n, *ia, *ja;
double *a;
double*   x;
{
  double  ff, 
          hf;
  int     ii, 
          jj,
          ibgn,
          iend;

  ff = 0.0;
  for (ii = 1; ii <= n; ii++)
    {
      hf = 0.0;
      ibgn = ia[ii];
      iend = ia[ii+1] - 1;
      for (jj = ibgn; jj <= iend; jj++)
	{
	  hf += a[jj] * x[ja[jj]];
	}
      hf = fabs(hf);
      if (ff < hf)
	{
	  ff = hf;
	}
    }
  return ff;

}


/**************************************************************************
  Name			: get_dimension

  Zweck			: Ermittlung der Dimension der P-matrix

  Eingabeparameter	: Modellname
    
  Rueckgabewert		: Dimension der P-matrix
---------------------------------------------------------------------------*/
int get_dimension(netname)
char *netname;
{
  register FILE *aux_file;
  char filename[MAXPATHLEN], line[200];
  int dim;

  sprintf (filename, "%s.AUX", netname);
  if (!(aux_file = fopen (filename, "r")))
    {
      perror (NULL);
      fprintf (stderr, "can't open file '%s' for reading!\n", filename);
      exit (1);
    }
  do
    fgets (line, 198, aux_file);
  while (!feof (aux_file) && strncmp (line, "TANGIBLE MARKINGS", 17));
  if (!strncmp (line, "TANGIBLE MARKINGS", 17))
    sscanf (line + 17, "%d", &dim);
  else
    {
      fprintf (stderr, "file '%s' doesn't contain a line 'TANGIBLE MARKINGS <number>' !\n",
	       filename);
      exit (1);
    }
  fclose (aux_file);
  return dim;
}


/**************************************************************************
  Name			: read_p_matrix

  Zweck			: liest die P-Matrix aus den Files <netname>.EMC00
                          bis <netname>.EMC<n> in ix,jx,x

  Eingabeparameter	: Modellname, x, ix
    
---------------------------------------------------------------------------*/
void read_p_matrix(netname,x,ix,jx)
double **x;
int    **ix,
       **jx;
char *netname;
{
  char emc_filename[MAXPATHLEN];
  FILE *emc_fp;
  register int diag_found;	   /* Diagonalelement in einer Zeile enthalten ? */
  register unsigned e_allocated,i;
  register unsigned files = 0;
  register unsigned row_nbr = 0;

  nonnull = 0;
  e_allocated = toptan;
  
  if((*x = (double*)malloc(e_allocated*sizeof(double)))==NULL)
     {
       perror("solve_LGS - for Matrix description");
       exit(1);
     }
  if((*ix = (int*)malloc((e_allocated+1)*sizeof(int)))==NULL)
     {
       perror("solve_LGS - for Matrix description");
       exit(1);
     }
  if((*jx = (int*)malloc((e_allocated+1)*sizeof(int)))==NULL)
     {
       perror("solve_LGS - for Matrix description");
       exit(1);
     }
  if ((min_ci=(int*)malloc(toptan*sizeof(int))) == NULL)
    {
      perror("solve_LGS - for Matrix description");
      exit(1);
    }
  for (i=0;i<toptan;i++)
    min_ci[i] = toptan;



  /* <netname>.EMC<k>-Files der Reihe nach lesen und dabei die Anzahl der
     Eintraege in jeder Spalte zaehlen: */
  while (1)
  {
    sprintf (emc_filename, "%s.EMC%02u", netname, files);

    if (!(emc_fp = fopen (emc_filename,"r")))
    {
      if (!files)
      {
	/* EMC00 ist nicht notwendigerweise vorhanden */
	++files;continue;
      }
      break;
    }
    ++files;

    /* File 'emc_filename' lesen: */
    while (1)
    {
      /* Header lesen */
      if (fread (&P_row_header, SIZEOF_P_row_header, 1, emc_fp) != 1)
      {
	if (ferror (emc_fp))
	{
	  fprintf (stderr, "can't read file '%s' !\n", emc_filename);
	  perror (NULL);
	  exit (1);
	}
	fclose(emc_fp);
	break;
      }
      row_nbr++;
      /* nonull ist die Anzahl der Eintraege der P-Matrix (inklusive Diagonalelemente) */
	nonnull += P_row_header.entries_in_row;

      if (nonnull + 1  > e_allocated)
	{
	  /* vorsichtshalber f"ur fehlenden Diagonaleintrag mit allozieren */
	  e_allocated += (P_row_header.entries_in_row + 1)* (toptan - P_row_header.row_index + 1); 
	  if((*x = (double*)realloc((char*)*x,e_allocated*sizeof(double)))==NULL)
	    {
	      perror("realloc");
	      exit(1);
	    }
	  if((*ix = (int*)realloc((char*)*ix,(e_allocated+1)*sizeof(int)))==NULL)
	    {
	      perror("solve_LGS - for Matrix description");
	      exit(1);
	    }
	  if((*jx = (int*)realloc((char*)*jx,(e_allocated+1)*sizeof(int)))==NULL)
	    {
	      perror("solve_LGS - for Matrix description");
	      exit(1);
	    }
	}
      
      diag_found = false;
      for ( i = nonnull - P_row_header.entries_in_row; 
	   i < nonnull;
	   i++)
	{
	  if (fread (&P_row_entry, (int)SIZEOF_P_row_entry,1,emc_fp) != 1)
	    {
	      fprintf (stderr, "can't read file '%s' !\n", emc_filename);
	      perror (NULL);
	      exit (1);
	    }
	  sta(P_row_entry.probability);
	  if (P_row_header.row_index == P_row_entry.column_index)
	    {
	      diag_found = true;
	      x[0][i] =  P_row_entry.probability - 1.0;
	    }
	  else
	    x[0][i] = P_row_entry.probability;

	  ix[0][i] = (int)P_row_entry.column_index ;
	  jx[0][i] = (int)P_row_header.row_index;
          if (P_row_entry.column_index < min_ci[P_row_header.row_index-1])
               min_ci[P_row_header.row_index-1] = P_row_entry.column_index;
	}

      if (!diag_found)
	{
	  x[0][nonnull] = -1.0;
	  jx[0][nonnull] = (int)P_row_header.row_index;
	  ix[0][nonnull] = (int)P_row_header.row_index;
          if (P_row_header.row_index < min_ci[P_row_header.row_index-1])
       	       min_ci[P_row_header.row_index-1] = P_row_header.row_index;
	  ++nonnull;
	}
    }
  }
}

/**************************************************************************
  Name			: mat2rpfm

  Zweck			: Konvertiert, die vorliegende Matrix in ix,jx,x
                          zum Zeilen-Pointer-Format

  Eingabeparameter	: x, ix, jx 
    
---------------------------------------------------------------------------*/
void mat2rpfm(x,ix,jx)
double **x;
int    **ix,
       **jx;
{
  register double xz;        /*Zwischensp. fuer x[0][i]*/  
  register int jz,           /*Zwischensp. fuer jx[0][i],ix[0][i]*/
      curr_pos,     /*index des betrachteten Elements*/
      new_pos,      /*index wohin mit Element*/
      act_row,      /* aktuelle Zeilennummer */
      fnd_row,      /* gefundene Zeilennummer */
      t_bgn,t_end,  /* Indexgrenzen wohin Element kommt */
      f_bgn,f_end;  /* Indexgrenzen woher Element kommt */
  int *cnt,         /* cnt[i]=Anzahl der Elemente von Zeile i */
      *rptr;        /* rptr[i]: Anfangsindex der Zeile in x,jx*/
  register int i,ind;
  int error = 0;

  if ((cnt = (int*)calloc((unsigned)toptan,sizeof(int))) == NULL)
      {
	perror("solve_LGS - for Matrix description");
	exit(1);
      }
      
  if ((rptr = (int*)malloc(((unsigned)(toptan+1))*sizeof(int))) == NULL)
    {
      perror("solve_LGS - for Matrix description");
      exit(1);
    }

  /* eintr"age zaehlen */
  for (i=0;i<nonnull;i++)
    cnt[ix[0][i]-1]++;

  /* rptr fuellen */
  ind = 1;
  for (i=0;i<toptan;i++)
    {
      rptr[i]=ind;
      ind += cnt[i];
    }
  rptr[toptan]=ind;

  /* tauschen */
  f_bgn = rptr[0] -1;
  f_end = rptr[1] -1;
  curr_pos = f_bgn;
  act_row = 1;

  while (1)
    {
      /* naechstes zu tauschendes Element finden */
      while (1)
	{
	  if (ix[0][curr_pos] == DONE)
	    curr_pos++;
	  else if (ix[0][curr_pos] == act_row)
	    {
	      ix[0][curr_pos] = DONE;
	      curr_pos++;
	    }
	  else
	    break;

	  if (curr_pos >= nonnull)
	    break;
	  
	  /* naechste Zeile zu bearbeiten ? */
	  if (curr_pos >= f_end)
	    {
	      act_row++;
	      f_bgn = rptr[act_row - 1] -1;
	      f_end = rptr[act_row] -1;
	    }
	}
      /* fertig? */
      if (curr_pos >= nonnull)
	break;

      fnd_row = ix[0][curr_pos];
      t_bgn = rptr[fnd_row - 1] -1;
      t_end = rptr[fnd_row] -1;
      new_pos = t_bgn;
      
      /* Stelle finden, wohin Element soll */
      while(1)
	{
	  if (ix[0][new_pos] == DONE)
	    new_pos++;
	  else if (ix[0][new_pos] == fnd_row)
	    {
	      ix[0][new_pos] = DONE;
	      new_pos++;
	    }
	  else
	    break;

	}
      if (new_pos >= t_end)
	{
	  error = 1;
	  break;
	}

      /*tauschen*/
      xz = x[0][new_pos];
      x[0][new_pos] = x[0][curr_pos];
      x[0][curr_pos] = xz;      
      jz = jx[0][new_pos];
      jx[0][new_pos] = jx[0][curr_pos];
      jx[0][curr_pos] = jz;
      ix[0][curr_pos] = ix[0][new_pos];
      ix[0][new_pos] = DONE;
    }
  free(cnt);
  if(error)
    {
      free((char*)rptr);
      printf("Fatal error occured in deriving row-pointer-format of emc-matrix; exiting\n");
      exit(1);
    }
  else
    {
      free((char*)*ix);
      *ix = rptr;
    }
}
	
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ 
  Name			: compare 

  Zweck			: Vergleichs Funktion fuer Quick-Sort-Algorithmus

  Eingabeparameter	: Pointer auf die Indizes der jeweiligen Zeilenanfaenge

  Rueckgabewert		: >0: zeilen_groesse1 > zeilen_groesse2
		   	  =0: zeilen_groesse1 = zeilen_groesse2
			  <0: zeilen_groesse1 < zeilen_groesse2
----------------------------------------------------------------------------*/

int compare (elem1,elem2)
int *elem1, *elem2;

{
  register int zeilen_groesse1,zeilen_groesse2;

  zeilen_groesse1 = ia[*elem1] - ia[*elem1 - 1];
  zeilen_groesse2 = ia[*elem2] - ia[*elem2 - 1];

  return (zeilen_groesse1 - zeilen_groesse2);
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   Name			: derive_solution

  Zweck			: Ermittlung derr Gelichgewichtsl"osung in nprob

-------------------------------------------------------------------------*/
void derive_solution()
{
  int ret;
  register int i;
  register double tmp,sum;
  double eps = epsilon;

  switch(method)
    {
    case ITERATIVE:
      break;
    case AUTOMATED:
      if (derive_direct_solution() == USE_ITERATIVE)
	{
	  printf("\nNot enough memory for direct solution found.\n\
Retry with iterative solution\n\n");
      fflush(stdout);
	  break;
	}
      else 
	return;
    default:
      fprintf(stderr,"solve_LGS: unknown solution Method");
      exit(1);
    }
  
  if (nprob==(double*)NULL)
    if ((nprob=(double*)malloc((unsigned)toptan*sizeof(double))) == NULL) 
      {
	perror("solve_LGS");
	exit(1);
      }
  for(i=0;i<toptan;i++)
    nprob[i] = 1.0 / (double)toptan;

  printf("\nSolution with SOR ...\n");
  fflush(stdout);

  ret = sor(toptan,ia,ja,a,nprob,maxiter,&epsilon,nonnull,netname);
  switch(ret) 
    {
    case ERROR:
      printf("\nFatal error in iterative solver; no solution available.\n");
      exit(1);
    case NOT_CONV:
      printf("\nNo convergence within %d iterations.\n",maxiter);
      printf("Achieved accuracy: %-15lg\n",epsilon);
      break;
    default:
      printf("\nConvergence in %d iterations with accuracy %lg\n",ret,epsilon);
      break;
    }
  fflush(stdout);
}

#define free_storage() \
  {\
     if(jn_row)\
       (void)free((char*)jn_row);\
     if(p)\
       (void)free((char*)p);\
     if(iu)\
       (void)free((char*)iu);\
     if(ju)\
       (void)free((char*)ju);\
     if(r)\
       (void)free((char*)r);\
     if(c)\
       (void)free((char*)c);\
     if(ic)\
       (void)free((char*)ic);\
     if(n_row)\
       (void)free((char*)n_row);\
     if(y)\
       (void)free((char*)y);\
     if(u)\
       (void)free((char*)u);\
     jn_row=p=iu=ju=r=c=ic = (int*)NULL;\
     n_row=y=u = (double*)NULL;}

#define free_partial_storage() \
  {\
     if(p)\
       (void)free((char*)p);\
     if(iu)\
       (void)free((char*)iu);\
     if(ju)\
       (void)free((char*)ju);\
     if(r)\
       (void)free((char*)r);\
     if(c)\
       (void)free((char*)c);\
     if(ic)\
       (void)free((char*)ic);\
     if(y)\
       (void)free((char*)y);\
     if(u)\
       (void)free((char*)u);\
     p=iu=ju=r=c=ic = (int*)NULL;\
     y=u = (double*)NULL;}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   Name			: derive_direct_solution

  Zweck			: Ermittlung derr Gelichgewichtsl"osung in nprob

-------------------------------------------------------------------------*/
int derive_direct_solution()
{
  int *jn_row,*p,*iu,*ju,n_row_size;
  double *n_row,*y,*u;
  int		ierr,		/* Error-Code */
 		*r,		/* Permutationsarray der Zeilen */
		*c,		/* Permutationsarray der Spalten */
		*ic;		/* inverses Permutationsarray der Spalten */
  int force_iterative;
  int nr_nbr,num_fill_in;

#ifdef DEBUG_IT
  force_iterative = 1;
#else
  force_iterative = 0;
#endif

  jn_row=p=iu=ju=r=c=ic = (int*)NULL;
  n_row=y=u = (double*)NULL;

  if ((nr_nbr = get_nrnbr(toptan,ia)) <= 0)
    {
      printf("ERROR: zero matrix dimension\n");
      exit(1);
    }
/* Normalisierungsbedingung einfuegen; entfernte Zeile in jn_row und n_row merken */
  { 

    int bgn,end,anf_ind,fip,lip,fcp,lcp;
    register int i,k,new_pos;

    /* set insert and copy positions */
    fip = ia[nr_nbr-1]-1;
    lip = fcp = ia[nr_nbr]-1;
    lcp = ia[toptan]-1;
    n_row_size = ia[nr_nbr] - ia[nr_nbr-1];
    num_fill_in = toptan - n_row_size;

    if ((n_row = (double*)malloc((unsigned)n_row_size*sizeof(double)))== NULL)
      {
	perror("malloc");
	return USE_ITERATIVE;
      }
    if ((jn_row = (int*)malloc((unsigned)n_row_size*sizeof(int)))== NULL)
      {
	perror("malloc");
	free_storage();
	return USE_ITERATIVE;
      }

    /* save values from row */
    bgn = ia[nr_nbr-1] -1;
    end = ia[nr_nbr] -1;
    for( k = bgn,i=0  ; k < end ; k++,i++) 
      {
	n_row[i] = a[k];
	jn_row[i] = ja[k];
      }

    anf_ind = ia[nr_nbr-1] - 1;
    nonnull += num_fill_in;

    if ((a=(double*)realloc((char*)a,nonnull*sizeof(double))) == NULL) 
      {
	perror("realloc: no mem for a");
	free_storage();
	return USE_ITERATIVE;
      }
    if ((ja=(int*)realloc((char*)ja,nonnull*sizeof(int))) == NULL) 
      {
	perror("realloc: no mem for ja");
	free_storage();
	return USE_ITERATIVE;
      }

    /* copy to end */
    for(i = fcp , new_pos = fcp + num_fill_in;
	i<lcp;
	i++,new_pos++)
      {
	a[new_pos] = a[i];
	ja[new_pos] = ja[i];
      }

    /* fill in */
    for( k = 1,i=fip; k<=toptan; k++,i++) 
      {
	a[i]= 1.0;
	ja[i] = k;
      }

    /* correct ia */
    for (i = nr_nbr;i<= toptan;i++)
      ia[i] += num_fill_in;

  }

#ifdef SHOW_MAT
  display_rp_matrix(toptan,a,ia,ja);
  fflush(stdout);
#endif

  /* Speicher allozieren */
  if ((r=(int*)malloc((unsigned)toptan*sizeof(int))) == NULL) 
    {
      perror("malloc: no mem for r");
      free_storage();
      return USE_ITERATIVE;
    }
  if ((c=(int*)malloc((unsigned)toptan*sizeof(int))) == NULL) 
    {
      perror("malloc: no mem for c");
      free_storage();
      return USE_ITERATIVE;
    }
  if ((ic=(int*)malloc((unsigned)toptan*sizeof(int))) == NULL) 
    {
      perror("calloc: no mem for ic");
      free_storage();
      return USE_ITERATIVE;
    }
  if ((nprob=(double*)malloc((unsigned)toptan*sizeof(double))) == NULL) 
    {
      perror("malloc: no mem for nprob");
      free_storage();
      return USE_ITERATIVE;
    }
  if ((y=(double*)malloc((unsigned)toptan*sizeof(double))) == NULL) 
    {
      perror("malloc: no mem for y");
      free_storage();
      return USE_ITERATIVE;
    }
  if ((iu=(int*)malloc((unsigned)(toptan+1)*sizeof(int))) == NULL) 
    {
      perror("malloc: no mem for iu");
      free_storage();
      return USE_ITERATIVE;
    }
  if ((p=(int*)calloc((unsigned)(toptan+1),sizeof(int))) == NULL) 
    {
      perror("calloc: no mem for p");
      free_storage();
      return USE_ITERATIVE;
    }
  
  { /* Felder r,i,ic,b initialisieren */
    register int k;

    /* Permutations arrays r,i,ic initialisieren */
    for ( k=0; k<toptan; k++) 
      {
	r[k]=c[k]=ic[k]=k+1;
      }
  }

  { /* Loesungsprozess */

    unsigned u_size = (unsigned)nonnull/2;

    if (structure == UNSTRUCTURED)
      (void)qsort((char *)r,toptan,sizeof(int),compare); 

    printf("\nSolution with sparse Gaussian-Elimination ...\n");
    fflush(stdout);

    while (drop_tolerance <= MAX_DROP_TOL)
      {
	if ((u=(double*)malloc(u_size*sizeof(double))) == NULL) 
	  {
	    perror("malloc: no mem for u");
	    free_storage();
	    return USE_ITERATIVE;
	  }
	if ((ju=(int*)malloc(u_size*sizeof(int))) == NULL) 
	  {
	    perror("malloc: no mem for ju");
	    free_storage();
	    return USE_ITERATIVE;
	  }
	dnspiv1_mod(toptan,
		    ia,ja,a,
		    nr_nbr,(int)u_size,
		    r,c,ic,
		    nprob,
		    y,p,iu,&ju,&u,
		    drop_tolerance,&num_drop,&ierr);

	(void)free((char *)u);
	u = (double*)NULL;
	(void)free((char *)ju);
	ju = (int*)NULL;

	if (ierr <=0) 
	  {
	    if ( ierr == 0) 
	      {
		printf("ERROR: Matrix Dimension = 0\n");
		exit(1);
	      }
	    if ( - toptan <= ierr ) 
	      {
		printf("ERROR: Row number %d of A is null\n",abs(ierr));  
		exit(1);
	      }
	    if ( - 2*toptan <= ierr) 
	      {
		printf("ERROR: Row number %d has a duplicate entry\n",abs(ierr/2)); 
		exit(1);
	      }
	    if ( -3*toptan <= ierr ) 
	      {
		printf("ERROR: Row number %d has a zero pivot\n",abs(ierr/3)); 
		exit(1);
	      }
	    if (  - 4*toptan <= ierr ) 
	      {
		if (LU_mode == EXACT)
		  {
		    LU_mode = APPROX;
		    drop_tolerance = epsilon*0.1;
		  }
		else
		  drop_tolerance *= 10.0;
		if (drop_tolerance <= MAX_DROP_TOL)
		  {
		    printf("\nStorage of U-matrix exceeds available storage.\n");
		    printf("Retrying solution with drop-tolerance %lg\n",drop_tolerance);
		    fflush(stdout);
		  }
	      }
	  }
	else
	  {	
	    printf("\nSolution process terminated\nNumber of Off-diagonal elements of U: %d\n",ierr);
	    if (LU_mode == APPROX)
	      {
		printf("Droptolerance: %g\n",drop_tolerance);
		printf("Number of dropped elements: %d\n",num_drop);
	      }
	    fflush(stdout);
	    break;
	  }
      }

    free_partial_storage();

    if (drop_tolerance > MAX_DROP_TOL)
      return USE_ITERATIVE;

      {
	int ret,fcp,lcp;
	register int i,k,bgn,end,new_pos;

	/* set insert and copy positions */
	fcp = ia[nr_nbr]-1;
	lcp = ia[toptan]-1;

	/* entnommene Zeile der Matrix wieder einf"ugen */
	bgn = ia[nr_nbr-1] -1;
	end = bgn + n_row_size;
	for( k = bgn,i=0  ; k < end ; k++,i++) 
	  {
	    a[k] = n_row[i];
	    ja[k] = jn_row[i];
	  }

	/* copy back from end */
	for(i = fcp , new_pos = fcp - num_fill_in;
	    i<lcp;
	    i++,new_pos++)
	  {
	    a[new_pos] = a[i];
	    ja[new_pos] = ja[i];
	  }

	/* correct ia */
	for (i = nr_nbr;i<= toptan;i++)
	  ia[i] -= num_fill_in;

        /* correct nonnull */
        nonnull -= num_fill_in;
#ifdef SHOW_MAT
	display_rp_matrix(toptan,a,ia,ja);
	fflush(stdout);
#endif
        direct_prec = accuracy(toptan,--a,--ia,--ja,--nprob);
        printf("Achieved accuracy:%g\n",direct_prec);

        ++ia,++ja,++a,++nprob;

	if (force_iterative || /*((LU_mode == APPROX || drop_tolerance != 0.0) 
            && (num_drop > 0)) ||*/ (direct_prec > epsilon)) 
        {

	/* store old solution */
        if ((direct_nprob = (double*)malloc((unsigned)toptan*sizeof(double))) == NULL) 
          {
	    perror("solve_LGS");
	    exit(1);
          }
        direct_nprob = (double*)memcpy((char*)direct_nprob, (char*)nprob,
                       (int)toptan*(int)sizeof(double)/(int)sizeof(char));

	printf("\nIterative refinement of solution with SOR ...\n");
	fflush(stdout);

	ret = sor(toptan,ia,ja,a,nprob,maxiter,&epsilon,nonnull,netname);
	switch(ret) 
	  {
	  case ERROR:
	    printf("\nFatal error in iterative refinement process; no solution available.\n");
	    break;
	  case NOT_CONV:
	    printf("\nNo convergence within %d iterations.\n",maxiter);
	    printf("Achieved accuracy: %-15lg\n",epsilon);
	    break;
	  default:
	    printf("\nConvergence in %d iterations with accuracy %lg.\n",ret,epsilon);
	    break;
	  }

        if (direct_prec < epsilon)
          {
          printf("\nIterative refinement was not succesfull ... keeping direct solution.\n\n");
          nprob = direct_nprob;
          }
	}
      }
    fflush(stdout);
    free((char*)jn_row);
    jn_row = (int*)NULL;
    free((char*)n_row);
    n_row = (double*)NULL;
  }
  return SOLUTION_OBTAINED;
}

/*****************************************************************************
  Name			:  renormal

  Zweck			: Konvertieren der EMC Zustandswahrscheinlichkeiten
			  in DSPN Zustandswahrscheinlichkeiten

----------------------------------------------------------------------------*/

void renormal()
{
  FILE *conv_fp,*mpdf;
  char probmark_filename[MAXPATHLEN],conv_filename[MAXPATHLEN];
  register int i, k, l;
  double zr, rr, sum = 0.0, mpdf_buf;
  unsigned filenumber = 0;

  if((mprob = (double*)calloc((unsigned)toptan,sizeof(double)))==NULL)
    {
      perror("calloc");
      exit(1);
    }

#ifdef DEBUG
  puts ("renormal");
#endif

  /* .CONV00 lesen: */
  {
    sprintf (conv_filename, "%s.CONV00", netname);
    if (/*!*/(conv_fp = fopen (conv_filename, "r")))
    /*
    {
      fprintf (stderr, "can't open file '%s' for reading !\n", conv_filename);
      exit (1);
    }
    */
    while (1)
    {
      if (fread ((char*)&C0_entry, SIZEOF_C0_entry, 1, conv_fp) != 1)
        if (ferror (conv_fp))
	{
	  perror ("renormal: fread");
	  exit (1);
	}
        else
	  break;

      rr = C0_entry.sojourn_time * fabs (nprob[C0_entry.state_number - 1]);
      sum += rr;
      mprob[C0_entry.state_number - 1] += rr;
    }
  }
  while (1)
  {
    sprintf (conv_filename, "%s.CONV%02u", netname, ++filenumber);

    if (!(conv_fp = fopen (conv_filename,"r")))
      break;

#ifdef DEBUG
    printf ("reading '%s' ...\n", conv_filename);
#endif
    while (1)
    {
      if (fread (&C_row_header, SIZEOF_C_row_header, 1, conv_fp) != 1)
        if (ferror (conv_fp))
	{
	  perror ("renormal: fread");
	  exit (1);
	}
        else
          break;
      else
      {
	zr = fabs (nprob[C_row_header.row_index - 1]);
	for (k = 1; k <= C_row_header.entries_in_row; ++k)
	{
	  if (fread (&C_row_entry, SIZEOF_C_row_entry, 1, conv_fp) != 1)
	  {
	    perror ("renormal: fread");
	    exit (1);
	  }
	  l = (int)C_row_entry.column_index - 1; 
	  rr = zr * C_row_entry.conversion_factor;
	  sum += rr; 
	  mprob[l] += rr;
	}
      }
    }
    fclose (conv_fp);
  }
  sum=1.0/sum;
  sprintf (probmark_filename, "%s.PROBMARK", netname);

  if ((mpdf = fopen(probmark_filename,"w")) == NULL) {
	perror("fopen: mpdf: renormal ");
	exit(1);
  }

  mpdf_buf= -prec;
  if (fwrite((char*)&mpdf_buf,sizeof(double),1,mpdf) != 1)
    if (ferror(mpdf)) 
      {
	perror("fwrite: 'renormal':");
	exit(1);
      }

  mpdf_buf=-1.0;
  if (fwrite((char*)&mpdf_buf,sizeof(double),1,mpdf) != 1)
    if (ferror(mpdf)) 
      {
	perror("fwrite: 'renormal':");
	exit(1);
      }

  for (i=0 ;i < toptan ;i++) 
    { 
      mpdf_buf = mprob[i]*sum; 
      if (fwrite((char*)&mpdf_buf,sizeof(double),1,mpdf) != 1)
	if (ferror(mpdf)) 
	  {
	    perror("fwrite: 'renormal':");
	    exit(1);
	  }
    }  /* for i */
  (void)fclose(mpdf);

} /* renormal */

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Name			: print_sta

  Zweck			: Ausgabe von statistischen Groessen ueber die 
			  eingelesene Matrix

  Eingabeparameter	: keine

  Rueckgabewert		: keiner
----------------------------------------------------------------------------*/

void print_sta()
{

  /* printf ("\n%-31s %d\n","Dimension :", toptan); */
  printf ("\n%-31s %d\n","Number of nonzeros :", nonnull - toptan);
    /* nonnull zaehlt Diagonalelemente mit, die eingefuegt werden */
  printf ("%-31s %.2lf\n","Number of nonzeros per row :",
	  ((double)toptan * (double)toptan) / (double)nonnull);
  printf ("%-31s %1.2e\n","Smallest entry :", abs_min);
}

/****************************************************************************
  Name			: sta

  Zweck			: Erfassen von Statistiken der Matrix

  Eingabeparameter	: rate	-> Eingelesene Rate

  Rueckgabewert		: keiner
----------------------------------------------------------------------------*/
void sta(rate)
double rate;
{
  if (rate < row_min)
    row_min = rate;
  if (rate > row_max)
    row_max = rate;
  if (rate < abs_min)
    abs_min = rate;
  if (rate > abs_max)
    abs_max = rate;
}

/****************************************************************************
  Name			: row_diff

  Zweck			: Berechnen von Statistiken der Matrix

  Eingabeparameter	: keine

  Rueckgabewert		: keiner
----------------------------------------------------------------------------*/
void row_diff()
{
#if 0
  double d;

  if (row_max == 0.0)
    d = -infinity ();
  else
    d = fabs (log10 (row_max / row_min));
  rmaxdiff = d > rmaxdiff ? d : rmaxdiff;
#endif

  row_min = infinity ();
  row_max = -infinity ();
}

/****************************************************************************
  Name			: get_nrnbr

  Zweck			: Berechnung der gr"o"sten Zeilennummer die
                          nicht leer ist.

  Eingabeparameter	: 
                          dimension: Dimension der Matrix
			  ia: Zeilenpointer der Matrix

  Rueckgabewert		: Zeilennummer, falls erfolgreich; sonst -1
----------------------------------------------------------------------------*/
int get_nrnbr(dimension,ia)
int dimension,*ia;
{

  register int i;

  for (i= dimension ; i>0;i--)
      if (ia[i]-ia[i-1] > 1)
	return i;
  return -1;
}

/* SOME NICE AND USEFUL PROCEDURES */
/************************************************************/

/* Ausgabe der Matrix im row-pointer Format */

void display_rp_matrix(dimension,a,ia,ja)
int dimension,*ia,*ja;
double *a;
{

  int i,j;
  double *val;
  int ibgn,iend;

  if ((val=(double*)calloc((unsigned)dimension,sizeof(double))) == NULL) 
    {
      perror("calloc: no mem for mprob");
      exit(1);
    }
  printf("\n\n");
  for (i=0;i<dimension;i++)
    {
      /*collect values of row i*/
      for (j=0;j<dimension;j++)
	val[j]=0.0;

      ibgn = ia[i]-1;
      iend = ia[i+1]-1;

      for (j=ibgn;j<iend;j++)
	{
	      val[ja[j]-1] = a[j];
	}
      for (j=0;j<dimension;j++)
	printf("% 1.2lf  ",val[j]);
      printf("\n");
    }
  free((char*)val);

}
	
/************************************************************/

/* Ausgabe der Matrix */
void display_matrix(dimension,nonnull,a,ia,ja)
int dimension,nonnull,*ia,*ja;
double *a;
{

  int i,j;
  double *val;

  if ((val=(double*)calloc((unsigned)dimension,sizeof(double))) == NULL) 
    {
      perror("calloc: no mem for mprob");
      exit(1);
    }

  printf("\n\n");
  for (i=0;i<dimension;i++)
    {
      /*collect values of row i*/
      for (j=0;j<dimension;j++)
	val[j]=0.0;

      for (j=0;j<nonnull;j++)
	{
	  if(ia[j]-1 == i)
	    {
	      val[ja[j]-1] = a[j];
	    }
	}
      for (j=0;j<dimension;j++)
	printf("% 1.2lf  ",val[j]);
      printf("\n");
    }
  free((char*)val);

}
	
/************************************************************/

/* Ausgabe des Musters der Matrix im row-pointer Format */
void display_pattern(dimension,nonnull,a,ia,ja)
int dimension,nonnull,*ia,*ja;
double *a;
{

  int i,j;
  char *val;
  int ibgn,iend;

  if ((val=(char*)calloc((unsigned)dimension + 1,sizeof(char))) == NULL) 
    {
      perror("calloc: no mem for mprob");
      exit(1);
    }
  printf("\n\n");
  for (i=0;i<dimension;i++)
    {
      /*collect values of row i*/
      for (j=0;j<dimension;j++)
	val[j]='-';

      ibgn = ia[i]-1;
      iend = ia[i+1]-1;

      for (j=ibgn;j<iend;j++)
	{
	      val[ja[j]-1] = 'x';
	}
      printf("%s\n",val);
    }
  free((char*)val);

}

/************************************************************/

/* "Uberpr"ufung der Matrix auf doppelte Eintr"age */
void check_matrix(dimension,nonnull,a,ia,ja)
int dimension,nonnull,*ia,*ja;
double *a;
{

  int i,j;
  char *val;
  int cnt = 0;

  if ((val=(char*)calloc((unsigned)dimension,sizeof(char))) == NULL) 
    {
      perror("calloc: no mem for mprob");
      exit(1);
    }

  printf("\n\n");
  for (i=0;i<dimension;i++)
    {
      /*collect values of row i*/
      for (j=0;j<dimension;j++)
	val[j]='\0';

      for (j=0;j<nonnull;j++)
	{
	  if(ia[j]-1 == i)
	    {
	      if (val[ja[j]-1] != 'x')
		val[ja[j]-1] = 'x';
	      else
		{
		  printf("DOUBLE ENTRY (rp format) in row %d column index: %d\n",i+1,ja[j]);
		  cnt++;
		}
	    }
	}
    }
  free((char*)val);
  if (cnt == 0)
    printf("MATRIX OK");

}

/* "Uberpr"ufung der Matrix im row-pointer-format auf doppelte Eintr"age */

void check_rp_matrix(dimension,nonnull,a,ia,ja)
int dimension,nonnull,*ia,*ja;
double *a;
{

  int i,j;
  char *val;
  int ibgn,iend;
  int cnt = 0;

  if ((val=(char*)calloc((unsigned)dimension,sizeof(char))) == NULL) 
    {
      perror("calloc: no mem for mprob");
      exit(1);
    }
  printf("\n\n");
  for (i=0;i<dimension;i++)
    {
      /*collect values of row i*/
      for (j=0;j<dimension;j++)
	val[j]='\0';

      ibgn = ia[i]-1;
      iend = ia[i+1]-1;

      for (j=ibgn;j<iend;j++)
	{
	  if (val[ja[j]-1] != 'x')
	    val[ja[j]-1] = 'x';
	  else
	    {
	      printf("DOUBLE ENTRY (rp format) in row %d column index: %d\n",i+1,ja[j]);
	      cnt++;
	    }
	}
    }
  free((char*)val);
  if (cnt == 0)
    printf("MATRIX OK");
}

