/******************************************************************************
 File:              weighter.c

 Purpose:           Calculation of Poisson-probabilities

 Programmer:        Enrik Baumann,
                    Tobias Bading

 History:           (Initially part of rand.c by E.Baumann, T.Bading)
                    Release 1: September 17, 1993
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include <math.h>
#include "weighter.h"

#define TRUE  1
#define FALSE 0

#define OUT_OF_MEMORY { \
   fprintf(stderr,"%s (line %u): out of memory !\n", __FILE__,__LINE__); \
   exit( -1 ); \
}


/******************************************************************************
 Funktion:	d

 Parameter:	k und lambda

 Rueckgabewert:	double-Wert d

 Zweck:		Wird von find_r() benoetigt
******************************************************************************/
static double d (k, lambda)
double k, lambda;
{
  return 1.0 / (1.0 - exp (-(2.0/9.0) * (k * sqrt (2.0 * lambda) + 3.0/2.0)));
}


/******************************************************************************
 Funktion:	find_r

 Parameter:	lambda, epsilon, Mode m, k_p

 Rueckgabewert:	Der rechte Abschnittspunkt

 Zweck:		Berechnet den rechten Abschnittspunkt
******************************************************************************/
static double find_r (lambda, epsilon, m, k_p)
double lambda, epsilon, m, *k_p;
{
   double a_lambda;
   double k;
   double q_k_lambda;
   double R;

   a_lambda = (1.0 + 1.0/lambda) * exp(1.0/16.0) * sqrt(2.0);

   /* corollary 1 of section 4 : k (bzw. R) berechnen */
   for (k = 3.0; /*k < 10.0*/; k++)
   {
      /* q_k_lambda berechnen */
      q_k_lambda = a_lambda * d(k,lambda) * exp(-pow(k,2.0)/2.0) / (k * sqrt(2.0*M_PI));

      /* R berechnen */
      R = ceil(m + k*sqrt(2.0*lambda) + 3.0/2.0);
      if (q_k_lambda < epsilon/2.0)
         break;
   }
   *k_p = k;

   return R;
} /* find_r */


/******************************************************************************
 Funktion:	finder

 Parameter:	lambda --> Wert von lambda
		epsilon --> Gewuenschte Genauigkeit
		tau --> Kleinster darstellbarer Wert
		omega --> Groesster darstellbarer Wert
		m --> Mode von Lambda
		L_p --> Linker Abschnittspunkt
		R_p --> Rechter Abschnittspunkt
		w_m_p --> Der Wert von w(m)
		F_p --> Flag ob Fehler aufgetreten sind

 Rueckgabewert:	---

 Zweck:		Berechnet den linken und rechten Abschnittspunkt und
		den Wert von w(m).
******************************************************************************/
static void finder (lambda, epsilon, tau, omega, m, L_p, R_p, w_m_p, F_p)
double lambda, epsilon, tau, omega, m, *L_p, *R_p, *w_m_p;
char *F_p;
{
   double L, R, w_m;
   char F = TRUE;
   double b_lambda;
   double t_k_lambda;
   double k;
   double c_m;
   double q_dach;
   double p_lambda_k;

   if (lambda <= 0.0)
   {
      L = R = 0.0;
      F = FALSE;

      goto EXIT;
   }

   R = find_r(lambda, epsilon, m, &k);

   if (lambda < 25.0)
   {
      L = 0.0;
   }
   else /* lambda >= 25.0 */
   {
      /* b_lambda berechnen */
      b_lambda = (1.0 + 1.0/lambda) * exp(1.0/(8.0*lambda));

      /* corollary 2 of section 4 : k (bzw. L) berechnen */
      t_k_lambda = b_lambda * exp(-pow(k,2.0)/2.0) / (k*sqrt(2.0*M_PI));
      if (t_k_lambda >= epsilon/2.0)
      {
         printf ("WARNING: finder(): t_k_lambda isn't less then epsilon/2 !\n");
      }
      L = floor(m - k*sqrt(lambda) - 3.0/2.0);
      if (L<0.0)
         L = 0.0;
   }

   /* w(m) berechnen */
   w_m = omega / (1.0e10 * (R - L));

   if (lambda >= 25.0)
   {
      /* corollary 3 of section 5 : p_lambda_k berechnen */
      c_m = (1.0 / sqrt(2.0*M_PI*m)) * exp(m - lambda - 1.0/(12.0*m));
      q_dach = k * sqrt(2.0) + 3.0/(2.0*sqrt(lambda));
      p_lambda_k = c_m * exp(- pow(q_dach + 1.0,2.0) / 2.0);
      if (p_lambda_k * w_m < tau)
      {
         F = FALSE;
         goto EXIT;
      }

      /* corollary 4 of section 5 : p_lambda_k berechnen */
      q_dach = k + 3.0/(2.0*sqrt(lambda));
      if (q_dach <= sqrt(lambda)/2.0)
      {
         p_lambda_k = c_m * exp( -(pow(q_dach,2.0)/2.0) - 
                                 pow(q_dach,3.0)/(3.0*sqrt(lambda)) );
      }

#if 0
      else
         fprintf(stdout,"Warning: Corollary 4 of section 5 (ii) & (iii) not implemented");
#endif

      if (p_lambda_k * w_m > tau)
         F = TRUE;
      else
         F = FALSE;
   }

EXIT:
   *L_p = L;
   *R_p = R;
   *w_m_p = w_m;
   *F_p = F;
}

/******************************************************************************
 Function:     weighter

 Input:        lambda -> Wert von Lambda
               epsilon --> geforderte Genauigkeit
	       tau --> kleinster darstellbarer Wert
	       omega --> groesster darstellbarer Wert
	       L --> Linker Abschnittspunkt
	       R --> Rechter Abschnittspunkt
	       w --> Die Gewichtungen
               W --> Summe der Gewichte
               F --> Flag ob Fehler aufgetreten sind

 Rueckgabewert:	---

 Zweck:		Berechnung der Poisson Wahrscheinlichkeiten fuer
		grosse Werte von lambda.
******************************************************************************/
void weighter (lambda, epsilon, tau,omega, L,R, w,W, F,run)
double	lambda;	/* in */
double	epsilon;
double	tau;
double	omega;
double	*L;	/* out */
double	*R;
double	**w;	/* w(L),...,w(R) */
double	*W;
char	*F;
unsigned run;
{
   double m, w_m;
   double j, q;

   unsigned i, s, t;

   /*Initialize:*/
   m = floor(lambda);

   finder(lambda, epsilon, tau, omega, m, L, R, &w_m, F);
   if (*F == FALSE)
      return;
   *L = (double )0.0;

   /* adapt right boundary to run-level */
   *R *= (double )(1 << run);

   if( (*w = (double *)calloc((size_t)(*R + 1),sizeof(double))) == NULL )
      OUT_OF_MEMORY;
   (*w)[(unsigned )m] = w_m;

   /* calculate downwards */
   for (j = m, i = (unsigned )j; j > 0.0; j--, i--)
      (*w)[i - 1] = (j / lambda) * (*w)[i];

   /* calculate upwards */
   if (lambda < 400) /* Special */
   {
#ifdef FORCE_600BOUNDARY
      if (*R > 600.0)
      {
         fprintf(stderr,"Panic: R > 600\n");
         *F = FALSE;
         return;
      }
#endif

      j = m; i = (unsigned)j;
      while (j < *R)
      {
         q = lambda / (j + 1.0);
         if ((*w)[i] > (tau / q))
         {
            (*w)[i+1] = q * (*w)[i];
            j++; i++;
         }
         else
         {
            *R = j;
            goto Compute_W;
         }
      }
   }
   else
   {
      for (j = m, i = (unsigned)j; j < *R; j++, i++)
         (*w)[i+1] = (lambda / (j + 1)) * (*w)[i];
   }

Compute_W:
   *W = 0.0;
   s = 0;
   t = (unsigned)*R;

   while (s < t)
   {
      if ((*w)[s] <= (*w)[t])
         *W += (*w)[s++];
      else
         *W += (*w)[t--];
   }
   *W += (*w)[s];
} /* weighter */
