/********************************************************************
Modul :             adapt_case.c

Purpose :           transientes Loesungsverfahren fuer Petri-Netze
                    (DOPRI5 with variable step-size)

Project Director :  Reinhard German

Programmer :        Armin Heindl

History :           18.5.95 ( Diplomarbeit )
************************************************************************/
/* #define DEBUG_AC */
/* #define KOSTEN */

#define DATA_BUFFER 200

#include <stdio.h>                     /* fuer NULL */
#include <math.h>        
#include "frontend.h"                  /* dort stehen Definitionen der sparse structures und
			                  forward declarations ihrer Operationen aus mat_proc.c */
#include "Options.h"
#include "shared.h"                    /* fuer ConcernsRandomisation etc.*/

/* Import aus main.c */
extern TransientOptions transientOptions;

/* Import aus shared.c */
extern FILE *open_solve();
extern void reset_digits();
extern void update_solve();
extern void close_adaptsolve();
extern double calculate_A_Matrix();

/* aus runge_kutta_methods.c */
extern double dopri5();
extern void rk2();
extern void rk3();
extern void set_k7_is_k1();

/* aus on_tx_plain.c */
extern int update_int_structures();
extern void set_old_stepsize();
extern void save_static_values();
extern void reaffirm_static_values();
extern void stetige_uebertragung();

/* aus interpolation_fields.c */
extern void empty_a();
extern void a_ab();
extern void abcc_int_points();
extern void abcc_y_values();
extern void abcd_cdbe();
extern void cdbe_cdef();
extern void rejected_update();
extern void repeated_update();

/* aus discontinuities.c */
extern void build_up_tree();
extern double get_next_discontinuity();

int ac_no_detT;                          /* ac steht fuer adaptive case */

double EXT_PRECISION;                    /* = transientOptions.precision/100.0 */
double LOW_PRECISION;                    /* = (ac_no_detT+1)*transientOptions.precision */
double off_sum;                          /* Schaetzwert fuer ertraegliche Abweichung der Summe aller
					    Wahrscheinlichkeiten von 1 */

/********************************************************************************************/
void boundary_conditions( daten_basis, res_values, int_fields)
/* Berechnung der Randbedingung */
/* Voraussetzungen : - in int_fields[i][0].aux_coefficients ist der ( meist interpolierte ) Vektor
                       der tau-Linie des aktuellen Zeitpunkts gespeichert
		     - in daten_basis enthalten p_E und p_d[i] die aktuellen Werte              */

solve_basis *daten_basis;
sp_vectors *res_values;
interpolation_field *int_fields;
{
  int d,c;
  sp_vector *v_help;

  /* Formel : p_d(t,0) = p_E(t)*Q_Ed + c_sum{p_c(t,tau_c)*Delta_cd + p_c(t)*Q_cd} */

  for( d=0; d<ac_no_detT; d++) {
    /* wichtig : res_values[d] ist NULL oder Zeiger bereits umgehaengt */
    res_values[d] = copy_vector( daten_basis->p_E );
    mult_vector_matrix( res_values[d], daten_basis->Q_Ed[d]); 
    
    for( c=0; c<ac_no_detT; c++) {
      v_help = copy_vector( int_fields[c][0].aux_coefficients);
      mult_vector_matrix( v_help, daten_basis->detMatrices[c].Delta_dc[d]);
      vector_add_mult( res_values[d], v_help, 1.0);
      del_vector(&v_help);

      v_help = copy_vector( (daten_basis->p_d)[c]);
      mult_vector_matrix( v_help, daten_basis->detMatrices[c].Q_dc[d]);
      vector_add_mult( res_values[d], v_help, 1.0);
      del_vector(&v_help);
    };
  };

}
/********************************************************************************************/
void save_areas( areas_ptr, careas_ptr)
/* legt eine Kopie der Impulsflaechen in careas an fuer den Fall der Schrittablehnung */

areas *areas_ptr;
areas *careas_ptr;
{
  int j;

  for(j=0; j<ac_no_detT; j++) {
    /* ueberholte careas-Struktur loeschen */
    while( careas_ptr[j].first != NULL) {
      careas_ptr[j].last = careas_ptr[j].first;
      careas_ptr[j].first = careas_ptr[j].first->next;
      del_vector( &careas_ptr[j].last);
    };
    /* areas-Struktur nach careas-Struktur kopieren */
    careas_ptr[j].first = copy_vector( areas_ptr[j].first);
    areas_ptr[j].last = areas_ptr[j].first;
    careas_ptr[j].last = careas_ptr[j].first;
    while( areas_ptr[j].last->next != NULL) {
      areas_ptr[j].last = areas_ptr[j].last->next;
      careas_ptr[j].last->next = copy_vector( areas_ptr[j].last);
      careas_ptr[j].last = careas_ptr[j].last->next;
    };
    careas_ptr[j].last->next = NULL;
  };
}
/********************************************************************************************/
void save_int_structures( where_to, add_where, no_upper_steps, up_boxes, plow_box, interpol_update, int_fields, c_no_upper_steps, c_up_boxes, pc_low_box, c_interpol_update, c_int_fields)
/* zweifache Verwendung in Abhaengigkeit von where_to :
   - falls 1 : Sicherheitskopien aller Daten im Zusammenhang mit Integration und Interpolation werden angelegt fuer den
               Fall, dass der folgende Schritt abgelehnt werden muss
   - falls 0 : nach abgelehntem Schritt werden die zuvor angelegten Sicherheitskopien wieder eingesetzt
               ( beachte Besonderheit fuer die low_box ! )                                                           */
/* Besonderheit : nach Schrittablehnung muessem die c_up_boxes-Zeiger auf die ersten step_inbetween-Elemente nach den
                  neu ermittelten und mit clear_int_reject() sofort wieder verworfenen upper_steps verweisen */
/* weitere Vorraussetzung : - Gruppen sind entsprechend ihrer determ. Schaltzeit aufsteigend indiziert 
                            - add_where wurde nach Schrittablehnung hinreichend gross gewaehlt               */

int where_to;
int add_where;
int *no_upper_steps;
step_inbetween **up_boxes;
lower_stepbox *plow_box;
int *interpol_update;
interpolation_field *int_fields;
int *c_no_upper_steps;
step_inbetween **c_up_boxes;
lower_stepbox *pc_low_box;
int *c_interpol_update;
interpolation_field *c_int_fields;
{
   int i,k, from;
   step_inbetween *ptr, *c_ptr;
   step_inbetween **parser;        /* markiert Enden der upper_steps, damit nur kopiert wird, was tatsaechlich
				      kopiert werden muss */

   parser = (step_inbetween **)emalloc(ac_no_detT*sizeof(step_inbetween *));

   if (where_to) 
     from = 0;    /* kopiere von (from) 0 nach 1 (where_to) : nach angenommenem Schritt */
   else
     from = 1;    /* kopiere von (from) 1 nach 0 (where_to) : nach abgelehntem Schritt */

   for( k=0; k<ac_no_detT; k++) {

     /* beachte : eine upper_steps-Box existiert immer ! */
     if (where_to) { /* Schrittannahme */
       parser[k] = up_boxes[k]->next_down;
       for( i=1; i<no_upper_steps[k]; i++) 
	 parser[k] = parser[k]->next_down;
       /* parser[k] zeigt auf das erste step_inbetween-Element nach den zugehoerigen upper_steps ! */
     }
     else { /* Schrittablehnung */
       parser[k] = c_up_boxes[k];
       /* parser[k] zeigt das step_inbetween-Elemente, das auf die wieder geloeschten upper_steps der Gruppe k folgte */
     };

     /* Sicherheitskopien der Anzahl der upper_steps, etc. */
     c_no_upper_steps[k] = no_upper_steps[k];
     c_up_boxes[k] = up_boxes[k];

     /* Sicherheitskopien der Interpolationsdaten */
     c_interpol_update[k] = interpol_update[k];     
     for(i=0;i<4;i++) {
       c_int_fields[k][i].int_point = int_fields[k][i].int_point;
       del_vector(&(c_int_fields[k][i].y_values));
       c_int_fields[k][i].y_values = copy_vector(int_fields[k][i].y_values);
       del_vector(&(c_int_fields[k][i].pol_coefficients));       
       c_int_fields[k][i].pol_coefficients = copy_vector(int_fields[k][i].pol_coefficients);
       del_vector(&(c_int_fields[k][i].aux_coefficients));
       c_int_fields[k][i].aux_coefficients = copy_vector(int_fields[k][i].aux_coefficients);
     };
   };

   /* Uebertragung der values-Vektoren in dafuer vorgesehene Doppelstrukturen */
   /* Bei obersten parser-Element beginnend werden nach unten hin alle zu kopierenden Vektoren aufgelesen
      und kopiert, z.B. braucht bei Schrittannahme ein values-Vektor der Gruppe k nicht kopiert werden, falls
      er in einem step_inbetween-Element auftaucht, das oberhalb des parser[k]-Elements liegt */
   ptr = parser[ac_no_detT-1];
   for( k=(ac_no_detT-2); k>=0; k--) {
      while (ptr != parser[k]) {
	for( i=k+1; i<ac_no_detT; i++) {
	  del_vector(&(ptr->values[where_to][i]));
	  ptr->values[where_to][i] = copy_vector( ptr->values[from][i]);
	};
	ptr = ptr->next_down;
      };
   };

   /* ab hier muss jeder Vektor der step_inbetween-Elemente kopiert werden */
   while (ptr != NULL) {
     for( i=0; i<ac_no_detT; i++) {
       del_vector(&(ptr->values[where_to][i]));
       ptr->values[where_to][i] = copy_vector( ptr->values[from][i]);
     };         
     ptr = ptr->next_down;
   };

   /* Uebertragung der values-Vektoren in low_box */
   /* beachte : add_where = add_here - 1 ( addhere aus update_int_structures() )*/
   for( i=0; i<=add_where; i++) {
     for( k=0; k<ac_no_detT; k++) {
       del_vector(&(plow_box->lower_steps[i].values[where_to][k]));
       plow_box->lower_steps[i].values[where_to][k] = copy_vector( plow_box->lower_steps[i].values[from][k]);
     };
   };
     
   /* Sicherheitskopie der low_box */
   for(i=0;i<=add_where;i++) {
     pc_low_box->lower_steps[i].step = plow_box->lower_steps[i].step;
     pc_low_box->lower_steps[i].integr_weight = plow_box->lower_steps[i].integr_weight;
     pc_low_box->lower_steps[i].lagrange3 = plow_box->lower_steps[i].lagrange3;
     pc_low_box->lower_steps[i].lagrange4 = plow_box->lower_steps[i].lagrange4;
     for( k=0; k<ac_no_detT; k++) {
       /* beachte : auf pc_low_box werden nur values[1][] zugewiesen, denn bei Schrittannahme
	  wird mit Originalbox weitergerechnet und bei Schrittablehnung wurden zuvor values_Vektoren
	  der Sicherheitskopie von low_box von 1 nach 0 kopiert ! */
       del_vector(&(pc_low_box->lower_steps[i].values[1][k]));
       pc_low_box->lower_steps[i].values[1][k] = copy_vector( plow_box->lower_steps[i].values[0][k]);
     };
   };

   pc_low_box->next_up = plow_box->next_up;
}
/*******************************************************************************************/
void clear_int_accept( no_upper_steps, up_boxes, old_no_upper_steps, old_up_boxes)
/* nach Schrittannahme : alte upper_steps werden geloescht, sowie ueberfluessig gewordene 
                         step_inbetween-Elemente */
/* Vorraussetzung : - Gruppen sind entsprechend ihrer determ. Schaltzeit aufsteigend indiziert */

int *no_upper_steps;
step_inbetween **up_boxes;
int *old_no_upper_steps;
step_inbetween **old_up_boxes;
{
  int i,k;
  step_inbetween *ptr, *hptr, *previous;


  hptr = up_boxes[ac_no_detT-1];
  for( i=0; i<no_upper_steps[ac_no_detT-1]; i++) 
    hptr = hptr->next_down;
  /* hptr zeigt auf erstes step_inbetween-Element nach der neuen (!) up_boxes[ac_no_detT-1]-Gruppe */

  /* Loeschen aller alten upper_steps : */
  for( k=0; k<ac_no_detT; k++) {
    ptr = old_up_boxes[k];
    for( i=0; i<old_no_upper_steps[k]; i++) {
      del_vector(&(ptr->values[0][k]));
      /* in upper_steps-Boxen ist nur der values[0][k]-Vektor belegt ! */
      efree( ptr->values[0]);
      previous = ptr;
      ptr = ptr->next_down;
      efree( previous);
    };
  };
  /* beachte : ptr zeigt auf das Element, das sich der old_up_boxes[ac_no_detT-1]-Gruppe anschloss */
  /*           von hier ausgehend muessen noch die Elemente bis hptr geloescht werden */
  while (ptr != hptr) {
    for( i=0; i<ac_no_detT; i++) {
      del_vector(&(ptr->values[0][i]));
      del_vector(&(ptr->values[1][i]));
    };
    efree( ptr->values[0]);
    efree( ptr->values[1]);
    previous = ptr;
    ptr = ptr->next_down;
    efree( previous);
  };
  
}
/**********************************************************************************************/
void clear_int_reject( no_upper_steps, up_boxes, old_no_upper_steps, old_up_boxes, pold_low_box )
/* nach Schrittablehnung : die neu erzeugten upper_steps muessen wieder geloescht werden, ebenso wie evtl. gerade 
                           aus der low_box ausgeworfene step_inbetween-Elemente */
/* Vorraussetzung : - Gruppen sind entsprechend ihrer determ. Schaltzeit aufsteigend indiziert */
/* Besonderheit : nach clear_int_reject() muessen in up_boxes die auf die zu loeschenden upper_steps-Gruppen 
                  folgenden step_inbetween-Elemente dereferenzierbar sein ( fuer save_int_structures() ) */
                  
int *no_upper_steps;
step_inbetween **up_boxes;
int *old_no_upper_steps;
step_inbetween **old_up_boxes;
lower_stepbox *pold_low_box;
{
  int i,k;
  step_inbetween *ptr, *previous;

  /* Loeschen der neuen upper_steps : */
  for( k=0; k<ac_no_detT; k++) {
    ptr = up_boxes[k];
    for( i=0; i<no_upper_steps[k]; i++) {
      del_vector(&(ptr->values[0][k]));
      /* in up_boxes ist nur der values[0][k]-Vektor belegt ! */
      efree( ptr->values[0]);
      previous = ptr;
      ptr = ptr->next_down;
      efree( previous);
    };
    up_boxes[k] = ptr; /* oben erwaehnte Besonderheit */
  };

  /* Loeschen der neu hinzugekommenen step_inbetween-Elemente */
  ptr = pold_low_box->next_up->next_down;
  /* nur so kann der Fall auftreten, dass das Ende einer upper_steps-Gruppe auf neu erzeugte step_inbetween-Elemente
     der Integrationskette trifft : 
            discontinuity = 1 und low_box enhielt alle Elemente von DISC_ANZEIGE
            bis DISC_ANZEIGE und unterste Gruppe mit no_upper_steps = 1 
     oder : erste Elemente nach Diskontinuitaet wurden gerade ausgeworfen und upper_steps fuer k=0 
            erreichen das Ende der obigen Diskontinuitaet */
  if (ptr == up_boxes[0])
    /* beachte : up_boxes[0] zeigt auf Nachfolgeelement der untersten upper_steps-Gruppe, u.U. neu erzeugt */
    up_boxes[0] = pold_low_box->next_up;

  /* neu hinzugekommenen Elemente wieder freigeben : */
  while (ptr != NULL) {
    for(i=0; i<ac_no_detT; i++) {
      del_vector(&(ptr->values[0][i]));
      del_vector(&(ptr->values[1][i]));
    };
    efree( ptr->values[0]);
    efree( ptr->values[1]);
    previous = ptr;
    ptr = ptr->next_down;
    efree( previous);
  };
  pold_low_box->next_up->next_down = NULL;

  /* Pruefen, ob letztes Element frueher ein dummy war : */
  ptr = old_up_boxes[0];
  for( i=0; i<old_no_upper_steps[0]; i++) 
    ptr = ptr->next_down;
  if (ptr == pold_low_box->next_up) {
    /* da war ein dummy, denn Nachfolgeelement der untersten upper_steps-Gruppe ist zugleich letztes der 
       Integrationskette */
    for(i=0; i<ac_no_detT; i++) {
      clear_vector(ptr->values[1][i]);
      /* wird durch save_int_structures()  auch auf values[0] kopiert ! */
    };
    ptr->integr_weight = 0.0;
    /* Kennzeichen eines dummy-Elements :*/
    ptr->step = 0.0;
    ptr->indicator = ZWISCHEN_ANZEIGE;
  };
}
      

/********************************************************************************************/
double starting_step( x_end, initial_normf)
/* Rueckgabewert ist eine moeglichst optimale Schaetzung fuer die erste Schrittweite nach Hairer/Wanner:
   Solving Ordinary Differential Equations I, die in unserem Fall jedoch erst nach der Anlaufrechnung eingeht */

double x_end;
double initial_normf; /* das f(t_0,y_0) des DGL-Systems y' = f(t,y) */
{
  double den;
  
  /* Formel nach Hairer/Wanner :*/
  den = pow((1.0/x_end),6.0) + pow(fabs(initial_normf),6.0);

  return (pow((transientOptions.precision/den),(1.0/6.0)));
}
/********************************************************************************************/
double normf_evaluation( daten_basis, int_fields, stored_k )
/* wertet die rechte Seite des DGL-Systems y' = f(t,y) aus, speichert diesen Vektor fuer rk2() in stored_k[1]
   und gibt die Unendlichnorm dieses Vektors fuer starting_steps() zurueck */
/* Voraussetzung : in int_fields[i][0].aux_coefficients ist der ( meist interpolierte ) Vektor
                   der tau-Linie des aktuellen Zeitpunkts gespeichert  */

solve_basis *daten_basis;
interpolation_field *int_fields;
sp_vector *stored_k[8];
{
  int i;
  sp_vector *hp_E;
  double n_help;

  /* Formel : f(t,y) '=' p_E(t)*Q_EE + d_sum{ p_d(t,tau_d)*Delta_dE + p_d(t)*Q_dE } */
  del_vector(&(stored_k[1]));
  stored_k[1] = copy_vector( daten_basis->p_E);

  mult_vector_matrix( stored_k[1], daten_basis->Q_EE);
  for( i=0; i<ac_no_detT; i++) {
    hp_E = copy_vector( int_fields[i][0].aux_coefficients );           /* ACHTUNG !!!! */
    mult_vector_matrix( hp_E, daten_basis->detMatrices[i].Delta_dE );
    vector_add_mult( stored_k[1], hp_E, 1.0);
    del_vector(&hp_E);
    
    hp_E = copy_vector( (daten_basis->p_d)[i]);
    mult_vector_matrix( hp_E, daten_basis->detMatrices[i].Q_dE );
    vector_add_mult( stored_k[1], hp_E, 1.0);
    del_vector(&hp_E);
  };
  /* Berechnung der Unendlichnorm */
  n_help = get_infnorm( stored_k[1]);
  return( n_help);
}
/********************************************************************************************/
int preemption_cases( daten_basis)
/* Rueckgabewert : 0 fuer Faelle analog dem MD1K-Fall ohne jegliche Iteration
                   1 fuer Normalfall ( Iteration schliesst gew. DGL fuer p_e(t) mit ein !)
		   2 fuer den Fall, dass nur Randbedingung und Integration iteriert werden muessen */
/* Anmerkung : 0 -> keine Transition, die eine deterministische deaktiviert 
               1 -> mind. eine Transition, die das System durch Deaktivierung einer determ. Transition 
	            in einen Zustand ueberfuehrt, in dem nur exponentielle Transitionen aktiviert sind 
               2 -> keine Deaktivierung wie unter eins, sondern hin zu einem Zustand, in dem wieder eine 
	            deterministische Transition aktiviert ist */
    
solve_basis *daten_basis;
{
  int i,j;

  /* Q_dE zuerst pruefen -> Deaktivierung zu S_E */
  for( i=0; i<ac_no_detT; i++) {
    if (daten_basis->detMatrices[i].Q_dE->rows != NULL)
      return ( 1);
  };

  /* Q_dc untersuchen -> Deaktivierung zu S_d  */
  for( j=0; j<ac_no_detT; j++) {
    for( i=0; i<ac_no_detT; i++) {
      if ( (daten_basis->detMatrices[j].Q_dc)[i]->rows != NULL )
	return( 2);
    };
  };

  return ( 0);
}
/********************************************************************************************/
void aufbau_und_initialisierung( daten_basis, no_upper_steps, up_boxes, plow_box, interpol_update, int_fields, areas_ptr)
/* baut initiale Datenstrukturen fuer Integration (step_inbetween-Kette, plow_box), Interpolation (int_fields) und fuer 
   die Verwaltung der Impulsflaechen (areas_ptr) auf */

solve_basis *daten_basis;
int *no_upper_steps;
step_inbetween **up_boxes;
lower_stepbox *plow_box;   
int *interpol_update;
interpolation_field *int_fields;
areas *areas_ptr;
{
  int i, j;
  step_inbetween *new_instep, *dummy;


  /* erstmal dummy-Element erzeugen : */
  /* Dummy wird dann wichtig, falls Nachfolgeelement der untersten upper_steps-Gruppe eigentlich in low_box laege.
     Letztes upper_steps-Element zeigte dann genauso wie das letzte Element der step_inbetween-Kette ins Leere. Aus 
     algorithmischen Gruenden muss aber plow_box->next_up auf das eindeutige letzte Element aller step_inbetween-
     Elemente zeigen, weshalb beide obigen, ins Leere weisende Zeiger auf ein dummy-Element gebuendelt werden, worauf
     dann auch plow_box->next_up verweist */
  dummy = (step_inbetween *)emalloc(sizeof(step_inbetween));
  dummy->integr_weight = 0.0;

  /* Kennzeichen eines dummy-Elements : */
  dummy->step = 0.0;
  dummy->indicator = ZWISCHEN_ANZEIGE;

  dummy->values[0] = (sp_vectors *)emalloc(ac_no_detT*sizeof(sp_vectors));
  dummy->values[1] = (sp_vectors *)emalloc(ac_no_detT*sizeof(sp_vectors));
  for(j=0; j<ac_no_detT; j++) {
    dummy->values[0][j] = new_vector();
    dummy->values[1][j] = new_vector();
  };
  dummy->next_down = NULL;


  for( i=0; i<ac_no_detT; i++) {
    
    no_upper_steps[i] = 2; /* anfaenglich besteht jede upper_steps-Gruppe aus zwei Elementen :
			      - oberstes Element auf der tau-Linie
			      - DISC_ANZEIGE-Element kurz vor der Diskontinuitaet 0.0 
			      ( darauf folgt fuer alle das dummy-Element )                   */

    up_boxes[i]  = (step_inbetween *)emalloc(sizeof(step_inbetween));
  
    up_boxes[i]->step = daten_basis->detMatrices[i].det_firing_time;
    up_boxes[i]->integr_weight = 0.0;
    up_boxes[i]->indicator = ZWISCHEN_ANZEIGE;
    up_boxes[i]->values[0] = (sp_vectors *)emalloc(ac_no_detT*sizeof(sp_vectors));
    up_boxes[i]->values[1] = (sp_vectors *)emalloc(ac_no_detT*sizeof(sp_vectors));
    for(j=0; j<ac_no_detT; j++) {
      up_boxes[i]->values[0][j] = new_vector();
      up_boxes[i]->values[1][j] = new_vector();
    };   
    up_boxes[i]->next_down = (step_inbetween *)emalloc(sizeof(step_inbetween));
    new_instep = up_boxes[i]->next_down;

    new_instep->step = 0.0;
    new_instep->integr_weight = 0.0;
    new_instep->indicator = DISC_ANZEIGE;
    new_instep->values[0] = (sp_vectors *)emalloc(ac_no_detT*sizeof(sp_vectors));
    new_instep->values[1] = (sp_vectors *)emalloc(ac_no_detT*sizeof(sp_vectors));
    for(j=0; j<ac_no_detT; j++) {
      new_instep->values[0][j] = new_vector();
      new_instep->values[1][j] = new_vector();
    };   
    new_instep->next_down = dummy;
  };


  /* low_box nullen.. */
  for(j=0; j<5; j++) {
      plow_box->lower_steps[j].step = 0.0; 
      plow_box->lower_steps[j].integr_weight = 0.0;
      plow_box->lower_steps[j].lagrange3 = 0.0;      
      plow_box->lower_steps[j].lagrange4 = 0.0;
      plow_box->lower_steps[j].values[0] = (sp_vectors *)emalloc(ac_no_detT*sizeof(sp_vectors));
      plow_box->lower_steps[j].values[1] = (sp_vectors *)emalloc(ac_no_detT*sizeof(sp_vectors));
      for(i=0; i<ac_no_detT; i++) {
	plow_box->lower_steps[j].values[0][i] = new_vector();
	plow_box->lower_steps[j].values[1][i] = new_vector();
      };   
    };
  plow_box->next_up = dummy;

  /* initiales Interpolationsfeld besteht nur aus Nullvektoren bis nach der ersten Diskontinuitaet (!= 0.0) */
  for( j=0; j<ac_no_detT; j++) {

    interpol_update[j] = 1;
    
    int_fields[j][0].int_point = (-1.0)*(daten_basis->detMatrices[j].det_firing_time);
    int_fields[j][1].int_point = ((-1.0)*(daten_basis->detMatrices[j].det_firing_time))/3.0;
    int_fields[j][2].int_point = ((-2.0)*(daten_basis->detMatrices[j].det_firing_time))/3.0;
    int_fields[j][3].int_point = (-1.0)*(daten_basis->detMatrices[j].det_firing_time);
    for(i=0;i<4;i++) {
      int_fields[j][i].y_values = new_vector();
      int_fields[j][i].pol_coefficients = new_vector();
      int_fields[j][i].aux_coefficients = new_vector();
    };
    /* damit ist auch in int_fields[j][0].aux_coefficients der aktuelle, interpolierte Wert gespeichert */
    
    /* wichtig bei Impulsflaechen-Vektoren : fuer jede erreichte Diskontinuitaet wird ein Vektor 
       (evtl. Nullvektor) ueber die vector->next-Strukturkomponente angehaengt; bei Operationen mit
       diesen Impulsflaechenvektoren (z.B. Randomisierung) darf diese Verkettung nicht geloest werden */
    areas_ptr[j].first = copy_vector( (daten_basis->p_d)[j]);
    areas_ptr[j].last = areas_ptr[j].first;
  };
}
/************************************************************************************************/
void conf_initialisierung( conf_no_upper_steps, conf_up_boxes, pconf_low_box, conf_int_fields, up_boxes, careas_ptr)
/* minimale Initialisierung der Sicherheitskopien fuer Datenstrukturen, da diese nach Anlaufrechnung
   sowieso nochmals erneuert werden muessen, denn erst danach koennen Schritte abgelehnt werden  */

int *conf_no_upper_steps;
step_inbetween **conf_up_boxes;
lower_stepbox *pconf_low_box;
interpolation_field *conf_int_fields;
step_inbetween **up_boxes;
areas *careas_ptr;
{
  int i,j;

  for( i=0; i<ac_no_detT; i++) {
    conf_no_upper_steps[i] = 2;    
    conf_up_boxes[i] = up_boxes[i];
  };

  for(j=0; j<5; j++) {
      pconf_low_box->lower_steps[j].values[0] = (sp_vectors *)emalloc(ac_no_detT*sizeof(sp_vectors));
      pconf_low_box->lower_steps[j].values[1] = (sp_vectors *)emalloc(ac_no_detT*sizeof(sp_vectors));
      for(i=0; i<ac_no_detT; i++) {
	pconf_low_box->lower_steps[j].values[0][i] = new_vector();
	pconf_low_box->lower_steps[j].values[1][i] = new_vector();
      };   
    };
  /* auch next_up muss neben step,integr_weight,lagrange3,lagrange4 nicht initialisiert werden */

  for( j=0; j<ac_no_detT; j++) {
    for(i=0;i<4;i++) {
      conf_int_fields[j][i].y_values = new_vector();
      conf_int_fields[j][i].pol_coefficients = new_vector();
      conf_int_fields[j][i].aux_coefficients = new_vector();
    };
    careas_ptr[j].first = new_vector();
    careas_ptr[j].last = careas_ptr[j].first;
  };
}  
/***********************************************************************************************/
void integralvervollstaendigung_a( daten_basis, incomplete_pdts, plow_box, add_where)
/* addiert zu den unvollstaendigen Werten fuer p_d(t) noch die neueste Randbedingung multipliziert mit
   dem zugehoerigen Integrationsgewicht und legt das Ergebnis in daten_basis->p_d ab */
/* Voraussetzung : Randbedingung an der Stelle `add_where' in low_box */

solve_basis *daten_basis;
sp_vector **incomplete_pdts;
lower_stepbox *plow_box;
int add_where;
{
  int i;

  for( i=0; i<ac_no_detT; i++) {
    del_vector( &(daten_basis->p_d[i]));
    daten_basis->p_d[i] = copy_vector( incomplete_pdts[i]);
    vector_add_mult( (daten_basis->p_d)[i], plow_box->lower_steps[add_where].values[0][i], plow_box->lower_steps[add_where].integr_weight);
  };
}
/********************************************************************************************/
void integralvervollstaendigung_b( daten_basis, incomplete_pdts, plow_box)
/* addiert zu den unvollstaendigen Werten fuer p_d(t) noch die neueste Randbedingung multipliziert mit
   dem zugehoerigen Integrationsgewicht und legt das Ergebnis in daten_basis->p_d ab */
/* Voraussetzung : Randbedingung in low_box->next_up-Element ( Aufruf nur vor Diskontinuitaet )*/

solve_basis *daten_basis;
sp_vector **incomplete_pdts;
lower_stepbox *plow_box;
{
  int i;

  for( i=0; i<ac_no_detT; i++) {
    del_vector( &(daten_basis->p_d[i]));
    daten_basis->p_d[i] = copy_vector( incomplete_pdts[i]);
    vector_add_mult( (daten_basis->p_d)[i], plow_box->next_up->values[0][i], plow_box->next_up->integr_weight);
  };
}
/**********************************************************************************************/
void boundary_check_a( piteration, new_boundaries, plow_box, add_where) 
/* liefert 0, falls relativen Abweichungen der neu ermittelten Randbedingungen von den vorher ermittelten 
   (bzw. extrapolierten) in jeder Komponente kleiner als transientOptions.precision sind, sonst 1 */
/* Voraussetzung : Randbedingung an der Stelle `add_where' in low_box */

int *piteration;
sp_vectors *new_boundaries;
lower_stepbox *plow_box;
int add_where;
{
  int i, epsilon_check;

  for(i=0; i<ac_no_detT; i++) {
    epsilon_check = rel_comp_vector( plow_box->lower_steps[add_where].values[0][i], new_boundaries[i], transientOptions.precision);
    if (!epsilon_check) {
      (*piteration)++;
      if (*piteration > 50) {
	printf("WARNING: no convergence in border conditions !");
	*piteration = 0;
	return;
      };
      *piteration = 1;
      return;
    };
  };
  *piteration = 0;
  return;
}
/*********************************************************************************************/
void boundary_check_b( piteration, new_boundaries, plow_box) 
/* liefert 0, falls Abweichungen der neu ermittelten Randbedingungen von den vorher ermittelten (bzw. extrapolierten)
   in jeder Komponente kleiner als transientOptions.precision sind, sonst 1 */
/* Voraussetzung : Randbedingung in low_box->next_up-Element ( Aufruf nur vor Diskontinuitaet )*/

int *piteration;
sp_vectors *new_boundaries;
lower_stepbox *plow_box;
{
  int i, epsilon_check;

  for(i=0; i<ac_no_detT; i++) {
    epsilon_check = rel_comp_vector( plow_box->next_up->values[0][i], new_boundaries[i], transientOptions.precision);
    if (!epsilon_check) {
      (*piteration)++;
      if (*piteration > 50) {
	printf("WARNING: no convergence in border conditions !");
	*piteration = 0;
	return;
      };
      *piteration = 1;
      return;
    };
  };
  *piteration = 0;
  return;
}
/*********************************************************************************************/
void set_boundaries_a( new_boundaries, plow_box, add_where)
/* haengt die Zeiger auf die neu ermittelten Randbedingungen an entsprechende Stelle in der low_box */
/* Voraussetzung : alte Randbedingung an der Stelle `add_where' in low_box */

sp_vectors *new_boundaries;
lower_stepbox *plow_box;
int add_where;
{
  int i;

  for( i=0; i<ac_no_detT; i++) {
    del_vector( &(plow_box->lower_steps[add_where].values[0][i]));
    plow_box->lower_steps[add_where].values[0][i] = new_boundaries[i];
    new_boundaries[i] = NULL;
  };
}
/********************************************************************************************/
void set_boundaries_b( new_boundaries, plow_box)
/* haengt die Zeiger auf die neu ermittelten Randbedingungen an step_inbetween-Element ueber low_box */
/* Voraussetzung : Randbedingung in low_box->next_up-Element ( Aufruf nur vor Diskontinuitaet )*/
/* ACHTUNG : wird falsch, falls dummy existiert ! */

sp_vectors *new_boundaries;
lower_stepbox *plow_box;
{
  int i;

  for( i=0; i<ac_no_detT; i++) {
    del_vector( &(plow_box->next_up->values[0][i]));
    plow_box->next_up->values[0][i] = new_boundaries[i];
    new_boundaries[i] = NULL;
  };
}
/********************************************************************************************/
int nonexp_initialization( daten_basis)
/* ermittelt zu Beginn der Rechnung, ob anfangs eine deterministische Transition initialisiert ist 
   falls ja -> Rueckgabewert gleich 1 */

solve_basis *daten_basis;
{
  double exp_sum;

  exp_sum = get_sum_components( daten_basis->p_E);
  if  ( fabs( exp_sum-1.0) < EXT_PRECISION)
	return( 0);
  return( 1);
}

/********************************************************************************************/
void addall( p_t, daten_basis)
/* addiert alle Wahrscheinlichkeitsvektoren in daten_basis zu einem Vektor auf   */
/* Bemerkung : wird nur fuer die Kommunikation mit Datenfiles benoetigt */
 
sp_vector *p_t;
solve_basis *daten_basis;
{
  int i;
  
  clear_vector( p_t);
  vector_add_mult( p_t, daten_basis->p_E, 1.0);
  for(i=0; i<ac_no_detT; i++) 
    vector_add_mult(p_t, (daten_basis->p_d)[i], 1.0);
}
/********************************************************************************************/
void reorganise_daten( c, daten_basis)
/* organisiert die vorgegebene Datenstruktur, in z.B. der alle Uebergangsmatrizen gespeichert sind, derart um, dass 
   - eine Gruppe mit kleinerer det. Schaltzeit auch einen kleineren Index erhaelt (bei der Umgruppierung wird 
     nicht kopiert, sondern Zeiger werden entsprechend gesetzt )
   - durch Zuweisung von new_vector() bzw. new_matrix() keine NULL-Zeiger mehr auftauchen */   

solve_basis *c;
solve_basis *daten_basis;
{
  int i,j;
  int *is_at;

  is_at = (int *)emalloc(c->no_detT*sizeof(int));

  for(i=0; i<c->no_detT; i++)
    is_at[i] = 0;
  for(i=0; i<c->no_detT; i++) {
    for(j=i+1; j<c->no_detT; j++) {
      if( c->detMatrices[i].det_firing_time <= c->detMatrices[j].det_firing_time)
	(is_at[j])++;
      else
	(is_at[i])++;
    };
  };

  daten_basis->Q_EE = (c->Q_EE != NULL) ? (c->Q_EE) : new_matrix(); /* = NULL, falls keine rein exponentiellen
								       Zustaende vorhanden ( preemption > 0 ) */
   
  daten_basis->Q_Ed = (sp_matrices *)emalloc(c->no_detT*sizeof(sp_matrices));
  for( i=0; i<c->no_detT; i++) 
    (daten_basis->Q_Ed)[is_at[i]] = ((c->Q_Ed)[i] != NULL) ? (c->Q_Ed)[i] : new_matrix();

  daten_basis->no_detT = c->no_detT;


  daten_basis->detMatrices = (sp_detMatrices *)emalloc((daten_basis->no_detT)*sizeof(sp_detMatrices));
  for( i=0; i<daten_basis->no_detT; i++) {

        daten_basis->detMatrices[is_at[i]].det_firing_time = c->detMatrices[i].det_firing_time;   

  	daten_basis->detMatrices[is_at[i]].Q_d = (c->detMatrices[i].Q_d != NULL) ? c->detMatrices[i].Q_d : new_matrix();
  	daten_basis->detMatrices[is_at[i]].Q_dE = (c->detMatrices[i].Q_dE != NULL) ? c->detMatrices[i].Q_dE : new_matrix();  
	daten_basis->detMatrices[is_at[i]].Delta_dE = (c->detMatrices[i].Delta_dE != NULL) ? c->detMatrices[i].Delta_dE : new_matrix();
 	daten_basis->detMatrices[is_at[i]].Q_dc = (sp_matrices *)emalloc(daten_basis->no_detT*sizeof(sp_matrices));
	daten_basis->detMatrices[is_at[i]].Delta_dc = (sp_matrices *)emalloc((daten_basis->no_detT)*sizeof(sp_matrices));
	for( j=0; j<daten_basis->no_detT; j++)  {
 		(daten_basis->detMatrices[is_at[i]].Q_dc)[is_at[j]] = (c->detMatrices[i].Q_dc[j] != NULL) ? c->detMatrices[i].Q_dc[j] : new_matrix();
		(daten_basis->detMatrices[is_at[i]].Delta_dc)[is_at[j]] = (c->detMatrices[i].Delta_dc[j] != NULL) ? c->detMatrices[i].Delta_dc[j] : new_matrix();
	};
  };

  daten_basis->no_markings = c->no_markings; 
  daten_basis->p_E = (c->Q_EE != NULL) ? copy_vector( c->p_E) : new_vector();   /* nicht nur Zeiger umgehaengt ! */

  daten_basis->p_d = (sp_vectors *)emalloc((daten_basis->no_detT)*sizeof(sp_vectors));
  for( i=0; i<daten_basis->no_detT; i++)  
	(daten_basis->p_d)[is_at[i]] = copy_vector( (c->p_d)[i]);
}
  
/********************************************************************************************/
double global_correcting( daten_basis )
/* liefert einen Korrekturfaktor, um den die vorgeschlagene, neue Schrittweite verkleinert werden sollte, um 
   angemessen zu beruecksichtigen, dass durch numerische Integration und Interpolation Fehler in die Rechnung
   eingeschleppt werden koennen, die mit der Schrittweitensteuerung des Dormand-Prince-Verfahrens micht erfasst 
   werden koennen. Ein zu grosser Korrekturfaktor (>=4.5), der heuristisch aus der Abweichung der Summe aller Wahr-
   scheinlichkeiten von 1.0 bestimmt wird, kann der ganze Schritt auch abgelehnt werden. */
/* Bemerkung : off_sum als globale Variable */

solve_basis *daten_basis;
{
  int i;
  double ideal_1;

  off_sum += LOW_PRECISION;

  ideal_1 = get_sum_components( daten_basis->p_E);
  for( i=0; i<ac_no_detT; i++) 
    ideal_1 += get_sum_components( daten_basis->p_d[i]);

  if(fabs(ideal_1 - 1.0)> off_sum) {
    ideal_1 = min(((fabs(ideal_1 - 1.0))/off_sum),4.5);
    return (ideal_1);
  };
  return( 1.0);
}
/********************************************************************************************/
void adaptive_stepcontrol( tnethome, modelpath, netname, c)
/* Aufruf aus frontend.c : adaptive_stepcontrol ist die zentrale Funktion fuer das numerische Loesungs-
   verfahren mit variabler Schrittweite fuer stochastische Petrinetze, bestehend aus dem Dormand-Prince-
   Verfahren fuenfter Ordnung, Randomisierung, numerischer Integration und Interpolation angemessener Ordnung.
   Zunaechst nur fuer DSPNs implementiert, laesst es sich begrenzt auch fuer allgemein verteilte Schaltzeiten
   erweitern. */

char *tnethome;
char *modelpath;
char *netname;
solve_basis *c;
{
  FILE *ftmp;           /* Datei zum Ablegen der errechneten Daten und fuer Kommunikation mit graphischer Ausgabe */
               
  /* Anmerkung zur log_kosten()-Funktion :
     Speicherplatzbedarf und Zeitkosten werden fuer den Algorithmus anteilsmaessig erfasst und wie folgt gespeichert :
     0 : Erzeugen des RRG
     1 : initiale Speicherallozierung (emalloc()'s), Berechnung der A-Matrizen fuer Randomisierung,
         Ablegen der errechneten Daten und dafuer notwendige Rechnungen ( addall(), update_solve() etc.),
	 sonstige Verwaltungsarbeiten fuer Datenstrukturen ( insbes. save_int_structures(), clear_int_reject() etc. )
     2 : Randomisierung ( ausser falls fuer Interpolationsfelder )
     3 : numerische Integration
     4 : Loesung der GDGL mit rk2(), rk3() und dopri5()
     5 : Randbedingung , Extrapolation
     6 : Verwaltung der Interpolationsfelder ( tau_densities, probabilities )
     Bemerkung : Kosten fuer eine Iteration bei preemption > 0 werden ebenfalls aufgegliedert                 */

  int i, j;                          /* Laufindices */
  int accept_step = 1;               /* zeigt an, ob gerade zuvor ein Schritt abgelehnt wurde (accept_step = 0) */
  int new_io_2 = 0;                  /* muss das Interpolationsfeld fuer p_d(t) ganz neu aufgebaut werden statt
					nur die zwei aeltesten Stuetzpunkte zu ersetzen ( normalerweise wird immer
					abwechselnd ersetzt und neu aufgebaut, wobei das Ersetzen wegen der Newton-
					Interpolation erheblich effektiver ist */
  int repeated = 0;                  /* zeigt an, ob wegen zu schwacher Konvergenz der Randbedingungen mit den neu
					ermittelten Werten die letzten Schritte nochmals wiederholt werden sollen,
					nur bei deaktivierten deterministischen Transitionen (preemption>0) */
  int preemption;                    /* bei deaktivierten deterministischen Transitionen > 0, wird durch 
					preemption_cases() ermittelt und bestimmt, welche Schritte bei notwendiger
					Wiederholung wiederholt werden muessen */
  int finish_loops = 0;              /* Ausstieg aus der Rechnung, falls transientOptions.transient_time erreicht wurde */
  int noexp = 0;                     /* gibt an, ob ueberhaupt Zustaende existieren, in denen nur exponentielle 
					Transitionen aktiviert sind ( dann = 0), falls 1, muss preemption > 0 sein 
					und rk2(), rk3() und dopri5() koennen immer uebersprungen werden */
  int init;                          /* = 1, falls deterministische Transitionen initial aktiviert sind */
  int numDatas;                      /* Zaehler fuer die Zeitpunkte, zu denen Werte errechnet wurden (fuer Datenablage)*/
  int show_discontinuity;            /* zeigt an, ob mit folgendem Schritt die naechste Diskontinuitaet erreicht wird */ 
  int add_where;                     /* weist die Stelle in low_box, an die die Randbedingungen angehaengt werden muessen*/
  int done_enough = 0;               /* zaehlt die Schrittablehnungen aufgrund eines zu grossen Korrektors, die auf
					maximal drei zwischen zwei Diskontinuitaeten beschraenkt sind, um Oszillationen
					des Interpolationspolynoms aufgrund zu dichter Stuetzpunkte zu vermeiden */
  double corrector;                  /* Korrekturfaktor fuer vorgeschlagene Schrittweite, der eingeschleppte Fehler
					durch num. Integration und Interpolation beruecksichtigen soll */
  double just_for_debug;             /* fuer Ermittlung der Summe aller Wahrscheinlichkeiten als Mass fuer Guete
					der Berechnungen ( nur bei gewaehlter DEBUG_AC-Option ) */
  double maxX;                       /* = transientOptions.transient_time - transientOptions.precision; noetig, weil
				        transientOptions.transient_time sich evtl. im Laufe der Rechnung entsprechend
					verkleinert, aber maximaler x_Wert bei update_solve() konstant bleiben muss */ 
  double h_max;                      /* maximale Schrittweite zwischen zwei Diskontinuitaeten, soll garantieren, dass
					immer zwei innere Stuetzpunkte zwischen zwei Diskontinuitaeten existieren, um
					Interpolationspolynome dritter Ordnung bilden zu koennen. */
  double next_discontinuity = 0.0;   /* dort wird als naechstes zu erreichende Diskontinuitaet gespeichert */
  

  solve_basis daten;                 /* dort werden an adaptive_stepcontrol() uebergebene Daten in fuer den Algorithmus
					notwendiger Form gespeichert */
  
  double transient_time = 0.0;       /* laufende Zeit t */
  double normvf;                     /* Wert fuer die Unendlichnorm von rechter Seite des GDGL-Systems zu Beginn jeder 
					Anlaufrechnung, dient zur Bestimmung der Anfangsschrittweite */
  double err_dopri;                  /* Fehler aus dopri5(), um ueber Schrittannahme oder -ablehnung zu entscheiden */
  double fac;                        /* errechnet sich aus err_dopri und dient zur Schrittweitensteuerung : */  
  double h_new, hh_new;              /* alte und neue Schrittweite : hh_new = h_new/fac */
  double rk2_step, rk3_step;         /* Schrittweiten fuer Anlaufrechnung (aus zwei Schritten, um Voraussetzungen zu
					schaffen, damit mit genuegend hoher Ordnung interpoliert und integriert werden 
					kann */
  double starting_steps;             /* = rk2_step + rk3_step */
  
  ConcernsRandomisation **randdata;  /* dort werden alle zur Randomisierung benoetigten Daten gespeichert, wie
					A-Matrix, maxq_ii, L, R */

  tnode **root;                      /* (*root) zeigt auf binaeren Baum, in dem die zur Laufzeit ermittelten 
					Diskontinuitaetszeitpunkte gespeichert werden */
  tnode **max_element;               /* (*max_element) zeigt aus Effektivitaetsgruenden auf momentan groesste 
					Diskontinuitaet im binaeren Baum */

  lower_stepbox *low_box;            /* low_box, conf_low_box und lbox_ptr zeigen alle auf eine Datenstruktur, in der */
  lower_stepbox *conf_low_box;       /* die letzten Integrationsstuetzpunkte gespeichert werden, bis feststeht aus    */
  lower_stepbox *lbox_ptr;           /* wieviel Stuetzpunkten die naechste Integrationsgruppe (3/4) besteht. low_box  */
                                     /* weist auf die zu bearbeitende Struktur, conf_low_box auf ihre Sucherheitskopie*/
                                     /* und lbox_ptr dient nur zum Vertauschen beider Zeiger */

  sp_vector *stored_k[8];            /* dient zur Speicherung der ki's der Runge-Kutta-Methoden, insbesondere ihrer 
                                        bei einer notwendigen Iteration (preemption>0) invarianten Anteile. Ausserdem
					geht bei dopri5() nach einer Schrittannahme das alte k7 als neues k1 fuer den
					folgenden Schritt ein */
  sp_vector *hp_E = new_vector();    /* Sicherheitskopie fuer daten_basis->p_E */
  sp_vector *p_t = new_vector();     /* Vektor aller Wahrscheinlichkeiten fuer Datenablage */
  sp_vectors *bound_results;         /* zur vorlaeufigen Speicherung neu ermittelter Randbedingungen 
					( nur bei preemption > 0 ) */
  areas *areas_ptr;                  /* Zeiger auf Struktur, deren Zeiger auf Anfang und Ende der Impulsflaechenvektoren */
  areas *careas_ptr;                 /* verweisen (careas_ptr ist Zeiger auf entsprechende Sicherheitskopie ) */

  step_inbetween **up_boxes;         /* enthaelt die Zeiger auf die obersten Elemente der upper_steps-Gruppen */
  step_inbetween **conf_up_boxes;    /* Sicherheitskopien dieser Zeiger */
  int *no_upper_steps;               /* Anzahl der Elemente der einzelnen upper_steps-Gruppen und ihrer Sicherheitskopien*/
  int *conf_no_upper_steps;          /* Je nachdem, wohin die tau-Linie einer Gruppe bezogen auf step_inbetween-Kette 
					faellt, wird eine neue erste Integrationsgruppe gebildet, deren Integrations-
					gewichte von den urspruenglichen abweichen. Das erst macht es notwendig, fuer
					jede Gruppe eine Anzahl eigener step_inbetween-Elemente (upper_steps) immer wieder
					zu bilden */ 

  sp_vector **incomplete_pdts;        /* unvollstaendiges Ergebnis fuer daten.p_d aus der numerischen Integration in
					 update_int_structures(); letzter Stuetzpunkt (Randbedingung) fehlt noch */

  interpolation_field *tau_densities; /* Interpolationsfelder fuer die p_d(t,tau_d)-Werte entlang der tau-Linien */
  interpolation_field *conf_tau_densities; /* Sicherheitskopien der tau_densities */
  int *interpol_update;               /* zeigt an, ob Interpolationsfelder 'tau_densities' durch Ersetzen von ein oder 
					 zwei Stuetzpunkten noch moeglich ist (effektiv, da Newton-Interpolation) oder
					 ob das Interpolationsfeld voellig neu aufgebaut werden muss (analog new_io_2) */
  int *conf_interpol_update;          /* Sicherheitskopie fuer interpol_update */

  interpolation_field *probabilities; /* Interpolationsfelder fuer die p_d(t)-Werte (keine Sicherheitskopie notwendig, 
					 da immer nur die letzte Neueintragung abgelehnt werden kann); haben nur 
					 Bedeutung, falls preemption > 0, sonst nur aus new_vectors() bestehend */
  
  /* Speicherplatzallozierung fuer alle moeglichen Datenstrukturen : */
  root = (tnode **)emalloc(sizeof(tnode *));
  max_element = (tnode **)emalloc(sizeof(tnode *));

  reorganise_daten( c, &daten);
  ac_no_detT = daten.no_detT;

  areas_ptr = (areas *)emalloc( ac_no_detT*sizeof( areas));
  careas_ptr = (areas *)emalloc( ac_no_detT*sizeof( areas));

  low_box = (lower_stepbox *)emalloc( sizeof(lower_stepbox));
  conf_low_box = (lower_stepbox *)emalloc( sizeof(lower_stepbox));
  
  up_boxes = (step_inbetween **)emalloc( ac_no_detT*sizeof(step_inbetween *));
  conf_up_boxes = (step_inbetween **)emalloc( ac_no_detT*sizeof(step_inbetween *));
  no_upper_steps = (int *)emalloc( ac_no_detT*sizeof(int));
  conf_no_upper_steps = (int *)emalloc( ac_no_detT*sizeof(int));

  tau_densities = (interpolation_field *)emalloc( ac_no_detT*sizeof(interpolation_field));
  conf_tau_densities = (interpolation_field *)emalloc( ac_no_detT*sizeof(interpolation_field));
  interpol_update = (int *)emalloc( ac_no_detT*sizeof(int));
  conf_interpol_update = (int *)emalloc( ac_no_detT*sizeof(int));
  
  probabilities = (interpolation_field *)emalloc( ac_no_detT*sizeof(interpolation_field));
  
  for( i=0; i<ac_no_detT; i++) {
    probabilities[i][0].int_point = daten.detMatrices[0].det_firing_time;
    probabilities[i][1].int_point = daten.detMatrices[0].det_firing_time/3.0;
    probabilities[i][2].int_point = (2.0*(daten.detMatrices[0].det_firing_time))/3.0;
    probabilities[i][3].int_point = daten.detMatrices[0].det_firing_time;
    for( j=0; j<4; j++) {
      probabilities[i][j].y_values = new_vector();
      probabilities[i][j].pol_coefficients = new_vector();
      probabilities[i][j].aux_coefficients = new_vector();
    };
  };
  for(i=0; i<8; i++) 
    stored_k[i] = new_vector();

  bound_results = (sp_vectors *)emalloc( ac_no_detT*sizeof(sp_vectors));
  incomplete_pdts = (sp_vector **)emalloc( ac_no_detT*sizeof(sp_vector *));
  randdata = (ConcernsRandomisation **)emalloc(ac_no_detT*sizeof(ConcernsRandomisation *));

  for( i=0; i<ac_no_detT; i++) {
    bound_results[i] = NULL;
    incomplete_pdts[i] = new_vector();

    randdata[i] = (ConcernsRandomisation *)emalloc(sizeof(ConcernsRandomisation));
    randdata[i]->A = copy_matrix(daten.detMatrices[i].Q_d);    
    /* Vorbereitung fuer Randomisierung : - A_Matrix = Einheitsmatrix + (Q_d/(1.02*max_el(Q_d)))
                                          - qmax_ii ist 1.02*max_element(Q_d)                  */
    randdata[i]->qmax_ii = calculate_A_Matrix( randdata[i]->A); 
  };
  /* Initialisierungen gelten nach Diskontinuitaeten-Uebertragung */
  EXT_PRECISION = transientOptions.precision*0.011;                  /* fuer erhoehte Genauigkeitsanforderungen */
  LOW_PRECISION = transientOptions.precision*((float)(ac_no_detT+1));/* fuer Schaetzung des zulaessigen globalen Fehlers*/
  off_sum = transientOptions.precision*5.0;                          /* tolerierte Abweichung nach erster Anlaufrechnung*/
  rk2_step = pow( transientOptions.precision, 0.55);   /* einmalige Bestimmung der Anlaufschritte, deren Groesse sich */
  rk3_step = pow( transientOptions.precision, 0.4);    /* aus der niedrigeren Ordnung ergeben */
  starting_steps = rk2_step + rk3_step;               

  preemption = preemption_cases( &daten);     /* welche Art der deaktivierten det. Transition liegt vor ? */
  init = nonexp_initialization( &daten);      /* sind det. Transitionen initialisiert ? */
  if (daten.Q_EE->rows == NULL)               /* gibt u.U. gar keine Zustaende, in denen nur exp. Transitionen */ 
    noexp = 1;                                /* aktiviert sind ? */

  /* Aufbau und Initialisierung aller Datenstrukturen */
  aufbau_und_initialisierung( &(daten), no_upper_steps, up_boxes, low_box, interpol_update, tau_densities, areas_ptr);
  conf_initialisierung( conf_no_upper_steps, conf_up_boxes, conf_low_box, conf_tau_densities, up_boxes, careas_ptr);
  /* conf_interpol_update[] braucht nicht initialisiert werden ! */

  set_border( &daten);                        /* setzt Obergrenze, ueber die hinaus keine Diskontinuitaeten 
						 beruecksichtigt werden brauchen */
  build_up_tree( &daten, root, max_element);
  h_max = (get_next_discontinuity( &daten, &next_discontinuity, root, max_element))/2.0;
  /* weil immer zwei Anlaufschritte, die nicht zu Diskontinuitaet fuehren, gemacht werden, ist so schon sichergestellt,
     dass mindestens zwei innere Stuetzpunkte zwischen zwei Diskontinuitaeten existieren */

  numDatas = 1;
  maxX = transientOptions.transient_time - transientOptions.precision; /* maxX bleibt dann unveraendert */
  ftmp = open_solve( tnethome, modelpath, netname, DATA_BUFFER, maxX); /* oeffnet temporaeres Datenfile */
  reset_digits();
  addall( p_t, &daten);                                                /* summiert alle Wahrscheinlichkeiten */
  update_solve( tnethome, modelpath, netname, ftmp, daten.no_markings, p_t, transient_time, maxX);

#ifdef DEBUG_AC
  show_vector( daten.p_E);
  show_matrix( daten.Q_EE);
  for( i=0; i<ac_no_detT; i++) {
    show_vector( (daten.p_d)[i]);
    show_matrix( (daten.Q_Ed)[i]);
    show_matrix( daten.detMatrices[i].Q_d);
    show_matrix( daten.detMatrices[i].Q_dE);
    show_matrix( daten.detMatrices[i].Delta_dE);
    for( j=0; j<ac_no_detT; j++) {
      show_matrix( (daten.detMatrices[i].Q_dc)[j]);
      show_matrix( (daten.detMatrices[i].Delta_dc)[j]);
    };
  };

  printf(" preemption = %d\n", preemption);
  printf(" init = %d\n", init);
  printf(" noexp = %d\n", noexp);

  just_for_debug = get_sum_components( p_t);
  printf(" p_t-Summe : %f\n", just_for_debug);
  printf(" rk2_step : %f\n", rk2_step);
  printf(" rk3_step : %f\n", rk3_step);
#endif

#ifdef KOSTEN
  log_kosten(1);
#endif

/* ausserste Schleife fuer Rechnung zwischen den Diskontinuitaeten */
for( ; ; ) {

#ifdef DEBUG_AC
  if (h_max < LOW_PRECISION) 
    printf("\n WARNING : h_max ist kleiner als 5.0*transientOptions.precision \n");
#endif

  /* Programmhygiene : in boundary_conditions() werden diese Vektoren nicht geloescht */
  for( i=0; i<ac_no_detT; i++)
    del_vector(&(low_box->lower_steps[0].values[0][i]));
  boundary_conditions( &daten, low_box->lower_steps[0].values[0], tau_densities);
#ifdef KOSTEN
  log_kosten(5);
#endif

  normvf = normf_evaluation( &daten, tau_densities, stored_k);
#ifdef DEBUG_AC
  printf(" normvf : %.9f\n", normvf);
#endif
  h_new = starting_step(next_discontinuity - transient_time , normvf);
  if (h_new > h_max) {
    h_new = h_max;
  };
  if ((transient_time + h_new) > transientOptions.transient_time) {
    h_new = transientOptions.transient_time - transient_time;
    finish_loops = 1;      /* falls spaeter rk2_step + rk3_step < h_new setze finish_loops wieder auf 0 */
  };

  hh_new = h_new;          /* wichtig, um gegebenenfalls finish_loops zurueckzusetzen, falls h_new < hh_new */
  show_discontinuity = 0;  /* naechste Diskontinuitaet kann mit Anlaufrechnung nicht erreicht werden */
#ifdef DEBUG_AC
  printf(" h_new : %f\n", h_new);
#endif

  if (noexp) {
    rk2_step = h_new/20.0;
    rk3_step = h_new/15.0;
    h_new = rk2_step + rk3_step;
    /* starting_steps = h_new; starting_steps spielt in diesem Fall keine Rolle */
#ifdef DEBUG_AC
    printf(" rk2_step : %f\n", rk2_step);
    printf(" rk3_step : %f\n", rk3_step);
#endif
  }
  else {
    if (starting_steps > h_new) {
      /* starting_steps duerfen nicht groesser als ermittelte Anfangsschrittweite sein */
      rk2_step *= h_new/starting_steps;
      rk3_step *= h_new/starting_steps;
      starting_steps = h_new;
#ifdef DEBUG_AC
    printf(" rk2_step : %f\n", rk2_step);
    printf(" rk3_step : %f\n", rk3_step);
#endif
    }
    else
      h_new = starting_steps;
  };
  /* es gilt : h_new = rk2_step + rk3_step ,aber h_new unter Umstaenden verkleinert*/
#ifdef KOSTEN
  log_kosten(1);
#endif
   
  /* update_int_structures() enthaelt Randomisierungsschritt und teilweise numerische Integration fuer p_d(t) */
  add_where = update_int_structures( show_discontinuity, rk2_step, preemption, init, no_upper_steps, conf_no_upper_steps, up_boxes, conf_up_boxes, low_box, interpol_update, tau_densities, incomplete_pdts, randdata, areas_ptr);
  
  if (preemption) {
    
    empty_a( rk2_step, &daten, probabilities);
#ifdef KOSTEN
    log_kosten(6);
#endif
    
    do {
      integralvervollstaendigung_a( &daten, incomplete_pdts, low_box, add_where);
#ifdef KOSTEN
      log_kosten(3);
#endif
      /* probabilities updaten ; */
      a_ab( &daten, probabilities);
#ifdef KOSTEN
      log_kosten(6);
#endif
      
      if (((preemption == 1)||(repeated == 0))&&(!noexp)) {
	rk2( repeated, rk2_step, &daten, tau_densities, probabilities, stored_k);
#ifdef KOSTEN
	log_kosten(4);
#endif
      }
      boundary_conditions( &daten, bound_results, tau_densities);
      boundary_check_a( &repeated, bound_results, low_box, add_where);
      set_boundaries_a( bound_results, low_box, add_where);
#ifdef KOSTEN
      log_kosten(5);
#endif
    }
    while (repeated);
    integralvervollstaendigung_a( &daten, incomplete_pdts, low_box, add_where);
#ifdef KOSTEN
    log_kosten(3);
#endif
    /* probabilities updaten ; */
    a_ab( &daten, probabilities);
#ifdef KOSTEN
    log_kosten(6);
#endif
  }
  else { /* preemption == 0 */
    /* analog MD1K-Fall */
    rk2( 0, rk2_step, &daten, tau_densities, probabilities, stored_k);
#ifdef KOSTEN
    log_kosten(4);
#endif
    for( i=0; i<ac_no_detT; i++)
      del_vector(&(low_box->lower_steps[add_where].values[0][i]));
    boundary_conditions( &daten, low_box->lower_steps[add_where].values[0], tau_densities);
#ifdef KOSTEN
    log_kosten(5);
#endif
    integralvervollstaendigung_a( &daten, incomplete_pdts, low_box, add_where);
#ifdef KOSTEN
    log_kosten(3);
#endif
  };

  set_old_stepsize( rk2_step);
  clear_int_accept( no_upper_steps, up_boxes, conf_no_upper_steps, conf_up_boxes);
  /* statt save_int_structures() : */
  for( i=0; i<ac_no_detT; i++) {
    conf_no_upper_steps[i] = no_upper_steps[i];
    conf_up_boxes[i] = up_boxes[i];
  };
  
  numDatas++;
  addall( p_t, &daten);
  update_solve( tnethome, modelpath, netname, ftmp, daten.no_markings, p_t, transient_time+rk2_step, maxX);
#ifdef DEBUG_AC  
  just_for_debug = get_sum_components( p_t);
  printf(" p_t-Summe : %f\n", just_for_debug);
#endif
  
#ifdef KOSTEN
  log_kosten(1);
#endif

  /* update_int_structures() enthaelt Randomisierungsschritt und teilweise numerische Integration fuer p_d(t) */
  add_where = update_int_structures( show_discontinuity, rk3_step, preemption, init, no_upper_steps, conf_no_upper_steps, up_boxes, conf_up_boxes, low_box, interpol_update, tau_densities, incomplete_pdts, randdata, areas_ptr);
  
  if (preemption) {
    
    abcc_int_points( rk3_step, probabilities);
#ifdef KOSTEN
    log_kosten(6);
#endif
    do {
      integralvervollstaendigung_a( &daten, incomplete_pdts, low_box, add_where);
#ifdef KOSTEN
      log_kosten(3);
#endif
      /* probabilities updaten ; */
      abcc_y_values( &daten, probabilities);
#ifdef KOSTEN
      log_kosten(6);
#endif
      
      if (((preemption == 1)||(repeated == 0))&&(!noexp)) {
	  rk3( repeated, rk3_step, &daten, tau_densities, probabilities, stored_k);
#ifdef KOSTEN
	  log_kosten(4);
#endif
      };
      boundary_conditions( &daten, bound_results, tau_densities);
      boundary_check_a( &repeated, bound_results, low_box, add_where);
      set_boundaries_a( bound_results, low_box, add_where);
#ifdef KOSTEN
      log_kosten(5);
#endif
    }
    while (repeated);
    integralvervollstaendigung_a( &daten, incomplete_pdts, low_box, add_where);
#ifdef KOSTEN
    log_kosten(3);
#endif
    /* probabilities updaten ; */
    abcc_y_values( &daten, probabilities);
#ifdef KOSTEN
    log_kosten(6);
#endif
  }
  else { /* preemption == 0 */
    /* analog MD1K-Fall */
    rk3( 0, rk3_step, &daten, tau_densities, probabilities, stored_k);
#ifdef KOSTEN
    log_kosten(4);
#endif
    for( i=0; i<ac_no_detT; i++)
      del_vector(&(low_box->lower_steps[add_where].values[0][i]));
    boundary_conditions( &daten, low_box->lower_steps[add_where].values[0], tau_densities);
#ifdef KOSTEN
    log_kosten(5);
#endif
    integralvervollstaendigung_a( &daten, incomplete_pdts, low_box, add_where);
#ifdef KOSTEN
    log_kosten(3);
#endif
  };
  
  /* Bearbeitung von stored_k[1] : besonders, da in stored_k[1] nach rk3() noch das alte p_E*Q_EE statt das neue
     p_E*Q_EE aufaddiert ist; altes p_E steht nach rk3() in stored_k[7] */
  /* Formel : neues stored_k[1] = {p_E(alt)*Q_EE + Rest(altes stored_k[1])} - (p_E(alt)-p_E(neu))*Q_EE */
  vector_add_mult( stored_k[7], daten.p_E, -1.0);
  mult_vector_matrix( stored_k[7], daten.Q_EE);
  vector_add_mult( stored_k[1], stored_k[7], -1.0);
#ifdef KOSTEN
  log_kosten(4);
#endif
  
  set_old_stepsize( rk3_step);
  clear_int_accept( no_upper_steps, up_boxes, conf_no_upper_steps, conf_up_boxes);
  
  transient_time += h_new;
  
  if (h_new < hh_new) {
    finish_loops = 0;
    h_new = ((3.0*h_new) + hh_new)/4.0;   /* hier geht ermittelte Startschrittweite ein, aber nur anteilig, da sonst */
    /* naechster Schritt zu gross im Vergleich zu rk3_step, was schlecht fuer Interpolation/Integration  waere */
  };
  
  if (finish_loops) 
    break;
  
  numDatas++;
  addall( p_t, &daten);
  update_solve( tnethome, modelpath, netname, ftmp, daten.no_markings, p_t, transient_time, maxX);
#ifdef DEBUG_AC  
  just_for_debug = get_sum_components( p_t);
  printf(" p_t-Summe : %f\n", just_for_debug);
#endif

  /* Anlaufrechnung hiermit abgeschlossen */

  /* dopri5() muss bei Schrittablehnung komplett neu berechnet werden, da sich Interpolationen von tau_densities
     mit Schrittweite aendern */
  
  while (!show_discontinuity) {

    if (transientOptions.transient_time > next_discontinuity) {
      /* naechste Diskontinuitaet vor Endzeitpunkt */
      if ((transient_time + h_new) >= (next_discontinuity - EXT_PRECISION)) {
	show_discontinuity = 1;  /* mit naechstem Schritt wird Diskontinuitaet erreicht */
	h_new = next_discontinuity - transient_time;
      }
      else if ((transientOptions.precision <= 0.001)&&((transient_time + h_new)>(next_discontinuity - 10.0*transientOptions.precision))) {
	/* damit letzter Schritt auf Diskontinuitaet nicht zu klein wird */
	h_new = 0.8*h_new;
#ifdef DEBUG_AC
	printf("\nVerkuerzung der Schrittweite, damit letzter Schritt vor Diskontinuitaet nicht zu klein wird !\n");
#endif
      };
    }
    else { /* transientOptions.transient_time <= next_discontinuity */
      if ((transient_time + h_new) >=  (transientOptions.transient_time - EXT_PRECISION)) {
	finish_loops = 1;           /* mit naechstem Schritt wird Endzeitpunkt erreicht */
	h_new = transientOptions.transient_time - transient_time;
	if ( transientOptions.transient_time == next_discontinuity)
	  show_discontinuity = 1;
      };
    };
    
    /* innerhalb von on_tx_plain.c werden `add_here' und `merke' gesichert */
    save_static_values();  
    if (init)
      save_areas( areas_ptr, careas_ptr);
      /* damit gegebenenfalls auch die Aenderungen aus stetiger Uebertragung kopiert ! */


    save_int_structures( 1, add_where, no_upper_steps, up_boxes, low_box, interpol_update, tau_densities, conf_no_upper_steps, conf_up_boxes, conf_low_box, conf_interpol_update, conf_tau_densities);

    del_vector( &hp_E);
    hp_E = copy_vector( daten.p_E);

#ifdef KOSTEN
  log_kosten(6);
#endif
        
    for( ; ; ) {

      /* update_int_structures() enthaelt Randomisierungsschritt und teilweise numerische Integration fuer p_d(t) */
      add_where = update_int_structures( show_discontinuity, h_new, preemption, init, no_upper_steps, conf_no_upper_steps, up_boxes, conf_up_boxes, low_box, interpol_update, tau_densities, incomplete_pdts, randdata, areas_ptr);

      if (preemption) {
	
	do {
	  if (show_discontinuity)
	    integralvervollstaendigung_b( &daten, incomplete_pdts, low_box);
	  else
	    integralvervollstaendigung_a( &daten, incomplete_pdts, low_box, add_where);
#ifdef KOSTEN
	  log_kosten(3);
#endif
    	  /* probabilities updaten : */
	  if (repeated) 
	    repeated_update( &daten, probabilities);
	  else {
	    if (accept_step) {
	      if (new_io_2) {
		new_io_2 = 0;
		abcd_cdbe( h_new, &daten, probabilities);
	      }
	      else { /* new_io_2 == 0 */
		new_io_2 = 1;
		cdbe_cdef( h_new, &daten, probabilities);
	      };
	    }
	    else { /* accept_step == 0 */
	      rejected_update( h_new, &daten, probabilities);
	    };
	  }; /* end if (repeated) */
#ifdef KOSTEN
	  log_kosten(6);
#endif

	  if (((preemption == 1)||(repeated == 0))&&(!noexp)) {	  
	    err_dopri = dopri5( repeated, h_new, &daten, tau_densities, probabilities, stored_k);
	    /* falls nicht durchlaufen, behaelt err_dopri seinen Wert aus erstem Aufruf ! */
#ifdef DEBUG_AC
	    printf(" err_dopri = %.9f \n",err_dopri);
#endif
#ifdef KOSTEN
	    log_kosten(4);
#endif
	  };
	  if (err_dopri < transientOptions.precision) {
	    /* Schritt wird erstmal angenommen, Iteration u.U. noch noetig */
	    boundary_conditions( &daten, bound_results, tau_densities);
	    if (show_discontinuity) {
	      /* wird falsch, falls dummy existiert ! */
	      boundary_check_b( &repeated, bound_results, low_box);
	      set_boundaries_b( bound_results, low_box);
	    }
      	    else { /* show_discontinuity == 0 */
	      boundary_check_a( &repeated, bound_results, low_box, add_where);
	      set_boundaries_a( bound_results, low_box, add_where);
	    };
	  }
	  else
	    repeated = 0;
#ifdef KOSTEN
	  log_kosten(5);
#endif
	}
	while (repeated);
	/* repeated = 0 */
	
	if (noexp) 
	  fac = 0.7;   /* Schrittweitensteuerung erfolgt in diesem Fall vor allem ueber corrector */
	else  
	  fac = max(0.2,min(5.0,pow((err_dopri/transientOptions.precision),0.2)/0.9));
	  /* Formel aus Hairer/Wanner */

	/* Berechnung von vorlaefig neuer Schrittweite, wobei h/5 <= h_neu <= 5*h */
	hh_new = h_new/fac;      

#ifdef DEBUG_AC
	printf(" fac = %.9f \n",fac);
#endif
        if (err_dopri < transientOptions.precision) {
	  /*  vorlaeufige Schrittannahme */
	  if (show_discontinuity)
	    integralvervollstaendigung_b( &daten, incomplete_pdts, low_box);
	  else
	    integralvervollstaendigung_a( &daten, incomplete_pdts, low_box, add_where);
#ifdef KOSTEN
	  log_kosten(3);
#endif
	  /* probabilities updaten : */
	  repeated_update( &daten, probabilities);
#ifdef KOSTEN
	  log_kosten(6);
#endif
	 
	  corrector = global_correcting( &daten);
#ifdef DEBUG_AC
	  printf("\n corrector = %.9f\n",corrector);
#endif
	  if ((corrector<4.5)||(done_enough > 7)) {   /* ||(!accept_step) kann ergaenzt werden */
	    /* wegen zu grossem corrector darf Schritt trotzdem nicht abgelehnt werden, falls gerade zuvor
	       ein Schritt abgelehnt wurde oder schon vier corrector-Ablehnungen seit letzter Diskontinuitaet
	       erfolgten */

	    /* endgueltige Schrittannahme */
	    del_vector(&(stored_k[1]));
	    stored_k[1] = stored_k[7];
	    stored_k[7] = new_vector();

	    if (accept_step) {
	      if (hh_new > h_max)
		hh_new = h_max;
	    }
	    else {
	      hh_new = min( hh_new, h_new); /* Vorsichtsmassnahme nach abgelehnten Schritt */
	      accept_step = 1;
	    };
	    clear_int_accept( no_upper_steps, up_boxes, conf_no_upper_steps, conf_up_boxes);
	    /* hh_new : Vorschlag fuer neue Schrittweite,
	       h_new  : tatsaechlich duechgefuehrte Schrittweite */

	    transient_time += h_new;
	    set_old_stepsize( h_new);
#ifdef DEBUG_AC
	printf(" h_new (last stepsize) : %.9f \n", h_new);
#endif
	    h_new = hh_new/corrector;
#ifdef DEBUG_AC
	printf(" h_new (proposal) : %.9f \n", h_new);
#endif

	    numDatas++;
	    addall( p_t, &daten);
	    update_solve( tnethome, modelpath, netname, ftmp, daten.no_markings, p_t, transient_time, maxX);
#ifdef DEBUG_AC
	    just_for_debug = get_sum_components( p_t); /* rausnehmen */
	    printf("\np_t-Summe : %f\n", just_for_debug); /* rausnehmen */
#endif
#ifdef KOSTEN
	    log_kosten(1);
#endif
	    break; /* nur so geht's raus aus der Endlos-for-Schleife */
	  }
	  else { /* corrector = 4.5 und done_enough <= 3 */
	    /* Schritt muss doch noch abgelehnt werden, da mit der ermittelten Schrittweite zu grosse Fehler
	       durch die numerische Interpolation und Integration eingeschleppt wuerden ! */
	    
#ifdef DEBUG_AC
            printf("\n corrector-REJECTION !!\n");
#endif
	    accept_step = 0;
	    done_enough++;

	    h_new = 0.5*h_new; 
	    clear_int_reject( no_upper_steps, up_boxes, conf_no_upper_steps, conf_up_boxes, conf_low_box);
	    /* low_box-Austausch : */
	    lbox_ptr = low_box;
	    low_box = conf_low_box;
	    conf_low_box = lbox_ptr;

	    if (show_discontinuity) {
	      show_discontinuity = 0;
	      add_where = 5;          
	      /* leider nur schwer zu unterscheiden, ob Uebertragung aus low_box aus weniger Stuetzpunkten bestand ! */
	      /* kleinster sinnvoller Wert : add_where = 3 (wegen -1 s.u.) */
	    };
	    if (finish_loops)
	      finish_loops = 0;
	    
	    if (add_where == 2) {
	      /* einziger Fall, dass neues add_where kleiner als altes ! */
	      /* unguenstig, wenn gleich nach rk3 Schritt abgelehnt wird, aber nicht zu aendern ! */
	      add_where = 4;
	    }
	    else
	      add_where -= 1;
	    
	    if (init) 
	      save_areas( careas_ptr, areas_ptr);
	    
	    save_int_structures( 0, add_where, conf_no_upper_steps, conf_up_boxes, low_box, conf_interpol_update, conf_tau_densities, no_upper_steps, up_boxes, conf_low_box, interpol_update, tau_densities);

	    /* Sicherheitskopien von `add_here' und `merke' in on_tx_plain.c wieder einsetzen */
	    reaffirm_static_values();

	    del_vector( &(daten.p_E));
	    daten.p_E = copy_vector( hp_E);
#ifdef KOSTEN
	    log_kosten(1);
#endif
	  };
	}
	else { /* err_dopri >= transientOptions.precision */
	  /* Schrittablehnung */
#ifdef DEBUG_AC
          printf("\n err_dopri-REJECTION !!\n");
#endif
	  accept_step = 0;
	  h_new = hh_new;     /* bei Ablehnung : hh_new < h_new -> deshalb kein h_max-Check ! */
	  clear_int_reject( no_upper_steps, up_boxes, conf_no_upper_steps, conf_up_boxes, conf_low_box);
	  /* low_box-Austausch : */
	  lbox_ptr = low_box;
	  low_box = conf_low_box;
	  conf_low_box = lbox_ptr;

	  if (show_discontinuity) {
	    show_discontinuity = 0;
	    add_where = 5;          
	    /* leider nur schwer zu unterscheiden, ob Uebertragung aus low_box aus weniger Stuetzpunkten bestand ! */
	    /* kleinster sinnvoller Wert : add_where = 3 (wegen -1 s.u.) */
	  };
	  if (finish_loops)
	    finish_loops = 0;

	  if (add_where == 2) {
	    /* einziger Fall, dass neues add_where kleiner als altes ! */
	    /* unguenstig, wenn gleich nach rk3-Schritt abgelehnt wird, aber nicht zu aendern ! */
	    add_where = 4;
	  }
	  else
	    add_where -= 1;

	  if (init) 
	    save_areas( careas_ptr, areas_ptr);

	  save_int_structures( 0, add_where, conf_no_upper_steps, conf_up_boxes, low_box, conf_interpol_update, conf_tau_densities, no_upper_steps, up_boxes, conf_low_box, interpol_update, tau_densities);

	  /* Sicherheitskopien von `add_here' und `merke' in on_tx_plain.c wieder einsetzen */
	  reaffirm_static_values();

	  del_vector( &(daten.p_E));
	  daten.p_E = copy_vector( hp_E);
#ifdef KOSTEN
	  log_kosten(1);
#endif
	};
      } 
      else { /* preemption  == 0 */
	/* analog MD1K-Fall */
	err_dopri = dopri5( 0, h_new, &daten, tau_densities, probabilities, stored_k);

	/* Berechnung von neuer Schrittweite, wobei h/5 <= h_neu <= 5*h */
	fac = max(0.2,min(5.0,pow((err_dopri/transientOptions.precision),0.2)/0.9));
	hh_new = h_new/fac;
#ifdef DEBUG_AC
	printf(" err_dopri : %.9f, fac : %f, hh_new : %f \n", err_dopri, fac, hh_new);
#endif
#ifdef KOSTEN
	log_kosten(4);
#endif
	if (err_dopri < transientOptions.precision) {
	  /* Schritt angenommen */
	  if (show_discontinuity) {

	    /* Programmhygiene : in boundary_conditions() werden diese Vektoren nicht geloescht */
	    for( i=0; i<ac_no_detT; i++)
	      del_vector(&(low_box->next_up->values[0][i]));
	    boundary_conditions( &daten, low_box->next_up->values[0], tau_densities);
#ifdef KOSTEN
	    log_kosten(5);
#endif
	    integralvervollstaendigung_b( &daten, incomplete_pdts, low_box);
#ifdef KOSTEN
	    log_kosten(3);
#endif
	  }
	  else {
	    /* Programmhygiene : in boundary_conditions() werden diese Vektoren nicht geloescht */
	    for( i=0; i<ac_no_detT; i++)
	      del_vector(&(low_box->lower_steps[add_where].values[0][i]));
	    boundary_conditions( &daten, low_box->lower_steps[add_where].values[0], tau_densities);
#ifdef KOSTEN
	    log_kosten(5);
#endif
	    integralvervollstaendigung_a( &daten, incomplete_pdts, low_box, add_where);
#ifdef KOSTEN
	    log_kosten(3);
#endif
	  };

	  del_vector(&(stored_k[1]));
	  stored_k[1] = stored_k[7];
	  stored_k[7] = new_vector();

	  if (accept_step) {
	    if (hh_new > h_max)
	      hh_new = h_max;
	  }
	  else {
	    hh_new = min(hh_new,h_new);/* Vorsichtsmassnahme nach abgelehnten Schritt */
	    accept_step = 1;
	  };
	  clear_int_accept( no_upper_steps, up_boxes, conf_no_upper_steps, conf_up_boxes);

	  transient_time += h_new;
	  set_old_stepsize( h_new);
#ifdef DEBUG_AC
	printf(" transient_time : %f, h_new (last stepsize) : %.9f \n", transient_time, h_new);
#endif
	  h_new = hh_new;
#ifdef DEBUG_AC
	printf(" h_new (proposal) : %.9f \n", h_new);
#endif

	  numDatas++;
	  addall( p_t, &daten);
	  update_solve( tnethome, modelpath, netname, ftmp, daten.no_markings, p_t, transient_time, maxX);
#ifdef DEBUG_AC
	  just_for_debug = get_sum_components( p_t);
	  printf("\np_t-Summe : %f\n", just_for_debug); 
#endif
#ifdef KOSTEN
	  log_kosten(1);
#endif
	  break;  /* nur so geht's raus aus der Endlosschleife */
	}  
	else {
	  /* Schritt abgelehnt */
#ifdef DEBUG_AC
	  printf(" REJECTION !!!\n");
#endif
	  accept_step = 0;
	  h_new = hh_new;     /* bei Ablehnung : hh_new < h_new */

	  clear_int_reject( no_upper_steps, up_boxes, conf_no_upper_steps, conf_up_boxes, conf_low_box);
	  /* low_box-Austausch : */
	  lbox_ptr = low_box;
	  low_box = conf_low_box;
	  conf_low_box = lbox_ptr;
	  
	
          if (show_discontinuity) {
	    show_discontinuity = 0;
	    add_where = 5;          
	    /* leider nur schwer zu unterscheiden, ob Uebertragung aus low_box aus weniger Stuetzpunkten bestand ! */
	    /* kleinster sinnvoller Wert : add_where = 3 (wegen -1 s.u.) */
	  };
	  if (finish_loops)
	    finish_loops = 0;

	  if (add_where == 2) {
	    /* dann ist naemlich, ausser direkt nach rk3, altes add_where groesser als neues ! */
	    add_where = 4;
	  }
	  else
	    add_where -= 1;

	  if (init) 
	    save_areas( careas_ptr, areas_ptr);

	  save_int_structures( 0, add_where, conf_no_upper_steps, conf_up_boxes, low_box, conf_interpol_update, conf_tau_densities, no_upper_steps, up_boxes, conf_low_box, interpol_update, tau_densities);

	  /* Sicherheitskopien von `add_here' und `merke' in on_tx_plain.c wieder einsetzen */
	  reaffirm_static_values();
	  del_vector( &(daten.p_E));
	  daten.p_E = copy_vector( hp_E);
#ifdef KOSTEN
	  log_kosten(1);
#endif
	};
      };

    };  /* Ende der leeren for-Schleife ! */

    if (finish_loops) 
      break;
  };   /* Ende der discontinuity-Schleife */

  /* spaeter hier aus transientOptions.transient_time-Schleife rausspringen */
  if (finish_loops) 
    break;
  
  /* (un)stetige Uebertragung und Vorbereitung fuer erneute Anlaufrechnung */
  printf("\nDiscontinuity (maybe only in the derivatives) occurred at time %.2f .\n",transient_time); 
  disk_uebertragung( &daten, init, no_upper_steps, up_boxes, conf_no_upper_steps, conf_up_boxes, interpol_update, tau_densities, randdata, areas_ptr);
  /* ohne boundary_conditions() */
  if (init) {
    /* in diesem Fall koennten sich durch die Uebertragung die Werte der Wahrscheinlichkeiten geaendert haben */
    numDatas++;
    addall( p_t, &daten);
    update_solve( tnethome, modelpath, netname, ftmp, daten.no_markings, p_t, transient_time, maxX);
#ifdef DEBUG_AC
    just_for_debug = get_sum_components( p_t);
    printf(" p_t-Summe : %f\n", just_for_debug);
#endif
  };

  h_max = ( get_next_discontinuity( &daten, &next_discontinuity, root, max_element)/2.0);
  /* weil immer zwei Anlaufschritte, die nicht zu Diskontinuitaet fuehren, gemacht werden, ist so schon sichergestellt,
     dass mindestens zwei innere Stuetzpunkte zwischen zwei Diskontinuitaeten existieren */
  if (fabs(transientOptions.transient_time-next_discontinuity) < transientOptions.precision)
    transientOptions.transient_time = next_discontinuity;
    /* falls Endzeitpunkt und naechste Diskontinuitaet zu nahe beieinander liegen, den Endzeitpunkt vorziehen, um
       sicheren Schleifenausstieg zu ermoeglichen */

  set_old_stepsize( 0.0);
  show_discontinuity = 0;
  new_io_2 = 0;
  done_enough = 0;
#ifdef KOSTEN
  log_kosten(1);
#endif

} /* leere for-Schleife fuer Rechnung zwischen Diskontinuitaeten */

close_adaptsolve( modelpath, netname, ftmp, numDatas);
}


