/*******************************************************************************
Modul:              frontend.h 

Purpose:            grundlegende Datenstrukturen werden definiert

Remark:             Matrix Indexe beginnen bei NULL also Matrix[1,2] ist 2.Zeile und 3.Spalte

Project Director:   Reinhard German

Programmer:         Joerg W. Mitzlaff

History:            10.94 (Dipl.Arbeit)
                    angelehnt an DSPNAnalysis/rand_types.h
*******************************************************************************/

#define MIN_REAL        1.0e-50         /* real values less than this are set to 0 */
#define is_zero(a)      ((a > - MIN_REAL) && (a < MIN_REAL))
 
#define BUFFERENLARGEMENT 1000
#define MEM_BUFFERS 1000
#define MEM_ITEMS  1000
#define GRENZBEREICH 5
#define DISC_ANZEIGE -1.0e10
#define ZWISCHEN_ANZEIGE 1.0e10
#define MAX_ABS_PREC 1.0e-16
  
        /* macros for short scanning of linear lists and arrays */
#define scan_list(walk,list)    for (walk = list; walk != NULL; walk = walk->next)
#define scan_array(count,num)   for (count = 0; count < num; count++)
 
 
/* types for sparse matrices and vectors */

/* entry of a sparse vector */
typedef struct sp_entry
{       int             num;            /* the column number */
        double          val;            /* the entry value */
        struct sp_entry *next;
 
}  sp_entry;
 
/* sparse vector or a column of a sparse matrix */
typedef struct sp_vector
{       int             num;            /* the row !!! number */
        int             no_entries;     /* number of entries */
        sp_entry        *entries;       /* linear list of entries */
        struct sp_vector *next;
}  sp_vector;
 
/* sparse matrix */
typedef struct sp_matrix
{       sp_vector       *rows;          /* linear list of rows */
}  sp_matrix;
 

/*************************************************************************************/
/*  expT ist eine exponentielle Transition und detT eine deterministische Transition */
/*  c und d sind aus S'D aber verschiedene detT                                      */
/*  S'x bezeichnen Zustaende in denen eine Transition vom Typ x angeschaltet ist     */
/*  Q und Delta (griechischer Buchstabe) sind die Generatormatrizen der CTMC,        */
/*              die das stochastische Verhalten beschreiben                          */
/*************************************************************************************/


typedef sp_matrix       *sp_matrices;


/* relevanten Daten detT mit Id >>>------> d <-----<<< */
typedef struct sp_detMatrices
{
        double          det_firing_time;        /* tau oder so */
        sp_matrix       *Q_d;                   /* expT schaltet von S'd nach S'd ohne ausschalten */
        sp_matrix       *Q_dE;                  /* expT schaltet von S'd nach S'E mit ausschalten der detT d */
        sp_matrix       *Delta_dE;              /* detT schaltet von S'd nach S'E */
        sp_matrices     *Q_dc;                 /* expT schaltet von S'd nach S'c mit ausschalten der detT d */
        sp_matrices     *Delta_dc;             /* detT schaltet von S'd nach S'c; Reihenfolge von Loesungsverfahren abhaengig */
} sp_detMatrices;


typedef sp_vector       *sp_vectors;

/* alle relevanten Daten zur transienten Berechnung */
typedef struct
{
        sp_matrix       *Q_EE;                  /* expT schaltet von S'E nach S'E */
        sp_matrices     *Q_Ed;                  /* expT schaltet von S'E nach S'd */
	int             no_detT;                /* Anzahl der detT */
        sp_detMatrices  *detMatrices;           /* es wird von S'D geschaltet; Reihenfolge von Verfahren abhaengig */
	int             no_markings;            /* Anzahl der greifbaren (tangible) Markierungen */
        sp_vector       *p_E;                   /* initial p'E(0) */
        sp_vectors      *p_d;                   /* initial p'd(0) fuer alle S'd */
} solve_basis;

/*****************************************************************************************************/
/* Die folgenden Datenstrukturen werden nur von dem Algorithmus mit adaptiver Schrittweite gebraucht */ 
/*****************************************************************************************************/

/* Datenstruktur fuer ein Element aus der Integrationskette */
typedef struct step_inbetween
{    double                 step;
     double                 integr_weight;
     double                 indicator;       /* ZWISCHEN_ANZEIGE fuer mittleren Integrationspunkt,
					        zweiter Lagrange-Faktor bei zweifachem Endpunkt,
					        DISC_ANZEIGE bei einfachem Endpunkt (Diskontinuitaet) */
     struct step_inbetween  *next_down;
     sp_vectors             *values[2];     /* Funktionswerte aus PDE, eigentlich fuer jede
					     determ. Gruppe ein Zeiger, bei MDK1 nur eine!*/
   } step_inbetween;

/* Datenstruktur fuer einzelne Komponenten des Typs lower_stepbox */
typedef struct step_at_edge
{    double                 step;
     double                 integr_weight;  /* aktueller Lagrange-Faktor*/
     double                 lagrange3;      /* Lagrange-Faktor bei 3 Stuetzstellen */
     double                 lagrange4;      /* Lagrange-Faktor bei 4 Stuetzstellen */
     sp_vectors            *values[2];
   } step_at_edge;

/* Datenstruktur fuer die low_box, eine Ansammlung der letzten Integrationsstuetzpunkte */
typedef struct lower_stepbox
{    struct step_at_edge    lower_steps[GRENZBEREICH];
     struct step_inbetween  *next_up;
   } lower_stepbox;

/* Datenstruktur fuer einzelne Komponenten eines Interpolationsfeldes */ 
typedef struct interpolation_row
{    double                 int_point;
     struct sp_vector       *y_values;
     struct sp_vector       *pol_coefficients;
     struct sp_vector       *aux_coefficients;
   } interpolation_row;

/* wichtig : fuer jede Gruppe (Zustaende, in denen eine bestimmte determ. Transition aktiviert ist)
             benoetigt man zwei Interpolationsfelder, die jeweils aus einem vierkomponentigen Feld 
	     von interpolation_row bestehen. 
	       - eins fuer Interpolation der Dichten p_d(t,x) entlang t bei x=tau, 
	       - eins fuer Interpolation von p_d(t) entlang t                               */

/* Datenstruktur fuer ein Interpolationsfeld */
typedef interpolation_row interpolation_field[4];

/* Datenstruktur fuer ein Element eines binaeren Baums, in dem die zur Laufzeit ermittelten
   Diskontinuitaeten gespeichert werden */
typedef struct tnode 
    { double discontime;
      struct tnode *left;
      struct tnode *right; } tnode; 
 
/* Datenstruktur fuer die Flaechen von den Impulsen in den Dichten p_d(t,x) einer Gruppe
   zu einem festen Zeitpunkt t parallel zur x-Achse */
typedef struct areas
    { sp_vector *last;
      sp_vector *first; } areas;

/***************************************************************/
/*   ACHTUNG : EINSCHUB  -  ENDE                               */
/***************************************************************/


/* foreward declaration of mat_proc.c prozeduren */
extern char      *emalloc();
extern void      efree();
extern double    min();
extern double    max();
extern double    get_infnorm();
extern double    get_sum_components();
extern void      set_negs_to_zero();

/* vector ops. */
extern sp_vector *new_vector    ();        
extern sp_vector *copy_vector   ( /* sp_vector *vec */ );
extern sp_entry  *add_to_vector ( /* sp_vector *vector, sp_entry *prev_item, int index, double value */ );     
extern void      clear_vector   ( /* sp_vector *vec */ );
extern void      del_vector     ( /* &(sp_vector *vec) */ );
extern int       comp_vector    ( /* sp_vector *vec1, sp_vector *vec2 */ );
extern int       rel_comp_vector( /* sp_vector *vec1, sp_vector *vec2 */ );

/* matrix ops. */
extern sp_matrix *new_matrix();        
extern sp_matrix *copy_matrix       ( /* sp_matrix *mat */ );        
extern void      add_entry_to_matrix( /* sp_matrix *mat, int row_nr, col_nr, double value */ );
extern void      clear_matrix       ( /* sp_matrix *mat */ );       
extern void      del_matrix         ( /* &(sp_matrix *mat) */ );         

/* komplexe Rechenoperationen */
extern void      vector_add_mult   ( /*	sp_vector *vec1, sp_vector *vec2, double factor */ );   /* vec1 = vec1 + vec2 * factor */
extern void      mult_vector_matrix( /* sp_vector *vec, sp_matrix *mat */ );                    /* vec  = vec  * mat */
extern void      mult_matrix_matrix( /* sp_matrix *mat1, sp_matrix *mat2 */ );                  /* mat1 = mat1 * mat2 */
extern void      matrix_add_mult   ( /* sp_matrix *mat1, sp_matrix *mat2, double factor */ );   /* mat1 = mat1 + mat2 * factor */
extern void      mult_vector_skalar( /* sp_vector *vec, double faktor */ );                     /* vec  = vec * factor */

/* for testing */
extern void      show_entry();
extern void      show_vector();
extern void      show_matrix();
extern int       count_elements( /* sp_matrix *mat */ );

