/******************************************************************************
 Program:           get_tmark.c

 Purpose:           .tmark daten einlesen und p(i) Vektor (Eintraege von p zur Zeit t=0)
                    in solveBasis Struktur ergaenzen.
                    bei Fehler abbrechen

 Project Director:  Reinhard German

 Programmer:        Joerg W. Mitzlaff

 History:           10.94 (Dipl.Arbeit)
                    angelehnt an Shared/src/convert_results/convert_results.c
*******************************************************************************/

#include <stdio.h> 
#include <string.h> 
#include "get_tmark.h"
#include "netio.h"
#include "get_rrg.h"          /* BUF_SIZE und mark_t definieren */
#include "frontend.h"         /* Vektor Typen/Operationen; emalloc decl. */


/* Import aus frontend.c */
extern NetDescr *net;

/* Import aus get_rrg.c */
extern mark_t *marking;       /* markings and its types */


/* exported variables to shared.c */
unsigned nplaces = 0;
place_t *place;
unsigned ntoken = 0;
int *tok_list;            /* Tabelle (nmarking x nplaces) */

/* local variables */
static unsigned nmarking = 0;

/* forward definition of private functions */
static void spool();
static void read_marking();

/* forward definition of functions exported to frontend.c */
solve_basis *build_p_Null_vector();
void free_tmark_mem();


/******************************************************************************/
solve_basis *build_p_Null_vector(modelpath,netname,component)
/******************************************************************************/
char *modelpath;
char *netname;
solve_basis *component;
{
   char tmark_file[256];
   FILE *tmark_fp;
   Place *foundplace;
   int i,j;
   int d;

   sprintf(tmark_file,"%s/%s.tmark",modelpath,netname);
   if( (tmark_fp = fopen(tmark_file,"r")) == NULL )
   {
      fprintf(stderr,"Unable to open file '%s'\n", tmark_file);
      exit(-1);
   }
 
   /* parse .tmark files */
   read_marking(tmark_fp);

   /* Konsistenz Check: nmarking muss gleich nmarking aus get_rrg.c sein !!! */
   if( component -> no_markings != nmarking )
   {
      fprintf(stderr,"Number of markings in .rrg and .tmark differ\n");
      exit(-1);
   }

   /* uebernehme aus struct NetDescr net->place->value bzw. net->place->markpar->value nach  place->init_ntoken */
   /* Konsistenz Check */
   if( net->num_places != nplaces )
   {
      fprintf(stderr,"Number of places in .TN and .tmark differ\n");
      exit( 1 );
   }
   for(i=0;i<nplaces;i++)
   {
     foundplace = find_place(net, place[i].name);
     if( foundplace -> is_value )
       place[i].init_ntoken = foundplace -> marking.value;
     else
       place[i].init_ntoken = foundplace -> marking.markpar -> value;
   }

   /* schreibe p(i) Vector */
   for(i=0;i<nmarking;i++)
   {
     /* finde initiale Markierung */
     for(j=0; j<nplaces;j++)
       if( tok_list[i*nplaces+j] != place[j].init_ntoken )
	 break;
     if(j==nplaces)  /* ist Initialzustand gleich akt. Markierung */
     {
       /* finde richtigen Vektor und setze Eintrag i auf 1 */
       if(marking[i].typeNr == -1)     /* S'E */
       {
	 add_to_vector(component -> p_E, NULL, i, 1.0);
       }
       else
       {
	 d = marking[i].typeNr;
	 add_to_vector(component -> p_d[d], NULL, i, 1.0);
       }
     }
     /* else tue nichts ! */
   }
   fclose(tmark_fp);

   return(component);
}

/*****************************************************************************/
/* spool: read lines from a file until a non-comment line is found */
/******************************************************************************/
static void spool(fp, buf)
FILE *fp;
char *buf;
{
   do
   {
      fgets(buf,BUF_SIZE,fp);
   } while( buf[0] == COMMENT );

   if( strlen(buf) > 1 )
      buf[strlen(buf) - 1] = '\0';
}


/******************************************************************************/
/* read_marking: read number of places, allocate matrix for number of token
 *               per marking per place, read data from file;
 *               reads .tmark */
/******************************************************************************/
static void read_marking(fp)
FILE *fp;              
{
   char buffer[BUF_SIZE], buf_name[256], *buf;
   int i,j;
   int mark,tok;

   /* read number of places */
   spool(fp,buffer);
   sscanf(buffer,"%d", &nplaces);

   /* allocate name & number list */
   if( (place = (place_t *)emalloc(nplaces * sizeof(place_t))) == NULL )
   {
      fprintf(stderr,"Unable to allocate places name list\n");
      exit( 1 );
   }

   /* read place info */
   for( i = 0; i < nplaces; i++ )
   {
      /* read place info */
      spool(fp,buffer);
      sscanf(buffer,"%s %d", buf_name,&place[i].number);

      /* copy place name */
      if( (place[i].name = (char *)strdup(buf_name)) == NULL )
      {
         fprintf(stderr,"Unable to allocate name for place\n");
         exit( 1 );
      }
   }

   /* read number of markings */
   spool(fp,buffer);
   sscanf(buffer,"%d", &nmarking);


   /* allocate token list */
   if( (tok_list = (int *)emalloc(nmarking * nplaces * sizeof(int))) == NULL )
   {
      fprintf(stderr,"Unable to allocate token list\n");
      exit( 1 );
   }

   /* read token distribution */
   for( i = 0; i < nmarking; i++ )
   {
      /* read marking number */
      spool(fp,buffer);
      buf = buffer;
      sscanf(buf,"%d", &mark);

      /* read number of token per place */
      for( j = 0; j < nplaces; j++ )
      {
         /* this spools buf to the next char after a the first space */
         while( *buf++ != ' ' )
            ;

         /* scan number of token */
         sscanf(buf,"%d", &tok);
         tok_list[mark * nplaces + j] = tok;

	 /* determine maximum number of tokens */
	 if(tok > ntoken)
	   ntoken = tok;
      }
   }
}


/*****************************************************************************/
/* free_tmark_mem: free allocated memory */
/*****************************************************************************/
void free_tmark_mem()
{
   int i;

   for( i = 0; i < nplaces; i++ )
      free(place[i].name);
   free(place);
   free(tok_list);
}
