/********************************************************************
Modul :             integrate&interpolate.c

Purpose :           Integration   -> Lagrange-Faktoren
                    Interpolation -> Newtondifferenzen
		    Extrapolation -> Aitken-Neville

Project Director :  Reinhard German

Programmer :        Armin Heindl

History :           18.5.95 ( Diplomarbeit )
*********************************************************************/
#define DEBUG_AC 
#include <stdio.h>      /* fuer NULL */
#include <math.h>
#include "frontend.h"

extern int ac_no_detT;
/*******************************************************************************
                   NUMERISCHE INTEGRATION

nach der Mittelwertformel mit Lagrangeschen Interpolationspolynomen :

 x_integral_von_x_0_bis_x_n{ f(x) } = i_sum{ f(x_i)*(x_integral_von_x_0_bis_x_n{ L_i(x) }) }

 - Summe laeuft ueber die Anzahl n+1 der Stuetzpunkte x_i, deren Funktionswerte f(x_i) 
   bekannt sind.
 - L_i(x) sind aus den Stuetzpunkten bildbaren Lagrange Interpolationspolynome
 - x_integral_von_x_0_bis_x_n{ L_i(x) } sind aus der relativen Lage der Stuetzpunkte
   zueinander bestimmbar und werden im folgende Koeffizienten genannt
 - bei Teilintegralen werden die Integrale ueber den entsprechenden Abschnitt 
   berechnet
*******************************************************************************/
void l3_integrale(l3, h0, h1)
/* berechnet die Koeffizienten fuer die Mittelwertformel mit Lagrangeschen Interpolationspolynomen
   zweiter Ordnung, wobei die Abstaende zwischen den drei Stuetzpunkten h0 und h1 betragen */
double l3[3];
double h0, h1;

{
  double sum1, quot0, quot1;

  /* Formeln : l3[0] = (h0+h1)*(2*h0-h1)/(6*h0)
               l3[1] = (h0+h1)^3/(6*h0*h1)
	       l3[2] = (h0+h1)*(2*h1-h0)/(6*h1)  */

  sum1 = (h0+h1)/6.0;
  quot0 = h0/h1;
  quot1 = h1/h0;

  l3[0] = sum1*(2.0-quot1);
  l3[1] = sum1*(quot0+2.0+quot1);
  l3[2] = sum1*(2.0-quot0);
  
}

/*****************************************************************************/

void l3_randintegrale(l3, h0, h1)
/* berechnet die Koeffizienten fuer die Mittelwertformel mit Lagrangeschen Interpolationspolynomen
   zweiter Ordnung, wobei die Abstaende zwischen den drei Stuetzpunkten h0 und h1 betragen und nur
   der Wert des Integrals zwischen dem zweiten und dritten Stuetzpunkt interessiert */

double l3[3];
double h0,h1;

{
  double ratio, part1;

  /* Formeln : l3[0] = -(h1/h0)*(h1/(h0+h1))*(h1/6)
               l3[1] = (h1/h0)*(h0/2+h1/6)
	       l3[2] = (h1/(h0+h1))*(h0/2+h1/3)    */

  part1 = h1/(h0+h1);
  ratio = h1/h0;
  h1 = h1/6.0;
  h0 = h0/2.0;
  
  l3[2] = part1*(h0+(2.0*h1));
  l3[1] = ratio*(h1+h0);
  l3[0] = (-1.0)*ratio*part1*h1;

}
  
/*****************************************************************************/

void l4_integrale(l4, h0, h1, h2)
/* berechnet die Koeffizienten fuer die Mittelwertformel mit Lagrangeschen Interpolationspolynomen
   dritter Ordnung, wobei die Abstaende zwischen den vier Stuetzpunkten h0, h1 und h2 betragen */

double l4[4];
double h0, h1, h2;

{
  double summand1, summand2, summand3, sum1a, sum1b, sum, help;


  /* Formeln : sum = h0+h1+h2

               l4[0] = (sum/2)*(1 - (sum/3)*( 1/(h0+h1) + 1/h0 - sum/(2*h0*(h0+h1) )))
               l4[1] = sum^3*(h0+h1-h2)/(12*h0*h1*(h1+h2))
	       l4[2] = sum^3*(h2+h1-h0)/(12*h2*h1*(h1+h0))
	       l4[3] = (sum/2)*(1 - (sum/3)*( 1/(h2+h1) + 1/h2 - sum/(2*h2*(h2+h1) )))  */
 
  sum1a = h0+h1;
  sum1b = h1+h2;
  sum   = sum1a + h2;
  summand1 = sum/2.0;
  summand2 = (summand1*sum)/3.0;
  summand3 = (summand2*sum)/2.0;

  help = summand3/sum1a;

  l4[0] = summand1 - ((summand2*(sum1a+h0))/(sum1a*h0)) + (help/h0);

  l4[2] = (help*(sum1b-h0))/(h1*h2);

  help = summand3/sum1b;

  l4[1] = (help*(sum1a-h2))/(h0*h1);

  l4[3] = summand1 - ((summand2*(sum1b+h2))/(sum1b*h2)) + (help/h2);

}

/*********************************************************************************
                     NEWTON-INTERPOLATION

  ( fuer Interpolationspolynome mit drei oder vier Stuetzpunkten )

Verwendet wird die in frontend.h definierte Datenstruktur des Interpolationsfelds.
Dessen Komponenten lauten :
 - int_point : relative Lage des Stuetzpunkts bezogen auf den Stuetzpunkt in Komponente 0
 - y_values  : Vektor von Funktionswerten an den Stuetzpunkten
 - pol_coefficients : enthaelt die Interpolationspolynomskoeffizienten in vektorieller Form
                      ( relevant : Komponenten 0, 1 und 2)
 - aux_coefficients : enthaelt Hilfsvektoren, die bei den Newtondifferenzen auftauchen.
                      Sie werden behalten, um schnell einzelne Punkte (nur die Stuetz-
		      punkte in Komponenten 2 und 3) ersetzen zu koennen.
		      ( relevant : Komponenten 2, 3 und 4)

Eine Auswertung des Interpolationspolynoms an bestimmten Stellen erhaelt man mit
polynom_auswertung() und polynom3_auswertung(), wobei die letztere Funktion
sich auf das Polynom aus den ersten drei Stuetzpunkten beschraenkt.

Beachte, dass in anderen Teilen des Algorithmus in field[0].int_point und
field[0].aux_coefficients spezielle Daten gespeichert, naemlich Lage eines
interessierenden Punktes und seine interpolierten Funktionswerte. Diese
Komponenten duerfen in den folgenden Funktionen nicht veraendert werden.
In den Auswertungsfunktionen dient wird um einen gewissen Betrag von diesem
interessierenden Punkt entfernt ausgewertet. Ausserdem werden dort alle
negativen Interpolationsergebnisse auf Null gesetzt, da es sich bei den
betrachteten Funktionen ja um Wahrscheinlichkeiten bzw. Dichten handelt !    

Die Komponente field[3].pol_coefficients bleibt voellig ungenutzt

Newton-Differenzen-Schema : Wofuer steht welche Komponente ?

0.0             ..[0].y_values  
                              ..[0].pol_coefficients
..[1].int_point ..[1].y_values                      ..[1].pol_coefficients
                              ..[1].aux_coefficients                      ..[2].pol_coefficients
..[2].int_point ..[2].y_values                      ..[3].aux_coefficients
                              ..[2].aux_coefficients
..[3].int_point ..[3].y_values

**********************************************************************************/
void newton_per_4( field)
/* berechnet fuer ein Interpolationspolynom mit vier Stuetzpunkten die Koeffizienten
   und legt diese in den pol_coefficients-Komponenten ab */
interpolation_row field[4];

{ 
  /* Nullpunkt ist immer x-Koordinate des ersten Stuetzpunkts */
  /* field[0].int_point darf nicht veraendert werden, denn sie traegt
     (hier nicht benoetigte) Information  */
   
  del_vector(&(field[0].pol_coefficients));
  field[0].pol_coefficients = copy_vector(field[1].y_values);
  vector_add_mult(field[0].pol_coefficients, field[0].y_values, -1.0);
  mult_vector_skalar(field[0].pol_coefficients, 1/(field[1].int_point));
  
  del_vector(&(field[1].aux_coefficients));
  field[1].aux_coefficients = copy_vector(field[2].y_values);
  vector_add_mult(field[1].aux_coefficients, field[1].y_values, -1.0);
  mult_vector_skalar(field[1].aux_coefficients, 1/((field[2].int_point)-(field[1].int_point)));
  
  del_vector(&(field[1].pol_coefficients));
  field[1].pol_coefficients = copy_vector(field[1].aux_coefficients);
  vector_add_mult(field[1].pol_coefficients, field[0].pol_coefficients, -1.0);
  mult_vector_skalar(field[1].pol_coefficients, 1/(field[2].int_point));
  
  /* Interpolation fuer drei Stuetzstellen abgeschlossen ... */
  
      del_vector(&(field[2].aux_coefficients));
      field[2].aux_coefficients = copy_vector(field[3].y_values);
      vector_add_mult(field[2].aux_coefficients, field[2].y_values, -1.0);
      mult_vector_skalar(field[2].aux_coefficients, 1/((field[3].int_point)-(field[2].int_point)));
  
      del_vector(&(field[3].aux_coefficients));
      field[3].aux_coefficients = copy_vector(field[2].aux_coefficients);
      vector_add_mult(field[3].aux_coefficients, field[1].aux_coefficients, -1.0);
      mult_vector_skalar(field[3].aux_coefficients, 1/((field[3].int_point)-(field[1].int_point)));
  
      del_vector(&(field[2].pol_coefficients));
      field[2].pol_coefficients = copy_vector(field[3].aux_coefficients);
      vector_add_mult(field[2].pol_coefficients, field[1].pol_coefficients, -1.0);
      mult_vector_skalar(field[2].pol_coefficients, 1/(field[3].int_point));
  }

/*************************************************************************************************/

void newton_per_3( field)
/* berechnet fuer ein Interpolationspolynom mit nur drei Stuetzpunkten die 
   Koeffizienten und legt diese in den pol_coefficients-Komponenten ab */

interpolation_row field[4];

{ 
  /* Nullpunkt ist immer x-Koordinate des ersten Stuetzpunkts */
  /* field[0].int_point darf nicht veraendert werden, denn sie traegt
     (hier nicht benoetigte) Information  */
  
  del_vector(&(field[0].pol_coefficients));  
  field[0].pol_coefficients = copy_vector(field[1].y_values);
  vector_add_mult(field[0].pol_coefficients, field[0].y_values, -1.0);
  mult_vector_skalar(field[0].pol_coefficients, 1/(field[1].int_point));
  
  del_vector(&(field[1].aux_coefficients));
  field[1].aux_coefficients = copy_vector(field[2].y_values);
  vector_add_mult(field[1].aux_coefficients, field[1].y_values, -1.0);
  mult_vector_skalar(field[1].aux_coefficients, 1/((field[2].int_point)-(field[1].int_point)));

  del_vector(&(field[1].pol_coefficients));
  field[1].pol_coefficients = copy_vector(field[1].aux_coefficients);
  vector_add_mult(field[1].pol_coefficients, field[0].pol_coefficients, -1.0);
    mult_vector_skalar(field[1].pol_coefficients, 1/(field[2].int_point));
  
  /* Interpolation fuer drei Stuetzstellen abgeschlossen ... */
  
}

/********************************************************************************************/

void newton_replace_1( field)
/* geht davon aus sich nur in_point und y_values in der letzten Komponente des Interpolationsfeld 
   geaendert haben und ersetzt diesen letzten Stuetzpunkt, indem es nur field[2].pol_coefficients
   neu berechnet */

interpolation_row field[4];

{ 
  /* Nullpunkt ist immer x-Koordinate des ersten Stuetzpunkts */
  /* field[0].int_point darf nicht veraendert werden, denn sie traegt
     (hier nicht benoetigte) Information  */

  del_vector(&(field[2].aux_coefficients));  
  field[2].aux_coefficients = copy_vector(field[3].y_values);
  vector_add_mult(field[2].aux_coefficients, field[2].y_values, -1.0);
  mult_vector_skalar(field[2].aux_coefficients, 1/((field[3].int_point)-(field[2].int_point)));
  
  del_vector(&(field[3].aux_coefficients));
  field[3].aux_coefficients = copy_vector(field[2].aux_coefficients);
  vector_add_mult(field[3].aux_coefficients, field[1].aux_coefficients, -1.0);
  mult_vector_skalar(field[3].aux_coefficients, 1/((field[3].int_point)-(field[1].int_point)));
  
  del_vector(&(field[2].pol_coefficients));
  field[2].pol_coefficients = copy_vector(field[3].aux_coefficients);
  vector_add_mult(field[2].pol_coefficients, field[1].pol_coefficients, -1.0);
  mult_vector_skalar(field[2].pol_coefficients, 1/(field[3].int_point));

}

/********************************************************************************************/

void newton_replace_2( field)
/* geht davon aus sich nur in_point und y_values in den letzten beidenKomponente des 
   Interpolationsfeld geaendert haben und ersetzt diese beiden Stuetzpunkt, indem 
   field[1].pol_coefficients und field[2].pol_coefficients neu berechnet */

interpolation_row field[4];

{ 
  /* Nullpunkt ist immer x-Koordinate des ersten Stuetzpunkts */
  /* field[0].int_point darf nicht veraendert werden, denn sie traegt
     (hier nicht benoetigte) Information  */
  
  del_vector(&(field[1].aux_coefficients));
  field[1].aux_coefficients = copy_vector(field[2].y_values);
  vector_add_mult(field[1].aux_coefficients, field[1].y_values, -1.0);
  mult_vector_skalar(field[1].aux_coefficients, 1/((field[2].int_point)-(field[1].int_point)));
  
  del_vector(&(field[1].pol_coefficients));
  field[1].pol_coefficients = copy_vector(field[1].aux_coefficients);
  vector_add_mult(field[1].pol_coefficients, field[0].pol_coefficients, -1.0);
  mult_vector_skalar(field[1].pol_coefficients, 1/(field[2].int_point));
  
  del_vector(&(field[2].aux_coefficients));
  field[2].aux_coefficients = copy_vector(field[3].y_values);
  vector_add_mult(field[2].aux_coefficients, field[2].y_values, -1.0);
  mult_vector_skalar(field[2].aux_coefficients, 1/((field[3].int_point)-(field[2].int_point)));
  
  del_vector(&(field[3].aux_coefficients));
  field[3].aux_coefficients = copy_vector(field[2].aux_coefficients);
  vector_add_mult(field[3].aux_coefficients, field[1].aux_coefficients, -1.0);
  mult_vector_skalar(field[3].aux_coefficients, 1/((field[3].int_point)-(field[1].int_point)));
  
  del_vector(&(field[2].pol_coefficients));
  field[2].pol_coefficients = copy_vector(field[3].aux_coefficients);
  vector_add_mult(field[2].pol_coefficients, field[1].pol_coefficients, -1.0);
  mult_vector_skalar(field[2].pol_coefficients, 1/(field[3].int_point));

  }
/******************************************************************************************/
sp_vector *polynom_auswertung( t_increment, int_field)
/* ACHTUNG : nur fuer nicht-negative Funktionen */
/* wertet das Interpolationspolynom mit vier Stuetzpunkten an der Stelle
   field[0].int_point + t_increment aus */

double t_increment;
interpolation_row int_field[4];
{
  double x_point,p_factor;
  sp_vector *vnap1;

  x_point = int_field[0].int_point + t_increment;

  vnap1 = copy_vector(int_field[0].y_values);
  vector_add_mult(vnap1, int_field[0].pol_coefficients, x_point);
  
  p_factor = x_point*(x_point-int_field[1].int_point);
  vector_add_mult(vnap1, int_field[1].pol_coefficients, p_factor);

  p_factor = p_factor*(x_point-int_field[2].int_point);
  vector_add_mult(vnap1, int_field[2].pol_coefficients, p_factor);

  set_negs_to_zero( vnap1);

  return (vnap1);
}    
 
/***********************************************************************/   
sp_vector *polynom3_auswertung( t_increment, int_field)
/* ACHTUNG : nur fuer nicht-negative Funktionen */
/* wertet das Interpolationspolynom mit drei Stuetzpunkten an der Stelle
   field[0].int_point + t_increment aus */

double t_increment;
interpolation_row int_field[4];
{
  double x_point;
  sp_vector *vnap1;

  t_increment = int_field[0].int_point + t_increment;

  vnap1 = copy_vector(int_field[0].y_values);
  vector_add_mult(vnap1, int_field[0].pol_coefficients, t_increment);
  
  t_increment = t_increment*(t_increment-int_field[1].int_point);
  vector_add_mult(vnap1, int_field[1].pol_coefficients, t_increment);

  set_negs_to_zero( vnap1);
 
  return (vnap1);
} 

/******************************************************************************************
                          EXTRAPOLATION                                                   

                       nach Aitken-Neville

Formeln : fuer drei Stuetzpunkte x_i und unter der Bedingung, dass der interessierende Punkt
          zeta auf der Zeitachse rechts von diesen Stuetzpunkten liegt.

	  x_0 <-h0-> x_1 <-h1-> x_2 <- h2 -> zeta

	  zet_a = ( f(x_1)*(h2+h1+h0) - f(x_0)*(h2+h1 )/h0

	  zet_b = ( f(x_2)*(h2+h1) - f(x_1)*h2/h1

	  -> zeta = ( zeta_b*(h2+h1+h0) - zeta_a*h2 )/(h0+h1)

	  ( zeta_a und zeta_b sind extrapolierte Werte fuer die gesuchte Stelle,
	    berechnet mit jeweils zwei Stuetzpunkten )
	    
	  Diese Rechnung leistet extrapol_routine(), wobei die auftretenden sechs
	  Koeffizienten vorher berechnet worden sein muessen.
******************************************************************************************/
void extrapol_routine( factors, values0, values1, values2)
/* *values0 wird ueberschrieben, alles andere bleibt unveraendert */

double factors[6];
sp_vector *values0;
sp_vector *values1;
sp_vector *values2;
{
  sp_vector *zeta;

  mult_vector_skalar( values0, factors[1]);
  vector_add_mult( values0, values1, factors[0]);
  
  zeta = copy_vector( values2);
  mult_vector_skalar( zeta, factors[2]);
  vector_add_mult( zeta, values1, factors[3]);
  
  mult_vector_skalar( values0, factors[5]);
  vector_add_mult( values0, zeta, factors[4]);

  set_negs_to_zero( values0);
  del_vector(&zeta);
}

/*****************************************************************************************/

void last_extrapolation( first_of_four)
/* Extrapolation bei erreichter Diskontinuitaet : Uebergabe der letzten vier step_inbetween-Elemente
   vor genullter low_box (evtl. dummy dazwischen); first_of_four zeigt auf das erste von ihnen ! /
/* implizite Voraussetzung : mindestens immer vier Stuetzpunkte zwischen Diskontinuitaeten ! */

step_inbetween *first_of_four;
{
  int i;
  double factors[6];
  step_inbetween *third_of_four, *second_of_four, *forth_of_four;

  second_of_four = first_of_four->next_down;
  third_of_four = second_of_four->next_down;
  forth_of_four = third_of_four->next_down;
  factors[1] = (-1.0)*((second_of_four->step + third_of_four->step)/first_of_four->step);
  factors[0] = 1.0 - factors[1];
  factors[3] = (-1.0)*(third_of_four->step/second_of_four->step);
  factors[2] = 1.0 - factors[3];
  factors[5] = (-1.0)*(third_of_four->step/(first_of_four->step + second_of_four->step));
  factors[4] = 1.0 - factors[5];

  for( i=0; i<ac_no_detT; i++) {
    /* leider wurde dieser Leervektor gerade in update_int_structures() erzeugt */
    del_vector( &(forth_of_four->values[0][i]));

    forth_of_four->values[0][i] = copy_vector( first_of_four->values[0][i]);
    extrapol_routine( factors, forth_of_four->values[0][i], second_of_four->values[0][i], third_of_four->values[0][i]);
  };
}
/*****************************************************************************************************/

void regular_extrapolation( add_here, plow_box, include)
/* extrapoliert mit moeglichst vielen Stuetzpunkten aus der low_box mit evtl. Hinzunahme eines
   weiteren Stuetzpunktes in dem plow_box->next_up-Element, aber mit maximal drei Stuetzpunkten */
/* darf nicht aufgerufen werden, falls Diskontinuitaet erreicht wird -> last_extrapolation() */

int add_here;
lower_stepbox *plow_box;
int include; /* soll plow_box->next_up miteinbezogen werden in Extrapolation, nur fuer add_here == 3 */
{
  int i;

  switch (add_here) {
  case 2 : for( i=0; i<ac_no_detT; i++) {
             /* letzter Stuetzpunkt ist bestmoegliche Schaetzung */
             del_vector(&(plow_box->lower_steps[1].values[0][i]));
	     plow_box->lower_steps[1].values[0][i] = copy_vector( plow_box->lower_steps[0].values[0][i]);
	   };
           break;
  case 3 : if (!include) {
             /* plow_box->next_up-Element kann nicht mit einbezogen werden, weil zwischen ihm und
		dem ersten low_box-Element eine Diskontinuitaet liegt */
             { /* lineare Extrapolation ( analog zeta_a/zeta_b -> siehe oben  */
	       double factor1, factor2;

	       factor2 = (-1.0)*(plow_box->lower_steps[1].step/plow_box->lower_steps[0].step);
	       factor1 = 1.0 - factor2;
	       for( i=0; i<ac_no_detT; i++) {
		 del_vector(&(plow_box->lower_steps[2].values[0][i]));	       
		 plow_box->lower_steps[2].values[0][i] = copy_vector( plow_box->lower_steps[1].values[0][i]);
		 mult_vector_skalar( plow_box->lower_steps[2].values[0][i], factor1);
		 vector_add_mult( plow_box->lower_steps[2].values[0][i], plow_box->lower_steps[0].values[0][i], factor2);
		 set_negs_to_zero( plow_box->lower_steps[2].values[0][i]);
	       };
	     };
	   }
           else { /* include == 1 */
	     { /* quadratische Extrapolation */
	       /* nochmal ueberlegen, ob ueber low_box kein dummy-Element sein kann ? */
	       double factors[6];

	       factors[1] = (-1.0)*((plow_box->lower_steps[0].step + plow_box->lower_steps[1].step)/plow_box->next_up->step);
	       factors[0] = 1.0 - factors[1];
	       factors[3] = (-1.0)*(plow_box->lower_steps[1].step/plow_box->lower_steps[0].step);
	       factors[2] = 1.0 - factors[3];
	       factors[5] = (-1.0)*(plow_box->lower_steps[1].step/(plow_box->next_up->step + plow_box->lower_steps[0].step));
	       factors[4] = 1.0 - factors[5];
				    
	       for( i=0; i<ac_no_detT; i++) {
		 del_vector(&(plow_box->lower_steps[2].values[0][i]));
		 plow_box->lower_steps[2].values[0][i] = copy_vector( plow_box->next_up->values[0][i]);
		 extrapol_routine( factors, plow_box->lower_steps[2].values[0][i], plow_box->lower_steps[0].values[0][i], plow_box->lower_steps[1].values[0][i]);
	       };
	     };
	   };
           break;
  case 4 : { /* quadratische Extrapolation */
	     double factors[6];

	     factors[1] = (-1.0)*((plow_box->lower_steps[1].step + plow_box->lower_steps[2].step)/plow_box->lower_steps[0].step);
	     factors[0] = 1.0 - factors[1];
	     factors[3] = (-1.0)*(plow_box->lower_steps[2].step/plow_box->lower_steps[1].step);
	     factors[2] = 1.0 - factors[3];
	     factors[5] = (-1.0)*(plow_box->lower_steps[2].step/(plow_box->lower_steps[1].step + plow_box->lower_steps[0].step));
	     factors[4] = 1.0 - factors[5];
	     
	     for( i=0; i<ac_no_detT; i++) {
	       del_vector(&(plow_box->lower_steps[3].values[0][i]));
	       plow_box->lower_steps[3].values[0][i] = copy_vector( plow_box->lower_steps[0].values[0][i]);
	       extrapol_routine( factors, plow_box->lower_steps[3].values[0][i], plow_box->lower_steps[1].values[0][i], plow_box->lower_steps[2].values[0][i]);
	     };
	   };
           break;
  case 5 : { /* quadratische Extrapolation */
	     double factors[6];

	     factors[1] = (-1.0)*((plow_box->lower_steps[2].step + plow_box->lower_steps[3].step)/plow_box->lower_steps[1].step);
	     factors[0] = 1.0 - factors[1];
	     factors[3] = (-1.0)*(plow_box->lower_steps[3].step/plow_box->lower_steps[2].step);
	     factors[2] = 1.0 - factors[3];
	     factors[5] = (-1.0)*(plow_box->lower_steps[3].step/(plow_box->lower_steps[1].step + plow_box->lower_steps[2].step));
	     factors[4] = 1.0 - factors[5];
	     
	     for( i=0; i<ac_no_detT; i++) {
	       del_vector(&(plow_box->lower_steps[4].values[0][i]));
	       plow_box->lower_steps[4].values[0][i] = copy_vector( plow_box->lower_steps[1].values[0][i]);
	       extrapol_routine( factors, plow_box->lower_steps[4].values[0][i], plow_box->lower_steps[2].values[0][i], plow_box->lower_steps[3].values[0][i]);
	     };
	   };
           break;
  default : 
#ifdef DEBUG_AC
            printf(" Fehler in regular_extrapolation : add_here = %d", add_here);
#endif
            break;
  };
  
}
/*******************************************************************************************************/
	       




