/******************************************************************************
 File:              weighter.c

 Purpose:           Calculation of Poisson-probabilities; verwendet von shared.c in randomi()

 History:           uebernommen aus EDSPNAnalysis/src/weighter.c
                    fuer Dipl.Arbeit in 10.94
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "frontend.h"
#include "Options.h"
#include "portab.h"

/* Import von main.c */
extern TransientOptions transientOptions;

/* procedures for the efficient and stable calculation of poisson probabilities */

/******************************************************************************

 Funktion:	d

 Parameter:	k und lambda

 Rueckgabewert:	double-Wert d

 Zweck:		Wird von find_r() benoetigt

******************************************************************************/
double d (k, lambda)
double k, lambda;
{
  return 1.0 / (1.0 - exp (-(2.0/9.0) * (k * sqrt (2.0 * lambda) + 3.0/2.0)));
}


/******************************************************************************

 Funktion:	find_r

 Parameter:	lambda, epsilon, Mode m, k_p

 Rueckgabewert:	Der rechte Abschnittspunkt

 Zweck:		Berechnet den rechten Abschnittspunkt

******************************************************************************/
double find_r (lambda, epsilon, m, k_p)
double lambda, epsilon, m, *k_p;
{
	double a_lambda;
	double k;
	double q_k_lambda;
	double R;

	a_lambda = (1.0 + 1.0/lambda) * exp(1.0/16.0) * sqrt(2.0);

	/* corollary 1 of section 4 : k (bzw. R) berechnen */
	for (k = 3.0; /*k < 10.0*/; k++)
	{
		/* q_k_lambda berechnen */
		q_k_lambda = a_lambda * d(k,lambda) * exp(-pow(k,2.0)/2.0) / (k * sqrt(2.0*M_PI));
		/* R berechnen */
		R = ceil(m + k*sqrt(2.0*lambda) + 3.0/2.0);
		if (q_k_lambda < epsilon/2.0)
			break;
	}
	*k_p = k;
	return R;
} /* find_r */


/******************************************************************************

 Funktion:	finder

 Parameter:	lambda --> Wert von lambda
		epsilon --> Gewuenschte Genauigkeit
		tau --> Kleinster darstellbarer Wert
		omega --> Groesster darstellbarer Wert
		m --> Mode von Lambda
		L_p --> Linker Abschnittspunkt
		R_p --> Rechter Abschnittspunkt
		w_m_p --> Der Wert von w(m)
		F_p --> Flag ob Fehler aufgetreten sind

 Rueckgabewert:	---

 Zweck:		Berechnet den linken und rechten Abschnittspunkt und
		den Wert von w(m).

******************************************************************************/
void finder (lambda, epsilon, tau, omega, m, L_p, R_p, w_m_p, F_p)
double lambda, epsilon, tau, omega, m, *L_p, *R_p, *w_m_p;
char *F_p;
{
	double L, R, w_m;
	char F = 1;
	double b_lambda;
	double t_k_lambda;
	double k;
	double c_m;
	double q_dach;
	double p_lambda_k;

	if (lambda <= 0.0)
	{
		L = R = 0.0;
		F = 0;
		goto EXIT;
	}

	R = find_r(lambda, epsilon, m, &k);

	if (lambda < 25.0)
	{
		L = 0.0;
	}
	else /* lambda >= 25.0 */
	{
		/* b_lambda berechnen */
		b_lambda = (1.0 + 1.0/lambda) * exp(1.0/(8.0*lambda));

		/* corollary 2 of section 4 : k (bzw. L) berechnen */
		t_k_lambda = b_lambda * exp(-pow(k,2.0)/2.0) / (k*sqrt(2.0*M_PI));
		if (t_k_lambda >= epsilon/2.0)
		{
			printf ("FINDER: WARNING: t_k_lambda isn't less then epsilon/2 !\n");
		}
		L = floor(m - k*sqrt(lambda) - 3.0/2.0);
		if (L<0.0)
		   L = 0.0;
	}

	/* w(m) berechnen */
	w_m = omega / (1.0e10 * (R - L));

	if (lambda >= 25.0)
	{
		/* corollary 3 of section 5 : p_lambda_k berechnen */
		c_m = (1.0 / sqrt(2.0*M_PI*m)) * exp(m - lambda - 1.0/(12.0*m));
		q_dach = k * sqrt(2.0) + 3.0/(2.0*sqrt(lambda));
		p_lambda_k = c_m * exp(- pow(q_dach + 1.0,2.0) / 2.0);
		if (p_lambda_k * w_m < tau)
		{
			F = 0;
			goto EXIT;
		}

		/* corollary 4 of section 5 : p_lambda_k berechnen */
		q_dach = k + 3.0/(2.0*sqrt(lambda));
		if (q_dach <= sqrt(lambda)/2.0)
		{
			p_lambda_k = c_m * exp( -(pow(q_dach,2.0)/2.0) - 
					       pow(q_dach,3.0)/(3.0*sqrt(lambda)) );
		}
#if 0
		else
		  puts ("ACHTUNG: corollary 4 of section 5 (ii) und (iii) noch nicht implementiert !!");
#endif

		if (p_lambda_k * w_m > tau)
			F = 1;
		else
			F = 0;
	}

EXIT:
	*L_p = L;
	*R_p = R;
	*w_m_p = w_m;
	*F_p = F;
} /* finder */


/******************************************************************************

 Funktion:	weighter

 Parameter:	lambda -> Wert von Lambda
		epsilon --> geforderte Genauigkeit
		tau --> kleinster darstellbarer Wert
		omega --> groesster darstellbarer Wert
		L --> Linker Abschnittspunkt
		R --> Rechter Abschnittspunkt
		w --> Die Gewichtungen
		W --> Summe der Gewichte
		F --> Flag ob Fehler aufgetreten sind

 Rueckgabewert:	---

 Zweck:		Berechnung der Poisson Wahrscheinlichkeiten fuer
		grosse Werte von lambda.

******************************************************************************/
void weighter (lambda, epsilon, tau, omega, L, R, w, W, F)
	double	lambda;	/* in */
	double	epsilon;
	double	tau;
	double	omega;
	double	*L;	/* out */
	double	*R;
	double	**w;	/* w(L),...,w(R) */
	double	*W;
	char	*F;
{
	double m;
	double w_m;
	double j;
	unsigned i, s, t;
	double q;

/*Initialize:*/
	m = floor(lambda);	/* mode */

#ifdef DEBUG
	printf ("before finder-call: lambda = %lg, epsilon = %lg, tau = %lg, m = %lg\n",
		lambda, epsilon, tau, m);
#endif

	finder (lambda, epsilon, tau, omega, m, L, R, &w_m, F);
	if (*F == 0)
	{
		printf ("FINDER fails !!\n");
		printf ("L=%.0f R=%.0f w_m=%e F=%d\n", *L, *R, w_m, (int)*F);
		return;
	}
	*w = (double *) emalloc ((int)(*R - *L + 1) * sizeof(double));
	(*w)[(int)(m-*L)] = w_m;
/*Down:*/
	for (j = m, i = (unsigned)(j - *L); j > *L; j--, i--)
	{
		(*w)[i-1] = (j / lambda) * (*w)[i];
	}
/*Up:*/
	if (lambda < 400.0) /* Special */
	{
#ifdef DEBUG
		puts ("Special:");
#endif
		if (*R > 600.0)
		{
			printf ("Special: R > 600 !\n");
			*F = 0;
			return;
		}
		j = m; i = (unsigned)(j - *L);
		while (j < *R)
		{
			q = lambda / (float)(j + 1);
			if ((*w)[i] > tau / q)
			{
				(*w)[i+1] = q * (*w)[i];
				j++; i++;
			}
			else
			{
				*R = j;
				goto Compute_W;
			}
		}
	}
	else
	{
		for (j = m, i = (unsigned)(j - *L); j < *R; j++, i++)
		{
			(*w)[i+1] = (lambda / (float)(j + 1)) * (*w)[i];
		}
	}
Compute_W:
	*W = 0.0;
	s = 0;
	t = (unsigned)(*R - *L);
	while (s < t)
	{
		if ((*w)[s] <= (*w)[t])
			*W += (*w)[s++];
		else
			*W += (*w)[t--];
	}
	*W += (*w)[s];
} /* weighter */


/******************************************************************************/
void call_weighter(lambda, L, R, Betas)
/******************************************************************************/
double	lambda;	
int *L;
int *R;
double **Betas;
{
  double L_d,R_d;  /* weighter args */
  double beta_sum;
  char error_flag;
  int k;


  if (fabs(lambda) < transientOptions.precision)
  {
    *L = *R = 0;
    *Betas = (double *)emalloc(sizeof(double));
    (*Betas)[0] = 1.0;
    return;
  };

  /* Beta(t*qmax_ii,k) Faktoren sowie Linke und Rechte Schranke berechnen */
  error_flag = 0;
  weighter(lambda, transientOptions.precision, 1.0e-300, 1.0e300,
	   &L_d,
	   &R_d,
	   Betas,
	   &beta_sum,
	   &error_flag);
  if (!error_flag)
  {	
    printf("weighter fails.\n");
    exit(-1);
  }
  *L = (int)(L_d + 0.5);
  *R = (int)(R_d + 0.5);
  for (k = 0; k <= *R - *L; k++)
  {
    (*Betas)[k] /= beta_sum;
    if( isinf( (*Betas)[k] ))
    {
      printf ("Error: Beta[%d] is infinite.\n", k);
      exit(-1);
    }
  }
}
/******************************************************************************/
void call_new_weighter(lambda, L, R, Betas, prec)
/* wie oben, nur kann absolute Genauigkeit vorgegeben werden */
/******************************************************************************/
double	lambda;	
int *L;
int *R;
double **Betas;
double prec;
{
  double L_d,R_d;  /* weighter args */
  double beta_sum;
  char error_flag;
  int k;

  if (*Betas != NULL) 
    efree( *Betas);

  if (fabs(lambda) < prec)
  {
    *L = *R = 0;
    *Betas = (double *)emalloc(sizeof(double));
    (*Betas)[0] = 1.0;
    return;
  };

  /* Beta(t*qmax_ii,k) Faktoren sowie Linke und Rechte Schranke berechnen */
  error_flag = 0;
  weighter(lambda, prec, 1.0e-300, 1.0e300,
	   &L_d,
	   &R_d,
	   Betas,
	   &beta_sum,
	   &error_flag);
  if (!error_flag)
  {	
    printf("weighter fails.\n");
    exit(-1);
  }
  *L = (int)(L_d + 0.5);
  *R = (int)(R_d + 0.5);
  for (k = 0; k <= *R - *L; k++)
  {
    (*Betas)[k] /= beta_sum;
    if( isinf( (*Betas)[k] ))
    {
      printf ("Error: Beta[%d] is infinite.\n", k);
      exit(-1);
    }
  }
}
/************************************************************************/
