/******************************************************************************
 Program:           net2tn.c

 Purpose:           Convert the .NET, .DEF, and .pmf files to a single .TN file.

 Project Director:  Reinhard German

 Programmer:        Ulrich Hertlein

 History:           Release 1: April 26, 1994
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*
 * defines
 */
#define BUFLEN 168

#define NO_SPN -1
#define GSPN   0
#define DSPN   1
#define EDSPN  2


/*
 * local variables
 */
char *progname;

char *net_name[] = { "GSPN", "DSPN", "EDSPN" };
int n_im = 0, n_exp = 0, n_det = 0, n_gen = 0;
int n_measures = 0;

char *delay_name[] = { "EXP_DELAY", "DET_DELAY", "WEIGHT" };

/*
 * usage:
 */
int usage()
{
   fprintf(stderr,"Usage: %s <Modelname>\n", progname);
   fprintf(stderr,"\t<Modelname> is the relative pathname of the model from\n");
   fprintf(stderr,"\tthe model-root directory. Model-root is either TNETZ_MODEL\n");
   fprintf(stderr,"\tor MODELDIR (one of them must be set, MODELDIR takes ");
   fprintf(stderr,"preference)\n");
   exit( -1 );
}


/*
 * fgetline:
 */
int fgetline(buf, fp)
char *buf;
FILE *fp;
{
   /* read line, strip newline */
   fgets(buf,BUFLEN,fp);
   buf[strlen(buf) - 1] = '\0';

   return( feof(fp) );
}


/*
 * translate_comment: translates the `#' comment into `--' comment;
 *                    returns 1 if a comment was recognized, else 0
 */
int translate_comment(buf, fp)
char *buf;
FILE *fp;
{
   /* translate comment */
   if( buf[0] == '#' )
   {
      fprintf(fp,"--%s\n", buf+1);
      return( 1 );
   }

   return( 0 );
}


/*
 * read_net: read .NET file, count number of transitions of each type, write to
 *           fragment file
 */
int read_net(fp)
FILE *fp;
{
   FILE *tf;
   char buf[BUFLEN];
   int skip = 0;

   char tname[42], delay[42], enabling[4], type[4];
   int prio,orient;
   double x,y, x1,y1, x2,y2;

   /* create file */
   if( !(tf = fopen("/tmp/frag0", "w")) )
   {
      fprintf(stderr,"Unable to create: /tmp/frag0\n");
      exit( -1 );
   }

   /* read .NET file */
   while( !feof(fp) )
   {
      /* get line */
      if( fgetline(buf,fp) )
         break;

      if( skip )
      {
         skip = 0;
         continue;
      }

      /* filter header & trailer */
      if( !strncmp(buf,"# FILE ",7) || !strncmp(buf,"# END ",5) )
      {
         skip = 1;
         continue;
      }

      /* translate comment */
      if( translate_comment(buf,tf) )
         continue;

      /* change transition description */
      if( !strncmp(buf,"TRANSITION ",11) )
      {
         sscanf(buf+11,"%s %s %s %s %d %d %lf %lf %lf %lf %lf %lf", tname,delay,enabling,type,&prio,&orient, &x,&y, &x1,&y1, &x2,&y2);

         /* convert <pmf> to <MD> */
         if( !strcmp(delay,"<pmf>") )
            strcpy(delay,"<MD>");

         /* count types */
         if( !strcmp(type,"IM") )
            n_im++;
         else if( !strcmp(type,"EXP") )
            n_exp++;
         else if( !strcmp(type,"DET") )
            n_det++;
         else if( !strcmp(type,"GEN") )
            n_gen++;
         else
	 {
            fprintf(stderr,"Unknown transition type: %s\n", type);
            exit( -1 );
	 }

         /* write transition */
         fprintf(tf,"TRANSITION %s %s %s %s RE %d %d %.2lf %.2lf %.2lf %.2lf %.2lf %.2lf\n",
            tname,delay,enabling,type,/* firing = RE */
            prio,orient, x,y, x1,y1, x2,y2);
         continue;
      }

      fprintf(tf,"%s\n", buf);
   }

   /* close file */
   fclose(tf);

   return( 0 );
}


/*
 * read_def: read .DEF file, write to fragment file
 */
int read_def(fp)
FILE *fp;
{
   FILE *tf, *t1,*t3;
   char buf[BUFLEN];
   int skip = 0, delay_type = 0;

   char tname[42];

   /* create file */
   if( !(t1 = fopen("/tmp/frag1", "w")) )
   {
      fprintf(stderr,"Unable to create: /tmp/frag1\n");
      exit( -1 );
   }
   if( !(t3 = fopen("/tmp/frag3", "w")) )
   {
      fprintf(stderr,"Unable to create: /tmp/frag3\n");
      exit( -1 );
   }
   tf = t1;

   /* read .DEF file */
   while( !feof(fp) )
   {
      /* get line */
      if( fgetline(buf,fp) )
         break;

      if( skip )
      {
         skip = 0;
         continue;
      }

      /* filter header & trailer */
      if( !strncmp(buf,"# FILE ",7) || !strncmp(buf,"# END ",5) )
      {
         skip = 1;
         continue;
      }

      /* filter FIRING DELAYS */
      if( !strncmp(buf,"# MARKING DEPENDENT FIRING DELAYS FOR EXP.",41) )
         delay_type = 1;
      else
         if( !strncmp(buf,"# MARKING DEPENDENT FIRING DELAYS FOR DET.",41) )
            delay_type = 2;

      /* filter WEIGHTS */
      if( !strncmp(buf,"# MARKING DEPENDENT WEIGHTS ",27) )
      {
         delay_type = 3;
         tf = t3;
      }

      /* translate comments */
      if( translate_comment(buf,tf) )
         continue;

      /* change transition description */
      if( !strncmp(buf,"TRANSITION ",11) )
      {
         sscanf(buf+11,"%s", tname);

         switch( delay_type )
	 {
         case 1:
         case 2:
         case 3:
            fprintf(tf,"%s %s ", delay_name[delay_type - 1],tname);
            break;
         default:
            fprintf(stderr,"Unknown delay_type: %d\n", delay_type);
            exit( -1 );
	 }

         continue;
      }

      /* change measure line */
      if( !strncmp(buf,"MEASURE ",8) )
      {
         fprintf(tf,"%s ", buf);
         n_measures++;
         continue;
      }

      fprintf(tf,"%s\n", buf);
   }

   /* close file */
   fclose(t1);
   fclose(t3);

   return( 0 );
}


/*
 * read_pmf: read .pmf file, write to fragment file
 */
int read_pmf(fp)
FILE *fp;
{
   FILE *tf;
   char buf[BUFLEN];

   char tname[42];

   /* create file */
   if( !(tf = fopen("/tmp/frag2", "w")) )
   {
      fprintf(stderr,"Unable to create: /tmp/frag2\n");
      exit( -1 );
   }

   fprintf(tf,"-- PROBABILITY MASS FUNCTION DEFINITIONS FOR GEN. TRANSITIONS:\n");

   /* read .pmf file */
   if( fp )
      while( !feof(fp) )
      {
         /* get line */
         if( fgetline(buf,fp) )
            break;

         /* translate comments */
         if( translate_comment(buf,tf) )
            continue;

         /* change transition description */
         if( !strncmp(buf,"TRANSITION ",11) )
         {
            sscanf(buf+11,"%s", tname);
            fprintf(tf,"GEN_DELAY %s ", tname);
            continue;
         }

         fprintf(tf,"%s\n", buf);
      }

   /* close file */
   fclose(tf);
   
   return( 0 );
}


/*
 * my_copy:
 */
int my_copy(file,fp)
char *file;
FILE *fp;
{
   FILE *in;
   char buf[BUFLEN];

   /* open input file */
   if( !(in = fopen(file,"r")) )
   {
      fprintf(stderr,"Unable to open: '%s'\n", file);
      return( -1 );
   }

   /* copy contents */
   while( !feof(in) )
      if( fgetline(buf,in) )
         break;
      else
         fprintf(fp,"%s\n", buf);

   /* close file */
   fclose(in);

   return( 0 );
}


/*
 * collect:
 */
int collect(tnfp)
FILE *tnfp;
{
   int type = NO_SPN;

   /* determine type */
   if( n_exp > 0 )
      type = GSPN;
   if( n_det > 0 )
      type = DSPN;
   if( n_gen > 0 )
      type = EDSPN;

   if( type == NO_SPN )
   {
      fprintf(stderr,"Invalid network description\n");
      exit( -1 );
   }

   fprintf(tnfp,"NET_TYPE:               %s\n", net_name[type]);
   fprintf(tnfp,"DESCRIPTION:            Generated by NET2TN\n");
   fprintf(tnfp,"REWARD_MEASURES:        %d\n", n_measures);

   /* copy fragment files */
   my_copy("/tmp/frag0", tnfp);
   my_copy("/tmp/frag1", tnfp);
   my_copy("/tmp/frag2", tnfp);
   my_copy("/tmp/frag3", tnfp);

   /* remove fragment files */
   system("rm /tmp/frag?");

   return( 0 );
}


/*
 * main:
 */
int main(argc, argv)
int argc;
char **argv;
{
   char *model, *model_dir, model_path[256];

   char net_name[256], def_name[256], pmf_name[256], tn_name[256];
   FILE *net_fp, *def_fp, *pmf_fp, *tn_fp;

   /* check arguments */
   progname = argv[0];
   model = argv[1];
   if( argc != 2 )
      usage();

   /* fetch modeldir */
   if( !(model_dir = getenv("MODELDIR")) )
   {
      if( !(model_dir = getenv("TNETZ_MODELS")) )
      {
         fprintf(stderr,"Environment not set: specify MODELDIR or TNETZ_MODELS\n");
         exit( -1 );
      }
   }
   fprintf(stdout,"Modeldir: %s\tModel: %s\n", model_dir,model);

   /* build filenames */
   sprintf(model_path,"%s/%s", model_dir,model);
   sprintf(net_name,"%s.NET", model_path);
   sprintf(def_name,"%s.DEF", model_path);
   sprintf(pmf_name,"%s.pmf", model_path);
   sprintf(tn_name,"%s.TN", model_path);

   /* open input files */
   if( (net_fp = fopen(net_name,"r")) == NULL )
   {
      fprintf(stderr,"File not found: '%s'\n", net_name);
      exit( -1 );
   }
   if( (def_fp = fopen(def_name,"r")) == NULL )
   {
      fprintf(stderr,"File not found: '%s'\n", def_name);
      exit( -1 );
   }
   if( (pmf_fp = fopen(pmf_name,"r")) == NULL )
   {
      fprintf(stderr,"File not found: '%s'\n", pmf_name);
      n_gen = -1;
   }

   /* open output file */
#ifdef DEBUG
   tn_fp = stdout;
#else
   if( (tn_fp = fopen(tn_name,"w")) == NULL )
   {
      fprintf(stderr,"Unable to create: '%s'\n", tn_name);
      exit( -1 );
   }
#endif

   /* parse files */
   read_net(net_fp);
   read_def(def_fp);
   read_pmf(pmf_fp);

   /* write header */
   fprintf(tn_fp,"-- FILE %s.TN CONTAINING DESCRIPTION OF THE NET\n", model);
   fprintf(tn_fp,"\n");

   /* collect results */
   collect(tn_fp);

   /* write trailer */
   fprintf(tn_fp,"\n");
   fprintf(tn_fp,"-- END OF SPECIFICATION FILE\n");

   /* close files */
   fclose(net_fp);
   fclose(def_fp);
   fclose(pmf_fp);
   fclose(tn_fp);

   return( 0 );
}
