/*****************************************************************************

	File Name	:	showSMC.c

	Parameter	:	Options, Net Name

	Return Value	:	0 -> Program successfully terminated
				1 -> Error occurred

	Purpose		:	showSMC displays data of subordinated
				markoff chains of a DSPN
 
	Input Files	:	<net>.TN, <net>.RG, <net>.CONV_i, <net>.EMC_i

	Project Director:	Christoph Lindemann
				Technical University of Berlin

	Programmer	:	Armin Zimmermann

	History		:	First version May 1993

*                                                                           *
*  Modifiziert	  :	Joachim Schneider				    *
*									    *
*  Datum	  :	July 1994 					    *
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "matrices_P_C.h"

#define	USAGE	"Usage: showSMC [] <netname>\n"
#define FILE_NAME	100

#include "portab.h"

typedef struct rg_type		/* <net>.RG */
{	unsigned	number;
}  rg_type;

typedef	struct m_type		/* a tangible marking */
{	int	number;		/* Zustandsnummer */
	int	*token;		/* token in places */
	int	used;
	struct m_type *next;
}  m_type;

/*	global variables */
int	no_places;			/* number of places in the net */
m_type	*markings = NULL;		/* tangible markings */
int	no_markings;			/* number of tangible markings */
char	**place_names;			/* names of places */
int	*place_bounds;			/* upper bounds of tokens in places */
FILE 	*net_file;			/* net definition file .TN */
FILE 	*rg_file;			/* reachability set file .RG */
FILE	*p_file;			/* EMC files */
FILE	*c_file;			/* CONV files */

/*	next_line skips to the beginning of the next data line
	and returns the number of the token; 0 = EOF */
#define	NEXT_LINE_BUF	500
#define NO_TOKENS	12
char	tokens[NO_TOKENS][30] = {"eof", "PLACES:", "TRANSITIONS:", 
			"DELAY_PARAMETERS:", "MARKING_PARAMETERS:",
			"MARKPAR", "PLACE", "DELAYPAR", "TRANSITION", 
			"INPARCS", "OUTPARCS", "INHARCS" };
int
next_line(in_file, buffer)
	FILE *in_file;
	char *buffer;
{	
	int	beg, tok, copy;
	char	in_token[50];

	while(1)
	{	if (fgets(buffer, NEXT_LINE_BUF, in_file) == NULL)
			return 0;		/* EOF reached */
		for (beg = 0; beg < NEXT_LINE_BUF; beg++)
			if (!isspace(buffer[beg]))
				break;
		if (buffer[beg] && (beg < NEXT_LINE_BUF) && (buffer[beg] != '#'))
			break;		/* line is not empty nor a comment */
	}
	for (copy = 0; buffer[beg] && !isspace(buffer[beg]); )
		in_token[copy++] = buffer[beg++];
	in_token[copy] = 0;
	for (tok = 1; tok < NO_TOKENS; tok++)
		if (!strcmp(in_token, tokens[tok]))
		{	for (copy = 0; buffer[beg]; )
				buffer[copy++] = buffer[beg++];
			buffer[copy] = 0;
			return tok;
		}
	return -1;			/* not found */
}

/*	read_net reads the net-description from net.TN */
void
read_net(net)
	FILE *net;
{	
	static	char buf[NEXT_LINE_BUF];
	char	help[50], h1[50], h2[50];
	int	code, i_place, i_trans;

	no_places = i_place = 0;
	while(code = next_line(net, buf))
	{	switch(code)
		{	case 1:	sscanf(buf, "%d", &no_places);
				place_names = (char **)malloc(no_places * sizeof(char *));
				break;
			case 6:	if (i_place >= no_places)
				{	printf("too many places, line %s\n", buf);
					exit(1);
				}
				sscanf(buf, "%s ", help);
				place_names[i_place] = strdup(help);
				i_place++;
				break;
			default:break;
		}
	}
}

/* look for a specific marking */
m_type *
m_lookup(num)
	int	num;	/* number of marking */
{
	m_type	*walk = markings;
	int	count = num;

	while(walk && --count)
		walk = walk->next;
	if (count || (walk->number != num))
	{	printf("Error: marking #%d not found\n", num);
		exit(1);
	}
	return walk;
}

/*	main procedure */
main(argc, argv)
	int argc;
	char ** argv;
{
	char	net_name[FILE_NAME], rg_name[FILE_NAME];
	char	p_name[FILE_NAME], c_name[FILE_NAME];
	char	options[10], topline[1000], help[100];
	int	arg_num = 1, i, ende, k, len, bit;
	int	file_num, nums, nums_c;
	rg_type	rg;
	m_type	*new_m, *walk;
	double	dummy;

	if (argc == 1)
	{	printf("ERROR: no netname found\n");
		printf(USAGE);
		exit(1);
	}
	if (argv[1][0] == '-')
	{	arg_num++;
		strcpy(options, argv[1]+1);
		for(i = 0; i < strlen(options); i++)
			switch(options[i])
			{	default : printf(USAGE);
					  exit(1);
			}
	}
	if (argc <= arg_num)
	{	printf("ERROR: no netname found\n");
		printf(USAGE);
		exit(1);
	}
	strcpy(net_name, argv[arg_num]);
	strcpy(rg_name, argv[arg_num]);
	strcat(net_name, ".TN");
	strcat(rg_name, ".RG");
	if ((net_file = fopen(net_name, "r")) == NULL)
	{	printf("ERROR: could not open %s for reading\n", net_name);
		exit(1);
	}
	if ((rg_file = fopen(rg_name, "r")) == NULL)
	{	printf("ERROR: could not open %s for reading\n", rg_name);
		exit(1);
	}
	printf("\nSubordinated Markoff chain data for %s:\n\n", argv[arg_num]);
	read_net(net_file);		/* read the net description file */
	if (no_places < 1)
	{	printf("Error: no places.\n");
		exit(1);
	}
	printf("The net contains %d places.\n", no_places);
	printf("Mapping numbers to places of the DSPN:\n");
	for (i=0; i<no_places; i++)
		printf("%3d => %s\n", i, place_names[i]);
	printf("\n");

	/* .RG lesen */
	place_bounds = (int *)malloc(no_places * sizeof(int));
	for (i=0; i<no_places; i++)
		place_bounds[i] = 0;
	ende = no_markings = 0;
	walk = markings;
        while (!ende)
	{	new_m = (m_type *)malloc(sizeof(m_type));
		new_m->number = ++no_markings;
		new_m->used = 0;
		new_m->token = (int *)malloc(no_places * sizeof(int));
		for(i=0; i<no_places; i++)
		{	if(fread(&rg, sizeof(rg), 1, rg_file)!=1)
                        {	printf("the reduced reachability set contains");
				printf(" %d markings.\n\n", --no_markings);
				ende = 1;
                                break;
                        }
			new_m->token[i] = rg.number;
			new_m->next = NULL;
			if (place_bounds[i] < rg.number)
				place_bounds[i] = rg.number;
		}
		if (!ende)
			if (walk == NULL)
				markings = new_m;
			else	walk->next = new_m;
		walk = new_m;
        }
	fclose(rg_file);

	/* Ausgabe vorbereiten */
	for (i=0; i<no_places; i++)
	{	if (place_bounds[i] < i)
			place_bounds[i] = i;
		if (!place_bounds[i])
			place_bounds[i] = 1;
		else place_bounds[i] = 1 + (int)floor(log10((double)place_bounds[i]));
	}
	strcpy(topline, "         PLACES\nMarking ");
	for (i=0; i<no_places; i++)
	{	sprintf(help, "%d", i);
		for (k=0; k<=place_bounds[i]-strlen(help); k++)
			strcat(topline, " ");
		strcat(topline, help);
	}
	strcat(topline, "   Paths\n");
	ende = strlen(topline) + 8;
	for (i=0; i<ende; i++)
		strcat(topline, "-");
	strcat(topline, "\n");

	/* files lesen */
	for (file_num = 0; 1; file_num++)
	{	sprintf(p_name, "%s.EMC%02u", argv[arg_num], file_num);
		sprintf(c_name, "%s.CONV%02u", argv[arg_num], file_num);
		if ((p_file = fopen(p_name, "r")) == NULL)
		{	if (!file_num)
				continue;
			for (walk=markings; walk!=NULL; walk=walk->next)
				if (!walk->used)
				{	*topline = 0;
					for (i=0; i<no_places; i++)
					{	sprintf(help, "%d ", walk->token[i]);
						strcat(topline, help);
					}
					printf("Warning: Marking %s unused.\n", topline);
				}
			exit(1);
		}
		if ((c_file = fopen(c_name, "r")) == NULL)
		{	printf("ERROR: could not open %s for reading\n", c_name);
			printf("(perhaps you should restart");
			printf(" %s.graph with option -f)\n", argv[arg_num]);
			exit(1);
		}
		bit = 0;
		while(1)
		{	if(fread(&P_row_header, SIZEOF_P_row_header, 1, p_file)!=1)
				break;
			nums = P_row_header.entries_in_row;
			walk = m_lookup(P_row_header.row_index);
			if (file_num)
			{	if(fread(&C_row_header, SIZEOF_C_row_header, 1, c_file)!=1)
				{	printf("read error 1");
					exit(1);
				}
				if (P_row_header.row_index != C_row_header.row_index)
				{	printf("read error 2");
					exit(1);
				}
				nums_c = C_row_header.entries_in_row;				
				if (!bit)
				{	printf("Markings with enabled ");
					printf("det. transition (MDET_%02u):\n", file_num);
					printf("%s", topline);
				}
			} else
			{	if(fread(&C0_entry, SIZEOF_C0_entry, 1, c_file)!=1)
					break;
				if (P_row_header.row_index != C0_entry.state_number)
				{	printf("read error 3\n");
					exit(1);
				}
				if (!bit)
				{	printf("Markings without enabled ");
					printf("det. transitions (MEXP):\n");
					printf("%s", topline);
				}
			}
			bit = 1;  len = 10;
			printf("%7d ", walk->number);
			for (i=0; i<no_places; i++)
			{	sprintf(help, "%d", walk->token[i]);
				for (k=0; k<=place_bounds[i]-strlen(help); k++)
				{	printf(" ");
					len++;
				}
				printf("%s", help);
				len += strlen(help);
			}
			if (walk->used)
			{	printf(" ERROR: double usage of marking!\n");
				exit(1);
			}
			walk->used = 1;
			for (i=0; i<len; i++)
				help[i] = ' ';
			help[len] = 0;
			if (file_num)
				printf("   State change probabilities:\n");
			else	printf("   Mean sojourn time: %f\n", (float)C0_entry.sojourn_time);
			for (i=0; i<nums; i++)
			{	if (fread(&P_row_entry, SIZEOF_P_row_entry, 1, p_file) != 1)
				{	printf("File eror\n");
					exit(1);
				}
				printf("%s prob=> %d : %f\n", help, 
					P_row_entry.column_index, (float)P_row_entry.probability);
			}
			if (file_num)
			{	printf("%s Conversion factors:\n", help);
				for (i=0; i<nums_c; i++)
				{	if (fread(&C_row_entry, SIZEOF_C_row_entry, 1, c_file) != 1)
					{	printf("File eror C\n");
						exit(1);
					}
					printf("%s conv=> %d : %f\n", help, 
							C_row_entry.column_index,
							(float)C_row_entry.conversion_factor);
				}
			}
		}
		if (!bit)
		{	printf("No markings without enabled det. transitions.\n");
			printf("Set MEXP is empty.\n");
		}
		printf("\n");
		fclose(p_file);
		fclose(c_file);
	}
}


